"""Listing and extracting files from TAR and ZIP archives
"""
import abc
import os
import tarfile
import zipfile

class Archive(object):
   """Abstract base class to manipulate a TAR or ZIP archive.
   """
   __metaclass__ = abc.ABCMeta

   def __init__(self, path):
      self.path = path

   def __enter__(self):
      """Context guard entry.
      """
      return self

   def __exit__(self, excType, excVal, traceback):
      """Context guard exit.
      """
      self.close()

   def extract(self, destDir=".", members=None):
      """Extract a list of files (or all files by default) from this archive.
      """
      for name in self.list():
         if name.startswith(("/", "..")):
            raise NameError("%s: malicious file %s in archive" %
                            (self.path, name))

      self._extract(destDir, members=members)

   @abc.abstractmethod
   def list(self, ignoreDirs=False):
      """Return the list of members in this archive.

      Setting the ignoreDirs option to True allows to filter out directories,
      such that only files are listed.
      """
      pass

   @abc.abstractmethod
   def _extract(self, destDir, members=None):
      """Extract a list of files (or all files by default) from this archive.
      """
      pass

   @abc.abstractmethod
   def close(self):
      """Close the archive file.
      """
      pass


class TarArchive(Archive):
   """Helper class to handle TAR family archives.
   """

   def __init__(self, path):
      super().__init__(path)
      self._tar = tarfile.open(path)

   def list(self, ignoreDirs=False):
      if ignoreDirs:
         return [m.name for m in self._tar.getmembers() if not m.isdir()]
      else:
         return self._tar.getnames()

   def _extract(self, destDir, members=None):
      if members:
         membersInfo = [self._tar.getmember(m) for m in members]
      else:
         membersInfo = None
      self._tar.extractall(path=destDir, members=membersInfo)

   def close(self):
      self._tar.close()


class ZipArchive(Archive):
   """Helper class to handle ZIP archives.
   """

   def __init__(self, path):
      super().__init__(path)
      self._zf = zipfile.ZipFile(path)

   def list(self, ignoreDirs=False):
      if ignoreDirs:
         try:
            return [m.filename for m in self._zf.infolist() if not m.is_dir()]
         except AttributeError:
            # Work around missing ZipInfo.is_dir() method in python < 3.6
            return [m for m in self._zf.namelist() if not m.endswith('/')]
      else:
         return self._zf.namelist()

   def _extract(self, destDir, members=None):
      self._zf.extractall(path=destDir, members=members)

   def close(self):
      self._zf.close()


def openArchive(path):
   """Open an archive of type Tar or Zip, and return an Archive object which can
   be used to decompress this archive.
   """
   if tarfile.is_tarfile(path):
      return TarArchive(path)
   elif zipfile.is_zipfile(path):
      return ZipArchive(path)
   else:
      raise AttributeError("%s: unsupported archive format" % path)
