# Copyright 2016-2017 VMware, Inc.
# All rights reserved. -- VMware Confidential

"""Bora source tree

Library to retrieve information about a #bora tree (aka. $VMTREE), such as the
location of the build deliverables, or the build settings (product, build type,
path to PXE directory, etc.).
"""
import os

class LocalSc(dict):
   """A class to parse the Local.sc configuration file.

   Local.sc is the optional configuration file used by developers to customize
   their build. If present, it is located at the root of the bora tree.
   """

   def __init__(self, filePath):
      try:
         self._conf = {}
         with open(filePath, 'r') as f:
            try:
               exec(f.read(), None, self._conf)
            except Exception:
               raise SyntaxError("Failed to load %s. Please check your "
                                 "Local.sc for syntax errors. Also note no"
                                 " python imports or side effects allowed "
                                 "in this file." % filePath)
      except IsADirectoryError:
         raise AttributeError('%s is a directory (not a valid file).' %
                              filePath)
      self._path = filePath

   def _parseStrKey(self, key, lower=False, dontStrip=False):
      """Parse a key of type string.
      """
      value = self._conf.get(key, None)
      if value:
         try:
            if not dontStrip:
               value = value.strip()
            if lower:
               value = value.lower()
         except AttributeError:
            raise SyntaxError("%s: '%s' value must be a string (or None)" %
                              (self._path, key))
      else:
         if value not in ("", None):
            raise SyntaxError("%s: '%s' value must be a string (or None)" %
                              (self._path, key))
      return value if value else None

   @property
   def product(self):
      """SCons product name (e.g.: 'esx', 'esxcore', ...)
      """
      return self._parseStrKey('PRODUCT', lower=True)

   @property
   def buildType(self):
      """Build type (e.g.: 'obj', 'beta', 'release', ...)
      """
      return os.getenv('BUILDTYPE', os.getenv('VMBLD',
         self._parseStrKey('BUILDTYPE', lower=True)))

   @property
   def pxeDir(self):
      """Path to the root of the PXE directory.
      """
      return self._parseStrKey('ESX_PXE_PROVISION_DIR')

   @property
   def pxeUrl(self):
      """HTTP URL pointing to the root of the (i)PXE directory.
      """
      return self._parseStrKey('ESX_PXE_HTTP_ROOT')


def isBoraTree(path):
   """Return True if @path points to a valid bora tree.
   """
   if os.path.basename(path) != 'bora' or not os.path.isdir(path):
      return False

   # Verify that 'scons' and 'vmkdrivers' sibling directories are present, not
   # to get fooled by a subdirectory that might also be called 'bora'.
   srcRoot = os.path.dirname(path)
   for sibling in ['scons', 'vmkdrivers']:
      if not os.path.isdir(os.path.join(srcRoot, sibling)):
         return False
   return True

def findBora(dirPath=None):
   """Traverse the parents of the given directory (or of the current working
   directory) until the root of a valid bora tree is found, and return a path to
   that bora root directory.
   """
   if dirPath is None:
      dirPath = os.getcwd()
   else:
      dirPath = os.path.realpath(dirPath)

   path = dirPath
   while not os.path.samefile(path, os.sep):
      if isBoraTree(path):
         return path
      path = os.path.dirname(path)

   raise FileNotFoundError('no bora tree found under %s' % dirPath)


class Bora(LocalSc):
   """A class which represents a bora source tree.
   """

   def __init__(self, vmtree=None):
      self._vmtree = findBora(vmtree)
      localSc = os.path.join(self._vmtree, 'Local.sc')
      if os.path.exists(localSc):
         super().__init__(localSc)

   @property
   def vmtree(self):
      """Path to the 'bora' directory (aka. $VMTREE).
      """
      return self._vmtree

   @property
   def pubDir(self):
      """Path to the directory where the build output is published.
      """
      publishDir = os.path.join(self._vmtree, 'build', self.product,
                                self.buildType, 'publish')
      if not os.path.exists(publishDir):
         raise EnvironmentError("%s: ESX publish dir not found (have you built your "
                                "tree?)" % publishDir)
      return publishDir


#
# Code below is for debugging purposes...
#
if __name__ == "__main__":
   bora = Bora()
   print("vmtree: %s" % bora.vmtree)
   print("product: %s" % bora.product)
   print("buildType: %s" % bora.buildType)
   print("pxeDir: %s" % bora.pxeDir)
   print("pxeUrl: %s" % bora.pxeUrl)
   print("pubDir: %s" % bora.pubDir)
