# Copyright 2017-2018 VMware, Inc.
# All rights reserved. -- VMware Confidential

"""Busybox-like utilities.
"""
from hashlib import md5, sha1
import os

from borautils.httputils import urlOpen, urlRead
import gzip

# Default IO size. Too small would impact performance (more reads), and too
# big would impact memory footprint. 64Kb is fairly high (shutils.filecopyobj()
# uses 16Kb by defaut), but at VMware we like to manipulate big files (ISOs,
# coredumps, ...).
READ_CHUNK_SIZE = 64 * 1024

def cat(filePath):
   """Iterate over each line in the given file.

   @filePath can be a regular file on the filesytem, or any type URL supported
   by urllib.
   """
   f = urlOpen(filePath)
   for line in f:
      yield line.decode('utf-8')
   f.close()

def md5sum(*filePaths):
   """Return the MD5 sum of the given (set of) file(s).
   """
   m = md5()
   for path in filePaths:
      with open(path, 'rb') as f:
         while True:
            data = f.read(READ_CHUNK_SIZE)
            if not data:
               break
            m.update(data)
   return m.hexdigest()

def sha1sum(*filePaths):
   """Return the SHA1 sum of the given (set of) file(s).
   """
   m = sha1()
   for path in filePaths:
      with open(path, 'rb') as f:
         while True:
            data = f.read(READ_CHUNK_SIZE)
            if not data:
               break
            m.update(data)
   return m.hexdigest()

def touch(path, mode='a'):
   """Imitate the touch unix command.
   """
   with open(path, mode):
      os.utime(path, None)

def wget(uri, dest, mode=None, computeMD5=False):
   """Download a file via HTTP.
   """
   if os.path.isdir(dest):
      dest = os.path.join(dest, os.path.basename(uri))

   if computeMD5:
      m = md5()

   with open(dest, 'wb') as fout:
      for data in urlRead(uri):
         if computeMD5:
            m.update(data)
         fout.write(data)

   if mode is not None:
      os.chmod(dest, mode)

   return m.hexdigest() if computeMD5 else None

def gunzip(inFile, outFile):
   """Runs gunzip on @inFile and stores the result in @outFile. Both arguments are
      file paths.
   """
   with open(outFile, "wb") as fout:
      with gzip.open(inFile) as gz:
         while True:
            data = gz.read(READ_CHUNK_SIZE)
            if not data:
               break
            fout.write(data)
