# Copyright 2015,2018 VMware, Inc.
# All rights reserved. -- VMware Confidential

"""UNIX daemon library

This library provides generic facilities to implement and control UNIX daemons.
"""
import os
from resource import getrlimit, RLIMIT_NOFILE
import signal
import sys
import time

from borautils.exception import DebugInfoMixIn
from borautils.flock import flockEx, funlockEx

# By convention, daemons write their PID file under /var/run/. This requires
# the daemon to run with superuser privileges, which is usually the case for
# system daemons. Other daemons (which don't run with 'root' privileges) may
# pick a different PID dir.
DEFAULT_DAEMON_PIDDIR = os.path.join(os.sep, 'var', 'run')

class DaemonError(DebugInfoMixIn, Exception):
   """Generic daemon exception.
   """
   _errType = "Daemon error"

   def __init__(self, errStr, daemon):
      super().__init__(errStr)
      self.addDebugInfo('daemon', str(daemon))
      pidFile = daemon.getPidFilePath()
      try:
         pid = open(pidFile, 'r').read()
      except IOError:
         self.addDebugInfo('PID file', '<none>')
      else:
         self.addDebugInfo("PID file", pidFile)
         self.addDebugInfo('PID (from PID file)', pid)

class NoSuchDaemon(DaemonError):
   """Attempted to stop a daemon, but no instance of this daemon is running.
   """
   _errType = "Daemon is not running"

class DaemonAlreadyRunning(DaemonError):
   """Attempted to start a daemon, but this daemon is already running.
   """
   _errType = "Another instance of this Daemon is already running"


class UnixDaemon(object):
   """Generic UNIX daemon base class.

   Generic base class to daemonize a UNIX process and to control the resulting
   daemon (start, stop, restart). There are two ways to specify the daemon's
   activity: by passing a callable object to the constructor, or by overriding
   the run() method in a subclass.

   Daemons created via this class use a file lock ("PID file") to guarantee only
   one instance of themselves.
   """

   def __init__(self, name, pidDir=None, target=None, args=(), kwargs=None):
      """Instantiate (but do not start) a UnixDaemon.

      @name: the daemon's name
      @target: callable object to be invoked by the run() method. Defaults to
               None, meaning nothing is called.
      @pidDir: where to store the PID file
      """
      self.__name = name
      if pidDir is None:
         self.__pidDir = DEFAULT_DAEMON_PIDDIR
      else:
         self.__pidDir = pidDir
      self.__target = target
      self.__args = args
      self.__kwargs = kwargs if kwargs is not None else {}
      self.__pidFile = None
      self.__fd = None

   def __str__(self):
      """Return a human-readable string to identify this daemon.
      """
      return self.__name

   def getPidFilePath(self):
      """Return a path to this daemon's PID file.
      """
      return os.path.join(self.__pidDir, self.__name + '.pid')

   def _forkAndExit(self):
      """Fork a child process, and terminate the calling process.
      """
      try:
         pid = os.fork()
      except OSError as e:
         raise DaemonError("fork() error: %s" % e, self)
      if pid > 0:
         sys.exit(0)
      assert(pid == 0)

   def _writePid(self):
      """Save the current process' PID in the daemon's PID files.
      """
      self.__pidFile.seek(0)
      self.__pidFile.truncate()
      self.__pidFile.write('%d\n' % os.getpid())
      self.__pidFile.flush()
      self.__pidFile.seek(0)

   def getPid(self):
      """Return this daemons's PID.
      """
      try:
         pidFileData = open(self.getPidFilePath(), 'r').read()
      except IOError:
         raise NoSuchDaemon("no PID file found", self)

      try:
         pid = int(pidFileData.strip())
      except ValueError:
         raise DaemonError("invalid PID file", self)

      return pid

   def _cleanupAndExit(self, *args):
      """Remove the daemon's PID file and exit.
      """
      funlockEx(self.__pidFile)
      self.__pidFile.close()
      os.unlink(self.getPidFilePath())
      sys.exit(0)

   def _daemonize(self):
      """Daemonize the current UNIX process using the 'double-fork' method.

      The double-fork method is explained in great details in W. Richard
      Stevens' book "Advanced Programming in the UNIX environment" (ISBN
      978-0-321-63773-4), Chapter 13.
      """
      # Try to acquire the daemon's PID file exclusive lock. This guarantees
      # that only one instance of the daemon can run at a time.
      self.__pidFile = open(self.getPidFilePath(), 'a+')
      try:
         flockEx(self.__pidFile)
      except IOError:
         raise DaemonAlreadyRunning("failed to lock daemon's PID file", self)

      # Reset default file creation mask
      os.umask(0)

      # Create a child process and let the current process (parent) exit. The
      # child process inherits its parent's GID, but gets a new PID. This
      # guarantees that the child process is not a process group leader, which
      # is necessary to call setsid().
      self._forkAndExit()

      # Create a new session with a single process group in it. The current
      # process becomes the leader of the new session, the process group leader
      # of the new process group, and has no controlling terminal.
      os.setsid()

      # Set the current working directory to '/', to avoid holding references
      # to a mounted filesystem (which would prevent from unmounting this
      # filesystem).
      os.chdir('/')

      # Close all inherited file descriptors but the PID file's.
      openedFdSoftLimit, _ = getrlimit(RLIMIT_NOFILE)
      for fd in range(openedFdSoftLimit):
         if fd != self.__pidFile.fileno():
            try:
               os.close(fd)
            except OSError:
               pass

      # Ignore signals coming from child processes or from a controlling
      # terminal. Since daemons don't have a controlling terminal, these signals
      # should never be received.
      signal.signal(signal.SIGHUP, signal.SIG_IGN)
      signal.signal(signal.SIGTSTP, signal.SIG_IGN)
      signal.signal(signal.SIGTTIN, signal.SIG_IGN)
      signal.signal(signal.SIGTTOU, signal.SIG_IGN)
      signal.signal(signal.SIGCHLD, signal.SIG_IGN)

      # Fork a grand-child process, and let the child process terminate. The
      # grand-child is guaranteed no to be a session leader, which prevents it
      # from acquiring a controlling terminal.
      self._forkAndExit()

      # Save the current process' PID into the daemon's PID file
      self._writePid()

      # Open all standard file descriptors, and redirect them to /dev/null. This
      # allows any library to read from <stdin> and to write to <stdout> and
      # <stderr>.
      si = open(os.devnull, 'r')
      so = open(os.devnull, 'a+')
      se = open(os.devnull, 'a+', 0)
      os.dup2(si.fileno(), sys.stdin.fileno())
      os.dup2(so.fileno(), sys.stdout.fileno())
      os.dup2(se.fileno(), sys.stderr.fileno())

      # Make sure to hold a reference to the standard file descriptors, so they
      # don't get closed by the python garbage collector.
      self.__fd = (si, so, se)

      # Attach the cleanup routine to the SIGTERM signal handler
      signal.signal(signal.SIGTERM, self._cleanupAndExit)

   def start(self):
      """Start the daemon.

      Start the daemon activity in a new background process. This function does
      not return (the calling process is terminated).
      """
      self._daemonize()
      self.run()
      self._cleanupAndExit()

   def stop(self):
      """Stop the daemon.

      Calling this function terminates the daemon execution by sending a SIGTERM
      signal against the PID that is contained in the daemon's PID file.
      """
      pid = self.getPid()

      while True:
         try:
            os.kill(pid, signal.SIGTERM)
         except OSError as e:
            if "No such process" in e.args:
               break
            else:
               raise
         time.sleep(0.1)

   def restart(self):
      """Restart the daemon.

      Re-start the daemon activity in a new background process. This function
      does not return (the calling process is terminated).
      """
      try:
         self.stop()
      except NoSuchDaemon:
         pass
      self.start()

   def run(self):
      """Daemon's main function.

      This method may be overridden in a subclass of UnixDaemon.
      """
      if self.__target is not None:
         self.__target(*self.__args, **self.__kwargs)
