# Copyright 2018 VMware, Inc.  All rights reserved. -- VMware Confidential

"""Functionality for defering cleanup operations."""


def defer():
   """Request a new defer context."""
   class _Function:
      """A function wrapper guarding against multiple executions."""
      def __init__(self, function):
         """Initialize a function wrapper."""
         self._function = function

      def __call__(self):
         """On a call of the object invoke the underlying function."""
         if self._function is not None:
            self._function()
            # Mark function as executed.
            self._function = None

      def release(self):
         """Release the function from the defer operation without invoking it."""
         self._function = None

   class _Defer:
      """Objects of this class act as a context with which to register deferred functions."""
      def __init__(self):
         """Initialize a defer object to make it ready for use."""
         self._functions = []

      def __enter__(self):
         """The block enter handler just returns a reference to this object."""
         return self

      def __exit__(self, type_, value, traceback):
         """The block exit handler destroys the object."""
         self.destroy()

      def defer(self, function, *args, **kwargs):
         """Register a deferred function invocation."""
         result = _Function(lambda: function(*args, **kwargs))
         self._functions += [result]
         return result

      def release(self):
         """Release all deferred functions without executing them."""
         self._functions = []

      def destroy(self):
         """Destroy the object, invoke all deferred functions."""
         # Always invoke in reverse order of registration.
         for function in reversed(self._functions):
            function()

   return _Defer()
