"""ELF64 binary parser

Provide a binding over libelf.so and a high level abstraction of an elf binary.
"""
from ctypes import *
import math
import os

from borautils.exception import DebugInfoMixIn
from borautils.infra.toolchain import loadTool

LIBELF_SO = loadTool("lin64/elfutils-0.94.1-2/lib/libelf.so")
libelf = CDLL(LIBELF_SO)

ASCII_NULL = b'\x00'
ELF_MAGIC = b'\x7FELF'

# Constants (from elf.h)
ELF_C_READ = 1
EI_NIDENT = 16
EV_CURRENT = 1
EV_NONE = 0

ET_DYN = 3
ET_CORE = 4

PT_NULL = 0
PT_LOAD = 1
PT_NOTE = 4

NT_AUXV = 6
AT_NULL = 0
AT_FLAGS = 8
AT_ENTRY = 9

R_AMD64_64 = 1

SHT_DYNSYM = 11
SHT_REL = 9
SHT_RELA =  4
SHT_SYMTAB = 2
STB_GLOBAL = 1
STT_NOTYPE = 0

def define(f, argtypes, restype):
   f.argtypes = argtypes
   f.restype = restype

# Types (from elf.h)
Elf_Cmd = c_int
Elf_IdentT = c_char * EI_NIDENT

Elf64_Addr = c_uint64
Elf64_Off = c_uint64
Elf64_Section = c_uint16
Elf64_Sxword = c_int64
Elf64_Xword = c_uint64
Elf64_Word = c_uint32
Elf64_Half = c_uint16

class PackedStructure(Structure):
   _pack_ = 1

class Elf(PackedStructure):
   _fields_ = []
ElfP = POINTER(Elf)

class Elf64_Ehdr(PackedStructure):
   _fields_ = [('e_ident', Elf_IdentT),
               ('e_type', Elf64_Half),
               ('e_machine', Elf64_Half),
               ('e_version', Elf64_Word),
               ('e_entry', Elf64_Addr),
               ('e_phoff', Elf64_Off),
               ('e_shoff', Elf64_Off),
               ('e_flags', Elf64_Word),
               ('e_ehsize', Elf64_Half),
               ('e_phentsize', Elf64_Half),
               ('e_phnum', Elf64_Half),
               ('e_shentsize', Elf64_Half),
               ('e_shnum', Elf64_Half),
               ('e_shstrndx', Elf64_Half)]
Elf64_EhdrP = POINTER(Elf64_Ehdr)

class Elf64_Phdr(PackedStructure):
   _fields_ = [('p_type', Elf64_Word),
               ('p_flags', Elf64_Word),
               ('p_offset', Elf64_Off),
               ('p_vaddr', Elf64_Addr),
               ('p_paddr', Elf64_Addr),
               ('p_filesz', Elf64_Xword),
               ('p_memsz', Elf64_Xword),
               ('p_align', Elf64_Xword)]
Elf64_PhdrP = POINTER(Elf64_Phdr)

class Elf_Data(PackedStructure):
   _fields_ = [('d_buf', c_void_p),
               ('d_type', c_int),
               ('d_version', c_uint),
               ('d_size', c_size_t),
               ('d_off', c_size_t),
               ('d_align', c_size_t)]
Elf_DataP = POINTER(Elf_Data)

class Elf64_Rel(PackedStructure):
   _fields_ = [('r_offset', Elf64_Addr),
               ('r_info', Elf64_Xword)]

class Elf64_Rela(PackedStructure):
   _fields_ = [('r_offset', Elf64_Addr),
               ('r_info', Elf64_Xword),
               ('r_addend', Elf64_Sxword)]

class Elf_Scn(PackedStructure):
   _fields_ = []
Elf_ScnP = POINTER(Elf_Scn)

class Elf64_Shdr(PackedStructure):
   _fields_ = [('sh_name', Elf64_Word),
               ('sh_type', Elf64_Word),
               ('sh_flags', Elf64_Xword),
               ('sh_addr', Elf64_Addr),
               ('sh_offset', Elf64_Off),
               ('sh_size', Elf64_Xword),
               ('sh_link', Elf64_Word),
               ('sh_info', Elf64_Word),
               ('sh_addralign', Elf64_Xword),
               ('sh_entsize', Elf64_Xword)]

class Elf64_Sym(PackedStructure):
   _fields_ = [('st_name', Elf64_Word),
               ('st_info', c_char),
               ('st_other', c_char),
               ('st_shndx', Elf64_Section),
               ('st_value', Elf64_Addr),
               ('st_size', Elf64_Xword)]

# Macros for relocations
def ELF64_R_SYM(i):
   return i >> 32

def ELF64_R_TYPE(i):
   return i & 0xffffffff

# Define functions with ctypes
define(libelf.elf_version, [c_int], c_int)

# Always call this before calling anything, read man elf_version(3)
if libelf.elf_version(EV_CURRENT) == EV_NONE:
   raise ValueError("%s: version mismatch" % LIBELF_SO)

define(libelf.elf_begin, [c_int, Elf_Cmd, ElfP], ElfP)
define(libelf.elf_end, [ElfP], c_int)
define(libelf.elf_errmsg, [c_int], c_char_p)
define(libelf.elf_errno, [], c_int)
define(libelf.elf_getdata, [Elf_ScnP, Elf_DataP], Elf_DataP)
define(libelf.elf_getident, [ElfP, POINTER(c_int)], POINTER(Elf_IdentT))
define(libelf.elf_getscn, [ElfP, c_size_t], Elf_ScnP)
define(libelf.elf_getshnum, [ElfP, POINTER(c_size_t)], c_int)
define(libelf.elf_getshstrndx, [ElfP, POINTER(c_size_t)], c_int)
define(libelf.elf_ndxscn, [Elf_ScnP], c_size_t)
define(libelf.elf_nextscn, [ElfP, Elf_ScnP], Elf_ScnP)
define(libelf.elf_rawdata, [Elf_ScnP, Elf_DataP], Elf_DataP)
define(libelf.elf_strptr, [ElfP, c_size_t, c_size_t], c_char_p)
define(libelf.elf64_getehdr, [ElfP], Elf64_EhdrP)
define(libelf.elf64_getphdr, [ElfP], Elf64_PhdrP)
define(libelf.elf64_getshdr, [Elf_ScnP], POINTER(Elf64_Shdr))


def intFromBytes(buf):
   """Convert an array of little-endian bytes into an integer.
   """
   return int.from_bytes(buf, byteorder='little')

def readAt(fd, size, offset=0, pos=os.SEEK_SET):
   """Read @size bytes from @fd starting at @offset.
   """
   os.lseek(fd, offset, pos)
   return os.read(fd, size)


class ElfError(DebugInfoMixIn, Exception):
   """ELF parsing base exception.
   """
   _errType = "ELF parsing error"

   def __init__(self, errStr, binary=None):
      super().__init__(errStr)
      if binary is not None:
         self.addDebugInfo('binary', binary.path)
      errno = libelf.elf_errno()
      if errno != 0:
         self.addDebugInfo('errno', errno)
         self.addDebugInfo('errmsg', libelf.elf_errmsg(errno))


class ElfBin(object):
   """High level abstraction of an elf binary
   """
   def __init__(self, filepath):
      self.path = filepath
      self._fd = -1
      self._elf = None
      self._ehdr = None
      self._phdr = None

      self.symbols = set()
      self.undefinedSymbols = set()

      self._sections = {}
      self.numSections = 0
      self._undefSymSectionBlacklist = set()
      self.sectionSymtab = None
      self._sectionContentCache = {}

      self._fd = os.open(filepath, os.O_RDONLY)

      if os.read(self._fd, len(ELF_MAGIC)) != ELF_MAGIC:
         os.close(self._fd)
         raise ElfError("file is not an ELF binary", self)

      self._elf = libelf.elf_begin(self._fd, ELF_C_READ, None)
      if self._elf is None:
         os.close(self._fd)
         raise ElfError("failed to get the ELF descriptor", self)

      self._loadEhdr()

      self._parseSections()
      self._parseSymbols()

   def __enter__(self):
      return self

   def __exit__(self, type, value, traceback):
      self.close()

   def _loadEhdr(self):
      """Parse a the ELF header (EHDR) and cache interesting fields.
      """
      ehdr = libelf.elf64_getehdr(self._elf)
      if not bool(ehdr):
         raise ElfError("failed to parse ELF header", self)

      self._type = ehdr.contents.e_type
      self._phentsize = ehdr.contents.e_phentsize
      self._phnum = ehdr.contents.e_phnum

   def _parsePhdr(self):
      """Parse and cache the ELF Program Header (PHDR).
      """
      if self._phdr is None:
         self._phdr = libelf.elf64_getphdr(self._elf)
         if not bool(self._phdr):
            raise ElfError("failed to parse ELF program header", self)
      return self._phdr

   def read(self, virtAddr, size):
      """Read @size bytes from the ELF image starting at virtual address
      @virtAddr.
      """
      phdr = self._parsePhdr()

      for n in range(self._phnum):
         entry = phdr[n]
         if entry.p_type == PT_NULL:
            break
         elif entry.p_type == PT_LOAD:
            vaddr = entry.p_vaddr
            if entry.p_vaddr <= virtAddr < entry.p_vaddr + entry.p_filesz:
               offset = entry.p_offset + virtAddr - entry.p_vaddr
               assert offset + size <= entry.p_offset + entry.p_filesz
               return readAt(self._fd, size, offset)

   def _sectionData(self, section):
      """Iterator over a section data
      """
      data = None
      while True:
         data = libelf.elf_getdata(section, data)
         if not bool(data):
            break
         yield data.contents

   def getSectionContent(self, name):
      """Returns the content of a section (in bytes)
      """
      if name in self._sectionContentCache:
         return self._sectionContentCache[name]

      section = self._sections[name]
      data = libelf.elf_rawdata(section, None)
      buf = cast(data.contents.d_buf, POINTER(c_char))
      result = buf[:data.contents.d_size]
      self._sectionContentCache[name] = result
      return result

   @staticmethod
   def getStringFromBuffer(buf, offset=0):
      """Get the null-terminated string starting at buffer[offset]
      """
      buf = buf[offset:]
      rOffset = buf.index(ASCII_NULL)
      return bytes(buf[:rOffset]).decode()

   def _parseRelocations(self, sectionName, symbolsSet):
      """Parse relocations in given section and add the symbols found to
      symbolsSet.
      """
      if sectionName not in self._sections:
         return

      section = self._sections[sectionName]
      shdr = libelf.elf64_getshdr(section)
      t = Elf64_Rel if shdr.contents.sh_type == SHT_REL else Elf64_Rela

      # We currently only support Elf64_Rela
      if t != Elf64_Rela:
         return

      for data in self._sectionData(section):
         relocs = cast(data.d_buf, POINTER(t))
         num_relocs = data.d_size // sizeof(t)
         for i in range(num_relocs):
            # For now we support only 64 bits relocs
            if ELF64_R_TYPE(relocs[i].r_info) == R_AMD64_64:
               # XXX: Assuming blindly here that the reloc is in .data,
               # we need to check properly that it is always the case
               data = self.getSectionContent('.data')
               sym = self.getStringFromBuffer(data, relocs[i].r_addend)
               symbolsSet.add(sym)

   def _iterSymbols(self):
      """Iterate over this ELF symbols.
      """
      if self.sectionSymtab is None:
         return

      shdr = libelf.elf64_getshdr(self.sectionSymtab)
      strndx = shdr.contents.sh_link
      for data in self._sectionData(self.sectionSymtab):
         syms = cast(data.d_buf, POINTER(Elf64_Sym))
         numSyms = data.d_size // sizeof(Elf64_Sym)
         for i in range(numSyms):
            symbolName = libelf.elf_strptr(self._elf, strndx, syms[i].st_name)
            yield symbolName.decode(), syms[i]

   def _parseSymbols(self):
      """Parse all symbols and put them in a set
      """
      for symbolName, sym in self._iterSymbols():
         if sym.st_shndx <= self.numSections:
            info = intFromBytes(sym.st_info)
            bind, symtype = (info >> 4, info & 0xF)
            if (symtype == STT_NOTYPE and bind == STB_GLOBAL and
                (sym.st_shndx not in self._undefSymSectionBlacklist)):
               self.undefinedSymbols.add(symbolName)
         self.symbols.add(symbolName)

   def _getSymbolVAddr(self, name):
      """get a symbol virtual adress and size.

      Return a (vAddr, size) tuple, where vAddr is the virtual address of
      symbol @name in the loaded ELF image, and size is the symbol size in
      bytes.
      """
      for symbolName, sym in self._iterSymbols():
         if symbolName == name:
            return (sym.st_value, sym.st_size)
      raise ElfError('%s: no such symbol' % name, self)

   def readSymbol(self, name):
      """Read the content of the given symbol.
      """
      vAddr, size = self._getSymbolVAddr(name)
      return self.read(vAddr, size)

   def _parseSections(self):
      """Find all sections and put them in a dict
      """
      section = libelf.elf_nextscn(self._elf, None)
      i = 1
      while bool(section):
         self.numSections += 1
         shstrndx = c_size_t()
         libelf.elf_getshstrndx(self._elf, byref(shstrndx))
         shdr = libelf.elf64_getshdr(section)
         name = libelf.elf_strptr(self._elf, shstrndx, shdr.contents.sh_name)
         self._sections[name.decode()] = section

         # Undefined symbol in .rodata* sections are skipped
         if b'.rodata' in name:
            self._undefSymSectionBlacklist.add(i)

         # XXX: do we need to handle dynamic symbols also?
         if shdr.contents.sh_type == SHT_SYMTAB:
            self.sectionSymtab = section

         section = libelf.elf_nextscn(self._elf, section)
         i += 1

   @property
   def sections(self):
      """Dict of ELF sections (SectionName => Elf_Scn)
      """
      return self._sections

   def parseAuxv(self):
      """Return a dictionary containing the Auxilliary Vector (AUXV) found in
      the NOTE section.
      """
      AUXV_PAIR_SIZE = 2 * sizeof(c_uint64)
      NOTE_ALIGN = 4

      auxv = {}
      phdr = self._parsePhdr()

      for n in range(self._phnum):
         entry = phdr[n]
         if entry.p_type == PT_NULL:
            break
         if entry.p_type == PT_NOTE:
            notes = readAt(self._fd, entry.p_filesz, entry.p_offset)
            while notes:
               noteType = intFromBytes(notes[8:12])
               nameSize = intFromBytes(notes[0:4])
               descSize = intFromBytes(notes[4:8])
               paddedNameSize = math.ceil(nameSize / NOTE_ALIGN) * NOTE_ALIGN
               paddedDescSize = math.ceil(descSize / NOTE_ALIGN) * NOTE_ALIGN

               if noteType == NT_AUXV:
                  descOffset = 3 * sizeof(c_uint32) + paddedNameSize
                  desc = notes[descOffset:descOffset + descSize]

                  # sizeof(auxv) must be a multiple of sizeof(auxv entry)
                  assert len(desc) % AUXV_PAIR_SIZE == 0

                  for i in range(descSize // AUXV_PAIR_SIZE):
                     offset = i * AUXV_PAIR_SIZE
                     key = intFromBytes(desc[offset:offset + 8])
                     if key == AT_NULL:
                        break
                     val = intFromBytes(desc[offset + 8:offset + 16])
                     auxv[key] = val
                  return auxv

               noteSize = 3 * sizeof(c_uint32) + paddedNameSize + paddedDescSize
               assert noteSize <= len(notes)
               notes = notes[noteSize:]
      return auxv

   def close(self):
      libelf.elf_end(self._elf)
      self._elf = None
      os.close(self._fd)
      self._fd = -1


class ElfCore(ElfBin):
   """Class representing an ELF core file.
   """
   def __init__(self, filepath):
      super().__init__(filepath)
      if self._type != ET_CORE:
         raise ElfError("not a valid ELF core file (bad ELF type: %#x)" %
                        self._type, self)
