"""Feature state switch lib

Utility lib for feature state switch dictionary class
"""
class FeatureState(dict):
   """Feature state switch dictionary class
   """
   ON = True
   OFF = False

   def __init__(self, disabledFeatures=None, enabledFeatures=None):
      super().__init__()
      if disabledFeatures:
         for featureName in disabledFeatures:
            self.disable(featureName)
      if enabledFeatures:
         for featureName in enabledFeatures:
            self.enable(featureName)

   def __setitem__(self, key, value):
      """Override __setitem__ function to avoid setting different values
      for the same feature
      """
      if value not in [self.ON, self.OFF]:
         raise ValueError("%s: invalid feature state value "
                          "(should be either ON or OFF)" % str(value))
      if self.__contains__(key):
         curValue = self.__getitem__(key)
         if curValue != value:
            raise ValueError("%s: cannot turn feature %s "
                             "(state switch already exists and is set to %s)" %
                             ("ON" if value else "OFF", key,
                              "ON" if curValue else "OFF"))
      super(FeatureState, self).__setitem__(key, value)

   def enable(self, featureName):
      """Turn a feature state switch on
      """
      self.__setitem__(featureName, self.ON)

   def disable(self, featureName):
      """Turn a feature state switch off
      """
      self.__setitem__(featureName, self.OFF)

   def isEnabled(self, featureName):
      """Get a feature state
      """
      return self.__getitem__(featureName)

   def generateVmkernelBootOpts(self):
      """Convert a feature state switch dictionary to a boot
      option string that can be passed in boot.cfg or vmkernel command line
      """
      bootOpts = []
      for name, value in self.items():
         bootOpts.append("FeatureState.%s=%s" %
                         (name, "enabled" if value else "disabled"))
      return ' '.join(bootOpts)
