# Copyright 2015 VMware, Inc.
# All rights reserved. -- VMware Confidential

"""File-based lock operations
"""
from contextlib import contextmanager
import fcntl
import time

def flockEx(fd, timeoutInSec=None):
   """Acquire an exclusive lock on the given file.

   The file is locked mutual exclusive among processes, _not_ among threads of
   the same process. This function is non blocking and raises an IOError
   "Resource temporarily unavailable" if some process already holds the file's
   lock.

   -- Mixing flock() and fork() --

   Underlying flock implementations do vary from OS to OS, particularly in the
   way a file lock is preserved - or not - across fork.

   On linux, file locks created with flock do persist across fork (on the
   contrary to those created with fcntl). When flock acquires a file lock from a
   parent process, and this process forks a child, then the child inherits the
   lock from its parent. As a result, both the parent and the child hold
   independent copies of the lock. Releasing the lock in the parent does not
   release the lock in the child, and vice versa.

   On ESX however (and other systems where flock is a wrapper to fcntl), flock
   does not preserve file locks across fork.
   """
   if timeoutInSec:
      start = time.time()

   while True:
      try:
         fcntl.flock(fd.fileno(), fcntl.LOCK_EX | fcntl.LOCK_NB)
         break
      except IOError:
         if (timeoutInSec is None or
             (timeoutInSec != 0 and time.time() - start >= timeoutInSec)):
            raise
         time.sleep(.1)

def funlockEx(fd):
   """Release a file's exclusive lock.
   """
   fcntl.flock(fd.fileno(), fcntl.LOCK_UN)

@contextmanager
def openEx(*args, **kwargs):
   """To open/close a file in exclusive mode in a 'with' context.
   """
   fd = open(*args)
   try:
      flockEx(fd, **kwargs)
      try:
         yield fd
      finally:
         funlockEx(fd)
   finally:
      fd.close()
