"""gdb - Interface to debug with gdb
"""
import os
import sys
import re
from shutil import which
from stat import S_ISCHR
import subprocess
from tempfile import NamedTemporaryFile


def isCharDev(filePath):
   """Return True if @filePath points to a chardev.
   """
   try:
      mode = os.stat(filePath).st_mode
   except Exception:
      return False
   return S_ISCHR(mode)

class Gdb(object):
   """Gdb wrapper for python.
   """

   @staticmethod
   def isRemoteTarget(target):
      """Check whether the given target is remote.

      Return True if the given target points to a serial device or to a remote
      url. Supported remote targets are:
        - /dev/ttyS0
        - [{udp|tcp}:]hostname:port
      """
      gdbRemoteHostRE = r'((tcp|udp):)?[^:]*:\d+'
      return (isCharDev(target) or
              re.match(gdbRemoteHostRE, target, flags=re.IGNORECASE))

   def __init__(self, path=None):
      self.path = which('gdb') if path is None else path
      if not os.path.exists(self.path):
         raise FileNotFoundError("%s: gdb not found", self.path)

      self._major = None
      self._minor = None
      self.prompt = None
      self.vars = {}
      self.macros = []
      self.srcPath = []
      self.debugPath = []
      self.soPath = []
      self._loadVersion()

   def _loadVersion(self):
      """Query and store gdb version.
      """
      p = subprocess.Popen([self.path, '--version'], stdout=subprocess.PIPE,
                           stderr=subprocess.PIPE)
      out, err = p.communicate()

      if p.returncode != 0:
         raise EnvironmentError("failed to retrieve gdb version: %s" %
                                err.decode())

      version = out.decode()
      m = re.search(r"(\d+)\.(\d+)", version.split('\n')[0])
      if m is None:
         raise EnvrionmentError("failed to parse gdb --version output)")

      self._major = int(m.group(1))
      self._minor = int(m.group(2))

   @property
   def version(self):
      """Gdb version as a (major, minor) tuple.
      """
      return (self._major, self._minor)

   def run(self, program, target=None, commands=None, options=None, batch=False,
           quiet=True, tui=False, outFile=None):
      """Run gdb.

      In batch mode, gdb runs all the command files specified with '-x' and
      quits.

      In quiet mode (default), gdb does not print introductory and copyright
      messages.

      In TUI mode, GDB splits the console window in 2 halves to display the
      source code in a separate text window. TUI assumes that the curses library
      is available.
      """
      init = []

      if outFile:
         init += ["set logging off",
                  "set logging file %s" % outFile,
                  "set logging overwrite on",
                  "set logging redirect on",
                  "set logging on"]

      init += ["set print thread-events off",
               "set confirm off",
               "file %s" % program]
      init += ["dir %s" % path for path in self.srcPath]

      if self.debugPath:
         init += ["set debug-file-directory %s" % ":".join(self.debugPath)]

      if self.soPath:
         init += ["set solib-search-path %s" % ":".join(self.soPath)]

      if self.prompt is not None:
         init += ["set prompt %s" % self.prompt]

      init += ["set $%s = %s" % (name, val) for name, val in self.vars.items()]

      if target is not None:
         if self.isRemoteTarget(target):
            init += ["set remotebaud 115200",
                     "set remote read-aux-vector-packet on",
                     "target remote %s" % target]
         else:
            init += ["target core %s" % target]

      for path in self.macros:
         if not os.path.exists(path):
            raise FileNotFoundError("%s: no such gdb macro" % path)
         if os.path.splitext(path)[1].lower() == '.py' and self._major < 7:
            raise Exception("GDB v%u.%u does not support python macros" %
                            self._major, self._minor)
         init += ["source %s" % path]

      if commands:
         init += commands

      cmd = [self.path, '--nw']
      if quiet:
         cmd.append('--quiet')
      if batch:
         cmd.append('--batch')
      if tui:
         cmd.append('--tui')
      if options:
         cmd.append(options)

      with NamedTemporaryFile(mode='w', delete=False) as tmpFile:
         tmpFile.write("\n".join(init) + '\n')
         tmpFile.flush()
         cmd.append("--command=%s" % tmpFile.name)
         subprocess.call(cmd)
      os.unlink(tmpFile.name)
