"""HTTP utils
"""
import re
import os

from http.client import OK as HTTP_OK, NOT_FOUND as HTTP_NOT_FOUND
from http.client import responses as http_responses
from urllib.parse import urlencode
from urllib.request import urlopen, Request
from urllib.error import HTTPError, URLError

from borautils.exception import DebugInfoMixIn


class HttpRequestError(DebugInfoMixIn, Exception):
   """An HTTP request has failed to execute, or has returned an error.
   """
   _errType = "HTTP Request error"

   def __init__(self, url, status, errMsg):
      super().__init__(self._errType)
      self.addDebugInfo('url', url)
      if status is not None:
         self.addDebugInfo('status', "%u (%s)" % (status,
                                                  http_responses[status]))
      self.addDebugInfo('error', errMsg)

class InvalidUrlError(HttpRequestError):
   """A URL for HTTP request is invalid.
   """
   _errType = "Invalid URL error"

def _urlFormat(url):
   """Format @url to urlopen()'s liking.
   """
   if not isinstance(url, Request):
      if not re.match(r'[a-zA-Z]+://', url):
         url = "file://" + os.path.realpath(url)
   return url

def urlOpen(url):
   """Open a URL, or any resource that is supported by urllib.
   """
   import ssl
   url = _urlFormat(url)
   try:
      # XXX: workaround PEP 476
      #      https://www.python.org/dev/peps/pep-0476/
      response = urlopen(url, context=ssl._create_unverified_context())
   except HTTPError as e:
      if e.code == HTTP_NOT_FOUND:
         raise InvalidUrlError(url, e.code, e.reason)
      else:
         raise HttpRequestError(url, e.code, e.reason)
   except URLError as e:
      if isinstance(e.reason, FileNotFoundError):
         raise e.reason
      elif "No such file or directory" in e.reason:
         raise InvalidUrlError(url, None, e.reason)
      else:
         raise HttpRequestError(url, None, e.reason)

   status = response.getcode()
   if status != HTTP_OK and status != None:
      try:
         errMsg = response.readline()
      except Exception:
         errMsg = "unknown error"
      raise HttpRequestError(response.geturl(), status, errMsg)
   return response

def urlRead(uri, readChunkSize=None):
   """Read data from an HTTP URL, or from any resource supported by urllib.
   """
   DEFAULT_READ_CHUNK_SIZE = 64 * 1024

   if readChunkSize is None:
      readChunkSize = DEFAULT_READ_CHUNK_SIZE

   with urlOpen(uri) as f:
      while True:
         data = f.read(readChunkSize)
         if not data:
            break
         yield data

def httpPost(url, params, encoding='utf-8'):
   """Send an HTTP POST request.
   """
   header = {"Content-type": "application/x-www-form-urlencoded",
             "Accept": "text/plain"}
   data = urlencode(params).encode(encoding)
   req = Request(url, data, header)
   return urlOpen(req)

def urlExists(url):
   """Check if a file (or any resource supported by urllib) exists.
   """
   notFoundExeptions = (InvalidUrlError, FileNotFoundError)

   try:
      response = urlOpen(url)
   except notFoundExeptions:
      exists = False
   else:
      response.close()
      exists = True
   return exists


#
# Code below is for debugging purposes...
#
if __name__ == "__main__":
   from argparse import ArgumentParser

   parser = ArgumentParser()
   parser.add_argument(type=str, metavar='URL', dest='url',
                       help=("URL to an HTTP resource"))
   args = parser.parse_args()

   for line in urlOpen(args.url):
      print(line.decode('utf-8'), end='')
   print('')
