"""Hardware DB API library

Library to query the Hardware DB [1] REST API [2].

[1] Hardware DB project:
    https://wiki.eng.vmware.com/Testesx/HWDB

[2] Hardware DB RestAPI documentation:
    TODO: Migrate RestAPI documentation to wiki [1]
    https://wiki.eng.vmware.com/VMkernelQE/SDLC/HWDB#Rest_API
"""
from __future__ import print_function

import getpass
import json
from threading import Event, Thread

from borautils.exception import DebugInfoMixIn
try:
   from borautils.httputils import httpPost, HttpRequestError
except ImportError:
   print('!!Please call bora/support/script/hwdb-cli instead of %s!!' % __file__)
   raise

HWDB_API = "http://vmkernel-infrastructure.eng.vmware.com/hwdb"

# The threshold needs to be adjusted according to CAT load
# in different phases.
CAT_POOL_IDLE_THRESHOLD = 0.5

class HwdbApiRequestError(DebugInfoMixIn, Exception):
   """A request to the HWDB API has failed.
   """
   _errType = "HWDB API request failure"

   def __init__(self, errStr, requestUrl, params):
      super().__init__(errStr)
      self.addDebugInfo("request", requestUrl)
      self.addDebugInfo("payload", params)

def _hwdbRequest(context, params):
   """Send a HWDB request.

   @param context: the context of hwdb request
   @param param: the json params for hwdb request
   """
   url = "%s/%s" % (HWDB_API, context)
   try:
      response = httpPost(url, params)
   except HttpRequestError:
      raise HwdbApiRequestError("HTTP post request failure", url, params)

   info = json.loads(response.read().decode("utf-8"))
   if "error" in info:
      raise HwdbApiRequestError(info["error"], url, params)
   return info

def _listMachines():
   """Get all machines' detailed info in hardware DB
   """
   return _hwdbRequest("list", {})

def listAllMachinesDetailedInfo():
   """List all hardware database machines detailed info
   """
   machines = _listMachines()
   for machine in machines:
      printMachineInfo(machine)

def listAllMachinesBriefInfo():
   """List all hardware database machines brief info
   """
   machines = _listMachines()
   # print title
   print('%-30s %-15s %-15s %-15s %-50s'
         % ('HostName', 'Status', 'Memory', 'CpuCores', 'CpuModel'))
   for machine in machines:
      print('%-30s %-15s %-15s %-15s %-50s'
            % (machine["hostname"], machine["status"],
               machine["physical_memory"],
               machine["cpu_cores"], machine["cpu_model"]))

def getCatMachineIdleRate(location):
   """Check current HWDB machine idle rate in CAT production pool

   @param location: machine suite location of the requested machine
   """
   machines = _listMachines()
   idleWDC = 0
   idlePA = 0
   totalWDC = 0
   totalPA = 0
   for machine in machines:
      if machine["pool"] == 'CAT':
         if machine["suite_location"] == 'WDC':
            totalWDC += 1
            if machine["status"] == 'idle' and  machine["wait_user"] == 'N/A':
               idleWDC += 1
         else:
            totalPA += 1
            if machine["status"] == 'idle' and machine["wait_user"] == 'N/A':
               idlePA += 1

   idleRateWDC = 0.0 if totalWDC == 0 else float(idleWDC / totalWDC)
   idleRatePA = 0.0 if totalPA == 0 else float(idlePA / totalPA)
   return idleRateWDC if location == 'WDC' else idleRatePA

def printMachineInfo(machine):
   """Print machine info retrieved from hardware DB
   """
   for key in sorted(machine.keys()):
      print("%s : %s" % (key.ljust(16), machine[key]))
   print("\n")


class HwdbMachine(object):
   """Class to operate machine from hardware DB
   """
   _lease = 48
   _user = getpass.getuser()

   def __init__(self, machineInfo):
      """Instantiate a HwdbMachine instance.

      Call classmethod queryMachine/reserveMachine to get HwdbMachine instance.
      Do not instantiate directly
      """
      assert machineInfo
      self._machineInfo = machineInfo
      self._hostName = self._machineInfo["hostname"]
      self._pool = self._machineInfo["pool"]
      self._location = self._machineInfo["suite_location"]

   @classmethod
   def queryMachine(cls, hostName):
      """Return hwdb instance which hostname is @hostName
      """
      params = {"hostname": hostName}
      machineInfo = _hwdbRequest("query", params)
      return HwdbMachine(machineInfo)

   @classmethod
   def reserveMachine(cls, hostName):
      """Reserve machine @hostName from hardware DB,
      then return hwdb instance which hostname is @hostName
      """
      params = {"hostname": hostName,
                "user": cls._user,
                "lease": cls._lease}
      machineInfo = _hwdbRequest("reserve", params)
      return HwdbMachine(machineInfo)

   def reserveMachinePrecheck(self):
      """XXX: This is a workaround
         @return True: when idle rate of CAT production machines is beyond the threshold
         @return False: when idle rate of CAT production machines is below the threshold
      """
      if self._pool == 'CAT':
         idleRate = getCatMachineIdleRate(self._location)

         if idleRate < CAT_POOL_IDLE_THRESHOLD:
            return False
      return True

   def releaseMachine(self):
      """Release machine back to hardware DB
      """
      params = {"hostname": self._hostName,
                "user": self._user}
      return _hwdbRequest("release", params)["result"]

   def renewLease(self):
      """Renew machine's lease

      Enforce default lease period after which the lease must be renewed,
      or the machine is automatically released.
      """
      params = {"hostname": self._hostName,
                "lease": self._lease}
      return _hwdbRequest("renewlease", params)["result"]

   def getMachineInfo(self):
      """Return machine info
      """
      return self._machineInfo

   def getHostName(self):
      """Return host name
      """
      return self._hostName

   def printReservationFailMsg(self):
      """Print error msg if reservation failed due to reserving server
         from CAT production pool when pool idle rate is low.
      """
      print("[ERROR] Failed to reserve HWDB machine.\
          \n%s belongs to %s pool and can't be reserved unless you have a blocker PR.\
          \nCAT production pool idle rate is below threshold (%s) for now.\
          \nPlease contact ldong or zhuojil(PA) and jdong or wsiqi(BJ).\
          \nThanks!" % (self._hostName, self._pool, CAT_POOL_IDLE_THRESHOLD))

class CatHwdbMachine(HwdbMachine):
   """Class to operate machines from hardware DB by CAT
   """
   _lease = 1

   def __init__(self, machineInfo):
      """Instantiate a CatHwdbMachine instance
      """
      super().__init__(machineInfo)
      self._updateLease = _RenewLeaseThread(self)
      self._user = machineInfo['reserve_user']
      self._updateLease.start()

   @classmethod
   def reserveMachine(cls, reserveUser, location):
      """Reserve machine from hardware DB by CAT testrun,
      which location is @location.

      then return hwdb instance which hostname is @hostName
      """
      params = {"location": location,
                "user": reserveUser,
                "lease": cls._lease}
      machineInfo = _hwdbRequest("reserve", params)
      return CatHwdbMachine(machineInfo)

   def releaseMachine(self):
      """Terminate renew lease thread and release machines back to hardware DB
      """
      self._updateLease.stop()
      super().releaseMachine()

class _RenewLeaseThread(Thread):
   """Thread to renew CAT reserved machine's lease periodically
   """
   RENEW_PERIOD = 30 * 60

   def __init__(self, hwdbMachine):
      """Instantiate a _RenewLeaseThread instance
      """
      super().__init__()
      self.daemon = True
      self._hwdb = hwdbMachine
      self._quitEvent = Event()

   def run(self):
      while not self._quitEvent.is_set():
         self._quitEvent.wait(self.RENEW_PERIOD)
         try:
            self._hwdb.renewLease()
         except Exception:
            pass

   def stop(self):
      """Terminate this renew lease thread
      """
      self._quitEvent.set()
      self.join()

#
# Code below is for user to operate machines in hardware DB...
#
if __name__ == "__main__":
   from argparse import ArgumentParser

   parser = ArgumentParser()

   group = parser.add_mutually_exclusive_group(required=True)

   group.add_argument('--listall', action='store_true', dest='listall',
                      help='List detailed info for machines in HWDB pool')
   group.add_argument('--list', action='store_true', dest='list',
                      help='List brief info for machines in HWDB pool')
   group.add_argument('--query', action='store', type=str, dest='query',
                      metavar='<HOST>',
                      help='Show detailed info for queried machine')
   group.add_argument('--release', action='store', type=str, dest='release',
                      metavar='<HOST>', help='Release a machine')
   group.add_argument('--renew', action='store', type=str, dest='renew',
                      metavar='<HOST>', help="Renew a machine's lease")
   group.add_argument('--reserve', action='store', type=str, dest='reserve',
                      metavar='<HOST>', help='Reserve a machine')

   args = parser.parse_args()

   def _getHostFromInput(args):
      """Return corresponding hostname according to user input
      """
      if args.reserve:
         return args.reserve
      elif args.query:
         return args.query
      elif args.release:
         return args.release
      elif args.renew:
         return args.renew

   if args.listall:
      listAllMachinesDetailedInfo()
   elif args.list:
      listAllMachinesBriefInfo()
   elif _getHostFromInput(args):
      hostname = _getHostFromInput(args)
      hwdb = HwdbMachine.queryMachine(hostname)
      if args.reserve:
         if hwdb.reserveMachinePrecheck():
            hwdb = HwdbMachine.reserveMachine(hostname)
            print("Reserved machine %s successfully" % hostname)
            printMachineInfo(hwdb.getMachineInfo())
         else:
            hwdb.printReservationFailMsg()
      if args.query:
         print("Show machine %s 's detailed info" % hostname)
         printMachineInfo(hwdb.getMachineInfo())
      if args.release:
         hwdb.releaseMachine()
         print("Release machine %s successfully" % hostname)
      if args.renew:
         hwdb.renewLease()
         print(("Renew machine %s 's lease (%d hours) successfully"
              % (hostname, hwdb._lease)))
