"""Intelligent Platform Management Interface (IPMI) library.

Library for controlling IPMI-enabled devices. This library relies on the linux
ipmitool utility.
"""
import logging
import os
import select
import subprocess
import time

from borautils.exception import DebugInfoMixIn
from borautils.ping import ping
from borautils.infra.toolchain import loadTool

IPMITOOL = loadTool("lin32/ipmitool-1.8.9/bin/ipmitool")
MAX_ATTEMPTS = 3
logger = logging.getLogger('ipmi')

class IPMIError(DebugInfoMixIn, Exception):
   """Generic IPMI exception.
   """
   _errType = "IPMI error"

   def __init__(self, errStr, host):
      super().__init__(errStr)
      self.addDebugInfo('ipmitool', IPMITOOL)
      self.addDebugInfo('host', host)

class IPMITransportError(IPMIError):
   """IPMI connection failure.
   """
   _errType = "IPMI transport error"

class IPMICommandError(IPMIError):
   """IPMI command failure.
   """
   _errType = "IPMI command error"

   def __init__(self, cmd, returnCode, stderr, host, reason):
      errStr = "command executed, but returned an error"
      super().__init__(errStr, host)
      self.addDebugInfo('command', cmd)
      self.addDebugInfo('exit status', returnCode)
      self.addDebugInfo('stderr', stderr)
      self.addDebugInfo('reason', reason)

class IPMIUnsupportedError(IPMIError):
   """IPMI unsupported feature/command.
   """
   _errType = "IPMI feature unsupported"

class IPMIConfigError(IPMIError):
   """IPMI incorrect host configuration.
   """
   _errType = "IPMI configuration error"

class IPMI(object):
   """Class to represent an IPMI session.
   """
   READ_TIMEOUT = 30

   def __init__(self, hostname, username, password):
      """Connect to a remote server IPMI interface.
      """
      super().__init__()
      self._hostname = hostname
      self._username = username
      self._password = password
      self._solProcess = None
      self._solPrivilege = None

      for isLanplus in [True, False]:
         self._isLanplus = isLanplus
         try:
            self._execCmd("power status")
         except IPMICommandError:
            continue
         else:
            return

      raise IPMIError('No IPMI interface found', self._hostname)

   def _doPopen(self, ctrlCmd, privilege="OPERATOR"):
      """Prepare and execute an IPMI command via Popen.
      """
      cmd = [IPMITOOL, '-H', self._hostname, '-L', privilege]
      if self._username:
         cmd += ['-U', self._username, '-P', self._password]
      if self._isLanplus:
         cmd += ['-I', 'lanplus']
      cmd += ctrlCmd.split()

      return subprocess.Popen(cmd, stdin=subprocess.PIPE,
                              stdout=subprocess.PIPE, stderr=subprocess.PIPE)

   def _execCmd(self, cmd, privilege="OPERATOR"):
      """Execute an IPMI command.
      """
      def byteToStr(byteString):
         """Convert an ascii byte string into a python native string.
         """
         return "".join(['%c' % c for c in byteString])

      # Retry up to MAX_ATTEMPTS times, if ipmi command has error due to bad
      # network (packet dropped) or too many users are connected to this IPMI.
      msg = 'error unknown'
      for _ in range(MAX_ATTEMPTS):
         p = self._doPopen(cmd, privilege)
         out, err = p.communicate()
         if p.returncode == 0:
            return byteToStr(out)

         try:
            ping(self._hostname)
         except Exception:
            raise IPMITransportError('host unreachable', self._hostname)

         errStr = byteToStr(err)

         if errStr.startswith('Address lookup for'):
            raise IPMITransportError('host unreachable', self._hostname)

         if 'Unable to establish LAN session' in errStr:
            raise IPMITransportError('unable to establish LAN session',
                                     self._hostname)

         if 'insufficient resources for session' in errStr:
            msg = 'too many users are connected to this IPMI'
         elif 'ipmi_lanplus_send_payload: Assertion' in errStr:
            msg = 'protocol error due to bad network (packet dropped)'
         elif 'Insufficient privilege level' in errStr:
            privilege = 'ADMINISTRATOR'
            msg = "ipmi command fail due to current user's privilege issue"

         time.sleep(1)

      raise IPMICommandError(cmd, p.returncode, errStr.strip(),
                             self._hostname, msg)

   def powerOn(self):
      """Power on a remote server via IPMI.
      """
      self._execCmd("power on")

   def powerOff(self):
      """Power off a remote server via IPMI.
      """
      self._execCmd("power off")

   def reset(self):
      """Reset a remote server via IPMI.
      """
      self._execCmd("power reset")

   def coldBoot(self):
      """Cold reboot a remote server via IPMI.
      """
      self._execCmd("power cycle")

   def isOn(self):
      """Query a remote server power status via IPMI.

      This function returns True is the server is powered on.
      """
      for line in self._execCmd("power status").splitlines():
         if line.strip() == "Chassis Power is on":
            return True
      return False

   def setPxeBoot(self):
      """Force PXE boot.
      """
      self._execCmd("chassis bootdev pxe")

   def setDiskBoot(self):
      """Set the default disk boot.
      """
      self._execCmd("chassis bootdev disk")

   def solActivate(self):
      """Use ipimitool to connect to SOL of a remote server.
      """
      solInfo = self.solGetInfo()
      # ipmitool sol can only be activated with sufficient privilege level,
      # some of machines sol privilege level is ADMINISTRATOR,but only TechOps
      # have the permission to access admin account. So let testesx to query
      # and use properly privilege level when activate sol
      self._solPrivilege = solInfo['Privilege Level']
      if solInfo['Enabled'] != "true":
         raise IPMIConfigError("SOL is not enabled on this host",
                                self._hostname)
      retry = 3
      command = 'sol activate'
      ret = 0
      errStr = None
      while retry:
         self._solProcess = self._doPopen(command, self._solPrivilege)
         # if sol session has been occupied by other process, you can still get
         # a valid _solProcess. But it will be terminated at once. So adding a
         # sleep here to capture this case.
         time.sleep(3)
         ret = self._solProcess.poll()
         if ret is not None and ret != 0:
            _, errStr = self._solProcess.communicate()
            self.solDeactivate()
            retry -= 1
         else:
            # Some machines send no output over SOL until a new line is
            # received
            self._solProcess.stdin.write(b'\n')
            self._solProcess.stdin.flush()
            self._epoll = select.epoll()
            self._epoll.register(self._solProcess.stdout, select.EPOLLIN)
            return

      msg = "Couldn't activate SOL after 3 retries."
      raise IPMICommandError(command, ret, errStr.strip(),
                             self._hostname, msg)

   def solDeactivate(self):
      """Terminate a SOL connection by sending the termination sequence
      """
      assert self._solProcess
      assert self._solPrivilege
      ret = self._solProcess.poll()
      if ret is None:
         assert self._epoll
         self._epoll.unregister(self._solProcess.stdout)
         self._epoll.close()
         self._epoll = None

         SOL_TERMINATION_SEQUENCE = b'~.'
         self._solProcess.stdin.write(SOL_TERMINATION_SEQUENCE)
         self._solProcess.communicate()
      else:
         try:
            self._execCmd('sol deactivate', self._solPrivilege)
         except Exception:
            pass
      self._solProcess = None

   def solGetInfo(self):
      """Return Serial Over Lan Status.

      Execute sol info from ipmitools and parse the input.
      Output is returned as a dictionary.
      """
      try:
         out = self._execCmd("sol info")
      except IPMICommandError:
         raise IPMIUnsupportedError("SOL is not supported by host",
                                    self._hostname)

      info = {}
      for line in out.split(os.linesep):
         if line.strip():
            key, _, val = line.partition(':')
            info[key.strip()] = val.strip()

      return info

   def solReadline(self, timeout=None):
      """Read one line from serial.

      Raise an exception if something went wrong or None process exited
      gracefully. This function will return None if it's timeout, otherwise
      block until there is something to read.
      @param timeout: timeout in seconds. Blocks if timeout is negative or None.
      """
      assert self._solProcess
      assert self._epoll

      if timeout is None:
         timeout = self.READ_TIMEOUT

      events = self._epoll.poll(timeout)
      if not events:
         # return none when timed out
         return None
      else:
         line = self._solProcess.stdout.readline()

      if line:
         try:
            # PR 2018133: garbage characters sometimes appear in serial logs.
            # Filter out them for now with a warning
            return line.decode('utf-8')
         except UnicodeDecodeError as e:
            logger.warn('Ignored illegal characters in serial log: %s', e)
            return line.decode('utf-8', errors='ignore')

      # Nothing to read here, the process might be dead
      ret = self._solProcess.poll()
      if ret is not None and ret != 0:
         # Something went wrong, let's read the error
         err = self._solProcess.stderr.read()
         raise IPMITransportError('lost connection to host : %s' % err,
                                  self._hostname)

      raise EOFError

   def solReadlines(self):
      """Endlessly read serial output from the server.
      """
      while True:
         yield self.solReadline()

#
# Code below is for debugging purposes...
#
if __name__ == "__main__":
   from argparse import ArgumentParser

   parser = ArgumentParser()
   parser.add_argument(type=str, metavar='HOST', dest='host',
                       help="Remote server hostname")
   parser.add_argument(type=str, metavar='COMMAND', dest='cmd',
                       help="IPMI power command")
   parser.add_argument('--user', type=str, metavar='USERNAME', dest='username',
                       help="IPMI username")
   parser.add_argument('--password', type=str, metavar='PASSWORD',
                       dest='password',
                       help="IPMI password")

   args = parser.parse_args()

   chassis = IPMI(args.host, args.username, args.password)

   cmd = args.cmd.lower()
   if cmd == 'on':
      chassis.powerOn()
   elif cmd == 'off':
      chassis.powerOff()
   elif cmd == 'reset':
      chassis.reset()
   elif cmd == 'status':
      print("%s is powered %s" % (args.host, 'on' if chassis.isOn() else 'off'))
   elif cmd == 'sol':
      print(chassis.solGetInfo())
   elif cmd == 'pxeboot':
      chassis.setPxeBoot()
   elif cmd == 'diskboot':
      chassis.setDiskBoot()

      chassis.solActivate()
      for i in range(5):
         line = chassis.solReadline()
         print(line, end='')
      chassis.solDeactivate()
