# Copyright 2017 VMware, Inc.  All rights reserved. -- VMware Confidential

"""logparser.py --

A line-wise log parsing framework, generally useful for parsing any log
file with useful messages which can be described using one or more
regular expressions.
"""

import io
import re


class LogParserMatch:
   """A successful match in LogParser.

   Attributes:
      lineNum: Log line number.
      name: Matching rule name.
      groups: Dictionary of regex groups.
   """
   def __init__(self, lineNum, ruleName, ruleREMatch):
      self.lineNum = lineNum
      self.name    = ruleName
      self.groups  = ruleREMatch.groupdict()


   def __repr__(self):
      return 'line %u: <%s> %s' % (self.lineNum, self.name, self.groups)


class LogParser:
   """Line-wise log parser.

   Consists of a set of regexes (or rules) that are matched against
   every line in a given stream of log lines. Successful matches are
   returned as LogParserMatch objects.
   """

   def __init__(self, strip=True):
      """Intialize LogParser.

      Args:
         strip: If True, strip left and right whitespaces in each line
            before performing the match.
      """
      self._rules = {}
      self._strip = strip


   def AddRules(self, rules):
      """Add regular-expressions to pattern match each line of a log
      with. The expressions are compiled, saved, and used in all
      invocations of the Parse* functions.

      Args:
         rules: A dictionary mapping rule names to regular expression
            strings.
      """
      self._rules.update(
         (name, re.compile(ruleRE))
            for name, ruleRE in rules.items())


   def OnMatch(self, lineNum, ruleName, ruleREMatch):
      """Function invoked for each match, returning an instance of
      LogParserMatch. Derived object may override this function to
      generate customized returns.
      """
      return LogParserMatch(lineNum, ruleName, ruleREMatch)


   def ParseStream(self, logStream):
      """Parse a log input stream, generating LogParserMatch objects
      for lines that match any rule.
      """
      lineNum = 0
      for line in logStream:
         lineNum += 1
         for ruleName, ruleRE in self._rules.items():
            if self._strip:
               line = line.strip()
            ruleREMatch = ruleRE.match(line)
            if ruleREMatch is not None:
                yield self.OnMatch(lineNum, ruleName, ruleREMatch)


   def ParseFile(self, logFile):
      """Parse a log file, generating LogParserMatch objects
      for lines that match any rule.
      """
      with open(logFile, "r", encoding='utf-8') as fh:
         for match in self.ParseStream(fh):
            yield match


   def ParseString(self, logString):
      """Parse a log string, generating LogParserMatch objects
      for lines that match any rule.
      """
      with io.StringIO(logString) as logStream:
         for match in self.ParseStream(logStream):
            yield match


if __name__ == '__main__':
   print('Running basic tests ...')
   lp = LogParser()
   lp.AddRules({
      'rule_a' : r'(?P<line_num>\d+): (?P<greeting>Hello|Goodbye) world!',
      'rule_b' : r'(?P<line_num>\d+): (?P<key>\w+) = (?P<value>\d+)',
   })
   loglines = (
      """1: What's up world!
         2: Hello World!
         3: a = 10
         4: Goodbye = blah
         5: Goodbye = 1234
         6: Goodbye world!
      """
   )
   match = lp.ParseString(loglines)

   # match "3: a = 10"
   m = next(match)
   assert (m.name == 'rule_b' and
           m.lineNum == int(m.groups['line_num']) and
           m.groups['key'] == 'a'
           and m.groups['value'] == '10'), m

   # match "4: Goodbye = 10"
   m = next(match)
   assert (m.name == 'rule_b' and
           m.lineNum == int(m.groups['line_num']) and
           m.groups['key'] == 'Goodbye' and
           m.groups['value'] == '1234'), m

   # match "7: Goodbye world!"
   m = next(match)
   assert (m.name == 'rule_a' and
           m.lineNum == int(m.groups['line_num']) and
           m.groups['greeting'] == 'Goodbye'), m

   try:
      m = next(match)
   except StopIteration:
      pass
   else:
      assert False, 'Unexpected match: %s' % m
   print('OK')

# vim: sw=3 ts=3
