# Copyright 2017 VMware, Inc.
# All rights reserved. -- VMware Confidential

"""MAC address handling

MAC address format is specified in RFC7042: "IANA Considerations and IETF
Protocol Usage for IEEE 802 Parameters":

https://tools.ietf.org/html/rfc7042
"""

class MacAddr(bytes):
   """Object representing a MAC address.
   """

   def __new__(cls, addr, allowMulticast=False):
      try:
         segments = addr.replace('-', ':').split(':')
      except AttributeError:
         raise TypeError("MacAddr() argument must be a string, not '%s'" %
                         type(addr).__name__)

      if len(segments) != 6:
         raise ValueError("%s: invalid MAC address" % addr)

      try:
         octets = [int(s, base=16) for s in segments]
      except ValueError:
         raise ValueError("%s: invalid MAC address" % addr)

      try:
         obj = super().__new__(cls, octets)
      except ValueError:
         raise ValueError("%s: invalid MAC address" % addr)

      if not allowMulticast and obj.isMulticast:
         raise ValueError("%s: invalid unicast MAC address" % addr)

      return obj

   def __str__(self):
      """Return a human-readable string to identify this object.
      """
      return ":".join(["%02x" % octet for octet in self])

   @property
   def isMulticast(self):
      """Return True if this MAC address is multicast ('group' bit is set).
      """
      _GROUP_BYTE = 0
      _GROUP_MASK = 0x1
      return self[_GROUP_BYTE] & _GROUP_MASK == _GROUP_MASK

   @property
   def isLocal(self):
      """Return True if this MAC address is locally administrated ('local' bit
      is set).
      """
      _LOCAL_BYTE = 0
      _LOCAL_MASK = 0x2
      return self[_LOCAL_BYTE] & _LOCAL_MASK == _LOCAL_MASK

   @property
   def oui(self):
      """The IANA OUI (Organizationally Unique Identifier) for this MAC address.
      """
      return self[:3]

   @property
   def isVMware(self):
      """Return True if this is a VMware MAC address (OUI is (0x00, 0x50, 0x56))
      """
      _IANA_OUI_VMWARE = bytes((0x00, 0x50, 0x56))
      return self.oui == _IANA_OUI_VMWARE


def isValidMacAddr(arg):
   """Return True if @arg is a string reprensenting a valid MAC address.
   """
   try:
      MacAddr(arg)
   except (ValueError, TypeError):
      return False
   return True
