# Copyright 2016 VMware, Inc.
# All rights reserved. -- VMware Confidential

"""NFS utils library
"""
import os
import re
import subprocess

def listFsMounts():
   """Return the list of mounted file systems in the form of (volume,
   mountPoint, fsType, attr) tuples.
   """
   mountRe = re.compile(r"(\S+)\son\s(\S+)\stype\s(\w+)\s\((.*)\)")

   mounts = []

   try:
      p = subprocess.run(['mount', '-l'], stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE, universal_newlines=True)
   except FileNotFoundError:
      if os.uname().sysname.lower() != 'linux':
         raise OSError("'mount' command not found (OS must be linux)")
      else:
         raise
   else:
      p.check_returncode()

   for line in p.stdout.split('\n'):
      m = mountRe.match(line)
      if m is not None:
         mounts.append(m.groups())
   return mounts

def getMountPoint(path):
   """Given a path, return the first mount point within the parent directories.
   """
   path = os.path.realpath(path)
   while path != os.sep and not os.path.ismount(path):
      path = os.path.dirname(path)
   return path

def getMountPointInfo(path):
   """Return a (volume, mountPoint, fsType, attr) tuple containing information
   about the closest mount point in the given path's parent directories.
   """
   mountPoint = getMountPoint(path)
   for mount in listFsMounts():
      if mountPoint == mount[1]:
         return mount
   raise OSError("%s: no mount point found" % os.path.realpath(path))


class NfsPath(object):
   """Class to retrieve info about an NFS path (host, share, mount point, etc.).
   """

   def __init__(self, localPath):
      self.localPath = os.path.realpath(localPath)
      self._scanned = False

   def _scanNfsMount(self):
      """Collect system info about this path's NFS mount point.
      """
      if self._scanned:
         return

      volume, mountPoint, fsType, attrs = getMountPointInfo(self.localPath)
      if fsType.lower() != 'nfs':
         raise ValueError("%s: not an NFS resource" % self.localPath)

      relPath = os.path.relpath(self.localPath, mountPoint)
      assert not relPath.startswith(os.pardir)

      self._host, sep, self._share = volume.partition(':')
      assert sep == ':'

      self._nfsPath = os.path.join(self._share, relPath)
      self._mountPoint = mountPoint
      self._attrs = attrs
      self._scanned = True

   def isNfs(self):
      """Return True if this path points to an NFS location.
      """
      try:
         self._scanNfsMount()
      except ValueError:
         return False
      else:
         return True

   @property
   def host(self):
      """Return hostname of the NFS server which serves this NFS resource.
      """
      self._scanNfsMount()
      return self._host

   @property
   def share(self):
      """Return the server-side path to the root of this NFS share.
      """
      self._scanNfsMount()
      return self._share

   @property
   def nfsPath(self):
      """Return the server-side path to this NFS resource.
      """
      self._scanNfsMount()
      return self._nfsPath

   @property
   def mountPoint(self):
      """Return the client-side mount point for this NFS path.
      """
      self._scanNfsMount()
      return self._mountPoint

   @property
   def attrs(self):
      """Return the mount attributes for this NFS path.
      """
      self._scanNfsMount()
      return self._attrs.split(',')


#
# Code below is for debugging purposes...
#
if __name__ == "__main__":
   from argparse import ArgumentParser

   desc = ("Print out information about the mount point of a given NFS path.")

   disclaimer = ("** DISCLAIMER ** This program is provided to help debugging "
                 "the %s library. It is NOT a production tool. CLI "
                 "may change at any time." % os.path.basename(__file__))

   parser = ArgumentParser(description=desc, epilog=disclaimer)
   parser.add_argument(type=str, metavar='PATH', dest="path", default='.',
                       nargs='?', help="")
   args = parser.parse_args()

   path = NfsPath(args.path)

   print("local path: %s" % path.localPath)
   if path.isNfs():
      print("is NFS: yes")
      print("nfs path: %s" % path.nfsPath)
      print("nfs host: %s" % path.host)
      print("nfs share: %s" % path.share)
      print("mount point: %s" % path.mountPoint)
      print("mount attrs: %s" % path.attrs)
   else:
      print("is NFS: no")
