# Copyright 2015-2018 VMware, Inc. All rights reserved. -- VMware Confidential

'''Library to obtain VMs via OVF from any URL.

OvfImporter class abstracts common functionalities for importing an OVF file.

To utilize this library, simply instantiate it and pass in the url from
which to deploy the OVF and the name of the VM.
'''

__author__ = "VMware, Inc."


import os
import ssl
import time
import contextlib
import http.client
import urllib

from borautils.exception import DebugInfoMixIn
from socket import gethostname
try:
   from pyVmomi import Vim
   from pyVim import connect, path
except ImportError as e:
   raise ImportError('pyVmomi/pyVim not found. Please make sure these '
                     'libraries are available on your system.')


class NoDatastoreFoundError(DebugInfoMixIn, Exception):
   """Datastore not found for the operation.
   """
   _errorType = "No datastore found."

   def __init__(self, errStr):
      super().__init__(errStr)


class Ovf(object):
   '''
   OVF base class.
   '''

   def __init__(self, url, name, datastore=None, network=None,
                resourcepool=None, diskmode='thin', mob=None, si=None):
      '''
      @param name: Name of the VM to be created.
      @param datastore: The datastore object to place the VM on.
                     default: First available datastore.
      @param network: The network object to connect to the VM.
                     default: First available VM network object.
      @param resourcepool: Resource pool object to place the VM in.
                           default: compute resource resource pool
      @param mob: The VM object, if the VM was previously created
      @param si: Service Instance. Use this if available, instead of default.
      '''
      super(Ovf, self).__init__()

      if mob is not None:
         self.__mob = mob
         return

      self.serviceInstance = connect.GetSi() if si is None else si
      assert self.serviceInstance is not None, 'VIM Service Instance required'
      self.__mob = None
      self.__url = url
      self.__name = name
      self.__datastore = datastore
      self.__network = network
      self.__resourcepool = resourcepool
      self.__diskmode = diskmode

      context = ssl._create_unverified_context()
      rsp = urllib.request.urlopen(self.__url, context=context)
      with contextlib.closing(rsp) as urlhandle:
         self.__data = urlhandle.read()
         self.__data = self.__data.decode("utf-8")

      serviceContent = self.serviceInstance.RetrieveContent()
      datacenter = serviceContent.rootFolder.childEntity[0]

      if self.__datastore is None:
         localDatastores = [ d for d in datacenter.datastoreFolder.childEntity
                               if hasattr(d.info, 'vmfs') and
                                  d.info.vmfs.local ]
         if len(localDatastores) == 0:
            raise  NoDatastoreFoundError('No VMFS local datastore found')
         localDatastores = sorted(localDatastores,
                                  key= lambda d: d.info.freeSpace,
                                  reverse=True)
         self.__datastore = localDatastores[0]

      if self.__network is None:
         self.__network = datacenter.networkFolder.childEntity[0]

      computeResource = datacenter.hostFolder.childEntity[0]
      self.__host = computeResource.host[0]
      if self.__resourcepool is None:
         self.__resourcepool = computeResource.resourcePool

      self.__importOvf()

   def __importOvf(self):
      '''
      Import the OVF from VM Library. Requires creating Import spec parameters
      passing it to the import spec and downloading the bits from the OVF URL
      and pushing it to the NFC Lease URL.
      '''
      networkMapping = [Vim.OvfManager.NetworkMapping(network=self.__network)]
      ovfMgrCisp = Vim.OvfManager.CreateImportSpecParams
      importSpecParams = ovfMgrCisp(entityName=self.__name,
                                    hostSystem=self.__host,
                                    diskProvisioning=self.__diskmode,
                                    networkMapping=networkMapping)

      serviceContent = self.serviceInstance.RetrieveContent()
      ovfManager = serviceContent.ovfManager
      spec = ovfManager.CreateImportSpec(ovfDescriptor=self.__data,
                                         resourcePool=self.__resourcepool,
                                         datastore=self.__datastore,
                                         cisp=importSpecParams)
      importSpec = spec.GetImportSpec()
      baseDir = os.path.dirname(self.__url)
      lease = self.__resourcepool.ImportVApp(importSpec)

      try:
         while lease.GetState() == Vim.HttpNfcLease.State.initializing:
            time.sleep(1)

         if lease.GetState() == Vim.HttpNfcLease.State.ready:
            self.__mob = lease.info.entity
            self.__downloadFiles(spec.fileItem, lease, baseDir)
            lease.Complete()
         else:
            raise lease.error
      except:
         lease.Abort()
         raise

   def __downloadFiles(self, fileItems, lease, baseDir):
      '''
      Download all file items
      @param fileItems: FileItems object from ImportSpec
      @param lease: NFC Lease
      @param baseDir: Directory URL path from OVF URL
      '''

      # Build a map, deviceId -> URL, of the URLs to which we can PUT disk
      # content
      uploadUrlMap = {}
      for deviceUrl in lease.info.deviceUrl:
         uploadUrlMap[deviceUrl.importKey] = (deviceUrl.key, deviceUrl.url)

      prog = 0
      for fileItem in fileItems:
         ovfDevId = fileItem.deviceId
         if ovfDevId not in uploadUrlMap:
            raise ValueError('OVF ID is not in upload URL Map')
         srcDiskPath = os.path.join(baseDir, fileItem.path)
         key, url = uploadUrlMap[ovfDevId]
         prog = self.__downloadFile(srcDiskPath, url, fileItem.create,
                                    lease, prevLeasePct=prog)
      lease.Progress(100)

   def __downloadFile(self, filePath, url, create, lease, prevLeasePct=0):
      '''
      Download and upload a single file to the NFC lease URL.
      @param filePath: File path, in this case is a URL.
      @param url: Device ID URL to upload contents of filePath to.
      @param create: File item is to be created or not.
      @param lease: NFC Lease object.
      @param prevLeasePct: previous lease progress percentage.
      '''

      o = urllib.parse.urlparse(url)
      protocol = o.scheme
      hostAndPort = o.netloc
      requestString = o.path

      forceHTTP = False
      # If the server returns * as the host name/ip we should use the
      # ip we used to connect
      if '*' == hostAndPort:
         if self.__host.name and self.__host.name != "localhost.eng.vmware.com":
            hostAndPort = hostAndPort.replace("*", self.__host.name)
            forceHTTP = gethostname() in hostAndPort
         else:
            hostAndPort = hostAndPort.replace("*", 'localhost')
            forceHTTP = True

      # For local host, no need to verify SSL certificates
      # Python3 throws ssl.CertificateError otherwise
      context = ssl._create_unverified_context()

      if protocol == 'https' and not forceHTTP:
         httpConn = http.client.HTTPSConnection(hostAndPort, context=context)
      else:
         httpConn = http.client.HTTPConnection(hostAndPort)

      try:
         request = urllib.request.Request(filePath)
         rsp = urllib.request.urlopen(request, context=context)

         prog = prevLeasePct
         with contextlib.closing(rsp) as fileObj:
            fileSize = int(fileObj.headers.get('Content-Length'))
            reqType = "PUT" if create else "POST"
            httpConn.putrequest(reqType, requestString)
            httpConn.putheader("Content-Length", fileSize)
            httpConn.endheaders()

            counter = 0
            chunkSize = 100 * 1024

            while True:
               data = fileObj.read(chunkSize)
               if len(data) == 0:
                  break
               httpConn.send(data)
               counter += 1

               if counter % 1000 == 0:
                  prog = min(99, prevLeasePct +
                             int((counter * chunkSize * 100) / fileSize))
                  lease.Progress(prog)
         state = lease.GetState()
         if state == Vim.HttpNfcLease.State.error:
            raise lease.error
         elif state not in [Vim.HttpNfcLease.State.ready,
                            Vim.HttpNfcLease.State.done]:
            raise RuntimeError('upload aborted, lease state=%s' % state)

         response = httpConn.getresponse()
         response.read()
      finally:
         httpConn.close()
      return prog

   def getMOB(self):
      """
      @return The managed object of this VM
      """

      return self.__mob

   def __getattr__(self, attrib):
      '''
      Aggregation of Virtual machine object with OVF object, such that calls to
      Virtual Machine objects can be made directly after deployment.
      This essentially makes OVF object a Virtual Machine object as well.

      @return Attribute of Virtual Machine managed object
      '''
      if self.__mob is None:
         raise ValueError('Managed Object unavailable')
      return getattr(self.__mob, attrib)
