# Copyright 2015-2018 VMware, Inc.
# All rights reserved. -- VMware Confidential

"""Ping utilities

Library to verify connectivity to a remote host.
"""
import socket
import subprocess

from borautils.exception import DebugInfoMixIn

class PingError(DebugInfoMixIn, ConnectionError):
   """Host or service is not responding.
   """
   _errType = "Failed to establish network connection"

   def __init__(self, errStr, host, port=None):
      super().__init__(errStr)
      self.addDebugInfo('host', host)
      if port is not None:
         self.addDebugInfo('port', port)

class NoRouteToHostError(PingError):
   """Host address is invalid, or DNS lookup failed.
   """

   def __init__(self, host):
      errStr = "no route to host (invalid host address, or DNS lookup failure)"
      super().__init__(errStr, host)

class ServiceUnreachableError(PingError):
   """Host is up, but service is not responding.
   """

   def __init__(self, host, port=None):
      errStr = "host is up, but service is not responding"
      super().__init__(errStr, host, port=port)

class HostUnreachableError(PingError):
   """Host is down or not responding.
   """

   def __init__(self, host):
      errStr = "host is unreachable"
      super().__init__(errStr, host)


def ping(host):
   """Issue a standard ICMP ping.

   Use the system's ping program to send an ICMP 'echo' packet. The ping command
   cannot be written in python, because sending an ICMP packet requires the
   creation of a raw socket, which only root can do (ping program is setuid'ed
   as root).
   """
   p = subprocess.Popen(['ping', '-c', '1', host],
                        stdout=subprocess.PIPE, stderr=subprocess.PIPE)
   p.communicate()
   if p.returncode == 0:
      return True
   else:
      raise HostUnreachableError(host)

def tcpPing(host, port, timeout=1):
   """Ping a TCP @host:@port.

   Try to connect a TCP socket to @host:@port, and return True if the connection
   succeeds, or raise a ConnectionError exception if the connection fails.
   """
   try:
      sock = socket.create_connection((host, port), timeout=timeout)
   except (socket.gaierror, socket.herror):
      raise NoRouteToHostError(host)
   except (socket.timeout, ConnectionError):
      raise ServiceUnreachableError(host, port=port)
   else:
      sock.shutdown(socket.SHUT_RDWR)
      sock.close()
      return True
