# Copyright 2018 VMware, Inc.  All rights reserved. -- VMware Confidential

"""Functionality evolving around the handling of TCP ports."""

from errno import (
   EADDRINUSE,
)
from functools import (
   partial,
   wraps,
)
from random import (
   randint,
)
from socket import (
   AF_INET,
   error as SocketError,
   SOCK_STREAM,
   socket,
)


FIRST_UNPRIVILEGED_PORT = 1025
LAST_UNPRIVILEGED_PORT = 65535


def isPortFree(port):
   """Check whether a given port is free."""
   with socket(AF_INET, SOCK_STREAM) as s:
      try:
         s.bind(("localhost", port))
         return True
      except SocketError as e:
         if e.errno != EADDRINUSE:
            raise
         return False


def noDuplicateResults(f):
   """A decorator making sure that we never return a result twice."""
   used = set()

   @wraps(f)
   def wrapper(used, *args, **kwargs):
      for _ in range(1000):
         value = f(*args, **kwargs)
         if value not in used:
            used |= {value}
            return value
      else:
         raise RuntimeError("all possible results exhausted?")

   return partial(wrapper, used)


@noDuplicateResults
def findRandomFreePort(start=FIRST_UNPRIVILEGED_PORT,
                       end=LAST_UNPRIVILEGED_PORT):
   """Find a free port to use in the range [start,end].

      This function is stateful in that it remembers returned free ports
      and will not return them again. That also means that there is the
      potential to run out of free ports. This case will be flagged by a
      RuntimeError.
      Use this function with caution: There is the potential for a race
      condition where a port is free at the point the function returns
      but it may have been reserved before the intended service could be
      bound to it.
   """
   while True:
      port = randint(start, end)
      if isPortFree(port):
         break

   return port
