# Copyright 2018 VMware, Inc.  All rights reserved. -- VMware Confidential

"""Functionality evolving around the spawning of processes.

This module provides advanced primitives on top of Python's subprocess
module.
"""
from subprocess import (
   CalledProcessError,
   Popen,
   TimeoutExpired,
)


def expectRunning(process):
   """Check whether a process is still running.

      This method can be used to check whether a process represented by
      a Popen object is still running. The assumption is that it is
      running and a `CalledProcessError` will be raised if that is not
      the case.
   """
   rc = process.poll()
   if rc is not None:
      raise CalledProcessError(rc, " ".join(process.args))

   return True


class DaemonPopen(Popen):
   """Popen wrapper to start and control long-live programs.
   """

   def terminateAndCleanup(self, forceKill=True, timeout=None):
      """Terminate this process and cleanup associated resources.

      Send a SIGTERM signal and wait until the process exits, or 'timeout'
      expires. The 'forceKill' parameter controls whether to kill the process
      with SIGKILL in case SIGTERM timed out.
      """
      DEFAULT_TERMINATE_TIMEOUT = 30

      if timeout is None:
         timeout = DEFAULT_TERMINATE_TIMEOUT

      try:
         self.terminate()
         try:
            self.wait(timeout=timeout)
         except TimeoutExpired:
            if forceKill:
               self.kill()
               self.wait()
            else:
               raise
      except ProcessLookupError:
         pass

   def __exit__(self, exc_type, exc_value, traceback):
      """Exit method for context manager syntax; kills this process.
      """
      self.terminateAndCleanup(forceKill=True)
      super().__exit__(exc_type, exc_value, traceback)
