# Copyright 2019 VMware, Inc.
# All rights reserved.  -- VMware Confidential

"""Library to handle .pyz executable archives.
"""
import os
import sys
from tempfile import TemporaryDirectory
from zipfile import BadZipFile, ZipFile


class _Pyz(ZipFile):
   """Singleton class to self-extract members from the current .pyz executable.
   """
   _instance = None
   _tmpDir = None

   def __new__(cls, *args, **kwargs):
      """Singleton magic.

      Create a unique instance of the _Pyz class, and initialize this instance
      with a dedicated temporary directory in which members of the .pyz archive
      can be extracted. The temporary directory (and its contents) are
      automatically garbaged collected upon exit.
      """
      if not cls._instance:
         cls._tmpDir = TemporaryDirectory(prefix='_PyzTmpDir')
         cls._instance = super().__new__(cls, *args, **kwargs)
      return cls._instance

   def __init__(self):
      try:
         super().__init__(sys.argv[0], mode='r')
      except BadZipFile:
         raise RuntimeError("%s: not a .pyz archive" % path)

   def extract(self, member, path=None, mode=None, **kwargs):
      """Extract a member from this .pyz archive.
      """
      if os.path.isabs(member):
         raise ValueError("%s: path to .pyz member must be relative" % member)

      if path is None:
         path = self._tmpDir.name

      try:
         extractedPath = super().extract(member, path=path, **kwargs)
      except KeyError:
         raise FileNotFoundError("%s: no such member in .pyz archive %s" %
                                 (member, self.filename))

      if mode is not None:
         os.chmod(extractedPath, mode)

      return extractedPath


def pyzExtract(member, mode=None):
   """Extract a member or the current .pyz archive.
   """
   return _Pyz().extract(member, mode=mode)
