# Copyright 2018 VMware, Inc. All rights reserved. -- VMware Confidential

"""REpresentational State Transfer (REST) client.

In this first version, the client assumes a JSON-backed REST API because it is
the most common data format. But this can be extended to support other formats.
"""
from base64 import b64encode
from http import HTTPStatus
from http.client import HTTPConnection, HTTPSConnection
from json import dumps as jsonDump, loads as jsonLoad
from ssl import CERT_NONE, CERT_REQUIRED, PROTOCOL_TLS, SSLContext
from urllib.parse import urlparse

from borautils.mime import MimeContentType

def createSslContext(certificate=None, allowUnsecure=False):
   """Create an SSL context that can be used to establish an HTTPS connection.
   """
   context = SSLContext(protocol=PROTOCOL_TLS)
   # The hostname will not match. Avoid checking for it.
   context.check_hostname = False
   context.verify_mode = CERT_NONE if allowUnsecure else CERT_REQUIRED
   context.load_default_certs()

   if certificate is not None:
      context.load_verify_locations(certificate)

   return context

def _toJson(obj):
   """Encode a python object to json.
   """
   if isinstance(obj, str):
      return obj
   else:
      return jsonDump(obj)

def authString(username, password):
   """Form an authorization string from a username and password.
   """
   return b"Basic " + b64encode(b":".join([username.encode(),
                                           password.encode()]))


class RestClient:
   """A client useful for connecting to a REST service.
   """

   def __init__(self, uri, port=None, username=None, password=None,
                certificate=None, allowUnsecure=False, **kwargs):
      """Initialize a new REST client object.
      """
      # This hidden argument allows for overwriting the security check.
      # It is meant to be used in tests only and allows for sending
      # credentials over non-encrypted connections.
      noSecurityCheck = kwargs.get("_noSecurityCheck", False)
      if (not noSecurityCheck and allowUnsecure and
          (username is not None or password is not None)):
         raise RuntimeError("Attempting to use username/password "
                            "authorization over unsecure connection.")

      if (username is None) != (password is None):
         raise RuntimeError("Username requires password and vice versa.")

      scheme = urlparse(uri).scheme

      # HTTP(S)Connection expects the scheme part of the url to be stripped off.
      host = uri[len(scheme + '://'):]

      connArgs = {'port': port}
      if 'timeout' in kwargs:
         connArgs['timeout'] = kwargs['timeout']

      if scheme.lower() == 'https':
         context = createSslContext(certificate=certificate,
                                    allowUnsecure=allowUnsecure)
         connArgs['context'] = context
         self._connection = HTTPSConnection(host, **connArgs)
      else:
         self._connection = HTTPConnection(host, **connArgs)

      if username is not None:
         self._authorization = authString(username, password)
      else:
         self._authorization = None

   def close(self):
      """Terminate the connection to the REST service.
      """
      self._connection.close()
      self._connection = None

   def __enter__(self):
      """Entry method for context manager syntax.
      """
      return self

   def __exit__(self, exc_type, exc_value, traceback):
      """Exit method for context manager syntax.
      """
      self.close()

   def _query(self, action, path, *args, headers=None, **kwargs):
      """Make an HTTP request through the client.
      """
      if headers is None:
         headers = dict()

      headers.setdefault("Accept", "application/json")

      if self._authorization is not None:
         headers.setdefault("Authorization", self._authorization)

      self._connection.request(action, path, *args, headers=headers, **kwargs)
      return self._connection.getresponse()

   def get(self, path):
      """Make a GET request to the given resource.
      """
      return self._query("GET", path)

   def put(self, path, data):
      """Make a PUT request to the given resource.
      """
      headers = {"content-type": "application/json"}
      return self._query("PUT", path, headers=headers, body=_toJson(data))

   def post(self, path, data):
      """Make a POST request to the given resource.
      """
      headers = {"content-type": "application/json"}
      return self._query("POST", path, headers=headers, body=_toJson(data))

   def patch(self, path, data):
      """Make a PATCH request to the given resource.
      """
      headers = {"content-type": "application/json"}
      return self._query("PATCH", path, headers=headers, body=_toJson(data))

   def delete(self, path):
      """Make a DELETE request to the given resource.
      """
      return self._query("DELETE", path)

   @staticmethod
   def readBody(response):
      """Retrieve the body of a REST response.

      Resources are decoded using the response header MIME type, and returned as
      python objects.
      """
      data = response.read()
      header = MimeContentType(response.getheader('Content-Type', ""))

      if (header.contentType, header.contentSubtype) == ('application', 'json'):
         if not data:
            return dict()

         try:
            charset = header.charset
         except AttributeError:
            charset = 'utf-8'

         body = data.decode(encoding=charset)
         return jsonLoad(body)
      else:
         return data
