# Copyright 2014-2016,2018,2020 VMware, Inc.
# All rights reserved. -- VMware Confidential

"""SSH library

This module provides a set of functions to execute remote commands over SSH.
Implementation is based on Paramiko.
"""
import io
import logging
import os
import socket
import subprocess
import sys
import time
import site

from borautils.infra.toolchain import loadTool

try:
   from select import PIPE_BUF
except ImportError:
   PIPE_BUF = 512
from datetime import datetime, timedelta
from borautils.exception import DebugInfoMixIn

MAX_CONNECTIONS_PER_ESX_HOST = 10

class SshConnectionError(DebugInfoMixIn, Exception):
   """SSH connection failure
   """
   _errType = "SSH connection error"
   def __init__(self, errStr, stderr):
      super(SshConnectionError, self).__init__(errStr)
      self.addDebugInfo("stderr", stderr)

class SshCmdTimeoutError(DebugInfoMixIn, Exception):
   """SSH command timeout
   """
   _errType = "SSH command timeout"

class SshCmdGenericError(DebugInfoMixIn, Exception):
   """SSH command generic failure
   """
   _errType = "SSH command error"
   def __init__(self, errStr, cmd, exitCode, stderr, stdout):
      super(SshCmdGenericError, self).__init__(errStr)
      self.addDebugInfo("command", cmd)
      self.addDebugInfo("exit code", exitCode)
      self.addDebugInfo("stderr", stderr)
      self.addDebugInfo("stdout", stdout)

class SshCalledProcessError(subprocess.CalledProcessError):
   """SSH process error.

   It enhances CalledProcessError by adding the hostname/IP address to
   the textual representation.
   """
   def __init__(self, host, *args, **kwargs):
      """Initialize the object."""
      super(SshCalledProcessError, self).__init__(*args, **kwargs)
      self.host = host

   def __str__(self):
      """Convert the object into a string."""
      s = super(SshCalledProcessError, self).__str__()
      return "%s: %s" % (self.host, s)

def getPythonPlatform():
   """Return the bitness and os type of the current python interpreter.
   """
   osType = sys.platform
   bits = "64" if sys.maxsize > 2 ** 32 else "32"
   return (osType, bits)

def loadToolParamiko():
   """Load paramiko from the toolchain
   """
   if sys.version_info[0] >= 3:
      PARAMIKO_PATH = 'noarch/paramiko-1.16.0-1/lib/python3.5/site-packages'
      ECDSA_PATH = 'noarch/ecdsa-0.13/lib/python3.3/site-packages'

      osType, bits = getPythonPlatform()
      if 'linux' in osType and bits == '64':
         arch = 'lin%s' % bits
      else:
         raise OSError('pyCrypto not available for %s-%s' % (osType, bits))

      pyCryptoPath = "%s/pycrypto-2.6.1/lib/python3.5/site-packages" % arch
   else:
      PARAMIKO_PATH = 'noarch/paramiko-1.16.0/lib/python2.7/site-packages'
      ECDSA_PATH = 'noarch/ecdsa-0.11/lib/python2.6/site-packages'

      osType, bits = getPythonPlatform()
      if osType == 'darwin':
         arch = 'mac32'
      elif 'linux' in osType:
         arch = 'lin%s' % bits
      else:
         raise OSError('pyCrypto not available for %s-%s' % (osType, bits))

      pyCryptoPath = '%s/pycrypto-2.6/lib/python2.7/site-packages' % arch

   for path in [PARAMIKO_PATH, pyCryptoPath, ECDSA_PATH]:
      sys.path.append(loadTool(path))

# For ESX, paramiko, ecdsa and pyCrypto are provided by esxcore-dev
def isInstalled(*args):
   p = site.getsitepackages()[0]
   return all([os.path.isdir(os.path.join(p, lib)) for lib in args])

if not isInstalled('paramiko', 'ecdsa', 'Crypto'):
   loadToolParamiko()
import paramiko


class _SshIn(io.IOBase):
   """Class to encapsulate the <stdin> stream of a paramiko channel into a
   python file-like object.
   """

   def __init__(self, channel):
      super(_SshIn, self).__init__()
      self._channel = channel

   def write(self, data):
      return self._channel.send(data)

   def close(self):
      self._channel.shutdown_write()

class _SshOut(io.IOBase):
   """Class to encapsulate the <stdout> stream of a paramiko channel into a
   python file-like object.
   """
   def __init__(self, channel):
      super(_SshOut, self).__init__()
      self._channel = channel
      self._recv = channel.recv

   def readall(self):
      """Read and return all the bytes from the stream until EOF.
      """
      out = bytes()
      while True:
         buf = self._recv(PIPE_BUF)
         if len(buf) == 0:
            break
         out += buf
      return out

   def read(self, size=-1):
      """Read up to size bytes from this ssh channel's <stdout>.

      As a convenience, if size is unspecified or -1, readall() is called. Fewer
      than size bytes may be returned. If 0 bytes are returned, and size was not
      0, this indicates end of file. If the object is in non-blocking mode and
      no bytes are available, None is returned.
      """
      if size == -1:
         return self.readall()
      return self._recv(size)

class _SshErr(_SshOut):
   """Class to encapsulate the <stderr> stream of a paramiko channel into a
   python file-like object.
   """

   def __init__(self, channel):
      super(_SshErr, self).__init__(channel)
      self._recv = self._channel.recv_stderr


class _SshPopen(object):
   """Class to encapsulate an ssh command into a Popen-like object.
   """

   def __init__(self, ssh, args, socketTimeout=None, stdin=None,
                stdout=None, stderr=None):
      """Execute a remote process in a new ssh channel.
      """
      self.args = args
      self.returncode = None
      if socketTimeout is None:
         socketTimeout = 0

      self._channel = ssh.openSession(socketTimeout=socketTimeout)

      if stdin == subprocess.PIPE:
         stdin = _SshIn(self._channel)
      if stdout == subprocess.PIPE:
         stdout = _SshOut(self._channel)
      if stderr == subprocess.PIPE:
         stderr = _SshErr(self._channel)

      self.stdin = stdin
      self.stdout = stdout
      self.stderr = stderr

      if isinstance(args, (list, tuple)):
         args = ' '.join(args)

      # exec_command is running the command in a shell, so use the shell
      # builtin exec to force the shell to execve the command.
      args = 'exec %s' % args
      self._channel.exec_command(args)

   def poll(self):
      """Check if remote process has terminated.

      Set and return returncode attribute. Warning: Note that some bad ssh
      servers may not return an exit status in some cases. If no exit status is
      provided by the server, returncode is set to -1.
      """
      if self._channel.exit_status_ready():
         self.returncode = self._channel.recv_exit_status()
         return self.returncode
      return None

   def communicate(self, input=None, timeout=None):
      """Interact with remote process: Send data to stdin. Read data from stdout
      and stderr, until end-of-file is reached. Wait for process to terminate.
      The optional input argument should be data to be sent to the child
      process, or None, if no data should be sent to the child. The type of
      input must be bytes.
      """
      if input is not None:
         self._channel.sendall(input)
         self.stdin.close()

      out = bytes()
      err = bytes()

      if timeout is None:
         timeout = 0

      deadline = time.time() + timeout if timeout > 0 else 0

      try:
         while self.poll() is None:
            if deadline != 0 and time.time() > deadline:
               raise SshCmdTimeoutError("Ssh command timeout")

            while self._channel.recv_ready():
               out += self._channel.recv(PIPE_BUF)
            while self._channel.recv_stderr_ready():
               err += self._channel.recv_stderr(PIPE_BUF)

            time.sleep(.1)

         # Make sure to flush stdout and stderr
         while self._channel.recv_ready():
            out += self._channel.recv(PIPE_BUF)
         while self._channel.recv_stderr_ready():
            err += self._channel.recv_stderr(PIPE_BUF)

      finally:
         self._channel.close()

      return out, err


class SSH(object):
   def __init__(self, host, username=None, password=None, key=None, keepalive=0,
                timeout=0, retryInterval=3, logger=None):
      """Connect to a remote SSH server.

      Initialize an SSH connection to a remote host.

      The 'keepalive' parameter can be set to control the period (in seconds)
      after which an out-of-band packet is sent to the SSH server. This allows
      to keep the session opened while there is no traffic, and very
      conveniently, it also provides a way to detect unresponsive connections
      (paramiko raises an exception if the keepalive packet is not ack'ed).

      The 'timeout' parameter specifies the amount of time (in seconds) to wait
      until the host is declared unreachable. When set, this parameter allows
      the client to repeatedly try to connect to the host until the timeout is
      reached. This provides a convenient way to connect to a host while it is
      booting up, and wait until SSH is started.
      """
      self._client = paramiko.SSHClient()

      if logger is None:
         logger = logging.getLogger("_sshNullLogger")
         logger.propagate = False

      self._client.set_log_channel(logger.name)

      self._client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

      if timeout > 0:
         expiration = datetime.now() + timedelta(seconds=timeout)
         interval = timedelta(seconds=retryInterval)

      error = Exception("Unknown SSH error")
      connected = False

      # Try to use the user's SSH config.
      config = paramiko.SSHConfig()
      sshConfigPath = os.path.expanduser(os.path.join('~', '.ssh', 'config'))
      try:
         with open(sshConfigPath, 'r') as f:
            config.parse(f)
            self._host = config.lookup(host).get('hostname', host)
      except Exception:
         self._host = host

      while True:
         try:
            self._client.connect(self._host, username=username,
                                 password=password,
                                 key_filename=key, look_for_keys=False,
                                 banner_timeout=10000)
         except (paramiko.BadHostKeyException,
                 paramiko.AuthenticationException) as e:
            # Authentication exception, no point to retry.
            error = e
            break
         except (paramiko.SSHException, socket.error) as e:
            # Connection error, we might want to retry.
            error = e
         else:
            connected = True
            break

         assert not connected
         if timeout == 0:
            break

         now = datetime.now()
         if now > expiration or ((expiration - now) < interval):
            break
         time.sleep(retryInterval)

      if connected:
         self._transport = self._client.get_transport()
         self._transport.set_log_channel(logger.name)
         if keepalive:
            self._transport.set_keepalive(keepalive)
      else:
         self._client.close()
         raise SshConnectionError("ssh connection failed", error)

   def __str__(self):
      """Return a human readable string to identify this object.
      """
      return "ssh://%s" % self._host

   def close(self):
      """Close this SSH connection.
      """
      self._client.close()

   def __enter__(self):
      """To enable usage of this class in conjunction with the 'with' statement.
      """
      return self

   def __exit__(self, unused_exc_type, unused_exc_value, unused_traceback):
      """__enter__() counterpart.
      """
      self.close()

   def openSession(self, blocking=True, socketTimeout=0, interactive=False):
      """Open a new SSH session.

      This function opens a new SSH session, and returns a Paramiko Channel to
      perform low-level operations on that session. The session can be made
      non-blocking by setting the 'blocking' parameter to 'False'.

      When interactive is set to True, a pseudo-terminal is created by the
      server, and connected to the returned channel.
      """
      timeout = socketTimeout if socketTimeout else 3600 # paramiko default 1hr
      channel = self._transport.open_session(timeout=timeout)
      if not blocking:
         channel.setblocking(0)
      elif socketTimeout > 0:
         channel.settimeout(socketTimeout)
      if interactive:
         channel.get_pty()
         channel.invoke_shell()
      return channel

   def Popen(self, *args, **kwargs):
      """Run an ssh command in a subprocess.Popen() fashion.
      """
      return _SshPopen(self, *args, **kwargs)

   def run(self, cmd, socketTimeout=0, check=False):
      """Execute a command over SSH.

      Run the given command and blocks until the command returns. If
      'check' is True the exit code of the executed program will be
      checked and a SshCalledProcessError will be raised should it
      indicate a command failure.
      """
      result = self._runInt(cmd, socketTimeout, blocking=True)
      if check:
         status, out, err = result
         if status != 0:
            raise SshCalledProcessError(self._host, status, cmd, output=out,
                                        stderr=err)

      return result

   def runAsync(self, cmd, socketTimeout=0):
      """Execute a command over SSH.

      Run the given command asynchronously and return the channel.
      """
      return self._runInt(cmd, socketTimeout, blocking=False)

   def _runInt(self, cmd, socketTimeout=0, blocking=None):
      """Execute a command over SSH.
      @param blocking: Block until the command returns
      """
      channel = self.openSession(blocking=blocking, socketTimeout=socketTimeout)

      try:
         channel.exec_command(cmd)
      except Exception:
         channel.close()
         raise

      if not blocking:
         return channel

      return self.getCmdStatus(channel, socketTimeout)

   def getCmdStatus(self, channel, timeoutInSec=0):
      """
      returns a 3-element tuple containing the command return status, stdout as
      a string, and stderr as a string. Channel will be closed
      @param channel: the channel which ssh command resides on
      @param timeoutInSec: maximum seconds to wait for cmd status
      """
      out = []
      err = []

      timeout = time.time() + timeoutInSec if timeoutInSec > 0 else 0

      try:
         while not channel.exit_status_ready():
            if timeout != 0 and time.time() > timeout:
               raise SshCmdTimeoutError("Ssh command timeout")

            while channel.recv_ready():
               out += channel.recv(PIPE_BUF)
            while channel.recv_stderr_ready():
               err += channel.recv_stderr(PIPE_BUF)
            time.sleep(.1)

         # Make sure to flush stdout and stderr
         while channel.recv_ready():
            out += channel.recv(PIPE_BUF)
            time.sleep(.1)
         while channel.recv_stderr_ready():
            err += channel.recv_stderr(PIPE_BUF)
            time.sleep(.1)

      except Exception:
         channel.close()
         raise

      status = channel.recv_exit_status()
      channel.close()

      out = "".join(["%c" % c for c in out])
      err = "".join(["%c" % c for c in err])
      return status, out, err

   def sftpGet(self, remotePath, localPath, removeSource=False,
               socketTimeout=0):
      """Download a file from a remote host over SFTP.

      If removeSource is 'True', this function tries to delete the remote file
      after the transfer has completed.
      """
      sftp = self._client.open_sftp()

      if socketTimeout > 0:
         channel = sftp.get_channel()
         channel.settimeout(socketTimeout)

      sftp.get(remotePath, localPath)

      if removeSource:
         try:
            sftp.unlink(remotePath)
         except IOError:
            pass

      sftp.close()

   def sftpPut(self, localPath, remotePath, socketTimeout=0, overwrite=False):
      """Upload a local file to a remote host over SFTP.

      Setting overwrite to 'True' will overwrite the remote file if it already
      exists.
      """
      sftp = self._client.open_sftp()

      if socketTimeout > 0:
         channel = sftp.get_channel()
         channel.settimeout(socketTimeout)

      if not overwrite:
         try:
            sftp.stat(remotePath)
         except IOError:
            pass
         else:
            raise IOError('%s: file already exists' % remotePath)

      sftp.put(localPath, remotePath)

      sftp.close()

   def isActive(self):
      """ssh connection is active or not?
      """
      transport = self._client.get_transport()
      return transport is not None and transport.is_active()

def sftpGet(host, remotePath, localPath, removeSource=False,
            username=None, password=None, key=None, socketTimeout=30):
   """Download a file from a remote host over SFTP.

   If removeSource is 'True', this function tries to delete the remote file
   after the transfer has completed.
   """
   with SSH(host, username=username, password=password, key=key) as ssh:
      ssh.sftpGet(remotePath, localPath, removeSource=removeSource,
                  socketTimeout=socketTimeout)

def sftpPut(host, localPath, remotePath, username=None, password=None,
            key=None, overwrite=False, socketTimeout=30):
   """Upload a local file to a remote host over SFTP.

   Setting overwrite to 'True' will overwrite the remote file if it already
   exists.
   """
   with SSH(host, username=username, password=password, key=key) as ssh:
      ssh.sftpPut(localPath, remotePath, overwrite=overwrite,
                  socketTimeout=socketTimeout)

def getPublicKey(keyFile):
   """Return the public key associated with the keyFile.
   """
   key = paramiko.rsakey.RSAKey(filename=keyFile)
   return "%s %s" % (key.get_name(), key.get_base64())
