"""Library to extract support bundles.
"""
import os
import re

from borautils.archive import openArchive
from borautils.busybox import READ_CHUNK_SIZE

FRAG_SUFFIX = '.FRAG-'
FRAG_SUFFIX_RE = re.compile(r"\.FRAG-(\d{5})$")

class SupportBundle(object):
   """Class to extract files from a support bundle.
   """
   _MANDATORY_FILES = ("action.log", "error.log", "errors-ignored.log",
                       "README")

   def __init__(self, filePath):
      self.path = filePath
      self._archive = openArchive(self.path)
      self._checkBundle()

   def _checkBundle(self):
      """Check whether this archive is a valid support bundle.

      A valid support bundle has the following properties:
      - all archive members must be contained in one single root directory
      - a specific set of files must be present at know locations within
        the archive.
      """
      members = self._archive.list()
      if not members:
         raise ValueError("%s: invalid support bundle (archive empty)" %
                          self.path)

      self.rootDir = members[0].split('/')[0]
      for path in members[1:]:
         if path.split('/')[0] != self.rootDir:
            raise ValueError("%s: invalid support bundle (multi-root archive)" %
                             self.path)

      for path in self._MANDATORY_FILES:
         if os.path.join(self.rootDir, path) not in members:
            raise ValueError("%s: invalid support bundle (missing %s file)" %
                             (self.path, path))

   def close(self):
      """Close this support bundle.
      """
      self._archive.close()

   def __enter__(self):
      """Context guard entry.
      """
      return self

   def __exit__(self, excType, excVal, traceback):
      """Context guard exit.
      """
      self.close()

   def list(self, ignoreDirs=False):
      """List all files in this support bundle. Fragmented files are listed as
      one reconstructed entry (without the .FRAG-* extension).
      """
      members = set()
      for member in self._archive.list(ignoreDirs=ignoreDirs):
         members.add(FRAG_SUFFIX_RE.sub(r"", member))
      return list(members)

   def _extractFragments(self, destDir, name, members):
      """Helper function to extract a fragmented file.
      """
      fragIndexes = []
      for member in members:
         if member.startswith(name):
            m = re.fullmatch(FRAG_SUFFIX_RE, member[len(name):])
            if m is not None:
               fragIndexes.append(int(m.group(1)))

      if not fragIndexes:
         raise FileNotFoundError("%s: no such file in support bundle" % name)

      fragIndexes.sort()
      numFragments = fragIndexes[-1] + 1

      if len(fragIndexes) != numFragments:
         raise RuntimeError("%s: wrong number of fragments" % name)
      elif fragIndexes != list(range(numFragments)):
         raise RuntimeError("%s: missing fragments" % name)

      dest = os.path.join(destDir, name)
      destDir = os.path.dirname(dest)
      os.makedirs(destDir, exist_ok=True)

      with open(dest, 'wb') as fout:
         for i in range(numFragments):
            member = "%s%s%05u" % (name, FRAG_SUFFIX, i)
            self._archive.extract(destDir, members=[member])

            fragmentPath = os.path.join(destDir, member)
            try:
               with open(fragmentPath, 'rb') as fin:
                  while True:
                     data = fin.read(READ_CHUNK_SIZE)
                     if not data:
                        break
                     numWrittenBytes = fout.write(data)
                     assert numWrittenBytes == len(data)
            finally:
               if os.path.exists(fragmentPath):
                  os.unlink(fragmentPath)

   def extract(self, destDir, members=None):
      """Extract a file from this support bundle. Fragmented files are referred
      to as their reconstructed name (with no .FRAG-* extension), and
      reconstructed on the fly.
      """
      archiveMembers = self._archive.list(ignoreDirs=True)

      if members is None:
         members = archiveMembers

      for member in members:
         if member not in archiveMembers:
            self._extractFragments(destDir, member, archiveMembers)
         else:
            self._archive.extract(destDir, members=[member])


def isSupportBundle(filePath):
   """Return True if @filePath points to a valid support bundle.
   """
   try:
      with SupportBundle(filePath) as _:
         return True
   except ValueError as e:
      return False
