"""Telnet facilities

This module relies on telnetlib to perform basic telnet operations such as:
 - executing remote commands over telnet
 - reading serial logs from a remote serial port concentrator.
"""
import string
import telnetlib

from borautils.exception import DebugInfoMixIn

TELNET_DEFAULT_PORT = 23

TELNET_EOL = '\r\n'
UNIX_EOL = '\n'

class TelnetError(DebugInfoMixIn, Exception):
   """Generic telnet error.
   """
   _errType = "Unknown telnet error"

   def __init__(self, errStr, session):
      super().__init__(errStr)
      self.addDebugInfo('host', session)

class ConnectionDroppedError(TelnetError):
   """Other end has closed the telnet connection.
   """
   _errType = "Telnet connection closed by peer"

class TelnetTimeout(TelnetError):
   """Timeout expired while receiving data.
   """
   _errType = "Telnet timeout"

   def __init__(self, errStr, session, timeout):
      super().__init__(errStr, session)
      self.addDebugInfo('timeout', timeout)


class _TelnetSession(telnetlib.Telnet):
   """A generic Telnet session.
   """
   READ_TIMEOUT = 30

   def __init__(self, host, port=TELNET_DEFAULT_PORT):
      """Initialize the telnet socket, and connect to the given host.
      """
      self._host = host
      self._port = port
      self._outputBuffer = ""

      try:
         telnetlib.Telnet.__init__(self, self._host, self._port)
      except Exception:
         raise TelnetError('failed to initialize telnet session', self)

   def __str__(self):
      """Return a human readable string to identify this telnet session.
      """
      return "telnet://%s:%s" % (self._host, self._port)

   def flush(self):
      """Flush the internal output buffer.
      """
      data = self._outputBuffer[:]
      self._outputBuffer = ""
      return data


class TelnetShell(_TelnetSession):
   """Class to execute remote commands over telnet.
   """

   def __init__(self, host, port=TELNET_DEFAULT_PORT, username=None,
                password=None):
      """Open and authenticate a new telnet connection.
      """
      _TelnetSession.__init__(self, host, port=port)

      if username is not None and password is not None:
         try:
            self.run(username, prompt=":")
            self.run(password, prompt=":")
         except Exception:
            self.close()
            raise

   def run(self, s, prompt=None):
      """Send a remote command over this telnet shell.
      """
      formattedInput = s.encode('latin-1') + TELNET_EOL

      try:
         if prompt:
            self._outputBuffer += self.read_until(prompt, self.READ_TIMEOUT)
            if not self._outputBuffer.endswith(prompt):
               raise TelnetTimeout("timeout while waiting for telnet prompt",
                                   self, self.READ_TIMEOUT)
         self.write(formattedInput)
      except EOFError:
         # Bad password may cause the other end to close the telnet connection
         raise ConnectionDroppedError("telnet connection closed by peer", self)
      except TelnetTimeout:
         raise
      except Exception:
         raise TelnetError('unknown telnet error', self)


class TelnetLog(_TelnetSession):
   """Class to read (serial) logs over telnet.
   """

   def __init__(self, host, port=TELNET_DEFAULT_PORT, eol=UNIX_EOL):
      """Initialize the telnet socket, and connect to the given host.
      """
      _TelnetSession.__init__(self, host, port=port)
      self._eol = eol

   def readline(self, timeout=None, raw=False):
      """Read a line from the telnet socket.

      This function blocks until a full line is read, or timeout has expired.
      """
      if timeout is None:
         timeout = self.READ_TIMEOUT

      line = self.read_until(b'\n', timeout)
      line = line.decode('utf-8', errors='replace')

      if line != "":
         if not raw:
            line = line.replace(TELNET_EOL, self._eol).strip('\r')
            line = ''.join(('.', c)[c in string.printable] for c in line)

         # read_until() won't return a full line if no match is found (for
         # example if a line is truncated), or if a timeout occurs.
         self._outputBuffer += line
         if self._outputBuffer.endswith(self._eol):
            return self.flush()

      return ""

   def readlines(self, timeout=_TelnetSession.READ_TIMEOUT):
      """Generator to iterate over log lines.
      """
      while True:
         try:
            line = self.readline(timeout=timeout)
         except EOFError:
            line = self.flush()
            if line != "":
               yield line
            break
         if line != "":
            yield line


#
# Code below is for debugging purposes...
#
if __name__ == "__main__":
   from argparse import ArgumentParser
   import time

   parser = ArgumentParser()
   parser.add_argument(type=str, metavar='HOST', dest='host',
                       help="Telnet host")
   parser.add_argument(type=int, metavar='PORT', dest='port',
                       help="Telnet port")
   parser.add_argument('-i', action="store_true", dest='interactive',
                       help="Enable interactive mode")
   parser.add_argument('-u', type=str, metavar='USER', dest='username',
                       help="Session username")
   parser.add_argument('-p', type=str, metavar='PASSWORD', dest='password',
                       help="Session password")
   args = parser.parse_args()

   if args.interactive:
      session = TelnetShell(args.host, port=args.port,
                            username=args.username, password=args.password)
      print(session.flush(), end='')
      time.sleep(3)
      print(session.read_very_eager(), end='')
      session.close()

   else:
      log = TelnetLog(args.host, port=args.port)
      for line in log.readlines():
         print(line, end='')
      log.close()
