"""TTY Terminal utilities

ANSI/ECMA control sequences are specified in the "Control Functions for Coded
Character Sets" spec, which is available at:

http://www.ecma-international.org/publications/files/ECMA-ST/Ecma-048.pdf
"""
import os
from select import select
import sys
import termios
import tty

# TTY color codes
BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(8)

# Valid XTerm types
TERM_XTERM = ['xterm', 'xterm-256color']
TERM_SCREEN = ['screen', 'screen-256color']

class ESC(str):
   """ANSI escape sequence.
   """
   def __new__(cls, value):
      return str.__new__(cls, "\033" + value)

class CSI(ESC):
   """ECMA-48 control sequence.
   """
   MAX_ESC_PARAMS = 16

   CUU = 'A'   # Cursor Up
   CUD = 'B'   # Cursor Down
   CUF = 'C'   # Cursor Forward
   CUB = 'D'   # Cursor Back
   CUP = 'H'   # Cursor Position
   SGR = 'm'   # Select Graphic Rendition
   ED  = 'J'   # Erase display

   # We couldn't find a reliable source/standard for the following keys but
   # we list them according to our empirical results. It might not work
   # depending on the terminal implementation.
   HOME   = '1~'
   END    = '4~'
   PGUP   = '5~'
   PGDOWN = '6~'

   def __new__(cls, cmd, *params):
      if params:
         if len(params) > 16:
            raise ValueError("too many escape sequence parameters (max=%u)" %
                             MAX_ESC_PARAMS)
         seq = ";".join(["%s" % p for p in params])
      else:
         seq = ""
      return ESC.__new__(cls, "[" + seq + cmd)

class DEC(CSI):
   """DEC Private Mode (DECSET/DECRST) sequence.
   """
   SET = 'h'    # Turn mode on
   RESET = 'l'  # Turn mode off

   AWM = 7      # Auto-Wrap Mode
   TCEM = 25    # Text Cursor Enable Mode

   def __new__(cls, mode, status):
      return CSI.__new__(cls, status, '?%u' % mode)

class SGR(CSI):
   """ECMA-48 Set Graphics Rendition sequence.
   """
   RESET = 0           # reset default rendition settings
   BOLD = 1            # bold or increased intensity
   FG_COLOR_BASE = 30  # foreground colors range from 30 to 37
   BG_COLOR_BASE = 40  # background colors range from 40 to 47
   FG_BRIGHT_BASE = 90 # bright foreground colors, 90 to 97

   def __new__(cls, *attrs):
      return CSI.__new__(cls, CSI.SGR, *attrs)

def ttyFormat(msg, fgColor=None, bgColor=None, bold=False, reset=False):
   """Format a string using TTY escape sequences.
   """
   args = list()

   if fgColor is not None:
      assert fgColor in range(8)
      args.append(SGR.FG_COLOR_BASE + fgColor)

   if bgColor is not None:
      assert bgColor in range(8)
      args.append(SGR.BG_COLOR_BASE + bgColor)

   if bold:
      args.append(SGR.BOLD)

   header = SGR(*args) if args else SGR(SGR.RESET)
   msg = str(msg)
   msg = header + msg
   if reset:
      msg += SGR(SGR.RESET)

   return msg


class TTY(object):
   """Class representing a TTY terminal.
   """

   def __init__(self, fd=None, updateTitle=False):
      if fd is None:
         self._stdin = sys.stdin.fileno()
         self._stdout = sys.stdout.fileno()
      else:
         self._stdin = fd
         self._stdout = fd

      self._isatty = os.isatty(self._stdin)
      if self._isatty:
         self._oldAttr = termios.tcgetattr(self._stdin)

      self._term = os.getenv('TERM')
      self._updateTitle = updateTitle

   def reset(self):
      """Restore the original terminal configuration.
      """
      if self._isatty:
         termios.tcsetattr(self._stdin, termios.TCSADRAIN, self._oldAttr)

   def __enter__(self):
      """Enter terminal context.
      """
      return self

   def __exit__(self, excType, excValue, traceback):
      """Exit terminal context, restoring all terminal original settings.
      """
      self.reset()

   def setcbreak(self):
      """Set terminal in cbreak mode.

      - Canonical (line-by-line) mode is disabled
      - Echo is disabled
      - Characters are read one at a time
      - Read is blocking (VTIME timeout is disabled).
      """
      assert self._isatty
      tty.setcbreak(self._stdin, when=termios.TCSANOW)

   def echo(self, msg):
      """Print a message to this terminal's <stdout>.
      """
      os.write(self._stdout, msg.encode())

   def setAutoWrapMode(self, on):
      """Toggle line Auto-Wrap Mode on/off.
      """
      self.echo(DEC(DEC.AWM, DEC.SET if on else DEC.RESET))

   def setEchoMode(self, on):
      """Enable/disable terminal echo mode.
      """
      attr = termios.tcgetattr(self._stdin)

      if on:
         attr[3] |= termios.ECHO | termios.ECHONL
      else:
         attr[3] &= ~(termios.ECHO | termios.ECHONL)

      termios.tcsetattr(self._stdin, termios.TCSAFLUSH, attr)

   def setTextCursorEnableMode(self, on):
      """Show/hide text cursor.
      """
      self.echo(DEC(DEC.TCEM, DEC.SET if on else DEC.RESET))

   def clearScreen(self):
      """Clear the screen completely and reset the cursor to 0,0
      """
      self.echo(CSI(CSI.ED, 2))
      self.echo(CSI(CSI.CUP))

   def setTitle(self, title):
      """ Set title of either XTerm or Screen window
      """
      if self._updateTitle:
         if self._term in TERM_XTERM:
            title = ESC(']0;') + title + '\a'
         elif self._term in TERM_SCREEN:
            title = ESC('k') + title + ESC('\\')
         else:
            raise Exception('Unsupported terminal type \'%s\'' % self._term)
         self.echo(title)

class CBreakTTY(TTY):
   """Class to interact with a terminal in cbreak mode.
   """

   def __init__(self, **kwargs):
      super().__init__(**kwargs)
      self.setcbreak()

   def getkey(self):
      """Poll the keyboard for key strokes.

      The function is non-blocking, and returns None is no key was pressed.
      """
      if self._isatty:
         i, _, _ = select([self._stdin], [], [], 0)
         if self._stdin in i:
            return os.read(self._stdin, 1)
      return None


#
# Code below is for debugging purposes...
#
if __name__ == "__main__":
   import time

   with CBreakTTY() as console:
      print(ttyFormat("Press <", fgColor=WHITE, bold=True) +
            ttyFormat("Q", fgColor=BLUE, bold=True) +
            ttyFormat("> or <", fgColor=WHITE, bold=True) +
            ttyFormat("CTRL-C", fgColor=BLUE, bold=True) +
            ttyFormat("> to quit...", fgColor=WHITE, bold=True, reset=True))

      while True:
         key = console.getkey()
         try:
            if key.lower() == b'q':
               break
         except AttributeError:
            pass
         print("Time: %s" % time.time())
         console.echo(CSI(CSI.CUU, 1))
         if key is None:
            time.sleep(.1)
