"""TestRun API library.

Library to abstract a testesxdeploy run.
"""
from contextlib import contextmanager
import os
import re
import sys
from tempfile import mkdtemp

from borautils.infra.buildapi import GobuildParserAction
from borautils.infra.dbc import DBC
from borautils.esx.build import EsxDbcBuild, EsxOfficialBuild, EsxSandboxBuild
from borautils.exception import DebugInfoMixIn

class TestRunError(DebugInfoMixIn, Exception):
   """TestRun base exception.
   """
   _errType = "TestRun error"

class TestRunEnvironmentError(TestRunError):
   """TestRun environment error.
   """
   _errType = "TestRun environment error"

   def __init__(self, errStr):
      super().__init__(errStr)

class TestRun(object):
   """Main class used in testesxdeploy, to be inherited and specialized.
   """
   # To ensure that names/id's can be used as filenames, for instance to create
   # a testrun-specific results directory.
   _filenameSafeRE = re.compile(r"[\w-]+")

   (PASS, FAIL, PSOD, TIMEOUT, INVALID) = list(range(5))

   def __init__(self, name, testrunId):
      try:
         self._name = self._filenameSafeRE.fullmatch(name).group(0)
      except Exception:
         raise ValueError("invalid name parameter for class %s(): '%s'"
                          % (self.__class__.__name__, name))

      try:
         if isinstance(testrunId, int):
            self._id = testrunId
         else:
            self._id = self._filenameSafeRE.fullmatch(testrunId).group(0)
      except Exception:
         raise ValueError("invalid testrunid parameter for class %s(): '%s'"
                          % (self.__class__.__name__, testrunId))

   def getResultsDir(self):
      """Return the results directory of this testrun.
      """
      return self._resultsDir

   def getEsxBuild(self, installer=False):
      """Return an ESXBuild object to be tested. This needs to be
      overridden in children classes.
      """
      return None

   def getId(self):
      """Return this testrun ID. To be overridden in children classes.
      """
      assert self._id is not None
      return self._id

   def _setupResultsDir(self, resultsDir=None):
      """Create the results directory.
      """
      if resultsDir is not None:
         self._resultsDir = resultsDir
      assert self._resultsDir is not None

      try:
         os.makedirs(self._resultsDir, exist_ok=True)
      except Exception:
         raise TestRunEnvironmentError("failed to create results directory: %s"
                                       % self._resultsDir)

   def reportTestResult(self, name, passed, runTime, bugReporthPath):
      """To be overridden by children classes.
      """
      pass

   def _uploadResults(self, result):
      """To be overridden by children classes.
      """
      pass

   @contextmanager
   def uploadTestResults(self, psodExc=None, timeoutExc=None, resultsDir=None):
      """Context manager to trap a program's exit paths and to convert exit
      codes or fatal exceptions into one of the following standard results:

      PASS:    program exited successfullly (exit code is 0),
      FAIL:    program exited with any non-zero exit code,
      INVALID: program was interrupted by a fatal exception.
      PSOD:    program was interrupted by a PSOD exception. An optional list of
               exception types may be passed to specify which types of
               exceptions should be treated as a PSOD rather than being reported
               as an INVALID testrun.
      TIMEOUT: pogram was interrupted by a TIMEOUT exception. An optional list
               of exception types may be passed to specify which types of
               exceptions should be treated as a TIMEOUT rather than being
               reported as an INVALID testrun.
      """
      try:
         self._setupResultsDir(resultsDir)
      except Exception:
         self._uploadResults(TestRun.INVALID)
         raise

      rc = os.EX_OK
      try:
         yield
      except Exception as e:
         if psodExc and isinstance(e, psodExc):
            self._uploadResults(TestRun.PSOD)
         elif timeoutExc and isinstance(e, timeoutExc):
            self._uploadResults(TestRun.TIMEOUT)
         else:
            self._uploadResults(TestRun.INVALID)
         rc = 1
      except SystemExit as e:
         rc = e.args[0]
         raise
      finally:
         if rc == os.EX_OK:
            self._uploadResults(TestRun.PASS)
         else:
            self._uploadResults(TestRun.FAIL)
            sys.exit(1)

   def __str__(self):
      """Return a human-readable string to identify this TestRun object.
      """
      return "%s-%s" % (self._name, self.getId())

class DevApp(DBC, TestRun):
   """A class to deploy an ESX build - and run ESX tests - from a DBC-like
   environment.
   """

   @classmethod
   def addArguments(cls, argsParser):
      """Register gobuild-specific CLI options.
      """
      argsParser.add_argument('--build', metavar='BUILDNUM', type=str,
                              dest='build', default=None,
                              action=GobuildParserAction,
                              help=("Gobuild build ID (e.g.: 'ob-1234567' "
                                    "or 'sb-1234567')"))

   def __init__(self, name, workingDir=None, build=None, **kwargs):
      self._workingDir = workingDir
      self._buildId = build

      try:
         DBC.__init__(self, self._workingDir)
      except ValueError:
         self._isDbc = False
         # XXX: we should pass a testrun ID rather than a build ID.
         testrunId = self._buildId.num
      else:
         self._isDbc = True
         testesxDir = os.path.join(self.rootDir, '.%s' % name)
         os.makedirs(testesxDir, exist_ok=True)

         # Create the results directory, and make sure its basename qualifies as
         # a unique testrun ID for this testrun.
         while True:
            resultsDir = mkdtemp(dir=testesxDir)
            testrunId = os.path.basename(resultsDir)
            if self._filenameSafeRE.fullmatch(testrunId) is not None:
               self._resultsDir = resultsDir
               break
            os.rmdir(resultsDir)

         # Results dir must be readable by all, and have write/execute
         # permissions for ESX to upload test results.
         os.chmod(self._resultsDir, 0o777)

      if not self._isDbc and self._buildId is None:
         raise EnvironmentError("No ESX build found. Please use --build to "
                                "specify a Gobuild build ID, or run from a DBC "
                                "bora tree")

      TestRun.__init__(self, name, testrunId)

   def getLocationName(self):
      """Return the location of this testrun. Since gobuilds are only

      - Gobuilds are always located in Palo Alto right now
      - XXX: Assume 'PA' for DBC (must be improved to support non-PA DBC's).
      """
      return 'PA'

   def getPxeDir(self):
      """Return the NFS path to the PXE directory for this testrun.
      """
      if self._isDbc:
         return (self.host, os.path.join(self.getResultsDir(), 'pxe'))
      return None

   def getPxeUrl(self):
      """Return an HTTP URL pointing to the PXE directory for this testrun.
      """
      if self._isDbc:
         return self.toUrl(os.path.join(self.getResultsDir(), 'pxe'))
      return None

   def getEsxBuild(self, installer=False):
      """Return an ESXBuild object containing info about this testrun's build.
      """
      if self._buildId is None:
         assert self._isDbc
         return EsxDbcBuild(self._workingDir)
      else:
         if self._buildId.type == 'sb':
            return EsxSandboxBuild(self._buildId.num, installer)
         else:
            assert self._buildId.type == 'ob'
            return EsxOfficialBuild(self._buildId.num, installer)
