# Copyright 2016-2018 VMware, Inc. All rights reserved. -- VMware Confidential

"""ThreadPool implementation based on python threads.
"""
from queue import Queue
from threading import Thread

from borautils.exception import DebugInfoMixIn


class WorkerThreadRunException(DebugInfoMixIn, Exception):
   """One or more worker threads failed
   """
   _errType = "Worker thread execution exception"
   def __init__(self, expMsg):
      super().__init__(self._errType)
      self.addDebugInfo('Exception(s)', expMsg)


class Worker(Thread):
   """Worker object used by ThreadPool. Represents a thread that will
   pick up work as it is available in the queue.
   """

   def __init__(self, workQueue, resultQueue, exceptionQueue):
      """Init function. Takes a queue as parameter.
      """
      super().__init__()
      self._workQueue = workQueue
      self._mustStop = False
      self._exceptionQueue = exceptionQueue
      self._resultQueue = resultQueue

   def stop(self):
      """Asks the worker to exit gracefully. Worker will exit after
      finishing it's current job.
      """
      self._mustStop = True

   def run(self):
      """Main worker loop. Wait for a job in the queue and execute it.
      """
      while not self._mustStop:
         func, args, kargs = self._workQueue.get(True)
         if self._mustStop:
            break

         try:
            ret = func(*args, **kargs)
         except Exception as e:
            self._exceptionQueue.put((func.__name__, args, kargs,
                                     type(e).__name__, str(e)))
         else:
            self._resultQueue.put(ret)
         finally:
            self._workQueue.task_done()


class ThreadPool(object):
   """ThreadPool class. It manages thread workers by giving them jobs to do.
   Implementation is essentially based on a queue.
   """

   def __init__(self, nbWorkers, queueSize=None):
      """Init function. User must specify the number of workers (i.e. threads).
      As an extra parameter user can specify the queue capacity, in other words
      the number of jobs that could wait in the queue.
      """
      if not queueSize:
         queueSize = nbWorkers * 2

      self._queue = Queue(queueSize)
      self._workers = []
      self._exceptionQueue = Queue()
      self._resultQueue = Queue()

      for _ in range(nbWorkers):
         self._workers.append(Worker(self._queue, self._resultQueue,
                                     self._exceptionQueue))

      for worker in self._workers:
         worker.start()

   def addJob(self, func, *args, **kargs):
      """Add job in the job queue. If the queue is full this function will
      block until one spot is freed in the queue.
      """
      self._queue.put((func, args, kargs), True)

   def waitForCompletion(self):
      """Wait for all jobs to complete. This function will make caller wait
      until the job queue is empty and all jobs are reported as done.
      """
      self._queue.join()

   def destroy(self, waitForCompletion):
      """Destroy the threadPool and all the workers. If waitForCompletion is
      True, will wait until all jobs are completed before destroying the pool.
      Otherwise jobs might be left in the queue.
      """
      if waitForCompletion:
         self.waitForCompletion()

      for worker in self._workers:
         worker.stop()

      # Add dummy jobs just to wake all threads
      for _ in range(len(self._workers)):
         self.addJob(lambda: None)

      for worker in self._workers:
         worker.join()

      exceptionList = self.getExceptionList()
      if exceptionList:
         raise WorkerThreadRunException(exceptionList)

      return self.getResultList()

   def getResultList(self):
      """Return result of the function called in parallel
      """
      resultList = []
      while not self._resultQueue.empty():
         resultList.append(self._resultQueue.get())

      return resultList

   def getExceptionList(self):
      """Return function name, args, kwargs and exception list recored in the
      exceptionQueue.
      """
      exceptionList = []
      while not self._exceptionQueue.empty():
         func, args, kargs, expType, expMessage = self._exceptionQueue.get()
         exceptionList.append({"Func": func, "Args": args, "Kargs": kargs,
                              "ExceptionType": expType,
                              "ExceptionMessage": expMessage})
      return exceptionList
