# Copyright 2014-2018 VMware, Inc.  All rights reserved. -- VMware Confidential

"""Printing logs on the terminal.

This module extends the standard logging module with a set of classes and
functions to print (colored) logs on the terminal console.
"""
import logging
import sys
import time

from borautils.terminal import *

def printProgress(msg, useColor=True):
   """Print a string followed by a carriage return ('\r').

   This function has no effect if <stdout> is redirected to a file or a pipe.
   """
   if sys.stdout.isatty():
      if useColor:
         msg = ttyFormat(msg, None, bold=True)
      print("%s\033[K\r" % msg, end='')
      sys.stdout.flush()


class TTYColorFormatter(logging.Formatter):
   """Logging formatter to colorize log records using TTY escape sequences.
   """
   _palette = {'DEBUG': (BLUE, BLUE),
               'INFO': (None, None),
               'WARNING': (YELLOW, YELLOW),
               'ERROR': (RED, RED),
               'CRITICAL': (MAGENTA, RED)}

   def __init__(self, fmt=None, datefmt=None, useColors=True):
      super().__init__(fmt=fmt, datefmt=datefmt)
      self._useColors = useColors

   def setLevelColor(self, level, name, prefixColor, msgColor):
      """Setup the display color of the given log level.
      """
      if name not in self._palette:
         logging.addLevelName(level, name)
      self._palette[name] = (prefixColor, msgColor)

   def format(self, record):
      """Do the formatting.
      """
      levelName = levelNameOrig = record.levelname
      msg = msgOrig = record.msg

      if self._useColors and levelName in self._palette:
         prefixColor, msgColor = self._palette[levelName]
         bold = (levelName == 'CRITICAL')
         levelName = ttyFormat(levelName, prefixColor, bold=True, reset=True)
         msg = ttyFormat(msg, msgColor, bold=bold, reset=True)
         # Clear up any garbage that might remain on the rest of the line,
         # for example if the previous log contained a '\r'.
         msg = '%s\033[K' % msg
      _, _, subLoggerName = record.name.partition('.')
      if subLoggerName:
         if self._useColors:
            subLoggerName = ttyFormat(subLoggerName, CYAN, bold=True,
                                      reset=True)
         msg = "[%s] %s" % (subLoggerName, msg)

      record.levelname = levelName
      record.msg = msg
      s = super().format(record)
      record.levelname = levelNameOrig
      record.msg = msgOrig
      return s

   def formatTime(self, record, datefmt=None):
      """Return a formatted timestamp for the given @record and @datefmt.
      """
      DATEFMT = '%Y-%m-%dT%H:%M:%S'

      ct = self.converter(record.created)
      if datefmt is None:
         timeStr = "%s.%03dZ" % (time.strftime(DATEFMT, ct), record.msecs)
      else:
         timeStr = time.strftime(datefmt, ct)
      if self._useColors:
         timeStr = ttyFormat(timeStr, BLUE, reset=True)
      return timeStr


class TTYHandler(logging.StreamHandler):
   """A generic TTY log handler.

   Stream handler to print logs on the terminal console. Color is enabled by
   default, although that can be turned off by setting the @useColors to False.
   """
   def __init__(self, stream=None, fmt=None, datefmt=None, useColors=True):
      super().__init__(stream=stream)
      useColors = useColors and self.stream.isatty()
      self._formatter = TTYColorFormatter(fmt, datefmt=datefmt,
                                          useColors=useColors)
      self.setFormatter(self._formatter)


def getDefaultLogLevel():
   """Define default logging level
   """
   return logging.DEBUG


def getDefaultLogFormat():
   """Define default logging format
   """
   return "%(asctime)s [%(levelname)s] %(message)s"


def getHandler(filename=None, filemode=None, format=None, datefmt=None,
               stream=None):
   """Configure the logger handler
   """
   if format is None:
      format = getDefaultLogFormat()

   if filename is not None:
      if filemode is None:
         filemode = 'w'
      handler = logging.FileHandler(filename, mode=filemode)
      formatter = logging.Formatter(fmt=format, datefmt=datefmt)
      handler.setFormatter(formatter)
   else:
      if stream is None:
         stream = sys.stdout
      handler = TTYHandler(stream=stream, fmt=format, datefmt=datefmt)
   assert handler is not None
   return handler


class TTYLogger(logging.Logger):
   """A generic logger to output logs to a TTY console.
   """
   @classmethod
   def basicConfig(cls, filename=None, filemode=None, format=None, datefmt=None,
                   level=None, stream=None):
      """Configure the TTYLoger, and set it up as the default logger so
      subsequent calls to logging.getLogger() return an instance of this class.
      """
      cls._rootHandler = getHandler(filename, filemode, format, datefmt, stream)
      logging.setLoggerClass(cls)

      root = logging.getLogger()
      for handler in root.handlers:
         root.removeHandler(handler)
      root.addHandler(cls._rootHandler)

      if level is None:
         level = getDefaultLogLevel()

      root.setLevel(level)

   @property
   def lock(self):
      """Return this logger's lock.

      Although functions in the logging module are already thread safe, a
      logging thread may need to hold the logger's lock to serialize a small
      batch of several log messages.

      The logger's lock is reentrant, allowing a thread which has acquired it to
      acquire it again without blocking. The thread must release the lock once
      for each time it has acquired it.
      """
      return self._rootHandler.lock

   def setLevelColor(self, level, name, prefixColor, msgColor):
      """Configure the name and display color of the given loglevel.
      """
      self._rootHandler.formatter.setLevelColor(level, name, prefixColor,
                                                msgColor)
