"""Small lib that implements a pager in python.

This code can be seen as reimplementation of a pager (similar to less or most)
in python. Here are the few differences/key points:
   - It supports page updates in the background. So the screen can be refreshed
   with new data if needed.
   - For now, we only support vertical scrolling. We have no need for
   horizontal scrolling at the moment. This is perfectly doable but would
   require a bit more code to handle all the cases.
   - We support having a header (i.e. pinned lines at the top). This is useful
   when scrolling stats and the name of the stats needs to be kept on screen.
   - We support arrow keys and vim-like commands.

   A typical usage of this lib would be:
   with Pager() as pager:
      pager.updatePage(content)
      pager.start()

      while True:
         content = updateContent()
         pager.update(content) # If needed
         pager.join(1) # Sleep for some time
         if not pager.is_alive():
            break # User asked to quit
"""

import os
import sys
import threading
import time

from . import terminal

def getLineSafe(page, i):
   """Helper function to avoid having to check if we reached the end of
   the list. Simply return an empty string instead and nothing will be
   printed on screen.
   """
   try:
      return page[i]
   except IndexError:
      return ""

class Pager(threading.Thread):
   """Pager class.

   It is implemented as a child of Thread so that we can refresh and process
   user inputs in the background. For more info about the usage and what we
   support please see the comment at the top of the file.
   """
   def __init__(self):
      self._tty = terminal.CBreakTTY()
      self._tty.setAutoWrapMode(False)
      self._tty.setTextCursorEnableMode(False)

      self._ttySize = None
      self._ESC = terminal.ESC("")
      self._UP = terminal.CSI(terminal.CSI.CUU)
      self._DOWN = terminal.CSI(terminal.CSI.CUD)
      self._HOME = terminal.CSI(terminal.CSI.HOME)
      self._END = terminal.CSI(terminal.CSI.END)
      self._PGUP = terminal.CSI(terminal.CSI.PGUP)
      self._PGDOWN = terminal.CSI(terminal.CSI.PGDOWN)
      self._curLine = 0
      self._needsRefresh = False

      # The lock protects the below values
      self._pageLock = threading.Lock()
      self._page = None
      self.headerLength = 0

      self._updateScreenSize()
      super().__init__()

   def _updateScreenSize(self):
      """Get the screen size
      """
      self._ttySize = os.get_terminal_size()

   def _computeCurLine(self, curLine, headerSize, numLines):
      """Adjust the line displayed depending on our page
      """
      # negative value indicates we want to go to the botttom
      if curLine < 0:
         usableLines = self._ttySize.lines - headerSize
         maxCurLine = numLines - usableLines
         curLine = maxCurLine

      # Cannot scroll higher than headerSize
      curLine = max(headerSize, curLine)
      # Cannot scoll lower than the last line
      curLine = min(numLines - 1, curLine)

      return curLine

   def _show(self, curLine):
      """Actual drawing function.

      First, we retrieve the new content. We then update the screen, print
      the header and shift to right content depending on the current line.
      """
      assert self._page is not None

      with self._pageLock:
         lines = self._page.rstrip(os.linesep).split(os.linesep)
         headerSize = self.headerLength

      self._tty.clearScreen()
      self._updateScreenSize()

      curLine = self._computeCurLine(curLine, headerSize, len(lines))

      # First, let's print the header.
      for i in range(headerSize):
         self._tty.echo(getLineSafe(lines, i) + os.linesep)

      # In the unlikely case of usableLines being negative,
      # we won't print anything
      usableLines = range(self._ttySize.lines - headerSize)

      for i in usableLines:
         line = getLineSafe(lines, curLine + i)
         # the last line should not have a newline at the end
         if i != usableLines[-1]:
            line += os.linesep
         self._tty.echo(line)

      return curLine

   def _readKey(self):
      """Read key stroke in a non-blocking fashion

      We support special keys like arrows hence we might
      read three characters.
      """
      ch = self._tty.getkey()
      if ch:
         ch = ch.decode()
         if ch == self._ESC:
            # We can expect two or three chars after ESC
            ch2 = self._tty.getkey()
            ch3 = self._tty.getkey()

            if not ch2 or not ch3:
               # We don't recognize the sequence.. Ignoring
               return None

            ch = ch + ch2.decode() + ch3.decode()

            # Home, End, PgUp, PgDown are 4-chars sequences
            ch4 = self._tty.getkey()
            if ch4:
               ch += ch4.decode()

      return ch

   def _userInput(self):
      """Get the user input and redraw the screen if needed.
      """
      refresh = self._needsRefresh
      key = self._readKey()

      while key:
         if key == "q":
            return False
         elif key in ("g", self._HOME):
            self._curLine = 0
         elif key in ("G", self._END):
            self._curLine = -1
         elif key in ("j", self._DOWN):
            self._curLine += 1
         elif key in ("k", self._UP):
            self._curLine -= 1
         elif key == self._PGUP:
            self._curLine -= self._ttySize.lines
            self._curLine = max(self._curLine, 0)
         elif key == self._PGDOWN:
            self._curLine += self._ttySize.lines

         refresh = True
         key = self._readKey()

      if refresh:
         self._needsRefresh = False
         self._curLine = self._show(self._curLine)

      return True

   def __enter__(self):
      """Enter the pager context.
      """
      return self

   def __exit__(self, excType, excValue, traceback):
      """Exit the pager context and restore the screen
      """
      self.cleanUp()

   def updatePage(self, page, headerLength=0):
      """Update the content to be displayed on screen.
      """
      with self._pageLock:
         self._page = page
         self.headerLength = headerLength

      self._needsRefresh = True

   def cleanUp(self):
      """Restore the screen to its original state.

      It is likely better to not call this directly and use the context
      API (enter, exit) instead. This will ensure cleanUp is called. If
      this is not called when exiting, the terminal will be left in an
      unstable state and would need to be reseted.
      """
      # Make sure we don't leave the next program in the middle of a line
      self._tty.echo(os.linesep)

      self._tty.setAutoWrapMode(True)
      self._tty.setTextCursorEnableMode(True)
      self._tty.reset()

   def run(self):
      """Start to accept user input in the background.

      Update page should have been called before calling this function. We
      assume we have some content to draw on screen. Additional calls to
      updatePage can be made while running to dynamically redraw the screen.
      """
      assert self._page

      while self._userInput():
         time.sleep(.1)
