# Copyright 2016-2019 VMware, Inc.
# All rights reserved. -- VMware Confidential

"""ESX build info interface

Library to query information relative to an ESX build options and deliverables.
"""
from hashlib import md5
import os
import re

from borautils.bora import Bora
from borautils.busybox import cat
from borautils.esx.vibDepot import VibDepot
from borautils.infra.dbc import DBC
from borautils.infra.buildapi import Build
from borautils.httputils import urlExists


def getVmtreeHash(vmtreePath):
   """Return a hash that identifies the given vmtree.

   The vmtree hash is computed at build time, based on the path to the vmtree,
   and using a recipe implemented in #bora/support/script/pxe-boot.
   """
   return "t-%s" % md5((vmtreePath + '\n').encode()).hexdigest()[:20]

class EsxBuild(object):
   """A class to load an ESX image from the publish directory of a standard ESX
   build.
   """
   _builderType = None

   def __init__(self, pubDir, installer=False, standard=False):
      """Load an ESX image from a standard publish directory.
      """
      self._pubDir = pubDir
      self._md5 = {}

      if hasattr(self, 'target') and self.target == 'esxcore':
         self._profile = 'esxcore'
         self._pxeCache = 'pxe-modules'
         self._depot = 'CUR-dev-depot/ESXi'
      elif hasattr(self, 'target') and self.target == 'petronas':
         self._profile = 'physical-amz-ena-nsxt'
         self._pxeCache = None
         self._depot = 'depots/physical-amz-ena-nsxt'
      else:
         if installer:
            if standard:
               self._profile = 'standard'
               self._pxeCache = 'installer-std-pxe-modules'
               self._depot = 'CUR-depot/ESXi'
            else:
               self._profile = 'dev'
               self._pxeCache = 'installer-pxe-modules'
               self._depot = 'CUR-dev-depot/ESXi'
         else:
            if standard:
               self._profile = 'no-tools'
               self._pxeCache = 'std-pxe-modules'
               self._depot = 'CUR-depot/ESXi'
            else:
               self._profile = 'dev-no-tools'
               self._pxeCache = 'pxe-modules'
               self._depot = 'CUR-dev-depot/ESXi'

      self._loadDeliverablesMd5sums()

      vibDepotPath = os.path.join(self._pubDir, self._depot)
      self._vibDepot = VibDepot(vibDepotPath)

   @property
   def buildNumStr(self):
      """Build number as a string (e.g. '123456' for OB/SB, or '00000' for DBC
      builds).
      """
      DEFAULT_BUILD_NUM_STR = "00000"

      if self.buildNum:
         return str(self.buildNum)

      return DEFAULT_BUILD_NUM_STR

   @property
   def pxeCache(self):
      """Path/URL to the pxe cache directory.
      """
      if self._pxeCache is None:
         raise RuntimeError("no PXE cache found for ImageProfile '%s' in depot "
                            "'%s'"  % (self._profile, self._depot))
      return os.path.join(self._pubDir, self._pxeCache)

   def _loadDeliverablesMd5sums(self):
      """Load deliverables MD5sums from $PUBLISH_DIR/MD5SUM.txt.

      Only Sandbox/Official builds have an MD5SUM.txt checksum file.
      Skipping this for developer builds.
      """
      md5LineRe = re.compile(r'([0-9a-fA-F]{32})\s+\*(.*)')

      md5sumTxt = os.path.join(self._pubDir, 'MD5SUM.txt')
      if urlExists(md5sumTxt):
         for line in cat(md5sumTxt):
            m = md5LineRe.match(line)
            if m is not None:
               self._md5[m.group(2)] = m.group(1)

   def _getProfile(self):
      """Return the image profile of the ESX
      """
      return self._vibDepot.getProfile(self._profile, buildNum=self.buildNum)

   def getDeliverableMd5(self, filePath):
      """Return the MD5sum of a build deliverable.
      """
      relPath = os.path.relpath(filePath, start=self._pubDir)
      if not relPath.startswith(os.pardir):
         return self._md5.get(relPath, None)
      return None

   def getBootloaderModules(self):
      """Return the list of bootloader modules.
      """
      profile = self._getProfile()
      return [os.path.join(self.pxeCache, m.name)
              for m in profile.iterPxePayloads()]

   def getBootModules(self, isInstaller=False):
      """Return the list of ESX boot modules.
      """
      profile = self._getProfile()
      return [os.path.join(self.pxeCache, m.name)
              for m in profile.iterBootPayloads(installer=isInstaller)]

   def getVibUrl(self, name):
      """Return the absolute path (or HTTP URL) to the VIB named @name.
      """
      return self._vibDepot.getVibUrl(name)

   def getInstallerIso(self):
      """Return a path to the ESX installer ISO.
      """
      # By convention, ESX version is inferred from the esx-base VIB version.
      version = self._vibDepot.getVibVersion('esx-base')
      m = re.match(r"(\d\.\d\.\d)-(\d\.\d+)\.(\d+)", version)
      if m is None:
         raise ValueError("%s: invalid version string in the esx-base VIB"
                          % version)
      esxVersion = m.group(1)

      isoName = ("VMware-VMvisor-Installer-%s-%s.x86_64.iso" %
                 (esxVersion, self.buildNumStr))
      return os.path.join(self._pubDir, isoName)

   def getBoraTree(self):
      """Return NFS path to the #bora directory as a (nfsHost, nfsPath,
      automountPath) tuple.
      """
      nfsHost, nfsPath, automountPath = self.getBuildTree()
      if nfsPath:
         nfsPath = os.path.join(nfsPath, 'bora')
      if automountPath:
         automountPath = os.path.join(automountPath, 'bora')
      return (nfsHost, nfsPath, automountPath)


class _EsxGobuild(EsxBuild):
   """Base class which represents an ESX build obtained via Gobuild. This class
   should not be instantiated directly.
   """
   def __init__(self, buildId, installer=False, standard=False):
      self._build = Build(buildId, self._builderType)
      self.buildNum = self._build.buildNum
      self.buildType = self._build.buildType
      self.target = self._build.getGobuildTarget()
      pubDir = self._build.getPublishDirUrl()
      super().__init__(pubDir, installer=installer, standard=standard)

   def getBuildTree(self):
      """Return the NFS path to the build tree as a (nfsHost, nfsPath,
      automountPath) tuple.
      """
      return self._build.getBuildTree('linux64')

   def getEsxDevVibUrl(self, vibName):
      """Return the url to the VIB under CUR-dev-depot/ESXi/vib20/@vibName.
      """
      if not vibName:
         raise ValueError("vibName must not be an empty string")

      vibPathPrefix = os.path.join('CUR-dev-depot', 'ESXi', 'vib20', vibName)
      vibFileRe = re.compile(r'.*%s.*\.vib' % vibName)
      for deliverable in self._build.deliverables:
         if vibPathPrefix in deliverable.url and \
            vibFileRe.match(os.path.basename(deliverable.url)):
            return deliverable.url

      raise FileNotFoundError("No VIB under %s" % vibPathPrefix)

   def __str__(self):
      """Return a human-readable string to identify this ESX build.
      """
      return str(self._build)

class EsxSandboxBuild(_EsxGobuild):
   """Class which represents an ESX Sandbox build.
   """
   _builderType = 'sb'


class EsxOfficialBuild(_EsxGobuild):
   """Class which represents an ESX official build.
   """
   _builderType = 'ob'


class EsxDbcBuild(Bora, DBC, EsxBuild):
   """Class which represents a DBC-based developer build of ESX.
   """
   _builderType = 'dbc'

   def __init__(self, workingDir=None, installer=False, standard=False):
      Bora.__init__(self, vmtree=workingDir)
      DBC.__init__(self, self.vmtree)

      EsxBuild.__init__(self, self.toUrl(self.pubDir), installer=installer,
                        standard=standard)
      self.buildNum = None

   def getBuildTree(self):
      """Return the NFS path to the build tree as a (nfsHost, nfsPath,
      automountPath) tuple.
      """
      buildRoot = os.path.dirname(self.vmtree)
      return (self.host, buildRoot, buildRoot)

   @property
   def target(self):
      """ESX build target.
      """
      if self.product == 'esx':
         return 'esxall'
      elif self.product == 'esxcore':
         return 'esxcore'
      else:
         raise AttributeError("unsupported product: %s" % self.product)

   def __str__(self):
      """Return a human-readable identifier for this build object.
      """
      return "%s-%s-%s" % (getVmtreeHash(self.vmtree), self.product,
                           self.buildType)


def getEsxBuildFromNum(buildNum, installer=False, standard=False):
   """Return a build object for the given build number.
   """
   if buildNum == 0:
      raise ValueError("0 is not a valid build number")

   try:
      ob = EsxOfficialBuild(buildNum, installer=installer, standard=standard)
   except Exception:
      ob = None
   try:
      sb = EsxSandboxBuild(buildNum, installer=installer, standard=standard)
   except Exception:
      sb = None

   if ob is not None and sb is not None:
      raise ValueError("failed to dis-ambiguate build number '%u' (both ob-%u "
                       "and sb-%u do exist)" % (buildNum, buildNum, buildNum))
   elif ob is not None:
      return ob
   elif sb is not None:
      return sb
   else:
      raise ValueError("%u: no such ESX build" % buildNum)

def getEsxBuildFromStr(buildIdStr, installer=False, standard=False):
   """Return a build object for the given build identifier.
   """
   if buildIdStr is None or os.path.exists(buildIdStr):
      return EsxDbcBuild(buildIdStr, installer=installer, standard=standard)

   if buildIdStr.lower().startswith('sb-'):
      return EsxSandboxBuild(int(buildIdStr[3:]), installer=installer,
            standard=standard)

   if buildIdStr.lower().startswith('ob-'):
      return EsxOfficialBuild(int(buildIdStr[3:]), installer=installer,
            standard=standard)

   try:
      buildNum = int(buildIdStr)
   except (ValueError, TypeError):
      pass
   else:
      return getEsxBuildFromNum(buildNum)

   raise ValueError("%s: invalid build identifier" % buildIdStr)
