# Copyright 2015-2018,2020 VMware, Inc.
# All rights reserved. -- VMware Confidential

"""ESX image customization lib

Utility lib to package various init scripts and configuration files into a boot
tardisk. This tardisk can be added to a pre-built image of ESX in order to
automatically configure this image with custom NFS shares, ssh keys, static IPs,
.bashrc, etc.
"""
import json
import os
import shutil
import stat
import tarfile
from tempfile import mkdtemp

from borautils.infra.automount import getAutomountInfoFromAbsPath
from borautils.infra.toolchain import TOOLCHAIN_HOST, TOOLCHAIN_NFS_PATH, TCROOT
from borautils.sshkeys import getEsxPublicKey


LOCKER_CONF = "/etc/vmware/locker.conf"
IP_CONF = "/etc/vmware/devel-host.conf"
NFS_INIT_SCRIPT = "/etc/rc.local.d/50_nfs"
VIB_INIT_SCRIPT = '/etc/rc.local.d/60_vibinstall'
POSTBOOT_SCRIPT = '/etc/rc.local.d/00_login'
BOOTDONE_SCRIPT = '/etc/rc.local.d/90_bootdone'
SSH_AUTHORIZED_KEYS = '/etc/ssh/keys-root/authorized_keys'
BASH_RC = '/.bashrc'
TESTESX_CONF='/var/lib/vmware/testesx/conf.json'

class EsxTarDisk(object):
   """Class to generate ESX tardisks
   """
   def __init__(self):
      self._files = []

   def addFile(self, src, dest, mode):
      """Add a file or a directory entry (non-recursive) to this tardisk.
      """
      self._files.append((src, dest, mode))

   def write(self, dest):
      """Generate (write) this tardisk.
      """
      with tarfile.open(dest, 'w:gz', format=tarfile.GNU_FORMAT) as t:
         for src, dst, mode in self._files:
            attr = t.gettarinfo(name=src, arcname=dst)
            attr.mode = mode
            if (stat.S_IWGRP | stat.S_IWUSR | stat.S_IWOTH) & mode:
               attr.mode |= stat.S_ISVTX
            attr.uid = 0
            attr.gid = 0
            attr.uname = 'root'
            attr.gname = 'root'
            if os.path.isfile(src): # add a file
               with open(src, 'rb') as f:
                  t.addfile(attr, fileobj=f)
            else: # add a directory entry
               t.addfile(attr)

class EsxConf(EsxTarDisk):
   """A class to load, customize and deploy an ESX image from the publish
   directory of a standard ESX build.
   """

   def __init__(self):
      """Load an ESX image from a standard publish directory.
      """
      super().__init__()
      self._bootVibs = []
      self._tmpDir = mkdtemp()
      self._nfsMounts = set()
      self._lockerDir = None
      self._uwSwapEnabled = False
      self._ipConf = None
      self._buildTree = None

   def __del__(self):
      """Cleanup temp files.
      """
      try:
         shutil.rmtree(self._tmpDir, ignore_errors=True)
      except Exception:
         pass

   def setHostIp(self, ipConf):
      """Setup the host IP configuration.

      The IP configuration is passed as a dictionary following the format below:

      ipConf['aa:bb:cc:dd:ee:ff'] = {
         'hostname': 'dev-box-1',
         'ip':       '10.20.xx.xx5',
         'netmask':  '255.255.252.0',
         'gateway':  '10.20.xx.xyz',
         'dns':      '10.20.xyz.a, 10.20.xyz.b'
      }
      """
      self._ipConf = ipConf

   def setNfsMount(self, host, share, mountDir, options):
      """Add an NFS mount to this ESX image.
      """
      self._nfsMounts.add((host, share, mountDir, options))

   def setBuildTreeMount(self, host, share, mountDir, options):
      """Set an build tree mount to this ESX image
      """
      self._buildTree = (host, share, mountDir, options)

   def addBootVib(self, vibUrl):
      """Add a boot VIB to this ESX image.

      Boot VIBs are automatically downloaded and installed from an ESX init
      script. The VIBs are transferred from the PXE directory once the system is
      up via NFS or HTTP. The former method requires the PXE directory to be
      mounted over NFS, which can easily be done with the EsxConf.setNfsMount()
      method. The latter method is recommended, if the server which serves the
      ESX image provides HTTP services.
      """
      self._bootVibs.append(vibUrl)

   def setLocker(self, lockerDir, uwSwapEnabled):
      """Configure the locker directory.

      Set the path to the locker directory. The volume containing that directory
      should have at least 1GB free for the uwswap. In addition to holding the
      userworld swap file, the locker directory is also used for /var/tmp and to
      hold various objects downloaded to the server.
      """
      self._lockerDir = lockerDir
      self._uwSwapEnabled = uwSwapEnabled

   @staticmethod
   def _writeIpConf(dest, ipConf):
      """Write the host IP configuration file.
      """
      import json
      with open(dest, 'w') as f:
         json.dump(ipConf, f)

   @staticmethod
   def _writeDevelNfsScript(nfsMounts, dest):
      """Generate the NFS mount script
      """
      with open(dest, 'w') as f:
         print("#!/bin/sh", file=f)

         defaultMounts = 8
         maxMounts = 32
         if len(nfsMounts) > defaultMounts:
            print('localcli system settings advanced set '
                  '-o /NFS/MaxVolumes -i %d' % maxMounts, file=f)

         multiHosts = {}
         knownHosts = {}
         for host, _, _, _ in nfsMounts:
            if host in knownHosts:
               multiHosts[host] = True
            knownHosts[host] = True

         for host, src, dest, opts in nfsMounts:
            if opts is None:
               opts = ''
            opts = opts.split(',')
            readonly = '-r' if 'ro' in opts else ''

            if host in multiHosts:
               mntName = "%s_%s" % (host, src.replace('/', '.'))
            else:
               mntName = host

            print("mkdir -p %s" % os.path.dirname(dest), file=f)
            print("localcli storage nfs add -H %s -s %s %s -v '%s'" %
                  (host, src, readonly, mntName), file=f)
            print("ln -sf '/vmfs/volumes/%s' %s" % (mntName, dest), file=f)

   @staticmethod
   def _generateVibInitScript(dest, vibs):
      """Generate an init script that auto-installs the boot VIBs

      @vibs is a list of paths or HTTP URLs.
      """
      script = [
         '# VIB auto-install',
         'for vib in %s; do' % ' '.join(vibs),
         '   localcli software vib install -v ${vib}',
         '   if [ $? -ne 0 ]; then',
         '      esxcfg-init -a "${vib}: VIB installation has failed, check '
         '/var/log/esxupdate.log for more details."',
         '   fi',
         'done'
         ]

      with open(dest, 'w') as f:
         print('\n'.join(script), file=f)

   @staticmethod
   def _generateBootDoneScript(dest):
      """Generate the bootDone init script

      The 'bootDone' init script is the last ESX init script to be executed. It
      prints an anchor message in the vmkernel logs to signal that ESX has
      finished initializing, and that it is safe to start test workloads.
      """
      script = ['#!/bin/sh',
                'vmdumper -g "ESX IS READY FOR TESTING"']
      with open(dest, 'w') as f:
         print('\n'.join(script), file=f)

   @staticmethod
   def _writeLockerConf(lockerConf, lockerDir, uwSwapEnabled):
      """Write the locker configuration (locker.conf)
      """
      with open(lockerConf, 'w+') as f:
         if lockerDir:
            print("%s %d" % (lockerDir, 1 if uwSwapEnabled else 0), file=f)

   def writeConfTardisk(self, dest):
      """Deploy this ESX image into the given PXE directory.
      """
      # TODO - support ipConf
      #if self._ipConf:
      #   ipConf = os.path.join(self._tmpDir, 'ip.conf')
      #   self._writeIpConf(ipConf, self._ipConf)
      #   self.addFile(ipConf, IP_CONF, 0o600)

      #lockerConf = os.path.join(self._tmpDir, 'locker.conf')
      #self._writeLockerConf(lockerConf, self._lockerDir, self._uwSwapEnabled)
      #self.addFile(lockerConf, LOCKER_CONF, 0o1600)
      conf = {'nfs':list(self._nfsMounts), 'vibs':self._bootVibs,
              'buildtree':self._buildTree}
      confFile = os.path.join(self._tmpDir, 'testesxconf.json')
      with open(confFile, 'w') as f:
         json.dump(conf, f)
      self.addFile(os.path.dirname(confFile),
                   os.path.dirname(TESTESX_CONF), 0o744)
      # add its directory entry into tardisk, otherwise mount failed.
      self.addFile(confFile, TESTESX_CONF, 0o644)
      super().write(dest)


class EsxDevConf(EsxConf):
   """An ESX image that is pre-configured for developers.
   """

   def __init__(self, vmtree=None):
      """Instantiate and initialize an ESX developer image.
      """
      super().__init__()
      self.setNfsMount('scm-trees', '/trees', '/build/trees', 'tcp,hard,intr')
      self.setNfsMount(TOOLCHAIN_HOST, TOOLCHAIN_NFS_PATH, TCROOT,
                       'tcp,hard,intr')
      self.setNfsMount('build-apps', '/apps', '/build/apps', 'tcp,hard,intr')

      if vmtree is not None:
         nfsShare = getAutomountInfoFromAbsPath(os.path.abspath(vmtree))
         if nfsShare is not None:
            nfsServer, nfsPath, mountAttrs = nfsShare
            self.setBuildTreeMount(nfsServer, nfsPath, nfsPath, mountAttrs)

   @staticmethod
   def _writePostBootLateScript(dest, extraScript=None):
      """Generate the post-boot late script.

      This script does all kinds of developer-only initializations. It is run at
      ESX initialization time, after NFS shares have been mounted.
      """
      script = [
         "#!/bin/sh",

         # Do this nice and early so it is always setup note we use localcli as
         # hostd is potentially not up yet.

         # Print ESX vmk0 IPv4 address in the vmkernel log at the beginning
         # This allows various scripts that monitor ESX vmkernel log
         # to obtain the IP of this host as early as possible
         "# Print ESX vmk0 IPv4 address into vmkernel log",
         "hostip=`localcli network ip interface ipv4 get | grep vmk0 | awk '{print $2}'`",
         'vmdumper -g "ESX vmk0 IPv4 address is $hostip"',
         "",

         '# Reset shell timeout to zero',
         "localcli system settings advanced set -o /UserVars/ESXiShellTimeOut -i 0",
         "",

         "# Enable local shell",
         "/etc/init.d/ESXShell start",
         "",

         "# Enable SSH",
         "/etc/init.d/SSH start",
         "localcli network firewall ruleset set -r sshServer -e true",
         "",
      ]

      if extraScript:
         script += extraScript

      with open(dest, 'w') as f:
         print('\n'.join(script), file=f)

   @staticmethod
   def _writeEsxPublicKey(dst):
      """Generate (write) the ESX SSH public key.

      The ESX ssh public key allows automation tools (like testesx) to connect
      to ESX without a password. This key is for developer-mode only. It must
      never be published with the product, or shared with anyone outside of
      VMware.
      """
      with open(dst, 'w') as f:
         f.write(getEsxPublicKey())


#
# Code below is for debugging purposes...
#
if __name__ == "__main__":
   from argparse import ArgumentParser

   parser = ArgumentParser()
   parser.add_argument(type=str, metavar='PUBDIR', dest='dest',
                       help="Path to publish (source) directory")
   args = parser.parse_args()

   tarDisk = EsxDevConf()
   tarDisk.writeConfTardisk(args.dest)
