"""
Copyright 2017 VMware, Inc. All rights reserved. -- VMware Confidential

Library that handles ESXi crypto needs.
"""

import re
import subprocess

#
# Constants
#

cryptoUtil = '/build/apps/bin/esx/crypto-util-full-repo/crypto-util-full'

class CryptoFile:
    """ The class for files used by this module
    """

    def __init__(self, pathName):
        if pathName is None or not pathName:
            raise ValueError("pathName is not valid")
        else:
            self._pathName = pathName

    @property
    def _cmdLineArgs(self):
        return [ self._pathName ]


class KeyFile(CryptoFile):
    """ The key file class

        Keys are kept in a key file. A key file is itself encrypted (wrapped)
        to keep the key matter secret. One must know the wrapping in order to
        unwrap the key and make use of it.

        A wrapping can be specified at instantiation time or at any later time.
    """

    def __init__(self, pathName, wrapping=None):
        CryptoFile.__init__(self, pathName)
        self._wrapping = wrapping

    def setWrapping(self, wrapping):
        """ Associate a wrapping for the key file

            If a wrapping is specified, the unwrapping of a key file will only
            be attempted via the specified wrapping.

            It is acceptable for a key file to not have a wrapping. While
            useless for encryption (the key file must be unwapped to gain
            access to its key), the lack of a wrapping can be used as an
            indicator for a decryption method to attempt built-in unwrapping
            procedures (e.g. well known passwords).
        """

        if wrapping is None:
            raise ValueError("no key file wrapping was specified")

        self._wrapping = wrapping

    @property
    def hasWrapping(self):
        """ Does the class instance have a wrapping?

            This allows a decryption method to know if it is free to attempt
            built-in unwrapping procedures.
        """
        return self._wrapping is not None

    def isWrappingValid(self):
        """ Is the wrapping associated with the key file valid?

            This is useful for testing wrappings (e.g. passwords) against a key
            file. Knowing the a key file cannot be unwrapped saves attempting
            a heavyweight cryptographic operation (e.g. encryption).
        """
        if self._wrapping is None:
            raise RuntimeError("no key file wrapping was specified")

        cmd = (
                  [ cryptoUtil, 'dictionary', 'decrypt' ] +
                  self._wrapping._cmdLineArgs +
                  self._cmdLineArgs +
                  [ '-' ]
              )

        p = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE,
                             universal_newlines=True)

        (outs, errs) = p.communicate()

        return p.returncode == 0

    def create(self, cipher, keyId=None):
        """ Create a key file that contains a key associated with the specified
            symmetric cipher.

            The key file will be wrapped via the wrapping method associated
            with the key file.
        """

        if cipher is None or not cipher:
            raise ValueError("cipher is not valid")

        if self._wrapping is None:
            raise RuntimeError("no key file wrapping was specified")

        cmd = (
                  [ cryptoUtil, 'keys', 'generatesymkey' ] +
                  self._wrapping._cmdLineArgs +
                  [ cipher ] +
                  self._cmdLineArgs
              )

        if keyId is not None:
            cmd.append(keyId)

        p = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE,
                             universal_newlines=True)

        (outs, errs) = p.communicate()

        if p.returncode != 0:
            raise RuntimeError("command '%s' failed: %s" % (cmd, errs))

    @property
    def keyID(self):
        """ Return the keyID associated with the key within the key file.

            A keyID is a unique string associated with a key, a way of
            referencing the key without having to have access to the key
            matter itself.
        """

        cmd = [ cryptoUtil, 'keys', 'describe' ] + self._cmdLineArgs

        p = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE,
                             universal_newlines=True)

        (outs, errs) = p.communicate()

        if p.returncode != 0:
            raise RuntimeError("command '%s' failed: %s" % (cmd, errs))

        keyID = None

        if p.returncode == 0 and outs:
            match = re.search("^.*Key file keyID is '(?P<keyID>[^']*)'.*$",
                              outs, re.MULTILINE)

            if match:
                keyID = match.group('keyID')

        if keyID is None:
            raise RuntimeError("possibly corrupt key file")

        return keyID

class KeyFilePasswordWrapping:
    """ A wrapping by means of a password.

        A key file can be wrapped via a password.
    """

    def __init__(self, password):
        self._password = password

    @property
    def _cmdLineArgs(self):
        value = []

        if self._password is not None:
            value.append("--password=%s" % self._password)

        return value

class KeyFilePublicKeyWrapping(CryptoFile, KeyFilePasswordWrapping):
    """ A wrapping by means of a public key encryption method

        A key file can be wrapped via a public key encryption method.

        When a key file is wrapped for encryption, the public portion of the
        public key encryption method (in a file) must be used. For decryption,
        the private portion (in a file) must be used.
    """

    def __init__(self, pathName, password=None):
        CryptoFile.__init__(self, pathName)
        KeyFilePasswordWrapping.__init__(self, password)

    @property
    def _cmdLineArgs(self):
        return (
                   [ "--asymkey" ] +
                   CryptoFile._cmdLineArgs.__get__(self, CryptoFile) +
                   KeyFilePasswordWrapping._cmdLineArgs.__get__(self, KeyFilePasswordWrapping)
               )

def createPublicKeyPair(cipher, publicPortionPath, privatePortionPath,
                        password=None, keyId=None):
    """ Create a public key encryption key pair
    """

    if cipher is None or not cipher:
        raise ValueError("cipher is not valid")

    if publicPortionPath is None or not publicPortionPath:
        raise ValueError("public portion path is not valid")

    if privatePortionPath is None or not privatePortionPath:
        raise ValueError("private portion path is not valid")

    if password is not None and not password:
        raise ValueError("password is not valid")

    cmd = [ cryptoUtil, 'keys', 'generatepublickeypair' ]

    if password is not None:
        cmd.append("--password=%s" % password)

    cmd.append(cipher)
    cmd.append(privatePortionPath)
    cmd.append(publicPortionPath)

    if keyId is not None:
        cmd.append(keyId)

    p = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE,
                         universal_newlines=True)

    (outs, errs) = p.communicate()

    if p.returncode != 0:
        raise RuntimeError("command '%s' failed: %s" % (cmd, errs))

