# Copyright 2017-2018 VMware, Inc.
# All rights reserved. -- VMware Confidential

"""Gdb wrapper to debug vmkernel core dumps.
"""
import glob
import os

from borautils.elfbin import ElfError
from borautils.gdb import Gdb
from .build import getEsxBuildFromStr
from .vmkbin import VmkBin, VmkCore


class VmkGdb(Gdb):
   """Class to setup and run vmkgdb on a given vmkernel core dump.

   This class scans the core dump, and auto-locate the vmtree and all the files
   required to configure vmkgdb (the vmkernel binary, vmkernel modules, gdb
   macros, ...).
   """

   def __init__(self, target):
      try:
         self.core = VmkCore(target)
      except ElfError:
         self.core = None
         self.vmkernel = VmkBin(target)
         self._locateVmtree(self.vmkernel.path)
      else:
         self._locateVmtree(self.core.buildNum)
         self.vmkernel = self._locateVmkBin()

      gdbBin = os.path.join(self._buildDir, 'vmkgdb', 'usr', 'bin', 'gdb')
      super().__init__(path=gdbBin)
      self._loadSoLibs()
      self._loadGdbMacros()

   def _locateVmtree(self, buildNum):
      """Infer the ESX source/build tree from the core build number.
      """
      buildNumStr = None if buildNum == 0 else str(buildNum)
      self._build = getEsxBuildFromStr(buildNumStr)

      _, _, self.vmtree = self._build.getBoraTree()
      if not os.path.isdir(self.vmtree):
         raise FileNotFoundError("%s: no such vmtree", self.vmtree)

      self._buildDir = os.path.join(self.vmtree, 'build', 'esx',
                                    self._build.buildType)
      if not os.path.isdir(self._buildDir):
         raise FileNotFoundError("%s: no such build directory" % self._buildDir)

   def _locateVmkBin(self):
      """Locate the vmkernel binary.
      """
      path = os.path.join(self._buildDir, 'vmkernel-vmvisor64', 'vmkernel')
      if not os.path.exists(path):
         raise FileNotFoundError("%s: no such vmkernel", path)

      vmkernel = VmkBin(path)
      assert self.core.buildNum == vmkernel.buildNum
      assert self.core.buildType == vmkernel.buildType
      return vmkernel

   def _loadGdbMacros(self):
      """Load the gdb macros from the source/build tree.
      """
      gdbMacros = os.path.join(self._buildDir, 'vmkernel-vmvisor64',
                               'gdbmacros', '*.gdb*')
      for macro in glob.glob(gdbMacros):
         self.macros.append(macro)

      gdbMacros = os.path.join(self._buildDir, 'vmkdrivers-gdbmacros',
                               'gdbmacros', '*.gdb*')
      for macro in glob.glob(gdbMacros):
         self.macros.append(macro)

   def _loadSoLibs(self):
      """Load the vmkernel modules.
      """
      vmkmodDir = os.path.join(self._buildDir, 'vmkmod-vmkernel64')
      vmkdriversDir = os.path.join(self._buildDir, 'vmkdrivers-stage64')

      for path in [vmkmodDir, vmkdriversDir]:
         if not os.path.isdir(path):
            raise FileNotFoundError("%s: no such directory" % path)
         self.soPath.append(path)

   def run(self, forceBreak=False, **kwargs):
      """Setup debug environment and run gdb.
      """
      self.vars['bin_vth0'] = "((uint64*)vmTreeHash)[0]"
      self.vars['bin_vth1'] = "((uint64*)vmTreeHash)[1]"
      self.vars['bin_vmtree'] = "*(uint32*)vmTree"
      self.vars['bin_bld'] = "*(uint32*)vmkernelBuild"
      self.vars['bin_start'] = "(uint32)&_start & 0xfff"
      self.vars['DEBUG_ESX_BLDNUM'] = self._build.buildNumStr
      self.vars['DEBUG_ESX_BLDTYPE'] = '"%s"' % self._build.buildType

      commands = ["echo ----\\n",
                  "echo 'info threads' to get a list of worlds\\n",
                  "echo 'help user' to get a list of custom GDB commands\\n",
                  "echo 'verifyConfigOptions' to list active non-default "
                  "config options\\n",

                  "VerifyVmkernelTarget"]

      options = ['--break'] if forceBreak else None

      target = self.core.path if self.core is not None else None

      super().run(self.vmkernel.path, target=target, commands=commands,
                  options=options, **kwargs)
