"""Debug bundle library

Library to collect debug bundle
"""
import os
import subprocess

from borautils.exception import DebugInfoMixIn
from borautils.nfsutils import NfsPath

DEFAULT_VMSUPPORT_TIMEOUT = 20 * 60

class VmSupportError(DebugInfoMixIn, Exception):
   """A request to collect vm-support has failed.
   """
   _errType = "vm-support command has failed"

   def __init__(self, cmd, stderr):
      errStr = "vm-support command executed, but returned an error"
      super().__init__(errStr)
      self.addDebugInfo("command", cmd)
      self.addDebugInfo("stderr", stderr)


def collectVmSupport(hostname, destDir, hostPassword=None):
   """Collect vm-support in server over ssh

   If destDir is a NFS directory, mount it in the target ESX host
   and save vm-support bundle to it, otherwise, save the
   vm-support to host local destDir
   """
   from borautils.ssh import SSH
   from borautils.sshkeys import getEsxKeyFile

   uploadDir = NfsPath(destDir)
   if uploadDir.isNfs():
      os.makedirs(destDir, exist_ok=True)

   if hostPassword:
      # Stateful deployed standard builds do not have devmodeint vib, hence
      # private key would not work. Need to use host password.
      key = None
   else:
      key = getEsxKeyFile()

   with SSH(hostname, username="root", password=hostPassword, key=key,
            keepalive=3, timeout=DEFAULT_VMSUPPORT_TIMEOUT) as ssh:
      if uploadDir.isNfs():
         cmd = ['localcli', 'storage', 'nfs', 'add', '-H', uploadDir.host,
                '-s', uploadDir.nfsPath, '-v', 'vmsupport']
         destDir = os.path.join(os.sep, 'vmfs', 'volumes', 'vmsupport')
      else:
         cmd = ['mkdir', '-p', destDir]

      p = ssh.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
      _, err = p.communicate()
      if p.returncode != 0:
         raise VmSupportError(cmd, err)

      outLog = os.path.join(destDir, "vmsupport.out.log")

      try:
         cmd = ['/bin/vm-support', '-l', '-w', destDir, '>>', outLog, '2>&1']
         p = ssh.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
         timeout = DEFAULT_VMSUPPORT_TIMEOUT
         _, err = p.communicate(timeout=timeout)
         if p.returncode != 0:
            raise VmSupportError(cmd, err)
      except Exception as e:
         raise VmSupportError(cmd, str(e))
      finally:
         if uploadDir.isNfs():
            cmd = ['localcli', 'storage', 'nfs', 'remove', '-v', 'vmsupport']
            p = ssh.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            _, err = p.communicate()
            if p.returncode != 0:
               raise VmSupportError(cmd, err)


#
# Code below is for debugging purposes...
#
if __name__ == "__main__":
   from argparse import ArgumentParser

   desc = ("Generate and collect the vm-support for the given ESX host.")

   disclaimer = ("** DISCLAIMER ** This program is provided to help debugging "
                 "the %s library. It is NOT a production tool. CLI "
                 "may change at any time." % os.path.basename(__file__))

   parser = ArgumentParser(description=desc, epilog=disclaimer)
   parser.add_argument(type=str, metavar='HOSTNAME', dest='hostname',
                       help='ESX hostname')
   parser.add_argument(type=str, metavar='DIR', dest='destDir',
                       help='path to vm-support upload directory')
   args = parser.parse_args()

   collectVmSupport(args.hostname, args.destDir)
