# Copyright 2015-2019 VMware, Inc.
# All rights reserved. -- VMware Confidential

"""ESX image deployment lib

Library to deploy pre-built images of ESX against a custom set of boot options
and boot modules.

The library supports both TFTP and HTTP transfer protocols. HTTP is recommended
because it is usually faster, safer (HTTP=TCP, TFTP=UDP), and also because it
scales better on the server side (TFTP servers easily collapse under pressure).
But using HTTP is not always possible, especially if the server that serves the
ESX image does not provide HTTP services. ESX also does not support HTTP-boot on
UEFI machines.

TFTP:
   The NIC firmware emits a PXE request, and downloads Pxelinux from TechoOps
   TFTP server. Once loaded in memory, Pxelinux executes against a configuration
   file which points to mboot.c32. So Pxelinux downloads mboot.c32 from Tech-ops
   TFTP server, and executes it. Mboot.c32 downloads the rest of the ESX boot
   modules via syscalls to the Pxelinux TFTP stack.

HTTP:
   The NIC firmware emits a PXE request, and downloads Pxelinux from Tech-ops
   TFTP server. Once loaded in memory, Pxelinux executes against a configuration
   file which points to iPXE. So Pxelinux downloads iPXE from Tech-ops TFTP
   server, and chainloads to it. iPXE overwrites Pxelinux and takes over the
   rest of the boot process. iPXE loads its configuration file, which points to
   mboot.c32. So iPXE downloads mboot.c32 from Tech-ops HTTP server, and
   executes it. Mboot.c32 downloads the rest of the ESX boot modules via
   syscalls to the iPXE HTTP stack.

This library also supports two different modes of deployment:

Full deployment
   In this mode, everything is copied into the PXE directory. That includes all
   of the ESX boot modules, as well as the bootloader, and anything else that is
   required to boot ESX. This mode of deployment is costly in storage, but has
   the advantage of being extremely simple.

Shallow deployment
   'ESX shallow deployments' allow to reduce the size of an ESX PXE directory
   from hundreds of MB down to just a few KB. In this deployment mode, only
   deployment-specific files are copied into the PXE directory. That is only the
   bootloader configuration files and any boot module that is not part of the
   original ESX image. The original ESX image is not copied in the PXE
   directory, but must be accessible via an HTTP-enabled component cache, so
   that the bootloader configuration files can be modified to refer to the boot
   modules by their HTTP URL in the cache. With this mechanism it is easy to
   point multiple PXE directories with different boot options to the same ESX
   build deliverables. Doing so is basically free on storage and provides
   instant deployment times because it avoids to stage a full copy of the ESX
   build deliverables in each PXE directory. However it is more complex because
   it configures the bootloader to fetch the deployment-specific files from one
   location, and the ESX-generic boot modules from another location.
"""
import os
import shutil

from borautils.busybox import md5sum, wget
from .build import EsxBuild
from .conf import EsxDevConf

STATEFUL_INSTALL_ROOT_PSW = 'ca$hc0w'


class EsxImage(EsxDevConf):
   """A class to load, customize and deploy an ESX image from the publish
   directory of a standard ESX build.
   """

   def __init__(self, esxBuild):
      """Load an ESX image from a standard publish directory.
      """
      self._build = esxBuild
      self._bootOpts = "autoPartitionOSDataSize=8192"
      self._extraBootModules = []
      self._autoInstalledVibs = []
      super().__init__()

   def addBootOpts(self, bootOpts):
      """Add a boot option to this image's boot command line.
      """
      if bootOpts:
         if self._bootOpts:
            self._bootOpts += ' %s' % bootOpts
         else:
            self._bootOpts = bootOpts

   def getBootOpts(self):
      """Return the ESX boot options as a string.
      """
      return self._bootOpts.strip()

   def addBootModule(self, filePath):
      """Add a file to the list of bootloader (mboot) modules.
      """
      self._extraBootModules.append(filePath)

   def addAutoInstalledVib(self, vibName):
      """Set VIB to be auto-installed at ESX startup.
      """
      self._autoInstalledVibs.append(self._build.getVibUrl(vibName))

   @staticmethod
   def _writeIpxeConf(dest, mbootUrl, bootCfgUrl, forceSerialLog=True):
      """Write an iPXE boot script that chainloads mboot.
      """
      opts = (mbootUrl, '-S1 ' if forceSerialLog else '', bootCfgUrl)
      conf = ['#!ipxe',
              'iseq ${platform} efi && goto efi ||',
              'chain %s.c32 %s-c %s BOOTIF=01-${netX/mac:hexhyp}' % opts,
              'exit',
              ':efi',
              'chain %s.efi %s-c %s' % opts]
      with open(dest, 'w') as f:
         print('\n'.join(conf), file=f)

   @staticmethod
   def writePxelinuxConf(deployDir, deployUrl=None, pubDirUrl=None,
                         forceSerialLog=True):
      """Generate pxelinux/iPXE configuration files.

      If an URL is passed in @deployUrl, then Pxelinux is configured to
      chainload iPXE, which in turn chainloads mboot. This allows mboot to
      download the ESX boot modules via the iPXE HTTP stack.

      Otherwise (if @deployUrl is None), Pxelinux is configured to directly
      chainload mboot. This setup is usually slower because it makes mboot
      download the ESX boot modules via the Pxelinux TFTP stack. It is useful
      nevertheless, when the server that serves the ESX image does not provide
      HTTP services.
      """
      pxelinuxDir = os.path.join(deployDir, 'pxelinux.cfg')
      os.makedirs(pxelinuxDir, exist_ok=True)

      pxelinuxConf = os.path.join(pxelinuxDir, 'default')

      conf = ['DEFAULT esx',
              'LABEL esx']

      if deployUrl is not None:
         if pubDirUrl is None:
            pubDirUrl = deployUrl
         mbootUrl = os.path.join(pubDirUrl, 'mboot')
         bootCfgUrl = os.path.join(deployUrl, 'boot.cfg')
         ipxeConf = os.path.join(deployDir, 'ipxe.conf')
         EsxImage._writeIpxeConf(ipxeConf, mbootUrl, bootCfgUrl, forceSerialLog)
         conf += ['KERNEL ipxe-undionly.0',
                  'EFI KERNEL ipxe-snponly.efi']

      else:
         tftpRoot = os.path.abspath(deployDir)
         mboot = os.path.join(tftpRoot, 'mboot.c32')
         bootCfg = os.path.join(tftpRoot, 'boot.cfg')
         conf += ['KERNEL %s' % mboot,
                  'APPEND %s-c %s' % ('-S1 ' if forceSerialLog else '',
                                      bootCfg),
                  'IPAPPEND 2']

      with open(pxelinuxConf, 'w') as f:
         print('\n'.join(conf), file=f)

   @staticmethod
   def writeBootCfg(dest, modules, title, prefix=None, bootOpts=None,
                    timeoutInSec=0):
      """Generate mboot's configuration file
      """
      conf = ['title=%s' % title]

      if prefix is not None:
         conf.append('prefix=%s' % prefix)

      if bootOpts:
         conf.append('kernelopt=%s' % bootOpts)

      conf += ['kernel=%s' % modules[0],
               'modules=%s' % ' --- '.join(modules[1:]),
               'timeout=%u' % timeoutInSec]

      with open(dest, 'w') as f:
         print('\n'.join(conf), file=f)

   def _deployFile(self, src, dst):
      """Deploy (copy) @src to its new location @dst in the deploy directory.

      Also lookup the MD5sums table to see if a checksum has been published (in
      MD5SUM.txt) for this deliverable. If an MD5sum is found, then @dst is
      checked against this MD5sum.
      """
      srcMd5 = self._build.getDeliverableMd5(src)

      if src.startswith("http://"):
         dstMd5 = wget(src, dst, (srcMd5 is not None))
      else:
         if os.path.exists(dst) and os.path.samefile(src, dst):
            return
         shutil.copy(src, dst)
         if srcMd5 is not None:
            dstMd5 = md5sum(dst)

      if srcMd5 is not None and srcMd5 != dstMd5:
         raise Exception('%s: MD5 mismatch (got: %s, expected: %s)' %
                         (dst, srcMd5, dstMd5))

      if os.path.isdir(dst):
         return os.path.join(dst, os.path.basename(src))
      else:
         return dst

   def _deployConfTardisk(self, deployDir, deployUrl=None, shallow=False):
      """Generate and deploy the ESX configuration tardisk.
      """
      for vib in self._autoInstalledVibs:
         if shallow:
            vibUrl = vib
         else:
            vibName = os.path.basename(vib)
            vibUrl = os.path.join(deployUrl, 'vibs', vibName)
         self.addBootVib(vibUrl)

      tarDisk = os.path.join(deployDir, 'testesxc.tgz')
      self.writeConfTardisk(tarDisk)
      if deployUrl and shallow:
         self.addBootModule(os.path.join(deployUrl, os.path.basename(tarDisk)))
      else:
         self.addBootModule(tarDisk)

   def writeKsCfg(self, dst, bootOpts, extraBootModules):
      """Generate kickstart config file to automate installation.
         It copies from a template ks.cfg (located in the same directory of
         this script. Only valid when boot media is pxe-installer.
         @param dst: the target path of kickstart configuration.
         @param bootOpts: the boot opts to add to the new esx
         @param extraBootModules: the boot modules to add to the new esx.
      """
      ksCfgDir = os.path.dirname(dst)
      os.makedirs(ksCfgDir, exist_ok=True)
      src = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'ks-template.cfg')
      catExtraBootModules = ''
      if extraBootModules:
         for bootModule in extraBootModules:
            catExtraBootModules = catExtraBootModules + bootModule + ' '

      with open(dst, "wt") as fout:
         with open(src, "rt") as fin:
            for line in fin:
               line = line.replace('@extraBootModules', catExtraBootModules)\
                          .replace('@bootOpts', bootOpts)\
                          .replace('@rootpw', STATEFUL_INSTALL_ROOT_PSW)
               fout.write(line)

   def pxeDeploy(self, deployDir, deployUrl=None, title="ESX", shallow=False,
                 isInstaller=False, forceSerialLog=True, configureTarDisk=True,
                 autoPartitionOnlyOnceAndSkipSsd=False):
      """Deploy this ESX image into the given PXE directory.
      """
      os.makedirs(deployDir, exist_ok=True)

      # For stateful customer install, configuration tardisk not required. Skip
      if configureTarDisk:
         self._deployConfTardisk(deployDir, deployUrl=deployUrl, shallow=shallow)

      bootModules = self._build.getBootModules(isInstaller=isInstaller)

      if not shallow:
         bootloaderModules = self._build.getBootloaderModules()
         for module in bootloaderModules + bootModules + self._extraBootModules:
            dest = os.path.join(deployDir, os.path.basename(module))
            self._deployFile(module, dest)

         bootVibs = self._build.getBootVibs()
         if bootVibs:
            vibDir = os.path.join(deployDir, 'vibs')
            os.makedirs(vibDir, exist_ok=True)
            for vib in bootVibs:
               dst = os.path.join(vibDir, os.path.basename(vib))
               self._deployFile(vib, dst)

         pubDirUrl = None
      else:
         pubDirUrl = self._build.pxeCache

      allBootModules = [os.path.basename(mod) for mod in bootModules]
      for mod in self._extraBootModules:
         if not shallow:
            mod = os.path.basename(mod)
         elif os.path.isabs(mod):
            relPath = os.path.relpath(mod, deployDir)
            if not relPath.startswith(os.pardir):
               mod = relPath
         allBootModules.append(mod)

      # Installer boot should not perform any autoPartition
      bootOpts = ''
      if not isInstaller:
         bootOpts = "autoPartition=TRUE bootUrl=%s" % pubDirUrl
         if autoPartitionOnlyOnceAndSkipSsd:
            bootOpts += " autoPartitionOnlyOnceAndSkipSsd=TRUE"

      if forceSerialLog:
         bootOpts += " vmbLog=TRUE debugLogToSerial=1 logPort=com1"
      self.addBootOpts(bootOpts)

      if isInstaller:
         # generate ks.cfg, perform autoPartition for the post-install boot to
         # create datastore(s) on additional disk(s).
         ksOpts = ' autoPartition=TRUE'
         if autoPartitionOnlyOnceAndSkipSsd:
            ksOpts += ' autoPartitionOnlyOnceAndSkipSsd=TRUE'
         self.writeKsCfg(dst=os.path.join(deployDir, 'ks.cfg'),
                         bootOpts=self.getBootOpts() + ksOpts,
                         extraBootModules=self._extraBootModules)

         # generate installer's boot.cfg
         self.addBootOpts("runweasel ks=%s" % os.path.join(deployUrl, 'ks.cfg'))
         self.writeBootCfg(os.path.join(deployDir, 'boot.cfg'), bootModules,
                           title, prefix=pubDirUrl, bootOpts=self.getBootOpts())
      else:
         self.writeBootCfg(os.path.join(deployDir, 'boot.cfg'), allBootModules,
                           title, prefix=pubDirUrl, bootOpts=self.getBootOpts())
      self.writePxelinuxConf(deployDir, deployUrl=deployUrl,
                             pubDirUrl=pubDirUrl)


def deployEsxImage(pubDir, deployDir, deployUrl=None, pxeRecord=None,
                   bootVibs=None, ipConf=None, nfsMounts=None, shallow=False,
                   extraBootOpts=None, configureTarDisk=True):
   """Wrapper function to download and customize an ESX image.

   @pubDir: path to the directory where the build deliverables have been
            published.
   @pxeRecord: path to a PXE record file.
   @deployDir: path to the directory where to deploy the ESX image.
   @deployUrl: HTTP url pointing to @deployDir.
   @bootVibs: list of VIBs to be automatically installed.
   @ipConf: IP configuration dictionary.
   @nfsMounts: List of NFS shares to be mounted at ESX startup.
   @extraBootOpts: extra boot options
   @configureTarDisk: Configure tar disks if not a customer build. Later
                      consumed by pxeDeploy()
   """
   esxBuild = EsxBuild(pubDir, pxeRecord)
   image = EsxImage(esxBuild)

   os.makedirs(deployDir, exist_ok=True)

   if ipConf:
      image.setHostIp(ipConf)

   if nfsMounts:
      for host, share, mountDir, options in nfsMounts:
         image.setNfsMount(host, share, mountDir, options)

   if bootVibs:
      for name in bootVibs:
         image.addAutoInstalledVib(name)

   image.addBootOpts(extraBootOpts)

   image.pxeDeploy(deployDir, deployUrl=deployUrl, shallow=shallow,
                   configureTarDisk=configureTarDisk)
