# Copyright 2016-2020 VMware, Inc.
# All rights reserved. -- VMware Confidential

"""Setup, boot, and monitor and ESX host.

This library allows to setup and boot an ESX host, and then monitor the status
of this host. The library also provides a simple mechanism to register callback
functions to be triggered when an ESX status change occurs.
"""
import os
import re
import sys
import ssl
from random import randint
from threading import RLock, Timer

from borautils.ssh import SSH, SshCmdGenericError
from borautils.sshkeys import getEsxKeyFile
from borautils.hwdbapi import CatHwdbMachine, HwdbApiRequestError
from borautils.ipmi import IPMI
from borautils.infra.nimbus import NimbusEsxVm, NimbusError
from borautils.infra.pxe import pxeConfig, pxeReset
from borautils.telnet import TelnetLog

try:
   from pyVmomi import Vim as vim
   from pyVim import vm, vmconfig, connect
except ImportError:
   # Nasty path business to import pyVmomi and pyVim from build directory
   try:
      from borautils.bora import Bora
      from borautils.infra.toolchain import loadTool

      bora = Bora()
      vmtree = bora.vmtree
      buildType = bora.buildType

      sys.path.append(os.path.join(vmtree,
                      'build/build/LIBRARIES/vmodl/generic/%s' % buildType))
      sys.path.append(os.path.join(vmtree, 'vim/py'))
      sys.path.append(loadTool('noarch/six-1.9.0/lib/python2.7/site-packages/'))

      from pyVmomi import Vim as vim
      from pyVim import vm, vmconfig, connect
   except Exception:
      # Failure is ignored in case that this is running in an environment
      # that does not have the bora tree. In that case, we won't be able
      # to use the 'vm' host type.
      pass

ESXHOST_STATE = (
   IDLE,
   CONFIGURING_PXE,
   POWERING_ON,
   BOOTLOADER_STARTED,
   BOOTLOADER_DONE,
   VMKBOOT_STARTED,
   VMKBOOT_DONE,
   VMKERNEL_STARTED,
   VMKERNEL_DONE,
   TESTESX_CONF_FAILED,
   GOT_IP_ADDR,
   GOT_BLD_NUM,
   BOOT_SUCCESSFUL,
   LOADESX_SHUTDOWN_START,
   LOADESX_SHUTDOWN_DONE,
   LOADESX_LAUNCH_ESX,
   ESX_INSTALLATION_START,
   ESX_INSTALLATION_DONE,
   ESX_STARTED,
   UW_COREDUMP,
   ESX_SHUTDOWN_STARTED,
   ESX_SHUTDOWN_DONE,
   PSOD,
   PSOD_COREDUMP_FAILED,
   PSOD_COREDUMP_STARTED,
   PSOD_COREDUMP_DONE
) = list(range(26))

BOOT_TIMEOUT = 1000
SHUTDOWN_TIMEOUT = 1001


class _EsxHost(object):
   """Helper class to monitor the status of an ESX host.

   This class provides helper functions to infer the status of an ESX host
   from the content of the serial logs.
   """
   bootloaderBiosStartRe = re.compile(r"COM32 v\d+\.\d+ \(\w+\)")
   bootloaderUefiStartRe = re.compile(r"UEFI v\d+\.\d+")
   bootloaderEndRe = re.compile(r"Relocating modules and starting up "
                                r"the kernel\.\.\.")
   vmkBootRe = re.compile(r"VMB_")
   ipv4Pattern = r"\.".join([r"\d{1,3}"] * 4)
   vmkIpPattern = r"(?P<ip>" + ipv4Pattern + r")"
   vmkIpEarlyRe = re.compile(r'"HostIPAddr" = "' + vmkIpPattern +
                             r'", Old value: "' + ipv4Pattern +
                             r'".*Status: 0x0')
   vmkIpLateRe = re.compile(r"ESX vmk0 IPv4 address is " + vmkIpPattern)
   uwCoreDumpRe = re.compile(r'Dumping cartel \d+ \(from world \d+\) to '
                             r'file (?P<path>.*/(?P<app>[^/]+)-zdump\.\d+)')
   haltPCPURe = re.compile(r'Halting PCPU \d+\.')
   testEsxConfFailedRe = re.compile(r'Failed to config testesx')
   vmkBldNumRe = re.compile(r'vmkernel build number = (\d+)', re.I)

   DEFAULT_FIRMWARE_TIMEOUT = None
   DEFAULT_ESX_START_TIMEOUT = 15 * 60
   DEFAULT_SHUTDOWN_TIMEOUT = 10 * 60
   ESX_SHUTDOWN_DONE_MSG = "This system has been halted"
   SSH_CMD_KEEPALIVE_DEFAULT = 30
   DEFAULT_ESX_INSTALL_TIMEOUT = 15 * 60

   def __init__(self):
      if self.DEFAULT_FIRMWARE_TIMEOUT is None:
         raise NotImplementedError("%s.DEFAULT_FIRMWARE_TIMEOUT abstract "
                                   "attribute must be overridden" %
                                   self.__class__.name)
      self._status = None
      self._statusCB = None
      self._statusCBLock = RLock()
      self._ip = None
      self._crashLog = None
      self._uwCoreDumps = []
      self._timers = {"boot": None, "install": None, "shutdown": None}
      self._bootTimeout = self.DEFAULT_FIRMWARE_TIMEOUT +\
                          self.DEFAULT_ESX_START_TIMEOUT
      self._shutdownTimeout = self.DEFAULT_SHUTDOWN_TIMEOUT
      self._exitWatchLoop = False
      self._stateful_install_complete = False

   def setStatusChangeCB(self, callback):
      """Register a status change callback function.

      The callback is triggered upon ESX status changes to perform any custom
      action that is relative to entering a new ESX state. Because the callback
      is run in the context of the serial log monitor, it must not block or take
      too long to run. The callback is invoked with two parameters: a reference
      to this ESX instance, and the new ESX status: callback(self, newStatus).
      """
      self._statusCB = callback

   def getBootTimeout(self):
      """Return this host's boot timeout.
      """
      return self._bootTimeout

   def setBootTimeout(self, timeout):
      """Set this host's boot timeout. Unit: second.
      """
      self._bootTimeout = timeout

   def getShutdownTimeout(self):
      """Return this host's shutdown timeout.
      """
      return self._shutdownTimeout

   def setShutdownTimeout(self, timeout):
      """Set this host's shutdown timeout.
      """
      self._shutdownTimeout = timeout

   def _onBootTimeoutCB(self):
      """Boot timeout callback.
      """
      self._updateStatus(BOOT_TIMEOUT)

   def _onShutdownTimeoutCB(self):
      """Called back when shutdown process timeout.
      """
      self._updateStatus(SHUTDOWN_TIMEOUT)

   def _cancelTimer(self, timer):
      """Cancel the timer if it exists.
      """
      if timer in self._timers and self._timers[timer] is not None:
         self._timers[timer].cancel()

   def _startTimer(self, timer, interval, callback):
      """Cancel the _bootTimer if it exists.
         Start the _bootTimer with the timeout and callback.
      """
      self._cancelTimer(timer)
      self._timers[timer] = Timer(interval, callback)
      self._timers[timer].start()

   def _updateTimer(self, status):
      """Update timer to the given status.
      """
      if status == POWERING_ON:
         self._startTimer("boot", self.getBootTimeout(), self._onBootTimeoutCB)
      elif status == ESX_STARTED:
         self._cancelTimer("boot")
      elif status == ESX_SHUTDOWN_STARTED:
         self._startTimer("shutdown", self.getShutdownTimeout(),
                          self._onShutdownTimeoutCB)
      elif status == ESX_SHUTDOWN_DONE:
         self._cancelTimer("shutdown")
      elif status == ESX_INSTALLATION_START:
         self._cancelTimer("boot")
         self._startTimer("install", self.DEFAULT_ESX_INSTALL_TIMEOUT,
                          self._onBootTimeoutCB)
      elif status == ESX_INSTALLATION_DONE:
         self._cancelTimer("install")
         self._startTimer("boot", self.getBootTimeout(), self._onBootTimeoutCB)

   def _updateStatus(self, status):
      """Update this ESX host to the given status.
      """
      self._updateTimer(status)
      # Lock to avoid races between the serial monitor and the bootTimer thread.
      with self._statusCBLock:
         if status != self._status:
            if status != UW_COREDUMP and status != BOOT_TIMEOUT:
               self._status = status
            if self._statusCB is not None:
               self._statusCB(self, status)

   def _parseVmkLog(self, log):
      """Parse a vmkernel log line, and deduce ESX status.
      """
      if self.bootloaderBiosStartRe.search(log) is not None or\
         self.bootloaderUefiStartRe.search(log) is not None:
         self._updateStatus(BOOTLOADER_STARTED)

      elif self.bootloaderEndRe.match(log) is not None:
         self._updateStatus(BOOTLOADER_DONE)

      elif self.vmkBootRe.match(log) is not None:
         self._updateStatus(VMKBOOT_STARTED)

      elif (self._status >= VMKBOOT_STARTED and
            "Alive and kickin, now jumping to " in log):
         self._updateStatus(VMKBOOT_DONE)

      elif self._status >= VMKBOOT_STARTED and log.startswith("TSC: 0 cpu0:"):
         self._updateStatus(VMKERNEL_STARTED)

      elif (self._status >= VMKERNEL_STARTED and
            'VMKernel loaded successfully' in log):
         self._updateStatus(VMKERNEL_DONE)

      elif self._status >= VMKERNEL_DONE and 'Boot Successful' in log:
         self._updateStatus(BOOT_SUCCESSFUL)

      elif (self._status >= VMKERNEL_DONE and
            'ESX IS READY FOR TESTING' in log):
         self._updateStatus(ESX_STARTED)

      elif (self._status == ESX_SHUTDOWN_STARTED and
            self.ESX_SHUTDOWN_DONE_MSG in log):
         self._updateStatus(ESX_SHUTDOWN_DONE)

      elif "LoadESX shutdown" in log:
         self._updateStatus(LOADESX_SHUTDOWN_START)

      elif (self._status >= LOADESX_SHUTDOWN_START and
            "LoadESX in progress" in log):
         self._updateStatus(LOADESX_SHUTDOWN_DONE)

      elif (self._status >= LOADESX_SHUTDOWN_DONE and
            "Launch ESX" in log):
         self._updateStatus(LOADESX_LAUNCH_ESX)

      elif self._status >= BOOT_SUCCESSFUL and 'ESX Installation Start' in log:
         self._updateStatus(ESX_INSTALLATION_START)

      elif (self._status >= ESX_INSTALLATION_START and
            'ESX Installation Done' in log):
         self._updateStatus(ESX_INSTALLATION_DONE)
         self._stateful_install_complete = True

      elif any(msg in log for msg in ['@BlueScreen',
                                      'Debugger waiting',
                                      'Panic. Halting.',
                                      'Panic: ASSERT',
                                      'Panic: vPanicInt',
                                      'Recursive panic depth too large',
                                      'Waiting for debugger...',
                                      ]):
         self._crashLog = log.strip()
         self._updateStatus(PSOD)

      elif self._status >= PSOD and 'Starting dump' in log:
         self._updateStatus(PSOD_COREDUMP_STARTED)

      elif (self._status >= PSOD and
            any(msg in log for msg in ['DiskDump: Successful.',
                                       'FileDump: Successful.',
                                       'NetDump: Successful.',
                                       'VsanDump: Successful.'])):
         self._updateStatus(PSOD_COREDUMP_DONE)

      elif (self._status >= PSOD and
            self._status < PSOD_COREDUMP_DONE and
            self.haltPCPURe.match(log) is not None):
         self._updateStatus(PSOD_COREDUMP_FAILED)

      else:
         for vmkIpRe in (self.vmkIpEarlyRe, self.vmkIpLateRe):
            if self._ip is None:
               m = vmkIpRe.search(log)
               if m is not None:
                  self._ip = m.group('ip')
                  self._updateStatus(GOT_IP_ADDR)

         m = self.testEsxConfFailedRe.search(log)
         if m is not None:
            self._updateStatus(TESTESX_CONF_FAILED)

         m = self.uwCoreDumpRe.search(log)
         if m is not None:
            self._uwCoreDumps.append((m.group('app'), m.group('path')))
            self._updateStatus(UW_COREDUMP)

         m = self.vmkBldNumRe.search(log)
         if m is not None:
            self._bldNum = m.group(1)
            self._updateStatus(GOT_BLD_NUM)

   def getBldNum(self):
      """Return the build number from the ESX running on the host.
      """
      return self._bldNum

   def getIp(self):
      """Return the management NIC IP address for this ESX.
      """
      return self._ip

   def setIp(self, ip):
      """Set the management NIC IP address for this ESX.
      """
      self._ip = ip

   def isEsxStarted(self):
      """Return True if ESX has started to boot.
      """
      if self._status is None:
         return False
      return self._status >= BOOTLOADER_STARTED

   def getUWCoredumps(self):
      """Return the list of UW coredumps for this host.
      """
      return self._uwCoreDumps

   def isPsoded(self):
      """Return True if this host has PSOD'ed.
      """
      if self._status is None:
         return False
      return self._status >= PSOD and self._status <= PSOD_COREDUMP_DONE

   def isDiskDumped(self):
      """Return True if this host generated disk dump.
      """
      if self._status is None:
         return False
      return self._status == PSOD_COREDUMP_DONE

   def getCrashLog(self):
      """Return the line in the vmkernel logs which contains the PSOD string.
      """
      return self._crashLog

   def getPxeLocation(self):
      """Return the pxe location for this ESX host.
      """
      return self._location

   def pxeCleanup(self):
      """Cleanup PXE configurations

      If hosts are stateful installed, cleanup PXE so that they won't boot from
      PXE anytime
      """
      for mac in self._macAddrs:
         pxeReset(mac, self._location)

   def _pxeConfig(self, pxeDir, ipxeConf=None):
      """Setup PXE server for this ESX host.

      Servers usually have more than one NIC. Since we don't know which one is
      the boot NIC, let's configure Tech-ops PXE server for all of them.
      """
      self._updateStatus(CONFIGURING_PXE)
      for mac in self._macAddrs:
         pxeConfig(mac, self._location, pxeDir, attempts=3, ipxeConf=ipxeConf)

   def pxeBoot(self, pxeDir, ipxeConf=None):
      """(Re-)boot this ESX host over (i)PXE.
      """
      raise NotImplementedError("%s.pxeboot() abstract method must be "
                                "overridden" % self.__class__.name)

   def _readTelnetLogs(self):
      """Iterate over each line in the telnet serial logs.
      """
      serial = TelnetLog(self._serialHost, self._serialPort)
      while not self._exitWatchLoop:
         try:
            log = serial.readline()
         except EOFError:
            break
         yield log
      serial.close()

   def getSerialReader(self):
      """Get ESX host serial port reader.
      """
      raise NotImplementedError("%s.getSerialReader() abstract method must be "
                                "overridden" % self.__class__.name)

   def startSerialMonitor(self, logPath=None):
      """Monitor this ESX's serial log.

      In a loop which iterates over each line in the serial port output, read a
      line and parse it. If a valid file path is passed in @logPath, then the
      serial output is also written line-by-line into this file. This function
      does not return until the serial connection is dropped by the remote ESX,
      or until another thread calls stopSerialMonitor().
      """
      self._exitWatchLoop = False
      f = open(logPath, 'wb', 0) if logPath is not None else None
      reader = self.getSerialReader()
      for log in reader():
         log = log.strip('\r')
         if f is not None:
            f.write(log.encode('utf-8'))
         self._parseVmkLog(log)

      if f is not None:
         f.close()

   def stopSerialMonitor(self):
      """Quit monitoring this machine's serial log.
      """
      self._exitWatchLoop = True

   def kill(self):
      """Clean up ESX
      """
      pass

   def _runSshCmd(self, cmd, timeout=0):
      """Run ssh command on this host.

      Return tuple (status, stdout, stderr) or raise SshCmdTimeoutError if
      cmd hasn't started in 'timeout' seconds.
      """
      if self.getIp() is None:
         raise Exception("Failed to issue SSH command (No Host IP)")
      with SSH(self.getIp(), username='root', key=getEsxKeyFile(),
               keepalive=self.SSH_CMD_KEEPALIVE_DEFAULT) as ssh:
         p = ssh.Popen(cmd)
         out, err = p.communicate(timeout=timeout)
         return (p.returncode, out, err)

   def shutdownAsync(self, timeout=0):
      """Shutdown host asynchronously or raise SshCmdTimeoutError if shutdown
      hasn't started in 'timeout' seconds.

      It runs /bin/poweroff by creating a SSH connection to the host by IP
      address. It's an async operation because poweroff command will triggers
      the shutdown process and return immediately.
      """
      cmd = "/bin/poweroff"
      status, out, err = self._runSshCmd(cmd, timeout=timeout)
      if status != 0:
         raise SshCmdGenericError("Failed to shutdown host",
                                  cmd, status, str(err), str(out))
      self._updateStatus(ESX_SHUTDOWN_STARTED)

   def restoreTestesxTgz(self):
      """Testesxc.tgz is the midwife of both testesxc.vib and testesx.vib.
      It will be cleared after any vib installation.
      Copy it back from another bootbank and add it into boot.cfg.
      """
      cmd = ('cur=/bootbank/;'
             '[ -f $cur/testesxc.tgz ] && cur=altbootbank;'
             'bcfg=$cur/boot.cfg;'
             'if test -f $bcfg && ! grep -q -- "--- testesxc.tgz" $bcfg; then'
             '   cp /*bootbank/testesxc.tgz $cur;'
             '   sed "/^modules=/ s,$, --- testesxc.tgz," -i $bcfg;'
             'fi'
            )
      with SSH(self.getIp(), username='root', key=getEsxKeyFile(),
               keepalive=self.SSH_CMD_KEEPALIVE_DEFAULT) as ssh:
         ssh.run(cmd, socketTimeout=60)


class EsxPhysHost(_EsxHost):
   """A class to setup, boot, and monitor an ESX physical host.

   Implementation of the _EsxHost() class that relies on live-parsing the
   vmkernel logs over a Telnet session. It is specifically designed to control
   and monitor lab-class hosts, i.e. ilo-enabled physical machines which are
   also connected to a serial port concentrator.
   """
   DEFAULT_FIRMWARE_TIMEOUT = 10 * 60

   def __init__(self, ilo, iloUser, iloPass, location, macAddrs,
                ipAddrs=None, serialHost=None, serialPort=None):
      def _sanitizeStr(inputStr):
         if inputStr is not None:
            inputStr = inputStr.strip()
         return inputStr if inputStr else None

      super().__init__()
      self._ilo = _sanitizeStr(ilo)
      self._iloUser = _sanitizeStr(iloUser)
      self._iloPass = _sanitizeStr(iloPass)
      self._location = _sanitizeStr(location)
      self._serialHost = _sanitizeStr(serialHost)
      self._serialPort = serialPort
      self._macAddrs = macAddrs
      self._ipAddrs = ipAddrs
      if self._macAddrs:
         self._macAddrs = [_sanitizeStr(mac) for mac in self._macAddrs
                           if _sanitizeStr(mac) is not None]

   def reset(self, forcePXE=False):
      """Hard reset this machine via IPMI.

      If the machine is not running, the machine is powered on instead
      of triggering a reset.
      """
      chassis = IPMI(self._ilo, self._iloUser, self._iloPass)
      self._updateStatus(POWERING_ON)
      if forcePXE:
         chassis.setPxeBoot()

      if chassis.isOn():
         chassis.reset()
      else:
         chassis.powerOn()

   def pxeBoot(self, pxeDir, ipxeConf=None):
      """(Re-)boot this machine over (i)PXE.
      """
      self._pxeConfig(pxeDir, ipxeConf=ipxeConf)
      self.reset(forcePXE=True)

   def setDiskBoot(self):
      """Force boot from default hard-drive.
      """
      chassis = IPMI(self._ilo, self._iloUser, self._iloPass)
      chassis.setDiskBoot()

   def getMachineIp(self):
      """Return physical machine's current IP address
      """
      return self._ipAddrs

   def _readSolLogs(self):
      """Iterate over each line in the SOL (IPMI Serial-Over-Lan) serial logs.
      """
      chassis = IPMI(self._ilo, self._iloUser, self._iloPass)
      chassis.solActivate()
      while not self._exitWatchLoop:
         try:
            log = chassis.solReadline()
            if log is None:
               continue
         except EOFError:
            break
         yield log
      chassis.solDeactivate()

   def __str__(self):
      """ESX physical hosts are identified by their ILO hostname.
      """
      return self._ilo

   def getSerialReader(self):
      """Get ESX physical host serial port reader.
      """
      if self._serialHost is None or self._serialHost == self._ilo:
         return self._readSolLogs
      else:
         return self._readTelnetLogs


class EsxPhyHwdbHost(EsxPhysHost):
   """A class to setup, boot, and monitor an ESX physical host
      by reserving machine from hardware DB.
   """
   def __init__(self, reserveUser, location):
      self._machine = CatHwdbMachine.reserveMachine(reserveUser, location)
      info = self._machine.getMachineInfo()
      super().__init__(info['ilo_hostname'], info['ilo_username'],
                       info['ilo_password'], info['suite_location'],
                       info['macAddrs'], serialHost=info['serial_host'],
                       serialPort=info['serial_port'])

   def kill(self):
      """Release machine back to hardware DB.
      """
      super().kill()
      self._machine.releaseMachine()


class EsxNimbusHost(_EsxHost):
   """A class to setup, boot, and monitor an ESX nimbus host.
   """
   DEFAULT_FIRMWARE_TIMEOUT = 5 * 60

   def __init__(self, name=None, cpus=None, memory=None, disk=None, nics=None,
                logFile=None, annotations=None, specFile=None, lease=None):
      super().__init__()
      self._nimbusVM = NimbusEsxVm(name=name, cpus=cpus, memory=memory,
                                   disk=disk, nics=nics, logFile=logFile,
                                   annotations=annotations, specFile=specFile,
                                   lease=lease)
      self._location = self._nimbusVM.location
      self._created = False

   def pxeBoot(self, pxeDir, ipxeConf=None):
      """(Re-)boot this nimbus vm over (i)PXE, the vm is created if necessary.
      """
      if not self._created:
         self._nimbusVM.create()
         self._created = True
         self._macAddrs = self._nimbusVM.macs
      self._pxeConfig(pxeDir, ipxeConf=ipxeConf)
      self._updateStatus(POWERING_ON)
      self._nimbusVM.powerOn()

   def reset(self, forcePXE=False):
      """Hard reset this nimbus VM.
      """
      self._updateStatus(POWERING_ON)
      self._nimbusVM.reset()

   def kill(self):
      """Clean up ESX nimbus vm.
      """
      self._nimbusVM.kill()

   def __str__(self):
      """ESX nimbus hosts are identified by their name.
      """
      return self._nimbusVM.__str__()

   def getSerialReader(self):
      """Get ESX nimbus host serial port reader.
      """
      return self._nimbusVM.readSerial

   def stopSerialMonitor(self):
      """Quit monitoring this machine's serial log.
      """
      super().stopSerialMonitor()
      self._nimbusVM.killReadSerialProcess()

   def setDiskBoot(self):
      """Force boot from default hard-drive.
      """
      self._nimbusVM.bootOrder(order='hdd')


class EsxVirtHost(_EsxHost):
   """A class to setup, boot, and monitor an ESX host as a VM on another ESX.
      The VM is created with a single vmxnet3 NIC, a single pvscsi controller,
      and a single disk. Default disk size capacity, memory size, and number
      of PCPUs are 12GiB, 6GiB, and 2, respectively.
      Assumes that the firewall on the parent ESX host has been disabled, or
      has ports 10000-40000 open and available. These are used for serial ports
      for the test ESX VMs.
   """
   DEFAULT_FIRMWARE_TIMEOUT = 5 * 60

   _defaultNumVcpus = 2
   _defaultMemSizeInMb = 6 * 1024
   _defaultDiskCapacityInKb = 12 * 1024 * 1024

   def __init__(self, host, user, password='', port='443', datastore=None,
                name=None, cpus=None, memory=None, disk=None,
                logFile=None, location=None, annotation='', hostType=None):
      super().__init__()

      if name is None:
         raise Exception('name is not specified')

      self._name = name
      self._serialHost = host
      self._serialPort = randint(10000, 40000)
      self._location = location

      sslContext = None
      if hasattr(ssl, '_create_unverified_context'):
         sslContext = ssl._create_unverified_context()

      self._connect = connect.SmartConnect(host=host, user=user, pwd=password,
                                           port=int(port), sslContext=sslContext)

      if memory is None:
         memory = self._defaultMemSizeInMb
      if cpus is None:
         cpus = self._defaultNumVcpus
      if disk is None:
         disk = self._defaultDiskCapacityInKb

      spec = vm.CreateQuickDummySpec(name,
                                     guest='vmkernel6Guest',
                                     memory=memory,
                                     vmxVersion='vmx-14',
                                     datastoreName=datastore)
      spec.numCPUs = cpus
      spec.SetAnnotation(annotation)
      spec.SetNestedHVEnabled(True)

      cnxInfo = vim.Vm.Device.VirtualDevice.ConnectInfo(connected=True,
                                                        startConnected=True,
                                                        allowGuestControl=True)
      vmconfig.AddNic(cspec=spec, nicType='vmxnet3', conInfo=cnxInfo)

      vmconfig.AddScsiCtlr(cspec=spec, ctlrType="pvscsi")
      vmconfig.AddScsiDisk(cspec=spec, thin=True, capacity=disk)

      serialURI = "tcp://%s:%d" % (self._serialHost, self._serialPort)

      vmconfig.AddURIBackedSerial(cspec=spec,
                                  serviceURI=serialURI,
                                  direction="server")

      self._vm = vm.CreateFromSpec(spec)

      # must power on the vm to initialize its mac address
      vm.PowerOn(self._vm)
      vm.PowerOff(self._vm)

      self._macAddrs = []
      for d in self._vm.config.hardware.device:
         try:
            self._macAddrs += [d.macAddress]
         except Exception as e:
            pass

      if len(self._macAddrs) == 0:
         raise NotImplementedError('Unable to identify mac address for %s' %
               name)

   def pxeBoot(self, pxeDir, ipxeConf=None):
      """(Re-)boot this vm over (i)PXE.
      """
      self._pxeConfig(pxeDir, ipxeConf=ipxeConf)
      self._updateStatus(POWERING_ON)
      try:
         vm.PowerReset(self._vm)
      except:
         vm.PowerOn(self._vm)
   def reset(self, forcePXE=False):
      """Hard reset this nimbus VM.
      """
      self._updateStatus(POWERING_ON)
      vm.Reset(self._vm)

   def kill(self):
      """Power off and destroy the VM to cleanup.
      """
      try:
         vm.PowerOff(self._vm)
      except:
         pass
      vm.Destroy(self._vm)

   def __str__(self):
      """ESX VM hosts are identified by their name.
      """
      return self._name

   def getSerialReader(self):
      """Get ESX serial port reader, it's just a simple TCP connection
      """
      return self._readTelnetLogs

   def setDiskBoot(self):
      """Force boot from default hard-drive.
      """
      spec = vim.vm.ConfigSpec()
      spec.bootOptions = vim.vm.BootOptions()
      disks = vmconfig.CheckDevice(self._vm.config, vim.vm.device.VirtualDisk)
      bootDisk = vim.vm.BootOptions.BootableDiskDevice(deviceKey=disks[0].key)
      spec.bootOptions.bootOrder = [bootDisk]
      return vm.Reconfigure(self._vm, spec)


def getEsxHost(hostInfo, name, logDir, logger):
   """Return a generic ESX host which matches the given @hostInfo.
   """
   host = None
   hostType = hostInfo['hostType']
   logFile = os.path.join(logDir, '%s-%s-esx.log' % (name, hostType))

   if hostType == 'nimbus':
      vm = hostInfo
      specFile = None
      if vm.get('hostSpec'):
         specFile = vm.get('hostSpec')
      try:
         annotation = vm['annotations'] if 'annotations' in vm else ''
         host = EsxNimbusHost(name=name, logFile=logFile,
                              annotations=annotation, specFile=specFile)
      except NimbusError as e:
         logger.exception('failed to create nimbus vm: %s', e)
         raise Exception('failed to create nimbus vm')

   elif hostType == 'physical':
      physHost = hostInfo
      host = EsxPhysHost(physHost['ilo_hostname'], physHost['ilo_username'],
                         physHost['ilo_password'], physHost['suite_location'],
                         physHost['macAddrs'], physHost['ipAddrs'],
                         serialHost=physHost['serial_host'],
                         serialPort=physHost['serial_port'])

   elif hostType == 'hwdb':
      phyHwdbHost = hostInfo
      try:
         host = EsxPhyHwdbHost(name, phyHwdbHost['location'])
      except HwdbApiRequestError as e:
         logger.exception('failed to reserve machine from hardware DB: %s', e)
         raise Exception('failed to reserve HWDB physical machine')

   elif hostType == 'vm':
      vm = hostInfo
      logger.info("VM SPECS: cpus=%u, memory=%uMB, disk=%uMB",
                  vm['cpus'], vm['memory'], vm['disk'] // 1024)
      try:
         host = EsxVirtHost(name=name, logFile=logFile,
                            annotation='testesx tester', **hostInfo)
      except Exception as e:
         logger.exception('failed to create vm: %s', e)
         raise Exception('failed to create vm')

   else:
      raise NotImplementedError('invalid hostType: %s' % hostType)

   assert host
   logger.info("ESX HOST: %s", host)
   logger.info("PXE SUITE: %s", host.getPxeLocation())
   return host


#
# Code below is for debugging purposes...
#
if __name__ == "__main__":
   from argparse import ArgumentParser
   from tempfile import mkstemp
   from threading import Thread

   def statusCallback(host, status):
      statusMsg = {CONFIGURING_PXE: "Configuring PXE...",
                   POWERING_ON: "Resetting machine %s..." % host,
                   BOOTLOADER_STARTED: "Bootloader started",
                   BOOTLOADER_DONE: "Bootloader done",
                   VMKBOOT_STARTED: "vmkBoot started",
                   VMKBOOT_DONE: "vmkBoot done",
                   VMKERNEL_STARTED: "VMKernel started",
                   VMKERNEL_DONE: "VMKernel loaded successfully",
                   GOT_IP_ADDR: "Got IP address",
                   GOT_BLD_NUM: "Got build number",
                   BOOT_SUCCESSFUL: 'Boot successful',
                   ESX_STARTED: "ESX is ready for testing",
                   PSOD: "PSOD: ESX has crashed :(",
                   PSOD_COREDUMP_FAILED: "PSOD: Failed to generate disk dump",
                   PSOD_COREDUMP_STARTED: "PSOD: Start to generate disk dump",
                   PSOD_COREDUMP_DONE: "PSOD: DiskDump successful"}
      print("=>", statusMsg[status])
      if status == ESX_STARTED:
         print("ESX IP: %s" % host.getIp())
      if status in [ESX_STARTED, PSOD]:
         host.stopSerialMonitor()

   def bootESX(host, pxeDir):
      ipxeConf = os.path.join(pxeDir, 'ipxe.conf')
      host.setStatusChangeCB(statusCallback)
      host.pxeBoot(pxeDir, ipxeConf=ipxeConf)
      host.startSerialMonitor()

   parser = ArgumentParser()
   parser.add_argument('--host', metavar='HOST', type=str, dest='host',
                       help="Host ILO/iDrac address")
   parser.add_argument('--user', metavar='USERNAME', type=str, dest='user',
                       help="Ilo/iDrac username")
   parser.add_argument('--password', metavar='PASSWORD', type=str,
                       dest='password',
                       help="Ilo/iDrac password")
   parser.add_argument('--location', metavar='LOCATION', type=str,
                       dest='location', default="PromB",
                       help="Location of the machine to boot")
   parser.add_argument('--serial-host', metavar='HOST', type=str,
                       dest='serialHost',
                       help="Serial port concentrator telnet host")
   parser.add_argument('--serial-port', metavar='PORT', type=int,
                       dest='serialPort',
                       help="Serial port concentrator telnet port")
   parser.add_argument('--mac', metavar='MAC', type=str, dest='macAddrs',
                       action='append', default=[],
                       help="Host MAC address")
   parser.add_argument('--pxedir', metavar='PXEDIR', type=str, dest='pxeDir',
                       help="Path to your PXE directory")
   parser.add_argument('--nimbus-vm', type=str, metavar='NIMBUS_VMNAME',
                       dest='nimbusVm', help="Name of ESX nimbus vm")

   args = parser.parse_args()

   if args.nimbusVm is None:
      host = EsxPhysHost(args.host, args.user, args.password, args.location,
                         args.macAddrs, serialHost=args.serialHost,
                         serialPort=args.serialPort)
   else:
      logFileFd, logFilePath = mkstemp(prefix='nimbus-esx-vm-', suffix='.log')
      os.close(logFileFd)
      host = EsxNimbusHost(args.nimbusVm, logFile=logFilePath)

   t = Thread(target=bootESX, args=(host, args.pxeDir))
   t.daemon = True
   t.start()
   t.join()
