# Copyright 2015-2018, 2020 VMware, Inc.
# All rights reserved. -- VMware Confidential

"""ESX host management

Automation lib to boot up an ESX host and to get notified when this host is up
and running, or if something went wrong ESX (PSODs, UW crashes). In case of a
crash, this library also handles collecting the vm-support bundle, rebooting the
crashed host if necessary.

TODO:
  More work is required to make this library multi-hosts ready:

   - Pick host-specific filenames for the serial logs / vm-support output, or
     make it configurable.
"""
import os
from threading import Thread, Event, Timer

from borautils.exception import DebugInfoMixIn
from .debugbundle import collectVmSupport
from borautils.esx.host import (
   BOOTLOADER_DONE,
   BOOTLOADER_STARTED,
   BOOT_SUCCESSFUL,
   BOOT_TIMEOUT,
   CONFIGURING_PXE,
   ESX_SHUTDOWN_DONE,
   ESX_SHUTDOWN_STARTED,
   LOADESX_SHUTDOWN_START,
   LOADESX_SHUTDOWN_DONE,
   LOADESX_LAUNCH_ESX,
   ESX_INSTALLATION_START,
   ESX_INSTALLATION_DONE,
   ESX_STARTED,
   GOT_BLD_NUM,
   GOT_IP_ADDR,
   POWERING_ON,
   PSOD,
   PSOD_COREDUMP_DONE,
   PSOD_COREDUMP_FAILED,
   PSOD_COREDUMP_STARTED,
   SHUTDOWN_TIMEOUT,
   TESTESX_CONF_FAILED,
   UW_COREDUMP,
   VMKBOOT_DONE,
   VMKBOOT_STARTED,
   VMKERNEL_DONE,
   VMKERNEL_STARTED,
)

DEFAULT_DISK_DUMP_TIMEOUT = 20 * 60


class EsxSessionError(DebugInfoMixIn, Exception):
   """Infrastructure (non-product) issue.
   """
   _errType = "Infrastructure issue"

   def __init__(self, errStr, hostname):
      super().__init__(errStr)
      self.addDebugInfo('hostname', hostname)

class EsxError(DebugInfoMixIn, Exception):
   """ESX (product) error.
   """
   _errType = "ESX product error"

   def __init__(self, errStr, hostname=None):
      super().__init__(errStr)
      if hostname is not None:
         self.addDebugInfo('hostname', hostname)

class EsxUWCoreDump(EsxError):
   """A userworld application has crashed.
   """
   _errType = "ESX userworld coredump"

   def __init__(self, errStr, app, zdump, hostname):
      super().__init__(errStr, hostname)
      self.addDebugInfo('app', app)
      self.addDebugInfo('zdump', zdump)

class EsxPsod(EsxError):
   """vmkernel has PSOD'ed.
   """
   _errType = "vmkernel PSOD"

   def __init__(self, errStr, psodLog, hostname):
      super().__init__(errStr, hostname)
      self.addDebugInfo('log', psodLog)

class EsxBootTimeout(EsxError):
   """ESX boot timed out.
   """
   _errType = "ESX boot timed out"

   def __init__(self, errStr, timeout, hostname):
      super().__init__(errStr, hostname)
      self.addDebugInfo("timeout", "%s seconds" % timeout)


class EsxShutdownTimeout(EsxError):
   """ESX shutdown timed out.
   """
   _errType = "ESX shutdown timed out"

   def __init__(self, errStr, timeout, hostname):
      super().__init__(errStr, hostname)
      self.addDebugInfo("timeout", "%s seconds" % timeout)

class EsxSession(Thread):
   """Class to configure, boot, and monitor an ESX physical machine.
   """

   def __init__(self, logger, host, pxeDir, logDir=None, killHost=True):
      super().__init__()
      self.daemon = True
      self._logger = None
      self._diskDumpTimer = None
      self._pxeDir = pxeDir
      self._logDir = logDir
      self._killHost = killHost
      self._ipxeConf = None
      if self._pxeDir is not None:
         cfg = os.path.join(self._pxeDir, 'ipxe.conf')
         self._ipxeConf = cfg if os.path.exists(cfg) else None
      self._hasEsxInitialized = False
      self._psodLog = None
      self._needVmSupport = False
      self._collectingVmSupport = False
      self._dumpingThread = None
      self._coreDumps = []
      self._ignoreCoreDumps = set()
      self._terminate = Event()

      self._setLogger(logger)
      self._host = host
      self._host.setStatusChangeCB(self._onStatusChangeCB)

   def _setLogger(self, logger):
      """Configure the logging channel for this ESX session.
      """
      self._logger = logger
      if not hasattr(self._logger, 'psod'):
         setattr(self._logger, 'psod', self._logger.error)

   def onErrorCB(self, error):
      """Misc errors callback, may be overridden in child class.
      """
      pass

   def onPsodCB(self, psodLog):
      """PSOD callback, may be overridden in child class.
      """
      pass

   def _onPsodCB(self, psodLog):
      """Internal PSOD callback.
      """
      if self._psodLog is None:
         self._psodLog = psodLog
         self._diskDumpTimer = Timer(DEFAULT_DISK_DUMP_TIMEOUT,
                                     self._onDiskDumpTimeoutCB)
         self._diskDumpTimer.start()
         self.onPsodCB(psodLog)

   def onUWCoreDumpCB(self, appName, coreDump):
      """UW coredump callback, may be overridden in child class.
      """
      pass

   def _onUWCoreDumpCB(self, host):
      """Internal UW coredump callback.
      """
      appName, coreDump = host.getUWCoredumps()[-1]
      if appName not in self._ignoreCoreDumps:
         self._coreDumps.append((appName, coreDump))
         self.onUWCoreDumpCB(appName, coreDump)
         self._needVmSupport = True

   def onTimeoutCB(self):
      """Timeout callback, may be overridden in child class.
      """
      pass

   def _onTimeoutCB(self):
      """Internal timeout callback.
      """
      if not self._host.isEsxStarted():
         error = EsxSessionError("Failed to start ESX "
                                 "(undiagnosed framework/infra timeout)",
                                 str(self._host))
      else:
         error = EsxBootTimeout("ESX has started, but was too slow to initialize"
                                " (or is it hung?)",
                                self._host.getBootTimeout(), str(self._host))
      self.onErrorCB(error)

   def _onEsxShutdownTimeoutCB(self):
      """ESX shutdown timeout callback.
      """
      self.onErrorCB(EsxShutdownTimeout("ESX shutdown timeout.",
                                        self._host.getShutdownTimeout(),
                                        str(self._host)))

   def onEsxInitializedCB(self):
      """ESX boot completed callback, may be overridden in child class.
      """
      pass

   def onBootSuccessfulCB(self):
      """Boot successful callback.
      """
      pass

   def _onBootSuccessfulCB(self):
      """Internal callback executed once ESX vital services have started.
      """
      self._logger.info("Boot Successful")
      self.onBootSuccessfulCB()

   def _onEsxInitializedCB(self):
      """This internal callback is executed when ESX has finished to initialize.
      """
      if self._collectingVmSupport:
         self._dumpingThread = Thread(target=self._runVmSupportAndExit)
         self._dumpingThread.start()
         return

      if self.getIp() is None:
         noIpError = EsxError("Failed to get ESX IP", str(self._host))
         self.onErrorCB(noIpError)
         return

      if self._host._stateful_install_complete:
         self._host.restoreTestesxTgz()

      self._hasEsxInitialized = True
      self.onEsxInitializedCB()

   def _onEsxShutdownDoneCB(self):
      """Callback when ESX has shutdown.
      """
      self.close()

   def _onStatusChangeCB(self, host, status):
      """ESX status change callback.

      This callback is registered to the underlying EsxPhysHost() status change
      notification system. It is called on every session's status change.
      """
      if status == CONFIGURING_PXE:
         self._logger.info("Configuring PXE against %s...", self._pxeDir)
      elif status == POWERING_ON:
         self._logger.info("Resetting machine %s...", self._host)
      elif status == BOOTLOADER_STARTED:
         self._logger.info("Bootloader started")
      elif status == BOOTLOADER_DONE:
         self._logger.info("Bootloader done")
      elif status == VMKBOOT_STARTED:
         self._logger.info("vmkBoot started")
      elif status == VMKBOOT_DONE:
         self._logger.info("vmkBoot done")
      elif status == VMKERNEL_STARTED:
         self._logger.info("VMKernel started")
      elif status == VMKERNEL_DONE:
         self._logger.info("VMKernel loaded successfully")
      elif status == LOADESX_SHUTDOWN_START:
         self._logger.info("loadESX: Shutdown Start")
      elif status == LOADESX_SHUTDOWN_DONE:
         self._logger.info("loadESX: Shutdown Done")
      elif status == LOADESX_LAUNCH_ESX:
         self._logger.info("loadESX: Launch ESX")
      elif status == ESX_INSTALLATION_START:
         self._logger.info("ESX Installation Start")
      elif status == ESX_INSTALLATION_DONE:
         self._logger.info("ESX Installation Done")
         self._logger.info("Setting boot from hard drive...")
         self._host.setDiskBoot()
         # XXX: DELL R630 can't honor the new boot order
         # without a hard reset. However, this may bring
         # ESX into inconsistent state. The ESX should be able
         # to perform shutdown or restart normally.
         # Filed PR 1724890 to fix this.
         self._host.reset()
      elif status == GOT_IP_ADDR:
         self._logger.info("Got IP address: %s", host.getIp())
      elif status == GOT_BLD_NUM:
         self._logger.info("Got build number: %s", host.getBldNum())
         self.onCheckBldNum(host.getBldNum())
      elif status == TESTESX_CONF_FAILED:
         self.onErrorCB(EsxSessionError('Failed to config testesx', self._host))
      elif status == BOOT_SUCCESSFUL:
         self._onBootSuccessfulCB()
      elif status == ESX_STARTED:
         self._onEsxInitializedCB()
      elif status == UW_COREDUMP:
         self._onUWCoreDumpCB(host)
      elif status == PSOD:
         self._onPsodCB(self._host.getCrashLog())
      elif status == PSOD_COREDUMP_STARTED:
         self._logger.info("Dumping PSOD coredump to disk...")
      elif status == PSOD_COREDUMP_FAILED:
         self._diskDumpTimer.cancel()
         self._logger.error("Failed to dump PSOD coredump to disk")
         self._host.stopSerialMonitor()
      elif status == PSOD_COREDUMP_DONE:
         self._diskDumpTimer.cancel()
         self._logger.info("PSOD coredump successfully dumped to disk")
         self._host.stopSerialMonitor()
      elif status == BOOT_TIMEOUT:
         self._onTimeoutCB()
      elif status == ESX_SHUTDOWN_STARTED:
         self._logger.info("ESX Shutdown Started")
      elif status == ESX_SHUTDOWN_DONE:
         self._logger.info("ESX Shutdown Done")
         self._onEsxShutdownDoneCB()
      elif status == SHUTDOWN_TIMEOUT:
         self._logger.error("ESX Shutdown Timeout")
         self._onEsxShutdownTimeoutCB()
      else:
         return False
      return True

   def onCheckBldNum(self, bldNum):
      """Subclass can override this method to check if the running ESX's
         build number is as expected.
      """
      pass

   def _onDiskDumpTimeoutCB(self):
      """Abort disk dump and bailout.

      This function is called when the disk dump timeout expires.
      """
      self._logger.warning("Disk dump timeout")
      self._host.stopSerialMonitor()

   def setIgnoreCoreDumps(self, appNames):
      """Ignore UW for the given set of apps.
      """
      self._ignoreCoreDumps |= appNames

   def isPsoded(self):
      """Return True if ESX has PSOD'ed.
      """
      return self._host.isPsoded()

   def getPsodError(self):
      """Return an exception that contains the PSOD error log.
      """
      if self._psodLog:
         return EsxPsod("PSOD detected during ESX run", self._psodLog, str(self._host))
      return None

   def getUWCoreDumpErrors(self):
      """Return the list of UW coredumps encountered during this ESX run
      """
      coreDumpErrors = []
      for coreDump in self._coreDumps:
         appName, zdump = coreDump
         error = EsxUWCoreDump('UW coredump detected during ESX run',
                               appName, zdump, str(self._host))
         coreDumpErrors.append(error)
      return coreDumpErrors

   def getIp(self):
      """Return ESX's IP address for this session.
      """
      return self._host.getIp()

   def _runVmSupportAndExit(self):
      """Run vm-support on this host, and terminate this ESX session.
      """
      self._logger.info("Collecting vm-support...")
      try:
         collectVmSupport(self.getIp(), self._logDir)
      except Exception as e:
         self._logger.exception('Failed to collect vm-support: %s', e)
      else:
         self._logger.info("vm-support was collected successfully")
      self._logger.info("Closing ESX session...")
      self._host.stopSerialMonitor()

   def _restart(self, logName=None, pxeDir=None, ipxeConf=None):
      """Restart this ESX session. The ESX host is rebooted.
      """
      try:
         if self._terminate.is_set():
            self._logger.warn("Session is terminated during restarting.")
            return
         if pxeDir:
            self._host.pxeBoot(pxeDir, ipxeConf=ipxeConf)
         else:
            pxeReboot = True
            if self._host._stateful_install_complete:
               pxeReboot = False
               self._logger.info("Reboot ESX from disk...")
               self._host.setDiskBoot()
            self._host.reset(forcePXE=pxeReboot)
      except Exception as e:
         self._logger.error(str(e))
         raise EsxSessionError("Failed to %s ESX host" %
                               'PXE-boot' if pxeDir else 'reset',
                               str(self._host))

      if logName:
         logPath = os.path.join(self._logDir, logName)
         if self._terminate.is_set():
            self._logger.warn("Session is terminated during restarting.")
            return
         try:
            self._host.startSerialMonitor(logPath=logPath)
         except Exception as e:
            self._logger.error(str(e))
            raise EsxSessionError("Failed to start serial log monitor",
                                  str(self._host))

   def _atExit(self):
      """Close this ESX session, but before, collect the vm-support.

      This function is executed on the exit path, after some other thread has
      called EsxSession.close(). It checks whether ESX has PSODed, and if it
      has, it reboots the host and collect the vm-support.
      """
      self._terminate.wait()

      if self.isPsoded():
         if not self._hasEsxInitialized:
            self._logger.warning("Cannot collect vm-support: "
                                 "ESX has failed to initialize")
            return

         if not self._host.isDiskDumped():
            self._diskDumpTimer.cancel()
            self._logger.warning("Failed to generate disk dump")

         self._logger.info('Rebooting host to collect vm-support...')
         try:
            self._collectingVmSupport = True
            self._terminate.clear()
            self._restart(logName='%s-vmkernel-reboot.log' % self._host)
         except Exception:
            self._collectingVmSupport = False
            self._logger.warning("Failed to collect vm-support (ESX host failed"
                                 " to reboot)")
            return

      if self._dumpingThread is not None and self._dumpingThread.is_alive():
         self._dumpingThread.join()

   def run(self):
      """Start this ESX session.

      Configure PXE, reset the ESX physical host, and parse the ESX serial logs
      until someone calls EsxSession.close(), or until a fatal ESX error occurs
      (boot failure, PSOD).
      """
      try:
         self._restart(logName='%s-vmkernel.log' % self._host,
                       pxeDir=self._pxeDir, ipxeConf=self._ipxeConf)
      except Exception as e:
         self.onErrorCB(e)

      self._atExit()
      if self._killHost:
         self._host.kill()
      self._logger.info("Successfully disconnected from ESX host")

   def shutdownHostAsync(self, timeout=60):
      """Shutdown host asynchronously or raise SshCmdTimeoutError if shutdown
      hasn't started in 'timeout' seconds.
      """
      self._host.shutdownAsync(timeout)

   def isHostShuttingDown(self):
      """Return True if the host is shutting down.
      """
      return self._host._status == ESX_SHUTDOWN_STARTED

   def close(self):
      """Terminate this ESX session.
      """
      if not self.isPsoded():
         if self._needVmSupport:
            # ESX is not PSODed, so we must still be listening to the serial
            # logs. Let's schedule a new thread here to run vm-support. This
            # thread will also stop the Serial Log Monitor when it is done
            # collecting the vm-support bundle.
            self._dumpingThread = Thread(target=self._runVmSupportAndExit)
            self._dumpingThread.start()
         else:
            self._host.stopSerialMonitor()

      # Notify EsxSession._atExit() that user has requested to terminate this
      # session
      self._terminate.set()

   def __str__(self):
      """Return a human-readable string to identify this ESX session.
      """
      return str(self._host)


class EsxSessionMonitor(EsxSession):
   """Class to monitor session of an already booted ESX physical machine.
   It assumes the state of ESX as ESX_STARTED.
   """
   def __init__(self, logger, host, logDir=None):
      super().__init__(logger, host, pxeDir=None, logDir=logDir,
                       killHost=False)

   def run(self):
      """Start this ESX session.

      Connect to esx session and parse the ESX serial logs until someone calls
      EsxSession.close(), or until a fatal ESX error occurs (boot failure,
      PSOD).
      """
      logPath = None
      if self._logDir is not None:
         logPath = os.path.join(self._logDir, '%s-vmkernel.log' % self._host)

      self._host._status = ESX_STARTED

      if self._terminate.is_set():
         self._logger.warn("Session is terminated.")
         return
      try:
         self._host.startSerialMonitor(logPath=logPath)
      except Exception as e:
         self._logger.error(str(e))
         raise EsxSessionError("Failed to start serial log monitor",
                               str(self._host))

      self._atExit()
      self._logger.info("Successfully disconnected from ESX host")


#
# Code below is for debugging purposes...
#
if __name__ == "__main__":
   from argparse import ArgumentParser
   import getpass
   import logging

   from borautils.ping import ping
   from borautils.esx.host import EsxNimbusHost, EsxPhysHost

   logging.basicConfig()
   logger = logging.getLogger()
   logger.setLevel(logging.INFO)

   stop = Event()

   class DummyEsxSession(EsxSession):
      def onErrorCB(self, error):
         logger.exception(error)
         stop.set()

      def onPsodCB(self, psodLog):
         logger.error('PSOD: %s' % psodLog)
         stop.set()

      def onEsxInitializedCB(self):
         logger.info('ESX is initialized!')
         stop.set()

   parser = ArgumentParser()
   parser.add_argument('--host', metavar='HOST', type=str, dest='host',
                       help="Host ILO/iDrac address")
   parser.add_argument('--user', metavar='USERNAME', type=str, dest='user',
                       help="Ilo/iDrac username")
   parser.add_argument('--password', metavar='PASSWORD', type=str,
                       dest='password',
                       help="Ilo/iDrac password")
   parser.add_argument('--location', metavar='LOCATION', type=str,
                       dest='location', default="PromB",
                       help="Location of the machine to boot")
   parser.add_argument('--serial-host', metavar='HOST', type=str,
                       dest='serialHost',
                       help="Serial port concentrator telnet host")
   parser.add_argument('--serial-port', metavar='PORT', type=int,
                       dest='serialPort',
                       help="Serial port concentrator telnet port")
   parser.add_argument('--mac', metavar='MAC', type=str, dest='macAddrs',
                       action='append', default=[],
                       help="Host MAC address")
   parser.add_argument('--pxedir', metavar='PXEDIR', type=str, dest='pxeDir',
                       help="Path to your PXE directory")
   parser.add_argument('--results-dir', metavar='DIR', type=str,
                       dest='resultsDir', default=os.path.join(os.sep, 'tmp'),
                       help="Path to the result (output) directory")
   parser.add_argument('--host-type', metavar='HOSTTYPE', type=str.lower,
                       dest='hostType', default='physical',
                       choices=['nimbus', 'physical'],
                       help="Type of the host under test, physical by default")

   args = parser.parse_args()

   if args.hostType == 'nimbus':
      ESX_DEFAULT_NUM_PCPU = 2
      ESX_DEFAULT_MEMORY_SIZE = 6 * 1024
      ESX_DEFAULT_DISK_CAPACITY = 10 * 1024 * 1024
      ESX_DEFAULT_NUM_VNICS = 1

      vmName = "%s-esxsession-test" % getpass.getuser()
      logPath = os.path.join(args.resultsDir, '%s.log' % vmName)

      host = EsxNimbusHost(name=vmName, cpus=ESX_DEFAULT_NUM_PCPU,
                           memory=ESX_DEFAULT_MEMORY_SIZE,
                           disk=ESX_DEFAULT_DISK_CAPACITY,
                           nics=ESX_DEFAULT_NUM_VNICS, logFile=logPath)
   else:
      host = EsxPhysHost(ilo=args.host, iloUser=args.user,
                         iloPass=args.password, location=args.location,
                         macAddrs=args.macAddrs, serialHost=args.serialHost,
                         serialPort=args.serialPort)


   # Instantiate, and start an ESX session
   esx = DummyEsxSession(logger, host, args.pxeDir, logDir=args.resultsDir)
   esx.start()

   # ESX is booting, wait until it has fully initialized
   logger.info('Waiting for ESX to initialize...')
   stop.wait()

   # ESX is up and running, let's poke it to see if it's alive...
   try:
      ping(esx.getIp())
      logger.info('ESX is alive!')
   except Exception:
      logger.error("ESX is not responding")

   # Signal that we no longer need ESX
   logger.info("Done with ESX, let's close the session...")
   esx.close()

   # Let the session terminate gracefully
   logger.info('Waiting for any vm-support to be collected...')
   esx.join()
   logger.info('ESX has finished, session closed.')
