""" Stateful install a given Physical host

This module subclasses EsxPhysHost, monitors vmkernel logs and
updates status to be consumed by statefulsession
"""
import re

from borautils.esx.host import (
   BOOT_SUCCESSFUL,
   ESXHOST_STATE,
   EsxPhysHost,
   VMKERNEL_STARTED,
)

# Extend ESXHOST_STATE status and add 4 new status levels
STATEFULHOST_STATE = (
   ESX_INSTALLATION_START,    # To signal the start of ESX installation
   ESX_INSTALLATION_DONE,     # For completion of ESX installation
   BOOT_FROM_DISK_SUCCESSFUL  # To confirm booting from disk (after 2nd reboot)
) = list(range(len(ESXHOST_STATE), len(ESXHOST_STATE) + 3))

# XXX
# Time out of 1 hour is little high for smaller servers
# Need to make this optional value in future
DEFAULT_ESX_INSTALL_TIMEOUT = 60 * 60  # 1 hour

class StatefulPhysHost(EsxPhysHost):

   def __init__(self, iloHostName, iloUserName, iloPassword, location, macAddrs,
                serialHost, serialPort):
      """Sub class EsxPhysHost

      @param name: Host ILO name
      @param
      """
      self._pxeBootComplete = False
      self._ip = None
      self._vmkIpRe = (
                   re.compile(r'"HostIPAddr" = "(?P<ip>\d+\.\d+\.\d+\.\d+)", '
                              r'Old value: "\d+\.\d+\.\d+\.\d+".*Status: 0x0'))
      super().__init__(iloHostName, iloUserName, iloPassword, location,
                       macAddrs, None, serialHost, serialPort)

   def getIpStateful(self):
      """Return IP

      @return: IP address
      """
      return self._ip

   def _updateTimer(self, status):
      """Update timer for install process to the given status.

      @param status: Current status based on logs
      """
      if status == ESX_INSTALLATION_START:
         self._cancelTimer("boot")
         self._startTimer("install", DEFAULT_ESX_INSTALL_TIMEOUT,
                          self._onBootTimeoutCB)
      elif status == ESX_INSTALLATION_DONE:
         self._cancelTimer("install")
      else:
         # XXX
         # Add Boot from disk timer
         #
         super()._updateTimer(status)

   def _parseVmkLog(self, log):
      """Parse a vmkernel log line, and update statfulSession status.

      @param log: Current snippet from vmkernel logs
      """
      if log.startswith("TSC: 0 cpu0:"):
         self._updateStatus(VMKERNEL_STARTED)

      elif self._status >= BOOT_SUCCESSFUL and 'ESX Installation Start' in log:
         self._updateStatus(ESX_INSTALLATION_START)

      elif (self._status >= ESX_INSTALLATION_START and
         'ESX Installation Done' in log):
         self._pxeBootComplete = True
         self._updateStatus(ESX_INSTALLATION_DONE)

      elif self._pxeBootComplete and 'Boot Successful' in log:
         self._updateStatus(BOOT_FROM_DISK_SUCCESSFUL)

      else:
         # If PXE install complete, try fetching IP address from logs
         if self._pxeBootComplete and self._ip is None:
            m = self._vmkIpRe.search(log)

            if m:
               self._ip = m.group('ip')
         super()._parseVmkLog(log)
