# Copyright 2018-2019 VMware, Inc.
# All rights reserved. -- VMware Confidential

"""Library to browse the content of an ESX VIB depot.

Helper classes to parse the XML metadata in a VIB depot (e.g. vendor-index.xml,
metadata.zip, etc.). This is useful to list the VIBs in a depot and to find the
path / URL to their actual location.
"""
import os
import re
from tempfile import TemporaryDirectory
import xml.etree.ElementTree as ET
from zipfile import ZipFile

from borautils.httputils import urlOpen
from borautils.busybox import wget

#
# The below constants are taken from the esximage library. It would be nice to
# consume them directly from esximage, but that would require to package and
# publish esximage as a build deliverable. We are not setup this way at this
# point.
#
TYPE_BOOT_COM32_BIOS = "boot_com32_bios"
TYPE_BOOT_LOADER_EFI = "boot_loader_efi"
TYPE_BOOT = "boot"
TYPE_VGZ = "vgz"
TYPE_TGZ = "tgz"
TYPE_INSTALLER_VGZ = "installer_vgz"

TYPE_BOOTBANK = "bootbank"
TYPE_LOCKER = "locker"

BOOT_ORDER_VGZ_BASE = 100
DATABASE_NAME = "imgdb.tgz"
IMGPAYLD_NAME = "imgpayld.tgz"

def getFatShortName(filename, ext):
   """Convert a regular filename to a FAT short name.
   """
   return "%s.%s" % (filename[:8].replace('-', '_').lower(), ext)


class XmlSpec(object):
   """Helper class to parse an XML document.

   Supports reading from the filesytem or from an HTTP URL.
   """

   def __init__(self, path):
      self.path = path

      with urlOpen(path) as f:
         tree = ET.parse(f)
         self._root = tree.getroot()


class VendorIndex(XmlSpec):
   """Class to parse a vendor-index.xml file.
   """

   def __init__(self, path):
      super().__init__(path)
      self._metadata = None
      self._version = None
      self._url = None

      for metadata in self._root.iter('metadata'):
         if metadata.find('productId').text in ('ESXi', 'embeddedEsx'):
            self._metadata = metadata
            return

      raise AttributeError("%s: invalid vendor-index "
                           "('ESXi' productId not found)" % path)

   @property
   def metadataRelPath(self):
      """Path to the metadata.zip file, relative to the root of the VIB depot.
      """
      if self._url is None:
         self._url = self._metadata.find('url').text
      return self._url


class PayloadMetadata(object):
   """Class to parse a VIB Payload metadata XML spec.
   """

   def __init__(self, xml):
      self._xml = xml
      self._type = None
      self._name = None
      self._bootOrder = ""

   @property
   def type(self):
      """Type of this payload.
      """
      if self._type is None:
         self._type = self._xml.get('type')
      return self._type

   @property
   def name(self):
      """Payload's filename as it would be named on the filesystem.

      Override with the 'vfat-name' payload attribute if it exists, otherwise
      fall back to the 'name' attribute, and coerce to FAT format if the payload
      is a boot module.
      """
      if self._name is None:
         vfatName = self._xml.get('vfat-name')
         if vfatName:
            self._name = vfatName
         else:
            name = self._xml.get('name')
            if self.type in (TYPE_BOOT, TYPE_VGZ, TYPE_TGZ):
               ext = "%c00" % self.type[0]
               self._name = getFatShortName(name, ext)
            else:
               self._name = name
      return self._name

   @name.setter
   def name(self, value):
      """To override the name property.
      """
      self._name = value

   @property
   def bootOrder(self):
      """Sequence number defining the order in which this (boot) payload should
      be loaded by the bootloader.
      """
      if self._bootOrder == "":
         self._bootOrder = self._xml.get('boot-order')
         if self._bootOrder is not None:
            self._bootOrder = int(self._bootOrder)
      return self._bootOrder


class ImgDbMetadata(PayloadMetadata):
   """Object representing the 'imgdb.tgz' payload.

   Imgdb.tgz is special in that it is not listed in the image profile although
   it is always part of an ESX image.
   """

   def __init__(self):
      super().__init__(None)
      self._type = TYPE_TGZ
      self._name = DATABASE_NAME
      self._bootOrder = None

class ImgPayldMetadata(PayloadMetadata):
   """Object representing the 'imgpayld.tgz' payload.

   Imgpayld.tgz contains the boot modules and has to be
   part of the installer, as it is not part of any vib
   in the image profile.
   """

   def __init__(self):
      super().__init__(None)
      self._type = TYPE_TGZ
      self._name = IMGPAYLD_NAME
      self._bootOrder = None


class VibMetadata(XmlSpec):
   """Class to parse a VIB metadata XML spec.
   """

   def __init__(self, path):
      super().__init__(path)
      self._name = None
      self._version = None
      self._type = None
      self._relPath = None
      self._overlay = None
      self._id = None

   @property
   def name(self):
      """This VIB's name.
      """
      if self._name is None:
         self._name = self._root.find('name').text
      return self._name

   @property
   def version(self):
      """This VIB's version string.
      """
      if self._version is None:
         self._version = self._root.find('version').text
      return self._version

   @property
   def isBootbank(self):
      """True if this VIB is to be installed in the bootbank.
      """
      if self._type is None:
         self._type = self._root.find('type').text
      return self._type == TYPE_BOOTBANK

   @property
   def isLocker(self):
      """True if this VIB is of type locker.
      """
      if self._type is None:
         self._type = self._root.find('type').text
      return self._type == TYPE_LOCKER

   @property
   def relPath(self):
      """Path to this VIB, relative to the root of the VIB depot.
      """
      if self._relPath is None:
         self._relPath = self._root.find('relative-path').text
      return self._relPath

   @property
   def id(self):
      """Unique identifier for this VIB.
      """
      if self._id is None:
         vibType = self._root.find('type').text
         vendor = self._root.find('vendor').text
         self._id = "%s_%s_%s_%s" % (vendor, vibType, self.name, self.version)
      return self._id

   @property
   def overlay(self):
      """True if this VIB contains overlay tardisks.
      """
      if self._overlay is None:
         self._overlay = self._root.find('overlay').text == 'true'
      return self._overlay

   def iterPayloads(self):
      """Iterates over the VIB's payloads.
      """
      payloads = self._root.find('payloads')
      for payload in payloads.findall('payload'):
         yield PayloadMetadata(payload)


class ProfileMetadata(XmlSpec):
   """Class to parse an Image Profile metadata XML spec.
   """

   def __init__(self, path, depot):
      super().__init__(path)
      self._depot = depot
      self._name = None
      self._vibs = []

      vibList = self._root.find('viblist')
      for vibXml in vibList.findall('vib'):
         vibId = vibXml.find('vib-id').text
         for vib in self._depot._vibs.values():
            if vib.id == vibId:
               self._vibs.append(vib)
               break
         else:
            raise ValueError("%s VIB is listed in profile %s but does not "
                             "match any VIB in depot %s" %
                             (vibId, path, depot.path))

   @property
   def name(self):
      """Name of this image profile.
      """
      if self._name is None:
         self._name = self._root.find('name').text
      return self._name

   def _iterPayloads(self):
      """Iterate over this image profile's payloads.
      """
      for vib in self._vibs:
         for payload in vib.iterPayloads():
            yield vib, payload

   def iterPxePayloads(self):
      """Iterate over this image profile's PXE payloads.
      """
      for _, payload in self._iterPayloads():
         if payload.type in (TYPE_BOOT_COM32_BIOS, TYPE_BOOT_LOADER_EFI):
            yield payload

   def iterBootPayloads(self, installer=False):
      """Iterate over this image profile's boot payloads, in the order they must
      be loaded by the bootloader.

      The boot payloads order is defined as follow:
        1. the boot modules, in ascending 'bootOrder' attribute
        2. the esx-base tardisks, in the order they appear in the VIB metadata
        3. non-overlay tardisks in the order they appear in the image profile
        4. overlay tardisks in the order they appear in the image profile
        5. image database
      """
      bootPayloads = []
      nonOverlayPayloads = []
      overlayPayloads = []
      imagePayload = []

      payloadTypes = [TYPE_VGZ, TYPE_TGZ]
      if installer:
         payloadTypes += [TYPE_INSTALLER_VGZ]
         imagePayload.append(ImgPayldMetadata())

      for vib, payload in self._iterPayloads():
         if not vib.isBootbank and not (vib.isLocker and installer):
            continue

         if payload.type == TYPE_BOOT:
            assert isinstance(payload.bootOrder, int)
            bootPayloads.append(payload)
         elif payload.type in payloadTypes:
            if vib.overlay:
               overlayPayloads.append(payload)
            else:
               if payload.bootOrder == BOOT_ORDER_VGZ_BASE:
                  nonOverlayPayloads.insert(0, payload)
               else:
                  nonOverlayPayloads.append(payload)

      bootPayloads.sort(key=lambda p: p.bootOrder)

      payloads = (bootPayloads + nonOverlayPayloads + overlayPayloads +
                  [ImgDbMetadata()] + imagePayload)

      # Fixup payload names to avoid duplicates. We use a 2-digit hex sequence
      # number suffix to disambiguate payloads with identical names.
      seqs = {}
      for payload in payloads:
         name = payload.name
         try:
            seqs[name] += 1
         except KeyError:
            seqs[name] = 0
         else:
            assert name.endswith("00")
            payload.name = name[:-2] + "%02x" % seqs[name]
         yield payload

   def iterExtraVibs(self):
      """Iterate over all the extra (non-bootbank) VIBs in this image profile.
      """
      for vib in self._vibs:
         if not vib.isBootbank:
            yield vib

class VibDepot(object):
   """Class to browse a VIB depot from a filesystem path or an HTTP URL.
   """

   def __init__(self, path):
      self.path = path
      self._version = None
      self._vibs = {}
      self._profiles = {}

      vendorIndex = VendorIndex(os.path.join(path, 'vendor-index.xml'))
      metadataZip = os.path.join(path, vendorIndex.metadataRelPath)

      with TemporaryDirectory() as tmpDir:
         if metadataZip.lower().startswith(('http://', 'https://')):
            wget(metadataZip, tmpDir)
            metadataZip = os.path.join(tmpDir, os.path.basename(metadataZip))

         with ZipFile(os.path.join(tmpDir, metadataZip)) as zf:
            zf.extractall(path=tmpDir)

         vibDir = os.path.join(tmpDir, 'vibs')
         for basename in os.listdir(vibDir):
            vib = VibMetadata(os.path.join(vibDir, basename))
            self._vibs[vib.name] = vib

         profilesDir = os.path.join(tmpDir, 'profiles')
         for basename in os.listdir(profilesDir):
            profile = ProfileMetadata(os.path.join(profilesDir, basename), self)
            self._profiles[profile.name] = profile

   @property
   def version(self):
      """Return the ESX version for this depot.

      The ESX version is extracted from the esx-base VIB.
      """
      if self._version is None:
         try:
            vib = self._vibs['esx-base']
         except KeyError:
            raise ValueError("'esx-base' VIB not found in depot %s" % self.path)
         m = re.match(r"(\d.\d.\d+)-(\d.\d+).(\d+)", vib.version)
         if m is None:
            raise ValueError("%s: invalid esx-base VIB version" % vib.version)
         self._version = m.group(1, 2, 3)
      return self._version

   def _getVib(self, name):
      """Return an VibMetadata object for the requested VIB.
      """
      try:
         return self._vibs[name]
      except KeyError:
         raise FileNotFoundError("%s: no such VIB in depot %s" %
                                 (name, self.path))

   def getVibUrl(self, name):
      """Return the absolute path/url to the requested VIB.
      """
      return os.path.join(self.path, self._getVib(name).relPath)

   def getVibVersion(self, name):
      """Return the version string of the requested VIB.
      """
      return self._getVib(name).version

   def getProfile(self, name, buildNum=None):
      """Return an ImageProfile.
      """
      version, _, bldNum = self.version
      if buildNum is not None:
         bldNum = buildNum

      profileName = "ESXi-%s-%s-%s" % (version, bldNum, name)
      try:
         return self._profiles[profileName]
      except KeyError:
         raise ValueError("%s: no such image profile in depot %s" %
                          (name, self.path))
