"""Interface to parse vmkernel ELF binaries.
"""
import re

from borautils.elfbin import ElfBin, ElfCore, ElfError, AT_ENTRY, AT_FLAGS

AT_MODLIST = 199

ASCII_NULL = b'\x00'
CODE_START_ADDR = 0x420000000000

def parseBuildStr(buildStr):
   """Extract the build number and type from a vmkernel binary/core build
   string.
   """
   m = re.search(r"(\w+)build-(\d+)", buildStr)
   if m is None:
      raise ValueError("%s: invalid build string" % buildStr)

   buildType = m.group(1).lower()
   buildNum = int(m.group(2))
   return (buildNum, buildType)


class VmkBin(ElfBin):
   """Abstraction of a vmk binary (vmkernel or a module)
   """
   def __init__(self, filepath):
      super().__init__(filepath)

      self._vmtree = None
      self._buildNum = None
      self._buildType = None

      # AFAIK we don't have any module exporting more than one namespace.
      self.providedNamespace = b''
      self.requiredNamespaces = set()

      # relevant for module only
      self.vmksymbolexports = set()
      self.vmksymbolaliases = set()

      self._parseProvidedNamespace()
      self._parseRequiredNamespaces()
      self._parseRelocations('.rela.vmksymbolexports', self.vmksymbolexports)
      self._parseRelocations('.rela.vmksymbolaliases', self.vmksymbolaliases)

   def _parseRequiredNamespaces(self):
      """Parse the .vmkrequiredns to find all the namespaces required.
      """
      if '.vmkrequiredns' not in self._sections:
         return

      vmkrequiredns = self.getSectionContent('.vmkrequiredns')
      for ns in vmkrequiredns.split(ASCII_NULL):
         self.requiredNamespaces.add(ns.decode())

   def _parseProvidedNamespace(self):
      """Parse the .vmkmodinfo section to find the namespace provided.
      """
      if '.vmkmodinfo' not in self._sections:
         return

      vmkmodinfo = self.getSectionContent('.vmkmodinfo')
      index = vmkmodinfo.find(b'nsProvides=')
      if index == -1:
         return

      namespace = self.getStringFromBuffer(vmkmodinfo,
                                           index + len(b'nsProvides='))
      self.providedNamespace = namespace

   def _parseBuildNum(self):
      """Extract the build info from this vmkernel binary.

      The build info string is contained in the buildVersion variable.
      """
      buildStr = self.readSymbol('buildVersion').decode()
      return parseBuildStr(buildStr)

   @property
   def buildNum(self):
      """This core dump build number.
      """
      if self._buildNum is None:
         self._buildNum, self._buildType = self._parseBuildNum()
      return self._buildNum

   @property
   def buildType(self):
      """This core dump build type (relase, obj, beta, ...).
      """
      if self._buildType is None:
         self._buildNum, self._buildType = self._parseBuildNum()
      return self._buildType

   @property
   def vmtree(self):
      """Path to the vmtree from which this vmkernel binary was built.
      """
      if self._vmtree is None:
         self._vmtree = self.readSymbol('vmTree').decode().rstrip('\x00')
      return self._vmtree

   def aslrOffset(self):
      """Returns the ALSR offset.
      """
      vAddr,_ = self._getSymbolVAddr("_start")
      if vAddr is None:
         raise ValueError("Cannot get ASLR offset of a non-vmkernel binary")

      return vAddr - CODE_START_ADDR

class VmkCore(ElfCore):
   """Class representing a vmkernel coredump.
   """

   def __init__(self, filePath):
      super().__init__(filePath)
      self._auxv = self.parseAuxv()
      self._buildNum, self._buildType = self._parseBuildNum()

   def parseAuxv(self):
      """Parse the auxv section of this vmkernel core dump.
      """
      CORE_TYPE_VMKERNEL = 1

      auxv = super().parseAuxv()
      try:
         flags = auxv[AT_FLAGS]
      except KeyError:
         raise ElfError("invalid vmkernel core (no auxv[AT_FLAGS] section)",
                        self)

      if flags != CORE_TYPE_VMKERNEL:
         raise ElfError("invalid vmkernel core (bad core type: %x)" % flags,
                        self)
      return auxv

   def _parseBuildNum(self):
      """Extract the build info from this core dump.

      The build info is contained in a string located at the beginning of the
      auxv[AT_MODLIST] section.
      """
      BUILD_STR_SZ = 0x20

      try:
         buildStrAddr = self._auxv[AT_MODLIST]
      except KeyError:
         raise ElfError("invalid vmkernel core (no auxv[AT_MODLIST] section)",
                        self)

      buildStr = self.read(buildStrAddr, BUILD_STR_SZ).decode()
      return parseBuildStr(buildStr)

   @property
   def buildNum(self):
      """This core dump build number.
      """
      return self._buildNum

   @property
   def buildType(self):
      """This core dump build type (relase, obj, beta, ...).
      """
      return self._buildType

   @property
   def entry(self):
      """Virtual address of the vmkernel entry point.
      """
      try:
         return self._auxv[AT_ENTRY]
      except KeyError:
         raise ElfError("invalid vmkernel core (no auxv[AT_ENTRY] section)",
                        self)
