# Copyright 2015-2017 VMware, Inc.
# All rights reserved. -- VMware Confidential

"""Zdump extraction lib
"""
from datetime import timedelta
import os
import re
import struct
import subprocess

from .vmkbin import parseBuildStr

#
# Public flags defined in dump_ext.h
#
ZDUMP_VER_MASK = 0xffff0fff
ZDUMP_TYPE_MASK = 0xf000
ZDUMP_TYPE_KERNEL = 0x1000
ZDUMP_TYPE_USER = 0x2000

#
# Each zdump version requires a matching version of the vmkdump_extract
# binary. vmkdump_extract binaries are stored in the following repository:
#
VMKDUMP_EXTRACT_REPO = "/build/apps/bin/esx/vmkdump_extract_repo"


class _Zdump(object):
   """Base class to extract zdumps.

   This class should not be instantiated directly, use the Zdump() function
   instead.
   """

   def __init__(self, path):
      """Load a zdump file.
      """
      self.path = path
      self._buildNum = None
      self._buildType = None
      self._extractBin = None
      self._isEncrypted = None
      self._header = None

      with open(self.path, 'rb') as f:
         data = f.read(4)

      try:
         version, = struct.unpack('<I', data)
      except struct.error:
         raise ValueError('%s: invalid zdump file' % self.path)

      self.version = version & ZDUMP_VER_MASK
      self.type = version & ZDUMP_TYPE_MASK
      if self.type not in (ZDUMP_TYPE_KERNEL, ZDUMP_TYPE_USER):
         raise ValueError("%s: invalid zdump type (%#x)" %
                          (self.path, version & ZDUMP_TYPE_MASK))

   @property
   def extractBin(self):
      """Path to the vmkdump_extract binary which matches this zdump version and
      type.
      """
      if self._extractBin is None:
         if not os.path.isdir(VMKDUMP_EXTRACT_REPO):
            raise FileNotFoundError("%s: no such directory (/build/apps/ not "
                                    "mounted)" % VMKDUMP_EXTRACT_REPO)

         # From version 59, we no longer have different version numbers for
         # vmkernel or UW dumps, they are all the same.
         if self.type == ZDUMP_TYPE_KERNEL or self.version >= 59:
            typeName = 'kernel'
         else:
            typeName = 'user'

         self._extractBin = os.path.join(VMKDUMP_EXTRACT_REPO, typeName,
                                         str(self.version), 'vmkdump_extract')
         if not os.path.exists(self._extractBin):
            if self.type == ZDUMP_TYPE_KERNEL:
               bzCategory = 'vmkcore'
            else:
               bzCategory = 'userworld'
            raise FileNotFoundError("vmkdump_extract missing for type '%s', "
                                    "version %u, please open a PR against "
                                    "ESX > ESX Server > %s" %
                                    (typeName, self.version, bzCategory))
      return self._extractBin

   def _extract(self, args):
      """Call vmkdump_extract on this zdump.
      """
      p = subprocess.Popen([self.extractBin] + args + [self.path],
                           stdout=subprocess.PIPE, stderr=subprocess.PIPE)
      out, err = p.communicate()
      if p.returncode != 0:
         raise RuntimeError("%s: failed to extract zdump (%s)" %
                            (self.path, err.decode()))
      return "" if out is None else out.decode().strip()

   @property
   def isEncrypted(self):
      """True if this zdump is encrypted.
      """
      if self._isEncrypted is None:
         self._isEncrypted = (self._extract(['-E']) == 'YES')
      return self._isEncrypted

   def _extractBuildInfo(self):
      """Extract the build number and type from this zdump.
      """
      if self._buildNum is None:
         assert self._buildType is None
         buildNumStr = self._extract(['-b'])
         self._buildNum, self._buildType = parseBuildStr(buildNumStr)
      return self._buildNum, self._buildType

   @property
   def buildNum(self):
      """Zdump build number.
      """
      return self._extractBuildInfo()[0]

   @property
   def buildType(self):
      """Zdump build type.
      """
      return self._extractBuildInfo()[1]

   def extractCore(self, dest):
      """Extract the coredump contained in this zdump.
      """
      destDir = os.path.dirname(dest)
      os.makedirs(destDir, exist_ok=True)
      self._extract(['-x', '-X', dest])


class UwZdump(_Zdump):
   """Class to extract Userworlds zdumps.
   """

   def __init__(self, *args, **kwargs):
      super().__init__(*args, **kwargs)
      if self.type != ZDUMP_TYPE_USER:
         raise ValueError("%s: invalid UW zdump" % path)
      self._progName = None

   @property
   def progName(self):
      """Path of the executable which crashed.
      """
      if self._progName is None:
         self._progName = self._extract(['-e'])
      return self._progName


class VmkZdump(_Zdump):
   """Class to extract vmkernel zdumps.
   """

   def __init__(self, path):
      super().__init__(path)
      if self.type != ZDUMP_TYPE_KERNEL:
         raise ValueError("%s: invalid vmkernel zdump" % path)
      self._isComplete = None

   def _extractHeader(self):
      """Extract the zdump header.
      """
      ZDUMP_FLAGS_TYPE_VMVISOR = 0x1

      if self._header is None:
         self._header = self._extract(['-h'])
         m = re.search(r"flags:\s*vmvisor\s*\((0x[0-9a-fA-F]+)\)", self._header)
         if m is None:
            raise ValueError("%s: failed to extract zdump flags, zdump might "
                             "be corrupted" % self.path)
         flags = int(m.group(1), base=16)
         if flags != ZDUMP_FLAGS_TYPE_VMVISOR:
            raise ValueError("%s: invalid vmkernel type (%u), zdump might be "
                             "corrupted" % (self.path, flags))
      return self._header

   @property
   def isComplete(self):
      """True if this zdump is complete (not truncated).
      """
      if self._isComplete is None:
         header = self._extractHeader()

         m = re.search(r"dumpSize:\s*(\d+)", header)
         if m is None:
            return False
         dumpSize = int(m.group(1))

         m = re.search(r"startOffset:\s*(\d+)", header)
         if m is None:
            return False
         offset = int(m.group(1))

         self._isComplete = (offset + dumpSize == os.stat(self.path).st_size)

      return self._isComplete

   @property
   def entryPoint(self):
      """Return the vmkernel entry point address.
      """
      info = self._extractHeader()
      m = re.search(r"codeEntryPoint:\s*(0x[0-9a-fA-F]+)", info)
      if m is None:
         raise ValueError("%s: failed to extract vmkernel entry point, zdump "
                          "might be corrupted" % self.path)
      return int(m.group(1), base=16)

   def extractVmkLog(self, dest):
      """Extract the vmkernel log contained in this zdump.
      """
      destDir = os.path.dirname(dest)
      os.makedirs(destDir, exist_ok=True)
      self._extract(['-l', '-L', dest])


def Zdump(path):
   """Interface function to load a Zdump.
   """
   for zdumpClass in (VmkZdump, UwZdump):
      try:
         return zdumpClass(path)
      except ValueError:
         pass
   raise ValueError("%s: invalid zdump file")
