#!/usr/bin/env python

# Copyright 2016-2020 VMware, Inc.  All rights reserved. -- VMware Confidential

"""ESX customization boot script

Boot script to be placed under /etc/rc.local.d/ in order to initialize NFS
mounts and install boot VIBs. This script expects to find the NFS/VIBs
configuration in /var/lib/vmware/testesx/conf.json.
"""
import json
import logging
import os
from esxutils import (installVibs, getEsxVersion, nfsMount, nfsUnmount,
                      runCli, vmkLog, NfsAlreadyExportedError)
from vmware import vsi

BUILD_TREE_NFS_SHARE = 'buildtree'
TESTESX_DIR = '/test'

def getLogger(name):
   """Create a logger with name. The logger will log messages to vmkernel log
      in format: {timestamp} {logger name} {log message}
   """
   class VmkHandler(logging.Handler):
      def emit(self, record):
         log = self.format(record)
         vmkLog(log)

   logger = logging.getLogger(name)
   logger.setLevel(logging.INFO)
   handler = VmkHandler()
   formatter = logging.Formatter(fmt='{0}: %(funcName)s:%(lineno)s: %(message)s'
                                     .format(name))
   handler.setFormatter(formatter)
   logger.addHandler(handler)
   return logger

logger = getLogger('testesxconf')

def setupNfsMounts(nfsMounts, isBuildTree=False):
   """Mount a list of NFS shares.
   """

   DEFAULT_MOUNTS_NUM = 8
   MAX_MOUNTS_NUM = 32

   errors = 0

   # unmount old build tree before mount a new one
   if isBuildTree:
      try:
         nfsList = runCli(['storage', 'nfs', 'list'], evalOutput=True)
      except Exception:
         logger.error("Failed to get NFS list")
         raise

      for item in nfsList:
         if item['Volume Name'] == BUILD_TREE_NFS_SHARE:
            try:
               nfsUnmount(BUILD_TREE_NFS_SHARE)
               break
            except Exception:
               logger.error("Failed to unmount old build tree")
               raise

   if len(nfsMounts) > DEFAULT_MOUNTS_NUM:
      try:
         runCli(['system', 'settings', 'advanced', 'set', '-o',
                 '/NFS/MaxVolumes', '-i', str(MAX_MOUNTS_NUM)])
      except Exception:
         logger.error("Failed to set NFS max volumes number to %u",
                      MAX_MOUNTS_NUM)
         raise

   for host, src, dest, opts in nfsMounts:
      if os.path.isdir(dest):
         logger.info('Skip symlink %s as it already exists' % dest)
         continue
      if os.path.islink(dest):
         logger.info('Removing inactive symlink %s' % dest)
         os.remove(dest)
      readOnly = opts and 'ro' in opts.split(',')
      if isBuildTree:
         mntName = BUILD_TREE_NFS_SHARE
      else:
         mntName = "%s_%s" % (host, src.replace('/', '.'))
      try:
         logger.info("Mounting %s:%s to %s" % (host, src, mntName))
         try:
            mntPath = nfsMount(host, src, mntName, readonly=readOnly)
         except NfsAlreadyExportedError as e:
            # reuse the existing NFS volume
            logger.info('Volume is already mounted as %s' %
                        e.volumeName)
            mntPath = os.path.join(os.sep, 'vmfs', 'volumes', e.volumeName.strip())
         os.makedirs(os.path.dirname(dest), exist_ok=True)
         logger.info('Symlink %s to %s' % (mntPath, dest))
         os.symlink(mntPath, dest)
      except Exception as e:
         errors += 1
         logger.error('Failed to setup NFS mount (%s:%s -> %s (%s)), err: %s',
                      host, src, dest, opts, e)

   return errors

def isPxeBooting():
   """Retrun True if this host is booted from pxe.
   """
   return vsi.get('/system/bootMAC')['macAddrStr'] != ''

def internalInstallVib(url):
   """Live mount large sized VIB by url.
   """
   # Note: this is a temp solution. see PR 2143443
   runCli(['--plugin',
           '/usr/lib/vmware/esxcli/int/ext/esxcli-softwareinternal.xml',
           'softwareinternal', 'vib', 'tardisk', 'mount', '--viburl',
           url])

def setupTestRamdisk():
   """Create a new ramdisk for /test as to not consume the existing
      ramdisks' inodes. See PR 2525097. The new ramdisk can accept
      upto 16k inodes and have upto 1GB of content. As of 3/2020,
      the testesx VIB has 8k inodes consuming over 700MB.
   """
   logger.warning("Creating testesx ramdisk")

   if os.path.isdir(TESTESX_DIR):
      # XXX This path is reached ONLY upon rebooting stateful host
      # This is a temp fix until we figure out how this can be handled
      # differently refer PR-2502650#c12
      logger.warning("Skip ramdisk creation: %s already exists" , TESTESX_DIR)
      return

   os.mkdir(TESTESX_DIR)
   runCli(['--plugin',
           '/usr/lib/vmware/esxcli/int/lib64/libsystem-internal.so',
           'systemInternal',
           'visorfs',
           'ramdisk',
           'add',
           '-I', '16384',   # max inodes
           '-M', '1024',    # max size (MB)
           '-m', '0',       # min size (0MB)
           '-n', 'testesx', # name
           '-p', '0777',    # permissions mask
           '-t', TESTESX_DIR])  # mount point

def setupVibs(vibUrls):
   """Install VIBs from a list of VIB URLs.
   """
   errors = 0

   try:
      setupTestRamdisk()
   except Exception as e:
      logger.warning("Failed to create ramdisk for tests: %s" % e)
      errors += 1

   for url in vibUrls:
      name = os.path.basename(url)
      try:
         if name.startswith('VMware_bootbank_testesx') and not isPxeBooting():
            try:
               installVibs([url])
            except Exception as e:
               # XXX ignore if the testesx vib is already installed
               # required for 5.x/6.x -> 7.x upgrade as there is not
               # enough space in the bootbank to install testesx vib
               # refer "Upgrade From Legacy Layout" of System Storage Next
               logger.warning("Failed to install VIB %s, "
                              "check /var/log/esxupdate.log for details "
                              "trying live mount...: %s " % (name, e))
               internalInstallVib(url)
         else:
            installVibs([url])
      except Exception as e:
         errors += 1
         logger.error(repr(str(e)))
         logger.error("Failed to install VIB %s, "
                      "check /var/log/esxupdate.log for more details.", name)
   return errors

def loadEsxConf(configPath):
   """Parse the ESX config file, and apply ESX customizations.
   """
   errors = 0

   try:
      with open(configPath) as f:
         conf = json.load(f)
   except Exception as e:
      logger.error("Failed to load ESX configuration file %s: %s",
                   configPath, e)
      raise

   if 'nfs' in conf:
      errors += setupNfsMounts(conf['nfs'])

   if 'vibs' in conf:
      errors += setupVibs(conf['vibs'])

   if BUILD_TREE_NFS_SHARE in conf and conf[BUILD_TREE_NFS_SHARE] is not None:
      errors += setupNfsMounts([conf[BUILD_TREE_NFS_SHARE]], isBuildTree=True)

   return errors

if __name__ == "__main__":
   import sys

   DEFAULT_ESXCONF = "/var/lib/vmware/testesx/conf.json"

   if os.path.isfile(DEFAULT_ESXCONF):
      logger.info("loading ESX customizations from %s", DEFAULT_ESXCONF)

      try:
         errors = loadEsxConf(DEFAULT_ESXCONF)
      except Exception as e:
         logger.error('Failed to config testesx: %s', e)
         sys.exit(1)
      else:
         if errors > 0:
            logger.error('Failed to config testesx: errors=%d' % errors)
            sys.exit(1)
