"""Query automount info from LDAP.
"""
import os
import subprocess
import sys

from borautils.exception import DebugInfoMixIn

DEFAULT_DNS_DOMAIN = "eng.vmware.com"
DEFAULT_LDAP_DOMAIN = "vmware.com"

def importLdap3():
   """Import ldap3 3rd party module.

   Try to import ldap3 module from the local environment. If not found, then try
   again from the toolchain.
   """
   try:
      import ldap3
   except ImportError:
      from borautils.infra.toolchain import loadTool
      LDAP3_PATH = "noarch/pyldap3-1.0.4/lib/python3.5/site-packages"
      PYASN1_PATH = "noarch/pyasn1-0.1.9/lib/python3.5/site-packages"

      for path in [LDAP3_PATH, PYASN1_PATH]:
         sys.path.append(loadTool(path))
      import ldap3

   return ldap3

ldap3 = importLdap3()

class LDAPError(DebugInfoMixIn, Exception):
   """Generic LDAP error.
   """
   _errType = "LDAP error"

   def __init__(self, errStr, url):
      super().__init__(errStr)
      self.addDebugInfo('server', url)

class LDAPRequestError(LDAPError):
   """An LDAP request has failed.
   """
   _errType = "LDAP request error"

   def __init__(self, errStr, url, request):
      super().__init__(errStr, url)
      self.addDebugInfo('base', request.base)
      self.addDebugInfo('scope', request.scope)
      self.addDebugInfo('filter', request.filterStr)
      self.addDebugInfo('attributes', request.attributes)

class LDAPResponseError(LDAPRequestError):
   """Got invalid response from LDAP server (protocol error).
   """
   _errType = "Invalid LDAP response"

class DNSError(DebugInfoMixIn, Exception):
   """A DNS lookup has failed.
   """
   _errType = "DNS lookup error"

   def __init__(self, errStr, cmd, output=None):
      super().__init__(errStr)
      self.addDebugInfo('lookup command', cmd)
      if output is not None:
         self.addDebugInfo('output', output)


def getDefaultLdapServer(domain):
   """Do a DNS lookup to retrieve the default LDAP server.

   For more details, See RFC 2782:
   "A DNS RR for specifying the location of services (DNS SRV)"
   """
   DNS_LOOKUP_CMD = "host -t srv _ldap._tcp.%s" % domain
   p = subprocess.Popen(DNS_LOOKUP_CMD.split(),
                        stdout=subprocess.PIPE, stderr=subprocess.PIPE)
   out, err = p.communicate()
   if p.returncode != 0:
      raise DNSError("failed to retrieve default LDAP server: %s" % err,
                     DNS_LOOKUP_CMD)
   out = out.decode('utf-8')
   try:
      return out.split()[-1].rstrip('.')
   except IndexError:
      raise DNSError("failed to retrieve default LDAP server",
                     DNS_LOOKUP_CMD, output=out)


class LdapRequest(object):
   """Class to encapsulate all components of an LDAP request.
   """
   def __init__(self, cn, ou, dc, scope, filterStr, attributes):
      self.scope = scope
      self.filterStr = filterStr
      self.attributes = attributes

      attr = []
      if cn:
         attr += ["cn=%s" % x for x in cn]
      if ou:
         attr += ["ou=%s" % x for x in ou]
      if dc:
         attr += ["dc=%s" % x for x in dc]
      self.base = ','.join(attr)


class _LdapServer(object):
   """LDAP server class.
   """

   def __init__(self, url):
      """Open a session to the LDAP server on the given domain.

      Turn off strict checking of the server certificate; analogous to setting
      OpenLDAP 2.1+ client setting: tls_checkpeer no. Only care about basic
      transport encryption and not identity verification (self-signed
      certificates).
      """
      self._url = url

      try:
         self._server = ldap3.Server(self._url, use_ssl=False)
         self._session = ldap3.Connection(self._server, auto_bind=True)
      except ldap3.LDAPException:
         raise LDAPError("LDAP connection failed", self._url)

   def search(self, request):
      """Do the query against the given LDAP session.
      """
      scopes = {'BASE': ldap3.BASE,
                'ONELEVEL': ldap3.LEVEL,
                'SUBTREE': ldap3.SUBTREE}

      self._session.search(request.base, request.filterStr,
                           search_scope=scopes[request.scope],
                           attributes=request.attributes)

      try:
         return self._session.response[0]['attributes']
      except IndexError:
         return None
      except KeyError:
         raise LDAPResponseError("no 'attribues' record in ldap response",
                                 self._url, request)


class LDAP(object):
   """A class to query LDAP
   """

   def __init__(self, domain, dnsDomain=None):
      """Open a session to the LDAP server on the given domain.

      Turn off strict checking of the server certificate; analogous to setting
      OpenLDAP 2.1+ client setting: tls_checkpeer no. Only care about basic
      transport encryption and not identity verification (self-signed
      certificates).
      """
      self._domain = domain
      self._dnsDomain = domain if dnsDomain is None else dnsDomain
      self._url = "ldaps://%s" % getDefaultLdapServer(self._dnsDomain)

      self._server = _LdapServer(self._url)

   def getAutomountInfo(self, absPath):
      """Get automount info for the given absolute path.
      """
      if not os.path.isabs(absPath):
         raise ValueError("%s: path is not absolute" % absPath)

      absPath = absPath.rstrip(os.sep)

      try:
         dirs = absPath.split(os.sep)[1:]
         cn = [dirs[1]]
         ou = ["auto.%s" % dirs[0], "automount"]
      except IndexError:
         raise ValueError("%s: invalid absolute path "
                          "(too short to determine automount info)" % absPath)

      dc = self._domain.split('.')

      request = LdapRequest(cn, ou, dc, scope='SUBTREE',
                            filterStr="(objectclass=automount)",
                            attributes=['automountInformation'])

      try:
         response = self._server.search(request)
      except Exception:
         raise LDAPRequestError("failed to retrieve automount info",
                                self._url, request)
      if response is None:
         return None

      try:
         mountAttrs, nfsShare = response['automountInformation'][0].split()
      except KeyError:
         raise LDAPResponseError("no 'automountInformation' record in ldap "
                                 "response", self._url, request)
      except IndexError:
         raise LDAPResponseError("no entry found in 'automountInformation' "
                                 "record", self._url, request)
      except ValueError:
         raise LDAPResponseError("invalid value in 'automountInformation' "
                                 "record", self._url, request)

      nfsServer, nfsPath = nfsShare.split(':')
      return nfsServer, nfsPath, mountAttrs


def getAutomountInfoFromAbsPath(absPath):
   """Return automount information for the given absolute path.

   This function queries LDAP to retrieve automount information for the given
   absolute path. It returns a (nfsHost, nfsPath, mountAttrs) tuple where
   'nfsHost' is the NFS server which host the NFS share, 'NfsPath' is the mount
   point on the NFS server, and 'mountAttrs' is a string which contains the
   mount attributes for the NFS share.
   """
   session = LDAP(DEFAULT_LDAP_DOMAIN, dnsDomain=DEFAULT_DNS_DOMAIN)
   return session.getAutomountInfo(absPath)


#
# Code below is for debugging purposes...
#
if __name__ == "__main__":
   from argparse import ArgumentParser

   parser = ArgumentParser()
   parser.add_argument(metavar='ABSPATH', dest='absPath',
                       help="an absolute path")
   args = parser.parse_args()

   print("Path:        %s" % args.absPath)
   print("LDAP server: %s" % getDefaultLdapServer(DEFAULT_DNS_DOMAIN))

   automountInfo = getAutomountInfoFromAbsPath(args.absPath)
   if automountInfo is None:
      infoStr = "not found"
   else:
      infoStr = "%s:%s (%s)" % automountInfo
   print("Automount:   %s" % infoStr)
