# Copyright 2017-2018 VMware, Inc.
# All rights reserved. -- VMware Confidential

"""Bugzilla API library

Library to query information about bugzilla PRs.
"""
import os
from shutil import copyfile

from borautils.busybox import wget

BZ_WEB_URL = "https://bugzilla.eng.vmware.com"
BZ_FILES_URL = "http://engweb.eng.vmware.com/bugs/files"
BZ_ARCHIVE_URL = "http://engweb.eng.vmware.com/bugs-archive/files"
BZ_NFS_ROOT = os.path.join(os.sep, 'mts', 'bugs', 'files')

# Valid Bugzilla PR numbers range from 1 to 99999999
BZ_MIN_PR_NUM = 1
BZ_MAX_PR_NUM = 99999999


class PR(object):
   """Class to query information about a bugzilla PR.
   """

   def __init__(self, bugNum):
      try:
         bugNum = int(bugNum)
      except (TypeError, ValueError):
         raise ValueError("%s: invalid PR number" % bugNum)

      if not BZ_MIN_PR_NUM <= bugNum <= BZ_MAX_PR_NUM:
         raise ValueError("%u: invalid PR number (out of range [%u, %u])" %
                          (bugNum, BZ_MIN_PR_NUM, BZ_MAX_PR_NUM))

      self.num = bugNum
      paddedBugNum = str(self.num).zfill(len(str(BZ_MAX_PR_NUM)))
      self._bugDir = os.path.join(*list(paddedBugNum))

   @property
   def uploadDir(self):
      """Path to the PR's upload directory.
      """
      uploadDir = os.path.join(BZ_NFS_ROOT, self._bugDir)
      if not os.path.isdir(uploadDir):
         raise FileNotFoundError("failed to find upload directory for PR%u" %
                                 self.num)
      return uploadDir

   def upload(self, src, dest):
      """Upload the @src attachment file to @dest under this PR's upload
      directory.
      """
      dest = os.path.join(self.uploadDir, dest.lstrip('/'))
      if os.path.isdir(dest):
         dest = os.path.join(dest, os.path.basename(src))
      else:
         os.makedirs(os.path.dirname(dest), exist_ok=True)
      copyfile(src, dest)
      return dest

   def download(self, src, dest, archive=False):
      """Download the file @src from this PR's upload directory.

      If 'archive' is True the PR's archive will be searched instead of
      the regular area. The files in older PRs will be archived after a
      while, in which case they are gzipped.
      """
      url = BZ_ARCHIVE_URL if archive else BZ_FILES_URL
      src = os.path.join(url, self._bugDir, src.lstrip('/'))
      if os.path.isdir(dest):
         dest = os.path.join(dest, os.path.basename(src))
      else:
         os.makedirs(os.path.dirname(dest), exist_ok=True)
      wget(src, dest)
      return dest

   def removeAttachment(self, path):
      """Remove an attachment file from this PR's upload directory.
      """
      path = os.path.join(BZ_NFS_ROOT, self._bugDir, path.lstrip('/'))
      os.unlink(path)

   @property
   def url(self):
      """Url to this PR.
      """
      return "%s/show_bug.cgi?id=%u" % (BZ_WEB_URL, self.num)
