# Copyright 2015-2019 VMware, Inc.
# All rights reserved. -- VMware Confidential

"""Build API library

Library to query the Build REST API.
"""
import json
import os
import re
import argparse

from borautils.busybox import md5sum, wget
from borautils.exception import DebugInfoMixIn
from borautils.httputils import urlOpen

VMWARE_ENG_DOMAIN = ".eng.vmware.com"
BUILDAPI_URL = "http://buildapi.eng.vmware.com"
BUILDWEBAPI_URL = "http://buildweb.eng.vmware.com/api"
BUILDSQUID_URL = "http://build-squid.eng.vmware.com"


class GobuildId(tuple):
   """Gobuild ID data type.
   """

   def __new__(cls, buildId):
      try:
         m = re.fullmatch(r"(sb|ob)-(\d+)", buildId, flags=re.IGNORECASE)
      except TypeError:
         raise TypeError("%s() argument must be a string, not '%s'" %
                         (cls.__name__, type(buildId).__name__))

      if m is None:
         raise ValueError("%s: invalid gobuild ID" % buildId)

      builderType = m.group(1).lower()
      buildNum = int(m.group(2))
      if buildNum == 0:
         raise ValueError("0 is not a valid %s build number" % builderType)

      return super().__new__(cls, (builderType, buildNum))

   @property
   def type(self):
      """Builder type ('ob' for Official Build, and 'sb' for Sandbox Build)
      """
      return self[0]

   @property
   def num(self):
      """Build number, not including the 'ob-' or 'sb-' prefix.
      """
      return self[1]

   def __str__(self):
      """Return a human-readable string for this build ID.
      """
      return "%s-%u" % self


class GobuildParserAction(argparse.Action):
   """Argparse action to parse sandbox and official builds arguments.
   """
   def __call__(self, parser, namespace, values, option_string):
      try:
         bld = GobuildId(values)
      except Exception:
         raise argparse.ArgumentTypeError("Only sandbox or official builds "
                                          "are supported")

      setattr(namespace, self.dest, bld)


class BuildAPIRequestError(DebugInfoMixIn, Exception):
   """A request to the Build API has failed.
   """
   _errType = "Build API request failure"

   def __init__(self, errStr, requestUrl):
      super().__init__(errStr)
      self.addDebugInfo('request', requestUrl)

class BuildAPIAttributeNotFound(BuildAPIRequestError):
   """An attribute is missing in the result of a Build API request.
   """
   _errType = "Invalid Build API response"

   def __init__(self, attrName, request):
      errStr = "attribute not found in request result"
      super().__init__(errStr, request)
      self.addDebugInfo('attribute', "'%s'" % attrName)


def jsonQuery(url):
   """Query the given HTTP @url, and parse the result as JSON.
   """
   jsonData = urlOpen(url).read()
   return json.loads(jsonData.decode('utf-8'))


class BuildDeliverable(object):
   """Class representing a build deliverable.
   """

   def __init__(self, attrs, request):
      self._request = request
      self._attrs = attrs

   def _getAttribute(self, name):
      """Return the given build deliverable attribute.
      """
      try:
         return self._attrs[name]
      except KeyError:
         raise BuildAPIAttributeNotFound(name, self._request)

   @property
   def url(self):
      """HTTP URL to this deliverable.
      """
      return self._getAttribute('_download_url')

   @property
   def md5(self):
      """MD5 hash of deliverable.
      """
      return self._getAttribute('md5')

   @property
   def sizeInMb(self):
      """The size of the deliverable, in MB.
      """
      return self._getAttribute('size_in_mb')

   def fetch(self, dest, checkMd5=True):
      """Download this deliverable.
      """
      wget(self.url, dest)

      if os.path.isdir(dest):
         dest = os.path.join(dest, os.path.basename(self.url))

      if checkMd5 and self.md5:
         assert self.md5 == md5sum(dest)

      return dest


class BuildApiObject(object):
   """Generic Build API object - do not instantiate directly.
   """

   def __init__(self, objType, buildNum, context, useBuildwebApi=False):
      buildId = GobuildId("%s-%s" % (context, buildNum))
      self._objType = objType
      self._context = buildId.type
      self._id = buildId.num

      if useBuildwebApi:
         self._request = "%s/%s/%u/%s" % (BUILDWEBAPI_URL, self._context,
                                          self._id, self._objType)
      else:
         self._request = "%s/%s/%s/%u" % (BUILDAPI_URL, self._context,
                                          self._objType, self._id)

      try:
         self._info = jsonQuery(self._request)
      except Exception:
         raise BuildAPIRequestError("JSON request failure", self._request)

   def _getAttribute(self, attrName):
      """Return the value of this object's @attrName attribute.
      """
      try:
         return self._info[attrName]
      except KeyError:
         raise BuildAPIAttributeNotFound(attrName, self._request)

   def _iterResources(self, name):
      """Iterate over a list of resources.
      """
      url = self._getAttribute(name)

      while url is not None:
         query = "%s%s" % (BUILDAPI_URL, url)
         try:
            response = jsonQuery(query)
         except Exception:
            raise BuildAPIRequestError("failed to query %s list resource" %
                                       name, query)

         try:
            resources = response['_list']
         except KeyError:
            raise BuildAPIAttributeNotFound('_list', query)

         for res in resources:
            yield res

         try:
            url = response['_next_url']
         except KeyError:
            raise BuildAPIAttributeNotFound('_next_url', query)

   @property
   def buildNum(self):
      """Build number, i.e. the 123456 part of 'sb-123456'.
      """
      return self._id


class BuildTree(BuildApiObject):
   """Class to query info about a Build tree.
   """

   def __init__(self, buildId, context):
      """Retrieve the build tree info for the given build.

      This function instantiates a BuildTree object, and intitializes it with
      the result of a Build API request for the build tree corresponding to the
      (@buildId, @context) build.
      """
      super().__init__('tree', buildId, context, useBuildwebApi=True)

   def getBuildTree(self, hostType=None):
      """Query the Build API for the path to the build tree.

      Return a (nfsHost, nfsPath, automountPath) tuple where (nfsHost, nfsPath)
      is the NFS server/path to the NFS share on which the build tree is stored,
      and automountPath is the local path at which this NFS share is mounted by
      automount (assuming automount is setup).
      """
      for tree in self._getAttribute('tree'):
         try:
            buildHost = tree['hosttype']
         except KeyError:
            raise BuildAPIAttributeNotFound('hosttype', self._request)

         if hostType is None or buildHost == hostType:
            nfsHost = tree.get('nfs_host', None)
            nfsPath = tree.get('nfs_path', None)
            path = tree.get('automount_path', None)
            if nfsHost and not nfsHost.endswith(VMWARE_ENG_DOMAIN):
               nfsHost += VMWARE_ENG_DOMAIN
            return (nfsHost, nfsPath, path)

      raise BuildAPIRequestError("no store tree found", self._request)


class Build(BuildApiObject):
   """Class to query info about a build.
   """

   def __init__(self, buildId, context):
      """Retrieve the given build info.

      This function instantiates a Build object, and intitializes it with the
      result of a Build API request for the (@buildId, @context) build.
      """
      super().__init__('build', buildId, context)
      self._tree = None
      self._deliverables = None

   def __str__(self):
      """Return a human readable string to identify this build object.
      """
      return "%s-%u" % (self._context, self._id)

   @property
   def buildType(self):
      """Build type (release, beta, obj,...).
      """
      return self._getAttribute('buildtype')

   def isZeroCopy(self):
      """Return True if this build is a 'zeroCopy' build.

      A zeroCopy build is a build for which the build tree is not copied to
      central storage, but is kept on the builder local storage instead.
      """
      try:
         return self._getAttribute('zerocopy') is True
      except Exception:
         return False

   def getBuildTree(self, hostType=None):
      """Return a (nfsHost, nfsPath, automountPath) tuple to the NFS share on
      which the build tree is stored.
      """
      if self._tree is None:
         self._tree = BuildTree(self._id, self._context)
      return self._tree.getBuildTree(hostType=hostType)

   def getPublishDirUrl(self):
      """Retrieve the publish directory for the given @buildId.

      First look if @buildId is a 'zeroCopy' build. If not, then just return the
      build tree URL for that build. Otherwise, lookup the path to the builder's
      local storage, and return an URL to that storage.
      """
      if not self.isZeroCopy():
         url = self._getAttribute('_buildtree_url').rstrip('/')

      else:
         nfsHost, nfsPath, automountPath = self.getBuildTree()
         if automountPath is None:
            url = "http://%s%s" % (nfsHost, nfsPath)
         else:
            url = "%s%s" % (BUILDSQUID_URL, automountPath)

      return url + '/publish'

   def getGobuildTarget(self):
      """Return the name of the gobuild target for this build.
      """
      return self._getAttribute('product')

   @property
   def deliverables(self):
      """Return the list of build deliverables.
      """
      if self._deliverables is None:
         self._deliverables = []
         for deliverable in self._iterResources('_deliverables_url'):
            self._deliverables.append(BuildDeliverable(deliverable,
                                                       self._request))
      return self._deliverables

   def fetch(self, path, dest, checkMd5=False):
      """Download a build deliverable.
      """
      for deliverable in self.deliverables:
         if deliverable.url.endswith(path):
            return deliverable.fetch(dest, checkMd5=checkMd5)
      raise FileNotFoundError("%s: no such deliverable in build %s" %
                              (path, self))


#
# Code below is for debugging purposes...
#
if __name__ == "__main__":
   parser = argparse.ArgumentParser()
   parser.add_argument(type=str, metavar='BLDNUM', dest='build',
                       help="Gobuild build number", action=GobuildParserAction)
   args = parser.parse_args()

   build = Build(args.build.num, args.build.type)

   print("Build: %s (%s)" % (build, build.buildType))
   print("Target: %s" % build.getGobuildTarget())
   print("Publish URL: %s" % build.getPublishDirUrl())
   nfsHost, nfsPath, automountPath = build.getBuildTree()
   print("Tree NFS: %s:%s" % (nfsHost, nfsPath))
   print("Tree automount: %s" % automountPath)
