# Copyright 2015-2018 VMware, Inc.
# All rights reserved. -- VMware Confidential

"""CAT API library

Library to query the CAT API.
"""
from datetime import datetime
import json
import os
import re
import shutil

from borautils.exception import DebugInfoMixIn
from borautils.esx.build import (
   EsxOfficialBuild,
   EsxSandboxBuild,
   getVmtreeHash,
)
from borautils.httputils import httpPost, HttpRequestError, urlOpen
from borautils.testrun import TestRun

CAT_URL = "http://cat.eng.vmware.com"
CAT_API_BASE_URL = "https://cat-api.eng.vmware.com"
CAT_API_VERSION = 'v3.0'
CAT_API = "%s/api/%s" % (CAT_API_BASE_URL, CAT_API_VERSION)
CAT_API_USER = 'svc.vmkernel-qe'
CAT_API_KEY = '1e51584f9d3ae31f0a87458daf8a96e6a6f87274'

NIMBUS_PXECACHE = "http://nimbus-pxecache.eng.vmware.com"

CAT_VALID_RESULTS = ('PASS', 'FAIL', 'PSOD', 'TIMEOUT', 'INVALID', 'INVALID-I')

class CATError(DebugInfoMixIn, Exception):
   """CAT base exception.
   """
   _errType = "CAT error"

class CATEnvironmentError(CATError):
   """CAT provided a bogus environment.
   """
   _errType = "CAT environment error"

   def __init__(self, errStr, app):
      super().__init__(errStr)
      self.addDebugInfo('testrunid', app.getId())

class CATRequestError(CATError):
   """A request to the CAT API has failed.
   """
   _errType = "CAT API request failure"

   def __init__(self, errStr, request):
      super().__init__(errStr)
      self.request = request
      self.addDebugInfo('API version', CAT_API_VERSION)
      self.addDebugInfo('request', self.request)

class CATAttributeError(CATRequestError):
   """The result of a CAT API request is erroneous.
   """
   _errType = "Invalid CAT API response"

   def __init__(self, errStr, attrName, request):
      super().__init__(errStr, request)
      self.addDebugInfo('attribute', "'%s'" % attrName)

class CATAttributeNotFound(CATAttributeError):
   """An attribute is missing in the result of a CAT API request.
   """
   def __init__(self, attrName, request):
      errStr = "attribute not found in request result"
      super().__init__(errStr, attrName, request)

class CATAttributeInvalid(CATAttributeError):
   """The result of an API request contains an invalid attribute.
   """
   def __init__(self, attrName, value, expected, request):
      errStr = "Request result contains an invalid attribute value"
      super().__init__(errStr, attrName, request)
      self.addDebugInfo('got', value)
      self.addDebugInfo('expected', expected)

class CATNoSuchResource(CATRequestError):
   """User requested an invalid CAT resource.
   """
   _errType = "Bad CAT API request"

   def __init__(self, objType, objId, request):
      errStr = "resource not found"
      super().__init__(errStr, request)
      self.addDebugInfo('resource type', objType)
      self.addDebugInfo('resource id', objId)

class CATDaemonUpdateError(DebugInfoMixIn, Exception):
   """Error while contacting the CAT daemonupdate.
   """
   _errType = "CAT daemonupdate error"

   def __init__(self, errStr, url, testrunId, result):
      super().__init__(errStr)
      self.addDebugInfo('request', url)
      self.addDebugInfo('testrun id', testrunId)
      self.addDebugInfo('result', result)


def jsonQuery(url):
   """Query the given HTTP @url, and parse the result as JSON.
   """
   jsonData = urlOpen(url).read()
   return json.loads(jsonData.decode('utf-8'))

def catTimestampToDatetime(timestamp):
   """Convert a CAT timestamp string to a Datetime object.
   """
   return datetime.strptime(timestamp, '%Y-%m-%dT%H:%M:%S')

def postTestResult(testrunId, result):
   """Send a test result to CAT.

   Post @result to CAT for the given @testrunId.
   """
   result = result.upper()
   if result not in CAT_VALID_RESULTS:
      raise ValueError("'%s': invalid test result (expect one of [%s])"
                       % (result, ', '.join(CAT_VALID_RESULTS)))

   url = "%s/tester/testruns/%d/daemonupdate/" % (CAT_URL, testrunId)
   params = {'result': result, 'testrunid': testrunId}
   try:
      response = httpPost(url, params)
   except HttpRequestError:
      raise CATDaemonUpdateError("failed to upload testrun result",
                                 url, testrunId, result)

   try:
      errMsg = response.read().decode('utf-8')
   except Exception:
      pass
   else:
      if "result already posted" in errMsg:
         raise CATDaemonUpdateError(errMsg, url, testrunId, result)

def generateTestInfo(results, resultsDir):
   """Write testinfo.csv providing per testcase information to CAT.

   Results is an array of tuples:
   [(name, success?, runTime, bugReport),
    (name, success?, runTime, bugReport)]

   See https://wiki.eng.vmware.com/ToolsEng/Projects/CAT/TestInfoCSV
   """
   TESTINFO = 'testinfo.csv'
   testinfoPath = os.path.join(resultsDir, TESTINFO)
   with open(testinfoPath, 'w') as testInfo:
      for (name, success, runTime, bugReport) in results:
         # For the testinfo 0 means success
         if success:
            successIndicator = 0
         else:
            successIndicator = 1

         if os.path.exists(bugReport):
            bugReportLink = "<a href=\"%s%s\">Logs</a>" % (CAT_URL, bugReport)
         else:
            bugReportLink = bugReport

         testInfo.write("N/A, %s, %u, %u, %s\n"
                        % (name, successIndicator, runTime, bugReportLink))
   catResultsDir = os.path.dirname(resultsDir)
   shutil.copyfile(testinfoPath, os.path.join(catResultsDir, TESTINFO))


class CATRequest(object):
   """CAT request object
   """

   def __init__(self, request, queryOne=True):
      """Instantiate and format a new CAT request.
      """
      namespace, _, args = request.partition('?')
      if 'format=json' not in args:
         request = "%s?format=json&%s" % (namespace, args)

      self.url = '%s/%s&username=%s&api_key=%s' %\
                 (CAT_API, request, CAT_API_USER, CAT_API_KEY)
      self.queryOne = queryOne

   def _queryOne(self):
      """Query one item from the CAT API.
      """
      try:
         return jsonQuery(self.url)
      except Exception:
         raise CATRequestError("JSON request failure", self)

   def _queryMulti(self, limit=None):
      """Query the CAT API and iterate over the result objects.
      """
      url = self.url

      if limit is not None:
         count = 0
         if re.search(r'limit=\d+', url) is not None:
            url = re.sub(r'limit=\d+', 'limit=%u' % limit, url)
         else:
            url += '&limit=%u' % limit

      while True:
         try:
            jsonData = jsonQuery(url)
         except Exception:
            raise CATRequestError("JSON request failure", self)

         try:
            objects = jsonData['objects']
         except KeyError:
            raise CATAttributeNotFound('objects', self)

         for obj in objects:
            yield obj
            if limit is not None:
               count += 1
               if count == limit:
                  break

         if limit is not None and count == limit:
            break

         try:
            nextRequest = jsonData['meta']['next']
         except KeyError:
            raise CATAttributeNotFound('meta.next', self)

         if nextRequest is None:
            break
         url = '%s/%s' % (CAT_API_BASE_URL, nextRequest)

   def query(self, limit=None):
      """Query the CAT API.

      Post this request to the CAT API, and waits for the result which is
      returned as a dictionnary.
      """
      if self.queryOne:
         return self._queryOne()
      else:
         return self._queryMulti(limit=limit)

   def __str__(self):
      """Return this request's URL as an identification string.
      """
      return self.url


class CatObject(object):
   """Generic CAT object - Do not instantiate directly.
   """
   _objType = None                   # Override me!
   _queryOne = True

   def __init__(self, objId=None, objInfo=None, requestInfo=None):
      """Retrieve info about the given CAT object

      If @objId is specified, this function queries the CAT API against @objId,
      and this object is initialized with the result of that request. Otherwise,
      @objInfo can be used to initialize this object with the result of a
      previous request.
      """
      if objId is not None:
         self._id = objId
         if self._queryOne:
            request = "%s/%u" % (self._objType, self._id)
         else:
            request = "%s/?id=%u" % (self._objType, self._id)

         self._request = CATRequest(request,
                                    queryOne=self._queryOne)
         if self._queryOne:
            self._info = self._request.query()
         else:
            try:
               self._info = next(self._request.query())
            except StopIteration:
               raise CATNoSuchResource(self._objType, self._id, self._request)

      elif objInfo is not None:
         self._request = requestInfo
         self._info = objInfo
         try:
            self._id = int(objInfo['id'])
         except KeyError:
            raise CATAttributeNotFound('id', self._request)
         except ValueError:
            raise CATAttributeInvalid('id', objInfo['id'], 'integer',
                                      self._request)

      else:
         raise ValueError("CatObject() expects a non-null value for either "
                          "@objId or @objInfo")

   @classmethod
   def search(cls, filterArgs=None, limit=None):
      """Query all CAT objects which match the given @filterArgs.
      """
      request = CATRequest("%s/?%s" % (cls._objType, filterArgs),
                           queryOne=False)
      for objInfo in request.query(limit=limit):
         yield cls(objInfo=objInfo, requestInfo=request)

   def getId(self):
      """Return this CAT object's ID.
      """
      return self._id

   def _getAttribute(self, attrName, subAttrName=None):
      """Return the value of this object's @attrName attribute.
      """
      try:
         value = self._info[attrName]
      except KeyError:
         raise CATAttributeNotFound(attrName, self._request)

      if subAttrName is not None:
         try:
            value = value[subAttrName]
         except KeyError:
            raise CATAttributeNotFound("%s.%s" % (attrName, subAttrName),
                                       self._request)
      return value

   def _getIdFromObjUrl(self, objType, url):
      """Return <id> from an url formatted as "/api/<version>/<objType>/<id>/".
      """
      if url:
         m = re.match(r'/api/%s/%s/(\d+)/' % (CAT_API_VERSION, objType), url)
         if m is not None:
            return int(m.group(1))
      return None

   def _getIdFromObjUrlInAttr(self, attrName, objType):
      """Parse this object's @attrName attribute as an object URL.

      Read this CAT object's @attrName attribute, and parse it as an URL
      pointing to another object of type @objType. This function returns the ID
      of the object pointed to by the parsed URL.
      """
      url = self._getAttribute(attrName)
      objId = self._getIdFromObjUrl(objType, url)
      if objId is None:
         raise CATAttributeInvalid(attrName, url,
                                   '/api/<version>/%s/<id>/' % objType,
                                   self._request)
      return objId

   def _getIdsFromObjUrlInAttr(self, attrName, objType):
      """Parse this object's @attrName attribute as an object URL.

      Read this CAT object's @attrName attribute, and parse it as a list of URLs
      pointing to objects of type @objType. This function returns a list of IDs
      for the objects pointed to by the parsed URL.
      """
      objIds = []
      for url in self._getAttribute(attrName):
         objId = self._getIdFromObjUrl(objType, url)
         if objId is None:
            raise CATAttributeInvalid(attrName, url,
                                      "/api/<version>/%s/<id>/" % objType,
                                      self._request)
         objIds.append(objId)
      return objIds


class CatViewType(CatObject):
   """Class to query info about a CAT view type.
   """
   _objType = 'viewtype'

   PRODUCTION = 1
   STAGING = 2
   EXPERIMENTAL = 3


class CatSLA(CatObject):
   """Class to query CAT SLA
   """
   _objType = 'sla'

   def getSLAName(self):
      """Retrun SLA name
      """
      return self._getAttribute('name')


class CatArea(CatObject):
   """Class to query info about a CAT area.
   """
   _objType = 'area'

   def getViewType(self):
      """Return the view type (production, staging or experimental) for this
      area.
      """
      viewTypes = {CatViewType.PRODUCTION: 'production',
                   CatViewType.STAGING: 'staging',
                   CatViewType.EXPERIMENTAL: 'experimental'}

      view = self._getIdFromObjUrlInAttr('view_type', 'viewtype')
      try:
         return viewTypes[view]
      except KeyError:
         raise CATAttributeInvalid("invalid area viewtype: %s" % str(view))

   def getSLAName(self):
      """Return SLA name for this area.
      """
      slaId = self._getIdFromObjUrlInAttr('sla_type', 'sla')
      sla = CatSLA(slaId)
      return sla.getSLAName()


class CatDeliverable(CatObject):
   """Class to query info about a CAT deliverable.
   """
   _objType = 'deliverable'

   def getBuildContext(self):
      """Return this deliverable's builder context ('cat', 'sb', or 'ob')
      """
      VALID_BUILD_CONTEXTS = {1: 'cat', 2: 'sb', 3: 'ob'}

      buildContextId = self._getIdFromObjUrlInAttr('builder_context',
                                                   'buildercontext')
      try:
         return VALID_BUILD_CONTEXTS[buildContextId]
      except KeyError:
         raise CATAttributeInvalid('bcontext', str(buildContextId),
                                   "One of %s" % str(VALID_BUILD_CONTEXTS),
                                   self._request)

   def getBuildId(self):
      """Return this deliverable's (builderContext, buildNum) build ID.
      """
      buildContext = self.getBuildContext()
      if buildContext == 'cat':
         buildNum = None
      else:
         buildNum = self._getAttribute('sandbox_build_id')
      return (buildContext, buildNum)

   def getChangeset(self):
      """Return the changeset number (CLN) at which this deliverable was built.
      """
      buildId = self._getIdFromObjUrlInAttr('build', 'build')
      cln =  CatBuild(buildId).getChangeSet()
      try:
         return int(cln)
      except (TypeError, ValueError):
         raise CATAttributeInvalid('build.changeset', cln, '<integer>',
                                   self._request)

   def getBuildStorage(self):
      """Get this deliverable's build storage NFS directory.
      """
      return (self._getAttribute('nfs_server'), self._getAttribute('path'))

   def _getCatBuildPxeSubdir(self, product, bldType):
      """Return the boot modules subdirectory for CAT builds.

      Return the subdirectory, relative to the PXE root directory, in which
      the boot modules have been published by the build.
      """
      _, path = self.getBuildStorage()
      bora = os.path.join(path, 'bora')
      bora = re.sub(r"\d+/bora", "current/bora", bora)
      treeHash = getVmtreeHash(bora)
      return os.path.join(treeHash, '%s-x64-26-%s' % (product, bldType))

   def getPxeSubdir(self, product=None):
      """Return the path of this build's PXE directory, relatively to the NFS
      share in which CAT stores all PXE directories.
      """
      buildContext = self.getBuildContext()
      buildId = self._getIdFromObjUrlInAttr('build', 'build')
      build = CatBuild(buildId)
      branch = build._getAttribute('branch_name')
      bldType = build._getAttribute('build_type_name')
      coverage = build._getAttribute('coverage')
      changeset = build._getAttribute('changeset')
      cov = 'cov' if coverage else 'clean'

      if buildContext in ['ob', 'sb']:
         dirPath = [buildContext, branch, bldType, cov, changeset]
         targets = self._getAttribute('targets')
         dirPath.append("%s-%s" % (targets.split(',')[0], product))
      elif buildContext == 'cat':
         pxeSubDir = self._getCatBuildPxeSubdir(product, bldType)
         dirPath = [branch, bldType, cov, changeset, product,
                    str(self._id), pxeSubDir]
      else:
         raise NotImplemented('Unsupported build context (%s)' % buildContext)
      return os.path.join(*dirPath)

   def getPxeDir(self, product=None):
      """Return an the NFS path of this build's PXE directory.
      """
      locationId = self._getIdFromObjUrlInAttr('location', 'location')
      location = CatLocation(locationId)
      nfsServer = location._getAttribute('nfs_server')
      pxeDir = location._getAttribute('pxe_dir')
      return (nfsServer, os.path.join(pxeDir, self.getPxeSubdir(product)))

   def getPxeUrl(self, product=None):
      """Return an HTTP URL pointing to this build's PXE directory.
      """
      locationId = self._getIdFromObjUrlInAttr('location', 'location')
      location = CatLocation(locationId)
      if location._getAttribute('name') == 'PA':
         httpServer = NIMBUS_PXECACHE
      else:
         httpServer = location._getAttribute('http_server')

      pxeDir = location._getAttribute('pxe_dir')

      if not httpServer.startswith("http://"):
         httpServer = "http://%s" % httpServer
      return os.path.join(httpServer, pxeDir.lstrip('/'),
                          self.getPxeSubdir(product))

   def getPxeBaseDir(self, product=None):
      """Return the NFS path of this build's PXE base directory.

      The PXE base directory is the build-specific root folder which contains
      all of the PXE directories for that build (there might be multiple PXE
      directories for one given build, if this build was deployed several times
      with different boot options).
      """
      nfsServer, deployDir = self.getPxeDir(product)
      buildContext = self.getBuildContext()
      if buildContext == 'cat':
         deployDir = os.path.dirname(os.path.dirname(os.path.dirname(deployDir)))
      return nfsServer, deployDir

   def getPxeBaseUrl(self, product=None):
      """Return an HTTP URL pointing to this build's PXE base directory.

      The PXE base directory is the build-specific root folder which contains
      all of the PXE directories for that build (there might be multiple PXE
      directories for one given build, if this build was deployed several times
      with different boot options).
      """
      buildContext = self.getBuildContext()
      deployUrl = self.getPxeUrl(product=product)
      if buildContext == 'cat':
         deployUrl = os.path.dirname(os.path.dirname(os.path.dirname(deployUrl)))
      return deployUrl

   def getBuildTargets(self):
      """Return the build target(s).
      """
      return self._getAttribute('targets').split(',')


class CatMachine(CatObject):
   """Class to query info about a CAT machine.
   """
   _objType = 'machine'

   def getMachineInfo(self):
      """Return this machine's info as a dictionary.
      """
      machine = {}
      for key in ('ilo_hostname', 'ilo_username', 'ilo_password',
                  'serial_host', 'serial_port', 'suite_location', 'hostname'):
         machine[key] = self._getAttribute(key)
      machine['macAddrs'] = self._getAttribute('mac_addresses')
      machine['ipAddrs'] = self._getAttribute('ip_address')

      try:
         machine['serial_port'] = int(machine['serial_port'])
      except Exception:
         # Serial information might be empty or invalid. Let's not raise an
         # exception here, so we can still get the rest of the machine info.
         pass

      return machine


class CatTester(CatObject):
   """Class to query info about a CAT tester.
   """
   _objType = 'tester'

   def getMachines(self):
      """Return this tester's machines.
      """
      machineIds = self._getIdsFromObjUrlInAttr('machines', 'machine')
      return [CatMachine(m) for m in machineIds]


class CatLocation(CatObject):
   """Class to query info about a CAT location.
   """
   _objType = 'location'

   def getLocationName(self):
      """Return the name of this location
      """
      return self._getAttribute('name')

   def getResultsDir(self):
      """Return the NFS path of this location's result directory
      """
      return (self._getAttribute('nfs_server'),
              self._getAttribute('results_path'))


class CatBuild(CatObject):
   """Class to query info about a CAT build.
   """
   _objType = 'build'

   def getChangeSet(self):
      """Return the changeset of this build
      """
      return self._getAttribute('changeset')


class CatTestrun(CatObject):
   """Class to query info about a CAT testrun.
   """
   _objType = 'testrun'

   def getArea(self):
      """Return the area id for this testrun.
      """
      return self._getIdFromObjUrlInAttr('area', 'area')

   def getViewType(self):
      """Return the view type for this testrun.
      """
      area = CatArea(objId=self.getArea())
      return area.getViewType()

   def getWorkunit(self):
      """Return the workunit id for this testrun.
      """
      return self._getIdFromObjUrlInAttr('workunit', 'workunit')

   def getWorkload(self):
      """Return the workload id for this testrun.
      """
      return self._getIdFromObjUrlInAttr('workload', 'workload')

   def getStartTime(self):
      """Return the time at which this testrun has started.
      """
      startTime = self._getAttribute('start_time')
      if startTime is not None:
         return catTimestampToDatetime(startTime)
      else:
         return None

   def getEndTime(self):
      """Return the time at which this testrun has finished.
      """
      endTime = self._getAttribute('end_time')
      if endTime is not None:
         return catTimestampToDatetime(endTime)
      else:
         return None

   def getRuntime(self):
      """Return this testrun's runtime.
      """
      endTime = self.getEndTime()
      if endTime is None:
         endTime = datetime.now()
      return endTime - self.getStartTime()

   def getResult(self):
      """Return this testrun's result.
      """
      return self._getAttribute('result')

   def getResultsDir(self):
      """Return this testrun's results directory.
      """
      return self._getAttribute('results_directory')

   def getResultsUrl(self):
      """Return an HTTP URL to this testrun's results directory.
      """
      return os.path.join(CAT_URL,
                          self._getAttribute('results_directory').lstrip('/'))

   def getDeliverables(self):
      """Return this testrun's deliverables as a list of deliverable IDs.
      """
      return self._getIdsFromObjUrlInAttr('deliverables', 'deliverable')

   def getTester(self):
      """Return the tester id for this testrun.
      """
      testerId = self._getIdFromObjUrlInAttr('tester', 'tester')
      return CatTester(testerId)

   def getMachines(self):
      """Return this testrun's test machines as a list of machine IDs.
      """
      return self.getTester().getMachines()

   def getBuildId(self):
      """Return this testrun's builds as a list of (buildContext, buildNum).

      Possible build contexts are:
        'cat': CAT local builder
        'sb':  Bobuild Sandbox Build
        'ob':  Gobuild Official Build
      """
      return [CatDeliverable(d).getBuildId() for d in self.getDeliverables()]

   def getProduct(self):
      """Return the name of the product tested by this testrun.
      """
      return self._getAttribute('host_platform')

   def getLocation(self):
      """Return the location id for this testrun.
      """
      locationId = self._getIdFromObjUrlInAttr('location', 'location')
      return CatLocation(locationId)

   def getLocationName(self):
      """Return the location name of this testrun.
      """
      return self.getLocation().getLocationName()

   def getResultsNfsDir(self):
      """Return this testrun's results NFS directory.
      """
      return self.getLocation().getResultsDir()

   def getSLAName(self):
      """Return SLA name for this testrun.
      """
      area = CatArea(objId=self.getArea())
      return area.getSLAName()


class CatWorkunit(CatObject):
   """Class to query info about a CAT workunit.
   """
   _objType = 'workunit'

   def isEnabled(self):
      """Return True if this workunit is enabled.
      """
      return 'tester' in self._info and self._getAttribute('enabled')


class CatApp(CatTestrun, TestRun):
   """Class to help writing CAT-ready applications with zero knowlegde of CAT
   internals, and without duplicating a whole bunch of CAT-specific glue code
   (or polluting applications with this glue code).
   """

   @classmethod
   def addArguments(cls, argsParser):
      """Add all the CAT specific CLI parameters.
      """
      argsParser.add_argument('--testrunid', type=int, dest='testrunId',
                              help="CAT testrun ID")

   def __init__(self, name, testrunId=None, product='visor', **kwargs):
      self._name = name
      self._product = product
      self._resultsDir = None
      self._deliverable = None
      self._testResults = []
      self._viewType = None

      if testrunId is None:
         raise EnvironmentError('process not running in CAT')

      try:
         CatTestrun.__init__(self, objId=testrunId)
      except Exception:
         try:
            self._uploadResults(testrunId, 'INVALID')
         except Exception:
            pass
         raise

      self._product = self.getProduct()
      if product and product != self._product:
         raise CATEnvironmentError("'%s': invalid product type "
                                   "(expected '%s')" % (self._product, product),
                                   self)

      TestRun.__init__(self, self._name, self.getId())

   @property
   def isProduction(self):
      """Return True if this testrun is a production workload.
      """
      if self._viewType is None:
         self._viewType = self.getViewType()
      return self._viewType == "production"

   def getResultsDir(self):
      """Return the local path to the NFS-mounted CAT results directory.
      """
      resultsDir = TestRun.getResultsDir(self)
      if resultsDir is None:
         resultsDir = CatTestrun.getResultsDir(self)
      return resultsDir

   def _setupResultsDir(self, resultsDir):
      """Check and sanitize the CAT results directory

      First assume that the results directory as reported by CAT is mounted. If
      it isn't, then hope that autofs is setup and supports auto-mounting NFS
      shares under /net/.

      TODO: Remove the empty '<machine>' subfolder, which is created by CAT at
            the root the results dir. We don't use this empty subfolder, and it
            pollutes the results directory unnecessarily.
      """
      if resultsDir is None:
         host, resultsDir = self.getResultsNfsDir()
         resultsDir = os.path.join(os.sep, 'net', host,
                                   resultsDir.lstrip(os.sep))

         if not os.path.isdir(resultsDir):
            raise CATEnvironmentError("%s: CAT results dir not mounted" %
                                      resultsDir, self)

      super()._setupResultsDir(resultsDir)

   def _getCachedEsxDeliverable(self):
      """Cache and return the ESX deliverable info for this testrun.
      """
      if not self._deliverable:
         for deliverableId in self.getDeliverables():
            self._deliverable = CatDeliverable(objId=deliverableId)
            break
      return self._deliverable

   def getPxeDir(self):
      """Return the NFS path to the PXE directory for this testrun.
      """
      deliverable = self._getCachedEsxDeliverable()
      host, pxeBaseDir = deliverable.getPxeBaseDir(self._product)
      pxeDir = os.path.join(pxeBaseDir, '%u-pxe' % self.getId())
      return host, pxeDir

   def getPxeUrl(self):
      """Return an HTTP URL pointing to the PXE directory for this testrun.
      """
      deliverable = self._getCachedEsxDeliverable()
      pxeBaseUrl = deliverable.getPxeBaseUrl(self._product)
      return os.path.join(pxeBaseUrl, '%u-pxe' % self.getId())

   def getEsxBuild(self, installer=False):
      """Return an ESXBuild object containing info about this testrun's build.
      """
      deliverable = self._getCachedEsxDeliverable()
      builderType, buildId = deliverable.getBuildId()
      if builderType == 'cat':
         raise CATEnvironmentError("WDC CAT builders are no longer supported")
      elif builderType == 'sb':
         esxBuild = EsxSandboxBuild(buildId, installer)
      elif builderType == 'ob':
         esxBuild = EsxOfficialBuild(buildId, installer)
      else:
         raise CATEnvironmentError("'%s': invalid builder type" % builderType,
                                   self)
      return esxBuild

   def reportTestResult(self, name, passed, runTime, bugReportPath):
      """Add a test result to the list of test results to be uploaded at the end
      of a testrun.
      """
      self._testResults.append((name, passed, runTime, bugReportPath))

   def _uploadResults(self, result):
      """Upload the list of test results and report final testrun status to CAT.
      """
      resultToCAT = {TestRun.PASS:    'PASS',
                     TestRun.FAIL:    'FAIL',
                     TestRun.PSOD:    'PSOD',
                     TestRun.TIMEOUT: 'TIMEOUT',
                     TestRun.INVALID: 'INVALID'}

      if self._testResults:
         try:
            generateTestInfo(self._testResults, self.getResultsDir())
         except Exception:
            # test results info is optional. Let's do our best to provide it,
            # but let's not abort a test workload if that fails.
            pass

      try:
         catResult = resultToCAT[result]
         postTestResult(self.getId(), catResult)
      except Exception:
         # CAT doesn't allow posting a testrun results twice. Ignore such
         # failures.
         pass


#
# Code below is for debugging purposes...
#
if __name__ == "__main__":
   from argparse import ArgumentParser

   parser = ArgumentParser()
   CatApp.addArguments(parser)
   args, _ = parser.parse_known_args()
   testrun = CatApp('fakeWorkload', testrunId=args.testrunId)

   print("Query: %s" % testrun._request)
   print("Testrun: %u" % testrun.getId())
   product = testrun.getProduct()
   print("Product: %s" % product)
   print("View: %s" % testrun.getViewType())

   for deliverableId in testrun.getDeliverables():
      print("Deliverable: %u" % deliverableId)
      deliverable = CatDeliverable(deliverableId)
      print("  Targets: %s" % ",".join(deliverable.getBuildTargets()))
      print("  CLN: %s" % deliverable.getChangeset())
      print("  Build: %s-%s" % deliverable.getBuildId())
      print("  Tree: %s:%s" % deliverable.getBuildStorage())
      print("  PXE dir: %s:%s" % deliverable.getPxeDir(product))
      print("  PXE URL: %s" % deliverable.getPxeUrl(product))
      print("  PXE base dir: %s:%s" % deliverable.getPxeBaseDir(product))
      print("  PXE base URL: %s" % deliverable.getPxeBaseUrl(product))

   print("Workunit: %u" % testrun.getWorkunit())
   print("Workload: %u" % testrun.getWorkload())
   print("Start: %s" % testrun.getStartTime())
   print("End: %s" % testrun.getEndTime())
   print("Runtime: %s" % testrun.getRuntime())
   print("Results dir: %s" % testrun.getResultsDir())
   print("Results URL: %s" % testrun.getResultsUrl())
   print("Results NFS dir: %s:%s" % testrun.getResultsNfsDir())
   print("Location: %s" % testrun.getLocationName())

   tester = testrun.getTester()
   print("Tester: %u" % tester.getId())

   for machine in tester.getMachines():
      print("Machine: %s" % machine.getId())
      for key, val in machine.getMachineInfo().items():
         print("  %s: %s" % (key, val))

   print("Similar workunits:")
   for workunit in CatWorkunit.search('workload__id=%u'
                                      % testrun.getWorkload()):
      print("   workunit: %u (%sabled)" %
            (workunit.getId(), "en" if workunit.isEnabled() else "dis"))
