"""DBC (Development Build Cluster) library
"""
import os

from .automount import getAutomountInfoFromAbsPath

class DBC(object):
   """Class to represent a DBC repository.
   """

   def __init__(self, path=None):
      if path is None:
         path = os.getcwd()
      path = os.path.normpath(path)

      try:
         dbc, self._host, self._user = path.split(os.sep)[1:4]
      except (IndexError, ValueError):
         raise ValueError("invalid DBC path: %s" % path)
      else:
         if dbc != 'dbc':
            raise ValueError("invalid DBC path: %s" % path)

      self._rootDir = "/dbc/%s/%s" % (self._host, self._user)
      self._rootUrl = "http://%s.eng.vmware.com/%s" % (self._host, self._user)
      self._nfs = None

   @property
   def host(self):
      """Name of the DBC server hosting this repository.
      E.g.: 'pa-dbc1234'
      """
      return self._host

   @property
   def user(self):
      """Username of this DBC repository.
      """
      return self._user

   @property
   def rootDir(self):
      """Absolute path to the root of this DBC repository.
      E.g.: '/dbc/pa-dbc1234/rvotlz'
      """
      return self._rootDir

   @property
   def rootUrl(self):
      """HTTP URL pointing to the root of this DBC repository.
      E.g.: 'http://pa-dbc1234.eng.vmware.com/rvotlz'
      """
      return self._rootUrl

   @property
   def nfs(self):
      """NFS share info as a (host, path, attr) tuple.
      E.g.: ('pa-dbc1234.eng.vmware.com',
             '/dbc/pa-dbc1234',
             '-rw,hard,intr,rsize=8192,wsize=8192,bg,tcp,vers=3')
      """
      if self._nfs is None:
         self._nfs = getAutomountInfoFromAbsPath(self._rootDir)

      if self._nfs is None:
         raise OSError("failed to retrieve NFS information for DBC host %s" %
                       self._host)

      return self._nfs

   def toUrl(self, path):
      """Convert a DBC path into an HTTP URL pointing to the same file/dir.
      """
      relPath = os.path.relpath(path, self._rootDir)
      if relPath.startswith(os.pardir):
         raise ValueError("invalid DBC path: %s" % path)

      if relPath == '.':
         return self._rootUrl
      else:
         return os.path.join(self._rootUrl, relPath)

   def __str__(self):
      """Return a human-readable string to identify this DBC repository.
      """
      return "%s@%s" % (self._user, self._host)


#
# Code below is for debugging purposes...
#
if __name__ == "__main__":
   dbc = DBC()
   print("%s" % str(dbc))
   print("  host: %s" % dbc.host)
   print("  user: %s" % dbc.user)
   print('  root: %s' % dbc.rootDir)
   print('  url: %s' % dbc.rootUrl)
   print("  nfs: %s:%s (%s)" % dbc.nfs)
