"""Email utilities

Library to send email to people.
"""
from borautils.exception import DebugInfoMixIn
from email.mime.text import MIMEText
from smtplib import SMTP, SMTPException

SMTP_SERVER = 'smtp.vmware.com'

class EmailError(DebugInfoMixIn, Exception):
   """Generic send email error.
   """
   _errType = 'Email Error'

class ServiceUnreachableError(EmailError):
   """SMTP service is unreachable
   """
   _errType = 'SMTP service unreachable'

   def __init__(self):
      super().__init__(self._errType)
      self.addDebugInfo("SMTP Server", SMTP_SERVER)

class InvalidEmailAddress(EmailError):
   """Email address is invalid.
   """
   _errType = 'Invalid email address'

   def __init__(self, recipent):
      super().__init__(self._errType)
      self.addDebugInfo('invalid email address', recipent)

class SendEmailError(EmailError):
   """Send email error.
   """
   _errType = 'Send email error'

   def __init__(self, subject, body, sender, recipients):
      errStr = 'Fail to send email'
      super().__init__(errStr)
      self.addDebugInfo('subject', subject)
      self.addDebugInfo('body', body)
      self.addDebugInfo('from', sender)
      self.addDebugInfo('to', recipients)


class Email:
   """Class to send emails
   """

   def __init__(self):
      """Connect to SMTP server
      """
      self._server = None
      try:
         self._server = SMTP(SMTP_SERVER)
      except Exception:
         raise ServiceUnreachableError()

   def __del__(self):
      """Disconnect from SMTP server
      """
      if self._server is not None:
         self._server.quit()

   def _checkEmailAddress(self, emailAddress):
      """Validate email address
      """
      if '@' not in emailAddress:
         raise InvalidEmailAddress(emailAddress)

   def sendEmail(self, subject, body, sender, recipients):
      """Send email to recipients.
      """
      recipients = list(set(recipients))
      self._checkEmailAddress(sender)

      for recipient in recipients:
         self._checkEmailAddress(recipient)

      msg = MIMEText(body, 'plain')
      msg['Subject'] = subject
      msg['From'] = sender
      msg['To'] = ', '.join(recipients)

      try:
         self._server.sendmail(sender, recipients, msg.as_string())
      except SMTPException:
         raise SendEmailError(subject, body, sender, recipients)


#
# Code below is for debugging purposes...
#
if __name__ == "__main__":
   from argparse import ArgumentParser

   parser = ArgumentParser()
   parser.add_argument(type=str, metavar='SUBJECT', dest='subject',
                       help="The subject of email")

   parser.add_argument(type=str, metavar='BODY', dest='body',
                       help="The body of email")

   parser.add_argument(type=str, metavar='SENDER', dest='sender',
                       help="Email address of the sender.")

   parser.add_argument('--recipient', type=str, metavar='RECIPIENT',
                       dest='recipients', action='append',
                       help="Email address of recipient. This option may be "
                            "passed multiple times.")

   args = parser.parse_args()

   email = Email()
   email.sendEmail(args.subject, args.body, args.sender, args.recipients)
