# Copyright 2017-2018 VMware, Inc.
# All rights reserved. -- VMware Confidential

"""Nimbus VMs creation and management

This library provides a python interface to deploy and manage ESX vms in Nimbus.
"""
from getpass import getuser
import json
import os
from random import choice
from string import ascii_lowercase, digits
import subprocess
import sys
from tempfile import mkstemp
import time

from borautils.exception import DebugInfoMixIn
from borautils.infra.toolchain import loadTool
from borautils.mac import MacAddr

YAML_PATH = 'noarch/python-pyyaml-3.11/lib/python3.3/site-packages'
sys.path.append(loadTool(YAML_PATH))
import yaml

class NimbusError(DebugInfoMixIn, Exception):
   """Generic Nimbus error.
   """
   _errType = "Nimbus error"

   def __init__(self, errStr, vm):
      super().__init__(errStr)
      self.addDebugInfo('vm name', str(vm))
      if vm.logFile is not None:
         self.addDebugInfo('logs', vm.logFile)
      self.addDebugInfo('site', vm.site)
      if vm.pod is not None:
         self.addDebugInfo('pod', vm.pod)


class NimbusVmNotFoundError(NimbusError):
   """Vm does not exist.
   """
   _errType = "No such nimbus vm"


class NimbusCmdError(NimbusError):
   """Exception to report a Nimbus CLI command error.
   """
   _errType = "Nimbus command error"

   def __init__(self, errStr, cmd, errorCode, vm):
      super().__init__(errStr, vm)
      self.addDebugInfo('command', cmd)
      if errorCode is not None:
         self.addDebugInfo('error code', errorCode)


class NimbusResultError(NimbusError):
   """Exception to report an invalid Nimbus result (json) file.
   """
   _errType = "Nimbus result file error"

   def __init__(self, errStr, resultFilePath, vm):
      super().__init__(errStr, vm)
      self.addDebugInfo('result file', resultFilePath)


class NimbusVmSpec(dict):
   """A class to load and store a nimbus vm spec.
   """
   @classmethod
   def fromFile(cls, path):
      """Create spec from a yaml file.
      """
      with open(path) as f:
         data = yaml.safe_load(f)
      return NimbusVmSpec(data)

   def updateNotNone(self, **kwargs):
      """A convenient method to update key/value pairs but ignore pairs where
         value is None.
      """
      self.update({k: v for k, v in kwargs.items() if v is not None})

   def toArgv(self):
      """Return an argv-like array of nimbus options for this spec.
      """
      argv = []
      for option, value in self.items():
         if isinstance(value, list):
            for inValue in value:
               argv.extend(['--%s' % option, str(inValue)])
         elif value is not None:
            argv.extend(['--%s' % option, str(value)])
         else:
            argv.append('--%s' % option) # non-value option
      return argv

   def toText(self):
      """Return current spec as text.
      """
      return "".join(['%s: %s\n' % (k, v) for k, v in self.items()])

   def toFile(self, path):
      """Write current spec to a file.
      """
      with open(path, 'w') as f:
         f.write(self.toText())


class NimbusVm(object):
   """Base class to instantiate and manage nimbus vms.
   """

   @staticmethod
   def _makeRandomName():
      """Create a random nimbus vm name.
      """
      hashStr = "".join(choice(ascii_lowercase + digits) for _ in range(10))
      return "%s-%s" % (getuser(), hashStr)

   def __init__(self, name=None, logFile=None, site='sc', annotations=None,
                specFile=None):
      """Instantiate, but do not create a Nimbus vm.
      """
      if name is None:
         self._name = self._makeRandomName()
      else:
         self._name = name.strip()

      if logFile:
         self._logFilePath = os.path.abspath(logFile)
         self._logFile = open(self._logFilePath, 'w+')
      else:
         self._logFilePath = None
         self._logFile = None

      self._site = site
      self._pod = None
      self._macs = None
      self._ip = None
      self._serialProcess = None
      self._annotations = annotations
      if specFile:
         self.spec = NimbusVmSpec.fromFile(specFile)
      else:
         self.spec = None

   def __del__(self):
      """Close this vm's log file, and destruct the vm object.
      """
      if self._logFile is not None:
         try:
            self._logFile.close()
         except Exception:
            pass

   def _nimbusPopen(self, argv, stdout=None, stderr=None, nimbusRoot=None):
      """Execute a Nimbus CLI command.
      """
      if stdout is None:
         stdout = subprocess.PIPE
      if stderr is None:
         stderr = subprocess.PIPE

      # Default Nimbus source tree is located under /mts/git/...
      if nimbusRoot is None:
         nimbusRoot = os.path.join(os.sep, 'mts', 'git')
      binPath = argv.pop(0)
      argv.insert(0, os.path.join(nimbusRoot, binPath))

      # Set NIMBUS env var to the vm's pod
      if self._pod:
         env = os.environ.copy()
         env['NIMBUS'] = self.pod
         env['NIMBUS_LOCATION'] = self.site
      else:
         env = None

      try:
         return subprocess.Popen(argv, stdout=stdout, stderr=stderr, env=env)
      except OSError as e:
         raise NimbusCmdError("failed to execute nimbus command: %s"
                              % e.strerror, ' '.join(argv), None, self)

   def create(self, cmd, tmpDir=None):
      """Create a nimbus vm, (but don't power it on).
      """
      tmpResultFd, tmpResultPath = mkstemp(dir=tmpDir, prefix='nimbus-result-',
                                           suffix='.json')
      os.close(tmpResultFd)

      cmd[1:1] = ['--result', tmpResultPath]
      if self._annotations:
         for k, v in self._annotations.items():
            cmd[1:1] = ['--annotation', '%s:%s' % (k, v)]
      p = self._nimbusPopen(cmd, stdout=self._logFile, stderr=self._logFile)
      p.communicate()
      if p.returncode != 0:
         raise NimbusCmdError('failed to deploy nimbus vm', ' '.join(cmd),
                              p.returncode, self)

      try:
         jsonResult = open(tmpResultPath, 'r')
      except OSError as e:
         raise NimbusResultError('failed to load result file (%s)'
                                 % e.strerror, tmpResultPath, self)

      try:
         vmInfo = json.load(jsonResult)
      except Exception as e:
         raise NimbusResultError("invalid result file (%s)" % str(e),
                                 tmpResultPath, self)
      finally:
         jsonResult.close()

      try:
         vmName = vmInfo['name'].strip()
      except KeyError:
         raise NimbusResultError("missing 'name' entry", tmpResultPath, self)

      if self._name not in vmName:
         raise NimbusResultError("inconsistent vm name (got:%s, expected:%s)"
                                 % (vmName, self._name), tmpResultPath, self)
      if self._name != vmName:
         self._name = vmName

      try:
         deployStatus = vmInfo['deploy_status'].lower().strip()
      except KeyError:
         raise NimbusResultError("missing 'deploy_status' entry",
                                 tmpResultPath, self)

      if deployStatus != 'success':
         raise NimbusError('Nimbus failed to deploy %s vm (status=%s)'
                           % (vmName, deployStatus), self)

      try:
         self._pod = vmInfo['pod'].strip()
      except KeyError:
         raise NimbusResultError("missing 'pod' entry", tmpResultPath, self)

      try:
         self._macs = []
         for mac in vmInfo['macaddrs']:
            try:
               self._macs.append(MacAddr(mac))
            except ValueError:
               raise NimbusResultError("invalid mac address: %s" % mac)
      except KeyError:
         raise NimbusResultError("missing 'macaddrs' entry",
                                 tmpResultPath, self)

      os.unlink(tmpResultPath)
      return vmInfo

   def _nimbusCtl(self, op, timeout=None, evalOutput=False):
      """Send a control command to this Nimbus vm.
      """
      NO_ARGS_OPS = ('list',)

      cmdLine = [os.path.join('bin', 'nimbus-ctl'),
                 '--nimbusLocation', self.site]
      if evalOutput:
         cmdLine += ['--outputFormat', 'json']
      cmdLine += op.split()
      if op not in NO_ARGS_OPS:
         cmdLine += [self._name]

      if evalOutput:
         p = self._nimbusPopen(cmdLine)
      else:
         p = self._nimbusPopen(cmdLine, stdout=self._logFile,
                               stderr=self._logFile)

      try:
         out, _ = p.communicate(timeout=timeout)
      except subprocess.TimeoutExpired:
         p.kill()
         p.communicate()
         raise

      if p.returncode != 0:
         raise NimbusCmdError("'%s' control command has failed" % op,
                              ' '.join(cmdLine), p.returncode, self)

      if evalOutput:
         out = out.decode().strip()
         try:
            # <WORKAROUND PR1795204>
            # Even with the '--outputFormat json' option, nimbus-ctl still
            # prints non-json logs on <stdout>. We have to ignore those, and
            # skip to the last line of <stdout>, which contains the json-
            # formatted result we care about.
            out = out.split('\n')[-1]
            # </WORKAROUND>

            jsonData = json.loads(out)
            result = jsonData['json_info']
         except Exception:
            raise NimbusCmdError("'%s' control command returned invalid json "
                                 "output: %s" % (op, out),
                                 ' '.join(cmdLine), p.returncode, self)
         return result

   def exists(self):
      """Return True is this vm exists.
      """
      try:
         result = self._nimbusCtl('list', evalOutput=True)
      except NimbusCmdError:
         # <WORKAROUND PR1795225>
         # 'nimbus-ctl list' prints out an error message and returns with
         # non-zero exit code if no vm is found. We have to interpret that as
         # if the command had succeeded, and just returned an empty list.
         return False
         # </WORKAROUND>

      try:
         _ = result[self._pod][self._name]
      except KeyError:
         return False
      return True

   def powerOn(self):
      """Power on this Nimbus vm.
      """
      self._nimbusCtl('on')

   def powerOff(self):
      """Power off this Nimbus vm.
      """
      self._nimbusCtl('off')

   def isOn(self):
      """Return True if this vm is powered on.
      """
      result = self._nimbusCtl('is_on', evalOutput=True)
      try:
         status = result[self._pod][self._name]
      except KeyError:
         raise NimbusVmNotFoundError("failed to query vm power status", self)
      return status == 'on'

   def reset(self):
      """Reset this Nimbus vm.
      """
      self._nimbusCtl('reset')

   def kill(self):
      """Power off and destroy this Nimbus vm.
      """
      self._nimbusCtl('kill')

   def waitForIp(self, timeout=None):
      """Wait until this VM gets assigned an IP addess.
      """
      result = self._nimbusCtl('ip', timeout=timeout, evalOutput=True)
      try:
         self._ip = result[self._pod][self._name]
      except KeyError:
         raise NimbusVmNotFoundError("failed to query vm IP", self)
      return self._ip

   def bootOrder(self, order='hdd,eth'):
      """Set the bootorder, hdd or eth first.
      """
      op = '-m %s --bootOrder=%s set-bootorder' % (self._pod, order)
      self._nimbusCtl(op)

   @property
   def ip(self):
      """This VM's IP address.
      """
      return self._ip

   def readSerial(self):
      """Iterate over each line of this vm's serial port output.

      This function does not return until the .killReadSerialProcess() method
      is called.
      """
      cmd = [os.path.join('bin', 'nimbus-vmserial'), self._name]
      p = self._nimbusPopen(cmd, stderr=self._logFile)
      self._serialProcess = p

      while p.poll() is None:
         line = p.stdout.readline()
         if line:
            yield line.decode(errors='replace').strip('\r')
         else:
            time.sleep(.1)

      if p.returncode > 0:
         raise NimbusCmdError("failed to connect to vm's serial port",
                              ' '.join(cmd), p.returncode, self)

   def killReadSerialProcess(self):
      """kill read serial process
      """
      if self._serialProcess is None:
         return

      p = self._serialProcess
      self._serialProcess = None

      p.terminate()
      try:
         p.wait(timeout=10)
      except subprocess.TimeoutExpired:
         # Just in case nimbus-vmserial fails to terminate within 10sec.
         p.kill()
         p.wait()

   @property
   def name(self):
      """Name of this nimbus vm.
      """
      return self._name

   @property
   def logFile(self):
      """Absolute path to this vm's log file.
      """
      return self._logFilePath

   @property
   def pod(self):
      """Name of this vm's Nimbus pod.
      """
      return self._pod

   @property
   def macs(self):
      """List of MAC addresses for this Nimbus vm.
      """
      return [str(mac) for mac in self._macs]

   @property
   def site(self):
      """Name of the geographical site this vm is deployed at.
      """
      return self._site

   @property
   def location(self):
      """PXE location of this Nimbus vm.
      """
      DEFAULT_PXE_LOCATIONS = {'sc': 'sc-rdops',
                               'wdc': 'wdc-rdops'}

      try:
         return DEFAULT_PXE_LOCATIONS[self.site]
      except KeyError:
         NimbusError('%s: unknown geographical site', self.site)

   def __str__(self):
      """Return a human-readable string to identify this Nimbus vm.

      The string matches the following format: "nimbus:<pod>:<vmName>", where:
      <pod> is the name of the nimbus pod on which the vm lives (or '_' if the
      vm hasn't been assigned a pod yet, and <vmName> is the name of the vm.
      """
      pod = self.pod
      if pod is None:
         pod = '_'
      return "nimbus:%s:%s" % (pod, self._name)


class NimbusWorkerVm(NimbusVm):
   """Class to deploy and manage Linux Dynamic Worker vms in nimbus.
   """
   USERNAME = 'worker'
   PASSWORD = 'ca$hc0w'

   def create(self, tmpDir=None):
      """Create a nimbus vm.
      """
      argv = [os.path.join('bin', 'nimbus-genericdeploy'),
              '--create-only', '--nimbusLocation', self.site,
              '--type', 'worker-template',
              self._name]
      super().create(argv, tmpDir=tmpDir)


class NimbusEsxVm(NimbusVm):
   """Class to deploy and manage an ESX vm in Nimbus.
   """
   _defaultNumVcpus = 2
   _defaultMemSizeInMb = 6 * 1024
   _defaultDiskCapacityInKb = 12 * 1024 * 1024
   _defaultCpuReservationInMHzPerVcpu = 200
   _defaultMemoryReservationInMb = 2 * 1024
   _defaultNumVnics = 1
   _defaultLease = 2

   def __init__(self, name=None, cpus=None, memory=None, disk=None,
                nics=None, lease=None, **kwargs):
      """Instantiate, but do not create an ESX Nimbus vm.
      """
      super().__init__(name=name, **kwargs)
      if self.spec is None:
         self.spec = NimbusVmSpec(cpus=self._defaultNumVcpus,
                                  cpuReservation=self._defaultNumVcpus *\
                                                 self._defaultCpuReservationInMHzPerVcpu,
                                  memory=self._defaultMemSizeInMb,
                                  memoryReservation=self._defaultMemoryReservationInMb,
                                  pvscsiDisk=self._defaultDiskCapacityInKb,
                                  nics=self._defaultNumVnics,
                                  lease=self._defaultLease)
      self.spec.updateNotNone(cpus=cpus, memory=memory, pvscsiDisk=disk,
                              nics=nics, lease=lease)

   def create(self, tmpDir=None):
      """Create an ESX nimbus vm, (but don't power it on).
      """
      FAKE_PXE_DIR = os.devnull

      cmd = [os.path.join('bin', 'nimbus-esxdeploy'),
             '--nimbusLocation', self.site, '--disableSasDisks',
             '--create-only']
      cmd.extend(self.spec.toArgv())
      cmd.extend([self._name, FAKE_PXE_DIR])
      super().create(cmd, tmpDir=tmpDir)


class NimbusNfsVm(NimbusVm):
   """Class to deploy and manage an NFS VM
   """
   USERNAME = 'root'
   PASSWORD = 'ca$hc0w'
   _defaultNumVcpus = 4
   _defaultMemSizeInMb = 4 * 1024
   _defaultDiskCapacityInKb = 4 * 1024 * 1024
   _defaultLease = 2
   _defaultMountPoint = "datastore11"

   def __init__(self, name=None, cpus=None, memory=None, disk=None, lease=None,
                mountPoint=None, **kwargs):
      """Instantiate, but do not create an NFS Nimbus vm.
      """
      super().__init__(name, **kwargs)
      if self.spec is None:
         self.spec = NimbusVmSpec(cpus=self._defaultNumVcpus,
                                  memory=self._defaultMemSizeInMb,
                                  disk=self._defaultDiskCapacityInKb,
                                  mountPoint=self._defaultMountPoint,
                                  lease=self._defaultLease)
      self.spec.updateNotNone(cpus=cpus, memory=memory, disk=disk,
                              mountPoint=mountPoint, lease=lease)

   def create(self, tmpDir=None):
      """Create an NFS nimbus vm
      """
      cmd = [os.path.join('bin', 'nimbus-nfsdeploy'),
             '--nimbusLocation', self.site]
      cmd.extend(self.spec.toArgv())
      cmd.append(self._name)

      super().create(cmd, tmpDir=tmpDir)


class NimbusIscsiVm(NimbusVm):
   """Class to deploy and manage an iscsi VM
   """
   _defaultNumVcpus = 4
   _defaultMemSizeInMb = 4 * 1024
   _defaultLunInGb = 4
   _defaultLun4kInGb = 6
   _defaultLease = 2

   def __init__(self, name=None, cpus=None, memory=None, lun=None, lun4k=None,
                lease=None, **kwargs):
      """Instantiate, but do not create an iscsi Nimbus vm.
      """
      super().__init__(name, **kwargs)
      if self.spec is None:
         self.spec = NimbusVmSpec(cpus=self._defaultNumVcpus,
                                  memory=self._defaultMemSizeInMb,
                                  lun=self._defaultLunInGb,
                                  lun4k=self._defaultLun4kInGb,
                                  lease=self._defaultLease)
      self.spec.updateNotNone(cpus=cpus, memory=memory, lun=lun, lun4k=lun4k,
                              lease=lease)

   def create(self, tmpDir=None):
      """Create an iscsi nimbus vm
      """
      cmd = [os.path.join('bin', 'nimbus-iscsideploy'),
             '--nimbusLocation', self.site]
      cmd.extend(self.spec.toArgv())
      cmd.append(self._name)

      super().create(cmd, tmpDir=tmpDir)


class NimbusOvfVm(NimbusVm):
   """Class to deploy VM from OVF and manage the VM
   """
   _defaultLease = 2

   def __init__(self, ovfFile, name=None, lease=None, **kwargs):
      """Instantiate, but do not create an vm.
      """
      super().__init__(name, **kwargs)
      self._ovfFile = ovfFile
      if self.spec is None:
         self.spec = NimbusVmSpec(lease=self._defaultLease)
      self.spec.updateNotNone(lease=lease)
      self._vmInfo = None

   def create(self, tmpDir=None):
      """Create a nimbus VM from OVF file.
      """
      cmd = [os.path.join('bin', 'nimbus-ovfdeploy')]
      cmd.extend(self.spec.toArgv())
      cmd.extend([self._name, self._ovfFile])
      self._vmInfo = super().create(cmd, tmpDir=tmpDir)

   @property
   def vmInfo(self):
      """Because all configurations are determined by OVF, return the whole
         deployment result to reflect VM information (CPU/Memory/Disk/...).
      """
      return str(self._vmInfo)


#
# Code below is for debugging purposes...
#
if __name__ == "__main__":
   from argparse import ArgumentParser
   from pxe import pxeConfig

   parser = ArgumentParser()
   parser.add_argument(type=str, metavar='VM_NAME', dest='vmName',
                       help="vm name")
   parser.add_argument(type=str, metavar='PXE_DIR', dest='pxeDir',
                       help="absolute path to PXE directory")
   args = parser.parse_args()

   logFileFd, logFilePath = mkstemp(prefix='nimbus-esx-vm-', suffix='.log')
   os.close(logFileFd)

   vm = NimbusEsxVm(args.vmName, logFile=logFilePath)

   print("Creating %s vm..." % args.vmName)
   vm.create()
   print("  vm: %s" % str(vm))
   location = vm.location
   print("  logs: %s" % vm.logFile)
   macs = vm.macs
   print("  boot MAC: %s" % macs)
   print("  pod: %s" % vm.pod)
   print("  location: %s" % location)

   print("Configuring PXE against %s..." % args.pxeDir)
   for mac in macs:
      pxeConfig(mac, location, args.pxeDir)

   print('Powering on...')
   vm.powerOn()

   print("Connecting to serial port...")
   try:
      for l in vm.readSerial():
         print(l, end='')
   except KeyboardInterrupt:
      print("Received keyboard interrupt, killing vm...")
      vm.kill()
   except Exception:
      print("Error detected, killing vm...")
      try:
         vm.kill()
      except Exception:
         pass
      raise
