"""PXE/iPXE server configuration

Library to setup Tech-ops DHCP tables in order to associate a MAC address with a
PXE directory. This module relies on the PXEconfig.pl tool provided by Tech-ops.
"""
import os
import time

from borautils.exception import DebugInfoMixIn
from borautils.mac import MacAddr
from borautils.ssh import SSH

SUITE_URI = 'suite1.eng.vmware.com'
SUITE_USERNAME = 'pxeuser'
SUITE_PASSWD = 'ca$hc0w'

class PxeconfigError(DebugInfoMixIn, Exception):
   """Generic Pxeconfig error.
   """
   _errType = "PXE configuration error"

class PxeconfigCommandError(PxeconfigError):
   """Pxeconfig returned an error.
   """
   _errType = "PXEconfig command error"

   def __init__(self, cmd, returnCode, err):
      errStr = "PXEconfig executed, but returned an error"
      super().__init__(errStr)
      self.addDebugInfo('command', cmd)
      self.addDebugInfo('exit status', returnCode)
      self.addDebugInfo('stderr', err)


def _runPxeConfig(sshSession, mac, location, pxeDir=None, ipxeConf=None,
                  reset=False):
   """Do the call to PXEconfig Tech-ops tool.
   """
   cmd = './PXEconfig.pl -m %s -l %s' % (str(mac), location)

   if ipxeConf:
      cmd += ' -i'

   if reset:
      cmd += ' -r'
   else:
      if pxeDir is not None:
         cmd += ' -p %s' % pxeDir
      if ipxeConf is not None:
         # PXEconfig.pl requires the iPXE configuration file to actually exist.
         if not os.path.exists(ipxeConf):
            raise PxeconfigError("%s: iPXE bootstrap not found" % ipxeConf)
         cmd += ' -f %s' % ipxeConf

   status, out, _ = sshSession.run(cmd)
   if status != 0:
      raise PxeconfigCommandError(cmd, status, out)

def pxeReset(macAddr, location):
   """Wipe out both PXE and iPXE server configurations for this MAC address.
   """
   mac = MacAddr(macAddr)

   with SSH(SUITE_URI, username=SUITE_USERNAME, password=SUITE_PASSWD) as ssh:
      _runPxeConfig(ssh, mac, location, reset=True)
      _runPxeConfig(ssh, mac, location, reset=True, ipxeConf=True)

def pxeConfig(macAddr, location, pxeDir, ipxeConf=None, attempts=3):
   """Setup PXE/iPXE for the given @mac address.

   Update the Tech-ops DHCP tables, so that @mac boots from to the given PXE
   directory.
   """
   mac = MacAddr(macAddr)

   for i in range(attempts):
      try:
         with SSH(SUITE_URI, username=SUITE_USERNAME,
                  password=SUITE_PASSWD) as ssh:
            _runPxeConfig(ssh, mac, location, pxeDir=pxeDir)
            if ipxeConf:
               cfg = os.path.join(pxeDir, ipxeConf)
               _runPxeConfig(ssh, mac, location, pxeDir=pxeDir, ipxeConf=cfg)
      except Exception:
         if i == (attempts - 1):
            raise
         else:
            time.sleep(5)
      else:
         return


#
# Code below is for debugging purposes...
#
if __name__ == "__main__":
   from argparse import ArgumentParser

   parser = ArgumentParser()
   parser.add_argument(type=str, metavar='MAC', dest='mac',
                       help="MAC address of the machine to boot")
   parser.add_argument('--location', type=str, metavar='LOCATION',
                       dest='location', default="PromB",
                       help="location of the machine to boot")
   parser.add_argument('--pxedir', type=str, metavar='DIR', dest='pxeDir',
                       help="absolute path to your PXE directory")
   parser.add_argument('--ipxe', type=str, metavar='FILE', dest='ipxeConf',
                       help="absolute path to your ipxe.conf")
   args = parser.parse_args()

   if args.pxeDir is None:
      pxeReset(args.mac, args.location)
   else:
      pxeConfig(args.mac, args.location, args.pxeDir, ipxeConf=args.ipxeConf)
