"""Toolchain access library

WARNING:
  The common desire is to slowly deprecate the toolchain. Please always consider
  a local approach rather than accessing the toolchain over NFS. When possible,
  also use packaged components (gobuild, cayman,...) to avoid depending on the
  toolchain.
"""
import os

from borautils.exception import DebugInfoMixIn

TOOLCHAIN_HOST = 'build-toolchain'
TOOLCHAIN_NFS_PATH = '/toolchain'
TCROOT = '/build/toolchain'
TCROOT_ENV = 'TCROOT'

class ToolchainError(DebugInfoMixIn, Exception):
   """A generic toolchain error.
   """
   _errType = "Toolchain error"

   def __init__(self, errStr, tcRoot=None):
      super().__init__(errStr)
      self.addDebugInfo('TCROOT', getTcRoot(tcRoot=tcRoot))

class ToolchainNotMounted(ToolchainError):
   """Toolchain is not accessible.
   """
   def __init__(self, tcRoot=None):
      errStr = "toolchain not mounted"
      super().__init__(errStr, tcRoot=tcRoot)

class ToolchainNotFound(ToolchainError):
   """A file (or directory) could not be found in the toolchain.
   """
   def __init__(self, tool, tcRoot=None):
      errStr = "No such file or directory"
      super().__init__(errStr, tcRoot=tcRoot)
      self.addDebugInfo('path', tool)

def getTcRoot(tcRoot=None):
   """Return the toolchain root directory.
   """
   if tcRoot is not None:
      return tcRoot
   tcRoot = os.environ.get(TCROOT_ENV, None)
   if tcRoot:
      tcRoot = tcRoot.strip()
   return tcRoot if tcRoot else TCROOT

def loadTool(relPath, tcRoot=None, checkExists=True):
   """Find a tool in the toolchain.

   If @checkExists is True (default), also make sure that @relPath points to an
   existing file (or directory) in the toolchain.
   """
   tcRoot = getTcRoot(tcRoot=tcRoot)
   relPath = relPath.strip(os.sep)
   path = os.path.join(tcRoot, relPath)

   if checkExists and not os.path.exists(path):
      if not os.path.isdir(tcRoot):
         raise ToolchainNotMounted(tcRoot=tcRoot)
      else:
         raise ToolchainNotFound(relPath, tcRoot=tcRoot)

   return path


#
# Code below is for debugging purposes...
#
if __name__ == "__main__":
   from argparse import ArgumentParser

   parser = ArgumentParser()
   parser.add_argument('--tcroot', type=str, metavar='DIR', dest='tcRoot',
                       help="path to the toolchain root directory")

   parser.add_argument(type=str, metavar='RELPATH', dest='tool',
                       help="path to a file relative to $TCROOT")

   args = parser.parse_args()

   tcRootEnv = os.environ.get(TCROOT_ENV, None)
   print("$TCROOT: %s" % ("<not set>" if tcRootEnv is None else tcRootEnv))
   print("TCROOT: %s" % getTcRoot(tcRoot=args.tcRoot))
   print("PATH: %s" % loadTool(args.tool, tcRoot=args.tcRoot))
