# Copyright 2017-2019 VMware, Inc.  All rights reserved. -- VMware Confidential

"""Linux dmesg parser.
"""

import os
import re
import tempfile
import borautils.logparser


class DmesgParser(borautils.logparser.LogParser):
   """dmesg log parser
   """

   # dmesg timestamp prefix
   PREFIX_RE = r' *\[ +(?P<timestamp>\d+\.\d+)\] *'

   # dmesg log message regex rules
   MSG_RULES = {
      # Linux version
      'version':
         r'Linux version (?P<version>.*)$',

      # TSC freq read from hypervisor
      'hv_tsc':
         r'TSC freq read from hypervisor : '
         r'(?P<hv_tsc_freq>\d+\.\d+) MHz',

      # vmware-stealtime info (Photon OS)
      'stealtime':
         r'vmware-stealtime: cpu\s+(?P<cpu>\d+),\s+'
         r'pa\s+(?P<pa>[0-9a-f]+).*$',

      # Host bus clock speed read from hypervisor
      'host_bus_clock':
         r'Host bus clock speed read from hypervisor : 66000000 Hz'
   }


   def __init__(self, **kwargs):
      super().__init__(**kwargs)
      rules = {}
      for name, msgRule in self.MSG_RULES.items():
         rules[name] = self.PREFIX_RE + '.*' + msgRule
      self.AddRules(rules)


if __name__ == '__main__':
   sampleDmesg = """
   [    0.000000] Initializing cgroup subsys cpuset
   [    0.000000] Initializing cgroup subsys cpu
   [    0.000000] Initializing cgroup subsys cpuacct
   [    0.000000] Linux version 4.4.8-esx
   [    0.000000] Command line: BOOT_IMAGE=/boot/vmlinuz-esx-4.4.8 root=/dev/sda2 ro net.ifnames=0 biosdevname=0 consoleblank=0 powersaved=off
   [    0.000000] Disabled fast string operations
   [    0.023836] vmware-stealtime: cpu 0, pa 7fc0c300
   [    0.037908] Console: colour VGA+ 80x25
   [    0.046920] console [tty0] enabled
   [    0.047290] clocksource: hpet: mask: 0xffffffff max_cycles: 0xffffffff, max_idle_ns: 133484882848 ns
   [    0.047531] hpet clockevent registered
   [    0.047554] TSC freq read from hypervisor : 1999.998 MHz
   [    0.047602] Host bus clock speed read from hypervisor : 66000000 Hz
   [    0.047655] tsc: Detected 1999.998 MHz processor
   [    0.047751] Calibrating delay loop (skipped) preset value.. 3999.99 BogoMIPS (lpj=7999992)
   [    0.047937] pid_max: default: 32768 minimum: 301
   """

   print("Running basic tests ...")
   match = DmesgParser().ParseString(sampleDmesg)

   m = next(match)
   assert (m.name == 'version' and
           m.groups['version'] == '4.4.8-esx'), m

   m = next(match)
   assert (m.name == 'stealtime' and
           m.groups['cpu'] == '0' and
           m.groups['pa']  == '7fc0c300'), m

   m = next(match)
   assert (m.name == 'hv_tsc' and
           m.groups['hv_tsc_freq'] == '1999.998'), m

   m = next(match)
   assert (m.name == 'host_bus_clock'), m

   print("OK")

# vim: sw=3 ts=3
