# Copyright 2018 VMware, Inc. All rights reserved. -- VMware Confidential

"""Dell/iDrac Redfish client.
"""
import time

from borautils.redfish import Client


class DellRedfish(Client):
   """A Redfish client that also implements Dell-specific (iDrac) features.
   """

   def exportScp(self, target="ALL"):
      """Export (extract) a Server Configuration Profile.

      This method blocks until the SCP is actually retrieved (or an error is
      returned)
      """
      actions = self.root.Managers.Members[0].Actions
      path = actions['Oem']['OemManager.v1_0_0#OemManager.ExportSystemConfiguration']['target']

      payload = {"ExportFormat": "JSON",
                 "ExportUse": "Clone",
                 "ShareParameters": {"Target": target}}
      task = self.post(path, payload)

      while task.poll() is None:
         time.sleep(1)

      return task.body

   def importScp(self, scp, target="ALL"):
      """Import (apply) a Server Configuration Profile.

      This method blocks until the SCP is actually applied (or an error is
      returned).
      """
      actions = self.root.Managers.Members[0].Actions
      path = actions['Oem']['OemManager.v1_0_0#OemManager.ImportSystemConfiguration']['target']

      payload = {"ImportBuffer": scp,
                 "ShareParameters": {"Target": target}}
      task = self.post(path, payload)

      # TODO: We need proper error handling and reporting here; the task
      #       may have failed and we do not detect that.
      while task.poll() is None:
         time.sleep(1)

      return task.body

   def rebootFromHttp(self, uri, bootNic="NIC.HttpDevice.1-1"):
      """Reboot this server over HTTP.

      Per Dell recommendation, we use an SCP to setup the server to reboot over
      UEFI HTTP. Applying the SCP automatically reboots the server.
      """
      # TODO: The "HTTP Device" is not enabled correctly. Enter the
      #       System Setup -> Boot Settings -> UEFI Boot Settings and
      #       deselect the HTTP Device. Applying the SCP will not reenable
      #       this setting!
      scp = self.exportScp(target="BIOS")

      for component in scp['SystemConfiguration']['Components']:
         if ('Attributes' not in component or
             component.get('FQDD') != 'BIOS.Setup.1-1'):
            continue

         for attr in component['Attributes']:
            # Only UEFI supports booting over HTTP (legacy BIOS doesn't)
            if attr['Name'] == 'BootMode':
               attr['Value'] = "Uefi"

            # Disable PXE
            elif attr['Name'] == 'PxeDev1EnDis':
               attr['Value'] = 'Disabled'

            # Enable HTTP boot
            elif attr['Name'] == 'HttpDev1EnDis':
               attr['Value'] = 'Enabled'

            # Setup boot URI
            elif attr['Name'] == 'HttpDev1Uri':
               attr['Value'] = uri

            # Set HTTP Boot as the first option in the boot sequence
            elif attr['Name'] == "UefiBootSeq":
               bootDevs = [s.strip() for s in attr['Value'].split(',')]
               try:
                  bootDevs.remove(bootNic)
               except ValueError:
                  pass
               bootDevs.insert(0, bootNic)
               attr['Value'] = ", ".join(bootDevs)

      self.importScp(json.dumps(scp), target="BIOS")
