########################################################################
# Copyright 2018 VMware, Inc. All rights reserved.
# -- VMware Confidential
########################################################################

"""A RedFish Python library

A Python library suitable for interacting with a RedFish service
instance. RedFish [1] is an industry standard created by the
Distributed Management Task Force (DMTF) that aims to replace IPMI and
certain parts of vendor dependent server system management solutions
such as iDRAC (Dell) and ILO (HPE). RedFish services expose a REST
based API to control a server system. Supported operations range from
previously possible actions such as resetting the system, through
changes of the boot device and/or order, to taking and applying
snapshots of the entire BMC configuration.

The design of this library aims to abstract away from the underlying
HTTP details and vendor specific paths wherever possible. It provides an
easy to use interface that directly harnesses Python's object notation
to navigate through the RedFish namespace.

The library comes with a set of tests. We make use of the DMTF RedFish
mockup creator [2]. This Python script can be used to capture the
RedFish related state of a live system. In order to serve this image
to run our tests against we use the DMTF RedFish mockup server [3]--a
simple Python script, located in the test/ directory.

[1] https://www.dmtf.org/standards/redfish
[2] https://github.com/DMTF/Redfish-Mockup-Creator
[3] https://github.com/DMTF/Redfish-Mockup-Server
"""
from http import (
   HTTPStatus,
)
from urllib.parse import (
   urljoin as urlJoin,
)

from borautils.rest import RestClient

# The protocol version we assume.
PROTOCOL_VERSION = "v1"
# The root resource from which traversal starts.
ROOT = "/redfish/%s/" % PROTOCOL_VERSION


class HTTPError(RuntimeError):
   """An exception object for HTTP related errors."""
   def __init__(self, status, message=""):
      """Initialize the error object."""
      super().__init__(message)
      self._status = status


   @property
   def status(self):
      """Retrieve the HTTP error code."""
      return self._status


def parameterized(decorator):
   """Decorator helper that allows for other decorators to accept parameters."""
   def outer(*args, **kwargs):
      def inner(decorated):
         return decorator(decorated, *args, **kwargs)
      return inner
   return outer


@parameterized
def allowableStatuses(decorated, statuses=(HTTPStatus.OK,)):
   """Allow for a function to return certain statuses, raise an error otherwise."""
   def overwritten(self, path, *args, statuses=statuses, **kwargs):
      # We expect the decorated function to return a requests-style
      # response object.
      response = decorated(self, path, *args, **kwargs)
      if response.status not in statuses:
         raise HTTPError(response.status, "Request for %s returned unexpected "
                         "status: %d (%s)" % (path, response.status, response.reason))

      return response

   return overwritten


@parameterized
def autoRead(decorated, read=True):
   """A decorator draining an HTTP response object."""
   def overwritten(self, path, *args, read=read, **kwargs):
      response = decorated(self, path, *args, **kwargs)
      # We use a single connection for all our requests. In order for
      # multiple requests to go through we need to ensure that the
      # response is read entirely, otherwise a ResponseNotReady error is
      # the result.
      if read:
         _ = response.read()
      return response

   return overwritten


# TODO: Should derive from AbstractContextManager once we have Python 3.6.
# TODO: Add support for advanced error reporting using resources from
#       the RedFish instance (to resolve message IDs and such).
class Client(RestClient):
   """A client useful for connecting to a RedFish instance."""

   @allowableStatuses((HTTPStatus.OK, HTTPStatus.ACCEPTED,
                       HTTPStatus.NO_CONTENT))
   def get(self, path):
      """Make a GET request to the given resource."""
      # TODO: We probably want to support redirects (302?) here
      #       transparently.
      return super().get(path)


   def post(self, path, data):
      """Make a POST request to the given resource."""
      response = super().post(path, data)
      return Task.fromPost(self, response)


   @autoRead()
   @allowableStatuses((HTTPStatus.OK, HTTPStatus.NO_CONTENT))
   def patch(self, path, data):
      """Make a PATCH request to the given resource."""
      return super().patch(path, data)


   @autoRead()
   @allowableStatuses()
   def delete(self, path):
      """Make a DELETE request to the given resource."""
      return super().delete(path)


   @property
   def root(self):
      """Retrieve the root RedFish resource from the client."""
      return Resource(self, ROOT)


class ResourceNotFound(RuntimeError):
   """An exception raised when a resource cannot be found."""
   pass


class ResourceMixin:
   def __init__(self, client, path, data):
      """Initialize the mixin."""
      super().__init__(data)
      # We need to set the attribute in the dictionary directly to avoid
      # going through __setitem__/__setattr__ during boot strap of the
      # object.
      self.__dict__["_client"] = client
      self.__dict__["_path"] = path


   def __getitem__(self, key):
      """Retrieve an element from the resource. Maps to HTTP GET."""
      try:
         result = super().__getitem__(key)
      except KeyError:
         raise ResourceNotFound(key)

      # We need special handling for keys that are slices. That is the
      # case when we use slice syntax on a list. In that case we just
      # want the default behavior.
      if isinstance(key, slice):
         return result
      elif isinstance(result, dict) and "@odata.id" in result:
         path = result["@odata.id"]
         return Resource(self._client, path)
      elif isinstance(result, list):
         return ResourceList(self._client, urlJoin(self._path, key), result)
      elif isinstance(result, dict):
         return ResourceDict(self._client, urlJoin(self._path, key), result)

      return result

   # We want to support native attribute syntax, so we map all attribute
   # retrieval calls to __getitem__.
   __getattr__ = __getitem__


   def __setitem__(self, key, value):
      """Updates an item. Maps to HTTP PATCH."""
      item = self[key]
      self._client.patch(item.path, value)

   __setattr__ = __setitem__


   def __delitem__(self, key):
      """Delete an item. Maps to HTTP DELETE."""
      # We have to "dereference" the item to retrieve its actual path.
      item = self[key]
      self._client.delete(item.path)

   __delattr__ = __delitem__


   @property
   def path(self):
      """Retrieve the resource's path."""
      return self._path


class ResourceList(ResourceMixin, list):
   """A RedFish resource JSON list."""
   pass


class ResourceDict(ResourceMixin, dict):
   """A RedFish resource JSON dict."""
   pass


class Resource(ResourceMixin, dict):
   """A RedFish resource."""
   def __init__(self, client, path):
      """Initialize a RedFish resource by retrieving it."""
      response = client.get(path)
      resource = client.readBody(response)
      super().__init__(client, path, resource)

class Task(object):
   """A Task object represents a synchronous or asynchronous operation being
   handled by the Redfish service in the result of a POST request.
   """

   def __init__(self, client, uri=None, body=None):
      self._client = client
      self._uri = uri
      self._body = body
      self._resource = None

   def _readTaskResource(self, response, uri):
      """Read a task resource fromt the given POST/GET response.
      """
      assert response.status == HTTPStatus.ACCEPTED

      task = Client.readBody(response)
      if task:
         # sanity check
         if 'TaskState' not in task:
            raise AttributeError("missing 'TaskState' attribute in response "
                                 "for %s" % uri)

      self._resource = task

   @classmethod
   def _handleSyncPost(cls, client, response):
      """Instantiate a Task object from the response of a synchronous POST
      request.
      """
      if response.status not in (HTTPStatus.OK, HTTPStatus.NO_CONTENT):
         raise HTTPError(response.status, "Request for synchronous POST "
                         "returned unexpected status: %d (%s)" %
                         (response.status, response.reason))

      body = client.readBody(response)

      return cls(client, body=body)

   @classmethod
   def _handleAsyncPost(cls, client, response):
      """Instantiate a Task object from the response of an asynchronous POST
      request.
      """
      taskUri = response.getheader('Location')
      if taskUri is None:
         raise AttributeError("missing 'Location' attribute in response for "
                              "asynchronous post request (%s)" % self._uri)

      task = cls(client, uri=taskUri)
      task._readTaskResource(response, taskUri)
      return task

   @classmethod
   def fromPost(cls, client, response):
      """Instantiate a Task object from the response of a POST request.
      """
      if response.status == HTTPStatus.ACCEPTED:
         return cls._handleAsyncPost(client, response)
      else:
         return cls._handleSyncPost(client, response)

   def poll(self):
      """Check whether a task operation has completed.

      This functions returns the response body of the scheduled operation,
      or None if the task has not completed yet.

      TODO: handle the Retry-After HTTP header.
      """
      if not self.isAsync:
         return self._body

      response = self._client.get(self._uri)

      if response.status in (HTTPStatus.OK, HTTPStatus.NO_CONTENT):
         self._body = self._client.readBody(response)
      elif response.status == HTTPStatus.ACCEPTED:
         self._readTaskResource(response, self._uri)

      return self._body

   @property
   def body(self):
      """Return the response body of the scheduled operation.

      This assumes that the task operation has completed, which can be verified
      using the Task.poll() method.
      """
      if self._body is None:
         raise RuntimeError("%s: failed to query the body of this "
                            "asynchronous task (task has not completed)." %
                            self._uri)
      return self._body

   @property
   def isAsync(self):
      """Return True is this task is asynchronous.
      """
      return self._uri is not None
