########################################################################
# Copyright 2018 VMware, Inc. All rights reserved.
# -- VMware Confidential
########################################################################

"""Higher-level RedFish based operations

This module builds on the `redfish` module to provide higher-level
abstractions. Examples include: system power operations, reconfiguration
of boot devices etc.
"""

from borautils.waitfor import (
   waitFor,
)
from datetime import (
   timedelta,
)
from enum import (
   Enum,
)
from http import (
   HTTPStatus,
)
from re import (
   search as regexSearch,
)
from redfish import (
   HTTPError,
)


class Stringable(Enum):
   """Base class for an enum that provides basic interaction with strings."""
   def __str__(self):
      """Convert the stringifiyable enum into a string."""
      return str(self.name)

   def __eq__(self, string):
      """Check for equality with a string."""
      return str(self.name) == string


class PowerState(Stringable):
   """An enum representing different power states a system can be in."""
   # TODO: Once we have Python 3.6 we should use auto().
   Off = 0
   On = 1


class PowerAction(Stringable):
   """An enum representing different power actions.

      Note that clients may not support all nodes.
   """
   ForceOff = 0
   On = 1
   GracefulRestart = 2
   PushPowerButton = 3
   NMI = 4


def powerAction(client, action):
   """Apply a power action to the system."""
   path = client.root.Systems.Members[0].Actions["#ComputerSystem.Reset"]["target"]
   data = {
      "ResetType": str(action),
   }
   client.post(path, data)


def ensureOff(client):
   """Make sure the system is powered off."""
   powerStatus = client.root.Systems.Members[0].PowerState
   if powerStatus == PowerState.On:
      powerAction(client, PowerAction.ForceOff)


def ensureReset(client):
   """Make sure to reset the system."""
   powerStatus = client.root.Systems.Members[0].PowerState
   if powerStatus == PowerState.Off:
      powerAction(client, PowerAction.On)
   else:
      powerAction(client, PowerAction.GracefulRestart)


def deleteCompletedJobs(client):
   """Delete all completed jobs.

   If there are too many jobs in the system the Redfish interface may
   not work properly anymore, responding to valid requests with error
   codes.

   TODO: We currently only support Dell systems.
   """
   jobs = client.root.Managers.Members[0].Links.Oem.Dell.Jobs
   # TODO: We need to support range syntax.
   for i in range(len(jobs.Members)):
      try:
         del jobs.Members[i]
      except HTTPError as e:
         # The job may just got deleted in which case it may no
         # longer be found.
         if e.status != HTTPStatus.NOT_FOUND:
            raise


def bootFromUrl(client, url):
   """Configure a server to boot over HTTP.

   Note that this operation is potentially very long running, as it
   involves multiple reboots (some explicitly asked for by us, others
   performed by the firmware).

   TODO: We currently only support Dell systems.
   TODO: We currently hardcode the boot source override target. The
         reason being that the Dell system seems to be buggy and does
         not accept our input correctly. Dell is investigating.
   """
   ensureOff(client)

   path = client.root.Systems.Members[0].path
   response = client.patch(path, {
      "Boot": {
         "BootSourceOverrideEnabled": "Continuous",
         "BootSourceOverrideMode": "UEFI",
         "BootSourceOverrideTarget": "UefiTarget",
         "UefiTargetBootSourceOverride": "VenHw(986D1755-B9D0-4F8D-A0DA-D1DB18672045)",
      }
   })

   jobIdLocation = response.headers["Location"]
   jobId = regexSearch("JID_.+", jobIdLocation).group()

   # TODO: OData services support the $filter syntax where we could
   #       make the server search for the desired item instead of
   #       doing that ourselves. E.g.,
   #       GET serviceRoot/People?$filter=FirstName eq 'Scott'
   jobs = client.root.Managers.Members[0].Links.Oem.Dell.Jobs
   for i in range(len(jobs.Members)):
      if jobs.Members[i].Id == jobId:
         break
   else:
      raise RuntimeError("Job %s not found" % jobId)

   waitFor(lambda: jobs.Members[i].JobState == "Scheduled")

   # We reset the system in order for the change to take effect. Note
   # that this will not happen instantly (not to say it takes forever),
   # and so we wait for the associated job to finish. Note furthermore
   # that the job only completes after an incredibly long time, way
   # after the UI actually shows the job as complete. Lastly, note that
   # the update of the UefiTargetBootSourceOverride variable happens at
   # the same time the job completes, so we could either wait for the
   # job or wait for the update, it really does not matter.
   ensureReset(client)

   waitFor(lambda: jobs.Members[i].PercentComplete >= 100,
           timeout=timedelta(minutes=5))
