#!/usr/bin/env python

########################################################################
# Copyright 2018 VMware, Inc. All rights reserved.
# -- VMware Confidential
########################################################################

from http.client import (
   ResponseNotReady,
)
from os import (
   environ,
)
from redFishClientTestCase import (
   RedFishClientTestCase,
)
from time import (
   sleep
)
from unittest import (
   main,
   SkipTest,
)
from urllib.parse import (
   urlparse as urlParse,
)


# Note: This test case runs on a live system. Please make sure that
#       the test is the only client to avoid spurious test failures.

class TestRedFishClient(RedFishClientTestCase):
   """Tests for the RedFish client functionality."""
   def testResourcePostSync(self):
      """Verify that we can post to a resource synchronously."""
      states = ["On", "Off"]

      with self.client() as client:
         powerState = client.root.Systems.Members[0].PowerState
         if powerState == states[0]:
            action = "ForceOff"
         else:
            assert powerState == states[1]
            action = "On"

         # TODO: Handling actions is still a bit cumbersome. The REST
         #       API design does not fit nicely into our overall scheme
         #       of things. We should check if there is a better way to
         #       accomplish the same task.
         path = client.root.Systems.Members[0].Actions["#ComputerSystem.Reset"]["target"]
         data = {
            "ResetType": action,
         }
         task = client.post(path, data)
         for _ in range(60):
            if task.poll() is not None:
               break
            sleep(1)
         else:
            raise TimeoutError("failed to reset platform in 60 sec.")

         powerState = client.root.Systems.Members[0].PowerState


   def testResourcePostAsync(self):
      """Verify that we can post to a resource asynchronously."""
      with self.client() as client:
         if client.root.Systems.Members[0].Manufacturer != "Dell Inc.":
            raise SkipTest("SCP only available on Dell platforms")

         actions = client.root.Managers.Members[0].Actions
         path = actions['Oem']['OemManager.v1_0_0#OemManager.ExportSystemConfiguration']['target']

         payload = {
            "ExportFormat": "JSON",
            "ExportUse": "Default",
            "ShareParameters": {"Target": "Bios"}
         }
         task = client.post(path, payload)
         for _ in range(60):
            if task.poll() is not None:
               break
            sleep(1)
         else:
            raise TimeoutError("failed to export SCP in 60 sec.")


   def testResourceDeletion(self):
      """Verify that resource deletion works."""
      with self.client() as client:
         sessions = client.root.SessionService.Sessions.Members
         if len(sessions) > 0:
            print("Deleting session %s" % sessions[0].Id)
            del sessions[0]

         # We removed the first session (if there was any). Make sure
         # there is one less session being reported.
         new_sessions = client.root.SessionService.Sessions.Members
         self.assertEqual(new_sessions, sessions[1:])


   def testResponseNotDrained(self):
      """Verify that the autoRead decorator works correctly."""
      for read in (True, False):
         with self.client() as client:
            path = client.root.Systems.Members[0].path
            client.patch(path, {
               "Boot": {
                  "BootSourceOverrideTarget": "None",
               }
            }, read=read)

            if read:
               # We drained the response object. The request should go
               # through just fine.
               _ = client.root.Systems.Members[0]
            else:
               # Because nobody read the response object completely, the
               # GET will fail.
               with self.assertRaises(ResponseNotReady):
                  _ = client.root.Systems.Members[0]


if __name__ == "__main__":
   try:
      host = environ["TEST_HOST_URL"]
   except KeyError:
      raise SkipTest("TEST_HOST_URL (http[s]://user:password@host) "
                     "environment variable not set")

   result = urlParse(host)
   RedFishClientTestCase.allowUnsecure = result.scheme != "https"
   RedFishClientTestCase.host = "%s://%s" % (result.scheme, result.hostname)
   RedFishClientTestCase.username = result.username
   RedFishClientTestCase.password = result.password
   RedFishClientTestCase.certificate = environ.get("TEST_HOST_CERT", None)

   main(exit=False)
