#!/usr/bin/env python

########################################################################
# Copyright 2018 VMware, Inc. All rights reserved.
# -- VMware Confidential
########################################################################

from borautils.port import (
   findRandomFreePort,
   isPortFree,
)
from borautils.process import (
   expectRunning,
   DaemonPopen,
)
from borautils.waitfor import (
   waitFor,
)
from contextlib import (
   contextmanager,
)
from datetime import (
   timedelta,
)
from http import (
   HTTPStatus,
)
from os import (
   listdir,
)
from os.path import (
   dirname,
   join,
)
from redfish.redfish import (
   # We peek into some internals here.
   ROOT as REDFISH_ROOT,
)
from redFishClientTestCase import (
   HTTPError,
   RedFishClientTestCase,
   ResourceNotFound,
)
from subprocess import (
   CalledProcessError,
   DEVNULL,
)
from sys import (
   executable as python,
)
from tarfile import (
   open as tarOpen,
)
from tempfile import (
   TemporaryDirectory,
)
from unittest import (
   main,
   SkipTest,
)


@contextmanager
def _startMockupServer(mockup, port, log=False):
   """Start a mockup server on the given port."""
   file_ = join(dirname(__file__), "redFishMockupServer.py")

   output = None if log else DEVNULL
   cmd = [python, file_, "-P", str(port), "-D", mockup]
   with DaemonPopen(cmd, stdout=output, stderr=output) as p:
      yield p


@contextmanager
def startMockupServer(mockup, log=False):
   """Start the RedFish mockup server."""
   attempts = 5

   for _ in range(attempts):
      try:
         port = findRandomFreePort()
         with _startMockupServer(mockup, port, log=log) as p:
            waitFn = lambda: expectRunning(p) and \
                         not isPortFree(port)
            waitFor(waitFn, timeout=timedelta(seconds=30))
            yield port
         return
      except CalledProcessError:
         # It is very likely that the server failed to start because the
         # port is already in use. That can happen if we race with other
         # instances being spun up concurrently. Retry.
         continue
   else:
      raise RuntimeError("Failed to start up mockup server in %u attempts." % attempts)


class TestRedFishMockupClient(RedFishClientTestCase):
   """Tests for the RedFish client functionality."""
   def testExpectedStatusOverwrite(self):
      """Check that we can overwrite the expected statuses."""
      with self.client() as client:
         regex = r"Request for .* returned unexpected status"
         with self.assertRaisesRegex(RuntimeError, regex):
            # Check that we can set multiple allowed statuses.
            client.get(REDFISH_ROOT, statuses=(HTTPStatus.BAD_REQUEST,
                                               HTTPStatus.NOT_FOUND))


   def testResourceNotFound(self):
      """Verify the behavior for when a resource cannot be found."""
      with self.client() as client:
         with self.assertRaisesRegex(ResourceNotFound, "XXX_Undef_XXX"):
            client.root.XXX_Undef_XXX


   def testResourceNavigation(self):
      """Verify that we can navigate the resource namespace."""
      with self.client() as client:
         self.assertIn("Name", client.root.Chassis)
         self.assertIn("Members", client.root.Chassis)
         self.assertIn("Boot", client.root.Systems.Members[0])
         self.assertIn("BootSourceOverrideEnabled", client.root.Systems.Members[0].Boot)


   def testResourceListSliceSyntax(self):
      """Verify that the slice syntax work properly for RedFish resource lists."""
      with self.client() as client:
         logs = client.root.Managers.Members[0].LogServices.Members[1].Entries.Members
         self.assertEqual(logs, logs[0:1] + logs[1:])


   def testResourceDelete(self):
      """Verify that we can delete a resource (well, make the attempt to...)."""
      with self.client() as client:
         sessions = client.root.SessionService.Sessions.Members

         if len(sessions) <= 0:
            raise SkipTest("No sessions found in mockup")

         # The mockups service, unfortunately, does not appear to be
         # dynamic enough to actually adhere to the request, so we
         # cannot check for an actual deletion.
         regex = r"returned unexpected status: 405 \(Method Not Allowed\)"
         with self.assertRaisesRegex(HTTPError, regex) as cm:
            del sessions[0]

         self.assertEqual(cm.exception.status, HTTPStatus.METHOD_NOT_ALLOWED)


   def testNumbericanResourceAttribute(self):
      """Verify that we can correctly access a numerical resource attribute."""
      with self.client() as client:
         jobs = client.root.Managers.Members[0].Links.Oem.Dell.Jobs

         if len(jobs) <= 0:
            raise SkipTest("No jobs found in mockup")

         # The real test here is the dereference of the PercentComplete
         # attribute, which is of type integer and that should be
         # handled correctly.
         self.assertIsInstance(jobs.Members[0].PercentComplete, int)


if __name__ == "__main__":
   MOCKUP_DIR = join(dirname(__file__), "mockups")

   for mockup in listdir(MOCKUP_DIR):
      mockup = join(MOCKUP_DIR, mockup)
      with TemporaryDirectory() as mockupDir:
         with tarOpen(mockup) as archive:
            archive.extractall(path=mockupDir)

         with startMockupServer(mockupDir, log=False) as port:
            RedFishClientTestCase.host = "http://localhost"
            RedFishClientTestCase.port = port

            main(exit=False)
