# Copyright 2017-2019 VMware, Inc.  All rights reserved. -- VMware Confidential

"""vmwarelog.py --

Best effort parser for vmware.log files. Based on borautils.logparser
functionality with specialization for vmware.log contents, including
common timestamp and thread prefixes to all log lines.

Attributes:

   MSGRULES (dict): A curated catalog of common rules describing
      vmware log messages, mapping rule names to regexes that are
      matched against the message portion of a log line.
"""

import datetime
import re
import types
import borautils.logparser

# Catalog of useful vmware.log message patterns.
MSGRULES = {
   # The first line of a vmware.log file.
   'log_info' :
       r'Log for VMware (?P<product>\w+)\ +'
       r'.* version=(?P<version>[0-9.]+)\ +'
       r'build=(?P<build>build-[0-9]+)\ +'
       r'option=(?P<option>\w+)',

   # TimeTracker conversion rates.
   'tt_rate' :
       r'TimeTracker host to guest rate conversion '
       r'(?P<inbase>\d+) @ (?P<inhz>\d+)Hz -> '
       r'(?P<outbase>\d+) @ (?P<outhz>\d+)Hz',

   # TimeTracker TSC properties.
   'tt_tsc' :
       r'TSC (?P<property>scaling|offsetting) enabled',

   # Guest resync stepping time
   'guest_resync_step' :
       r'Guest resync: stepping time',

   # W32time resync executing
   'guest_resync_exec' :
       r'ProcMgr_ExecAsync: Executing async command: w32tm /resync',

   # Guest resync timeout
   'guest_resync_timeout' :
       r'Guest resync timeout handler: stepping time',

   # Guest onetime stepping time
   'guest_onetime_step' :
       r'One time synchronization: stepping time',

   # Guest periodic stepping time
   'guest_periodic_step' :
       r'Periodic synchronization: stepping time',

   # VMX config values
   'vmx_config' :
       r'DICT\s+(?P<vmx_flag>.*) = "(?P<value>.*)"',

   # Guest hard reset
   'guest_reset_hard' :
       r'Chipset: The guest has requested that the virtual machine be hard '
       r'reset'
}


class VMwareLogParserMatch(borautils.logparser.LogParserMatch):
   """A successful match in VMwareLogParser.

   Attributes:
      timestamp: datetime.datetime object corresponding to the log.
   """

   def __init__(self, lineNum, ruleName, ruleREMatch):
      """Intialize VMwareLogParseMatch object.

      Args:
         lineNum: Log line number.
         ruleName: Matching rule name.
         ruleREMatch: The regex match object.
      """
      super().__init__(lineNum, ruleName, ruleREMatch)
      self.timestamp = self._matchToTimestamp(ruleREMatch)


   def _matchToTimestamp(self, match, fractional_mult=1000):
      """Extract a vmware lib/log flavor of iso8601 timestamp from a
      given regex match object.

      Args:
          match: regex result matching VMwareLogParser.LOGLINEPREFIX_RE.

      Returns:
          An 'aware' datetime.datetime object corresponding to the
          timestamp. None, if the timestamp is not valid.
      """
      year        = int(match.group('year'))
      month       = int(match.group('month'))
      day         = int(match.group('day'))
      hour        = int(match.group('hour'))
      minute      = int(match.group('minute'))
      second      = int(match.group('second'))
      frac_usec   = match.group('fractional_usec')
      frac_msec   = match.group('fractional_msec')
      microsecond = 0
      if frac_usec:
         microsecond = int(frac_usec)
      elif frac_msec:
         microsecond = int(frac_msec) * 1000
      if match.group('tzsign'):
         tzsign = int("%s1" % match.group('tzsign'))
         tzhour = tzsign * int(match.group('tzhour'))
         tzminute = tzsign * int(match.group('tzminute'))
         tz = datetime.timezone(datetime.timedelta(hours=tzhour,
                                                   minutes=tzminute))
      else:
         tz = datetime.timezone.utc
      return datetime.datetime(year=year, month=month, day=day,
                               hour=hour, minute=minute, second=second,
                               microsecond=microsecond,
                               tzinfo=tz)


class VMwareLogParser(borautils.logparser.LogParser):
   """vmware.log parser
   """

   # Log-line prefix regular expression. A log line looks something
   # like this:
   #
   #     2017-07-19T09:24:12.871Z| vmx| I125: The process is 64-bit.
   #
   # The characters until and including 'I125:' are treated as the
   # common prefix prepended to all rules.
   #
   ISO8601TIMESTAMP_RE = (
       r'(?P<year>[0-9]{4})-(?P<month>[0-9]{2})-(?P<day>[0-9]{2})T'
       r'(?P<hour>[0-9]{2}):(?P<minute>[0-9]{2}):(?P<second>[0-9]{2})'
       r'([.]((?P<fractional_usec>[0-9]{6})|(?P<fractional_msec>[0-9]{3})))?'
       r'((?P<zulu>Z)'
       r'|((?P<tzsign>[-+])(?P<tzhour>[0-9]{2}):?(?P<tzminute>[0-9]{2})))?'
   )
   LOGLEVELINFO_RE     = r'(?P<loglevel>[A-Z]+[0-9]+):'
   THREADNAME_RE       = r'(?P<thread>[ a-zA-Z0-9\-_]+)'
   LOGLINEPREFIX_RE    = (ISO8601TIMESTAMP_RE + r'\ *\|\ *' +
                          THREADNAME_RE       + r'\ *\|\ *' +
                          LOGLEVELINFO_RE     + r'\ *')

   def __init__(self, msgRules=MSGRULES, **kwargs):
      """Initialize the parser.

      Args:
         msgRules: A dictionary of pattern matching rules, mapping rule
            names to regular expression strings. Defaults to MSGRULES.
      """
      super().__init__(**kwargs)
      self.AddMsgRules(msgRules)


   def AddMsgRules(self, msgRules):
      """Add rules matching the message portion of vmware.log log lines.
      This is a specialized extension of AddRules implicitly prepending
      the common prefix to all regexes. To add rules that match the
      raw log line, use AddRules.

      Args:
         msgRules: A dictionary of pattern matching rules, mapping rule
            names to regular expression strings. Defaults to MSGRULES.
      """
      rules = {}
      for name, msgRule in msgRules.items():
         rules[name] = self.LOGLINEPREFIX_RE + '.*' + msgRule
      self.AddRules(rules)


   def OnMatch(self, lineNum, ruleName, ruleREMatch):
      """Returns VMwareLogParseMatch object on a successful match,
      overriding the base object's callback.
      """
      return VMwareLogParserMatch(lineNum, ruleName, ruleREMatch)


if __name__ == '__main__':
   print("Running basic tests ...")
   loglines ="""
   2017-07-19T09:24:12.871Z| vmx| I125: Log for VMware ESX pid=2114349 version=6.6.2 build=build-6078998 option=Release
   2017-07-19T09:24:12.871Z| vmx| I125: The process is 64-bit.
   2017-07-19T09:24:12.871Z| vmx| I125: Host codepage=UTF-8 encoding=UTF-8
   2017-07-19T09:24:12.871Z| vmx| I125: Host is VMkernel 6.6.2
   2017-07-19T09:24:13.347Z| vmx| I125: TimeTracker host to guest rate conversion 20883022009711 @ 2400029000Hz -> 0 @ 2400029000Hz
   2017-07-19T09:24:13.347Z| vmx| I125: TimeTracker host to guest rate conversion ((x * 2147483648) >> 31) + -20883022009711
   2017-07-19T09:24:13.347Z| vmx| I125: TSC scaling enabled.
   2017-07-19T09:24:13.347Z| vmx| I125: TSC offsetting enabled.
   2017-07-20T03:53:40.193Z| vmx| I125: TimeTrackerStats behind by 0 us; running at 100.0%; mode 0; catchup limited 0 us; 1 stops, 0 giveups, 0 steps, 0 large steps; apparentOffset -6384232865808
   2017-07-20T03:53:40.193Z| vmx| I125: TimeTrackerStats APIC0 aperiodic 2262 ints, 226.14/sec; 1626909 tot; 0 loprg, 0 rtry
   """
   lp = VMwareLogParser()
   match = lp.ParseString(loglines)

   # match: log_info
   m = next(match)
   assert (m.name == 'log_info' and
           m.groups['product'] == 'ESX' and
           m.groups['version'] == '6.6.2' and
           m.groups['build'] == 'build-6078998' and
           m.groups['option'] == 'Release'), m

   # match: tt_rate
   m = next(match)
   assert (m.name == 'tt_rate' and
           m.groups['inbase'] == '20883022009711' and
           m.groups['inhz'] == '2400029000' and
           m.groups['outbase'] == '0' and
           m.groups['outhz'] == '2400029000'), m

   # match: tt_tsc
   m = next(match)
   assert (m.name == 'tt_tsc' and
           m.groups['property'] == 'scaling'), m
   m = next(match)
   assert (m.name == 'tt_tsc' and
           m.groups['property'] == 'offsetting'), m

   print('OK')

# vim: sw=3 ts=3
