# Copyright 2017-2019 VMware, Inc.  All rights reserved. -- VMware Confidential

"""Windows w32time log parser.
"""

import os
import re
import datetime
import tempfile
import borautils.logparser


class W32TimeLogParser(borautils.logparser.LogParser):
   """W32time log parser.
   """

   # W32time log timestamp prefix
   PREFIX_RE = (
       r'(?P<days>\d+) '
       r'(?P<hour>[0-9]{2}):(?P<minute>[0-9]{2}):(?P<second>[0-9]{2})'
       r'([.](?P<hns>[0-9]{7})s)'
   )

   # W32time log message regex rules
   MSG_RULES = {
      # W32time resync request
      'w32tm_resync_request':
         r'W32TmServiceMain: resync req, user requested',

      # W32time setting guest time
      'w32tm_set_time':
         r'ClockDispln Discipline: \*SET\*TIME\* - KPho:(?P<div>-?\d+)',

      # W32time skewing guest time
      'w32tm_skew_time':
         r'ClockDispln Discipline: \*SKEW\*TIME\* .*KPhO:(?P<kpho>-?\d+)',

      # Provider found
      'provider_found':
         r"Found provider '(?P<provider>\S+)':",

      # Provider enabled
      'provider_enabled':
         r"'Enabled'=0x00000001",

      # Provider started
      'provider_started':
         r"Time provider '(?P<provider>\S+)' has started successfully",

      # Provider selected
      'provider_selected':
         r"The time service is now synchronizing the system time with "
         r"the time source (?P<provider>\S+).",

      # Provider removed
      'provider_removed':
        r"RemoveProviderFromList: (?P<provider>\S+)"
   }


   def __init__(self, **kwargs):
      super().__init__(**kwargs)
      rules = {}
      for name, msgRule in self.MSG_RULES.items():
         rules[name] = self.PREFIX_RE + '.*' + msgRule
      self.AddRules(rules)


   def OnMatch(self, lineNum, ruleName, ruleREMatch):
      """Returns W32TimeLogParserMatch object on a successful match,
      overriding the base object's callback.
      """
      return W32TimeLogParserMatch(lineNum, ruleName, ruleREMatch)


class W32TimeLogParserMatch(borautils.logparser.LogParserMatch):
   """A successful match in W32TimeLogParser.

   Attributes:
      timestamp: datetime.datetime object corresponding to the log.
   """

   def __init__(self, lineNum, ruleName, ruleREMatch):
      """Intialize W32TimeLogParserMatch object.

      Args:
         lineNum: Log line number.
         ruleName: Matching rule name.
         ruleREMatch: The regex match object.
      """
      super().__init__(lineNum, ruleName, ruleREMatch)
      self.timestamp = self._matchToTimestamp(ruleREMatch)


   def _matchToTimestamp(self, match, fractional_mult=1000):
      """Extract a w32time flavor of timestamp from a given regex match object.

      Args:
          match: regex result matching w32TimeLogParser.PREFIX_RE.

      Returns:
          An datetime.datetime object corresponding to the timestamp.
      """
      days = int(match.group('days'))
      hour = int(match.group('hour'))
      minute = int(match.group('minute'))
      second  = int(match.group('second'))
      hns = int(match.group('hns'))

      # 01/01/1601 is the start of Windows epoch time
      log_time = datetime.datetime(1601, 1, 1) + datetime.timedelta(days=days)
      return datetime.datetime(year=log_time.year, month=log_time.month,
                               day=log_time.day, hour=hour, minute=minute,
                               second=second, microsecond=int(hns / 10),
                               tzinfo=datetime.timezone.utc)


if __name__ == '__main__':
   sampleW32tm = """
   152223 15:47:39.9610296s - 5 Age:5 Ofs:+00.0000000s COfs:+3599.7383401s Dly:+00.0000000s RDly:+00.0000000s Dsp:16.0000000s RDsp:00.0000000s Pnt:00.0000000s Dst:16.0000000s FDsp:00.2500000s Jitter:00.0000000s AgeTime:+13152124059.9610241s
   152223 15:47:39.9610654s - 4 Age:4 Ofs:+00.0000000s COfs:+3599.7383401s Dly:+00.0000000s RDly:+00.0000000s Dsp:16.0000000s RDsp:00.0000000s Pnt:00.0000000s Dst:16.0000000s FDsp:00.7500000s Jitter:00.0000000s AgeTime:+13152124059.9610241s
   152223 15:47:39.9611001s - 3 Age:3 Ofs:+00.0000000s COfs:+3599.7383401s Dly:+00.0000000s RDly:+00.0000000s Dsp:16.0000000s RDsp:00.0000000s Pnt:00.0000000s Dst:16.0000000s FDsp:01.7500000s Jitter:00.0000000s AgeTime:+13152124059.9610241s
   152223 15:47:39.9613435s - 2 Age:2 Ofs:+00.0000000s COfs:+3599.7383401s Dly:+00.0000000s RDly:+00.0000000s Dsp:16.0000000s RDsp:00.0000000s Pnt:00.0000000s Dst:16.0000000s FDsp:03.7500000s Jitter:00.0000000s AgeTime:+13152124059.9610241s
   152223 15:47:39.9613864s - 1 Age:1 Ofs:+00.0000000s COfs:+3599.7383401s Dly:+00.0000000s RDly:+00.0000000s Dsp:16.0000000s RDsp:00.0000000s Pnt:00.0000000s Dst:16.0000000s FDsp:07.7500000s Jitter:00.0000000s AgeTime:+13152124059.9610241s
   152223 15:47:39.9614226s - 0 Age:0 Ofs:-00.0235397s COfs:-00.0235397s Dly:+00.0007290s RDly:+00.0041351s Dsp:00.0000001s RDsp:00.0614166s Pnt:00.0000000s Dst:00.0007290s FDsp:07.7500000s Jitter:00.0000000s AgeTime:+00.0006696s
   152223 15:47:39.9614577s - Peer jitter: 00.0000000s Filter Dispersion: 07.7500000s
   152223 15:47:39.9614951s - Reachability:  peer ntp1-pao12.eng.vmware.com,0x1 (ntp.m|0x1|0.0.0.0:123->10.20.145.1:123) is reachable.
   152223 15:47:39.9615191s - Logging information: NtpClient is currently receiving valid time data from ntp1-pao12.eng.vmware.com,0x1 (ntp.m|0x1|0.0.0.0:123->10.20.145.1:123).
   152223 15:47:39.9615156s - W32TmServiceMain: resync req, user requested, get samples as soon as possible.
   152223 15:47:39.9615569s - UpdateTimerQueue1: TN:1175780000::: LRT:1175310000  LUT:1175780000 LAFLTNS:0 TSLGT:31476 LTNS:500000000000
   152223 15:47:39.9615770s - W32TmServiceMain: waiting i0.000s (1023.996s)
   152223 15:47:39.9616022s - Reachability: Attempting to contact peer ntp2-pao12.eng.vmware.com,0x1 (ntp.m|0x1|0.0.0.0:123->10.20.145.2:123).
   152223 15:47:39.9616298s - Polling peer ntp2-pao12.eng.vmware.com,0x1 (ntp.m|0x1|0.0.0.0:123->10.20.145.2:123)
   152223 15:47:39.9616530s - Sending packet to ntp2-pao12.eng.vmware.com,0x1 (ntp.m|0x1|0.0.0.0:123->10.20.145.2:123) in Win2K detect mode, stage 1.
   152223 15:44:01.2161598s - ClockDispln Discipline: *SET*TIME* - KPho:-35997383401
   152223 15:44:01.2162910s - Logging information: The time service is now synchronizing the system time with the time source ntp2-pao12.eng.vmware.com,0x1 (ntp.m|0x1|0.0.0.0:123->10.20.145.2:123).
   152657 09:20:18.4544139s - ReadConfig: Found provider 'vmTimeProvider':
   152657 09:20:18.4544407s - ReadConfig:   'Enabled'=0x00000001
   152651 09:17:29.5737282s - Starting 'vmTimeProvider', dll:'c:\Program Files\VMware\VMware Tools\vmTimeProvider\vmTimeProvider.dll'
   152651 09:17:29.6234704s - Logging Info: Time provider 'vmTimeProvider' has started successfully
   152665 09:14:40.9976104s - Logging information: The time service is now synchronizing the system time with the time source vmTimeProvider.
   """
   print("Running basic tests ...")
   match = W32TimeLogParser().ParseString(sampleW32tm)
   m = next(match)
   assert m.name == 'w32tm_resync_request'
   assert m.timestamp.tzinfo == datetime.timezone.utc
   m = next(match)
   assert m.name == 'w32tm_set_time'
   m = next(match)
   assert m.name == 'provider_selected'
   m = next(match)
   assert m.name == 'provider_found'
   m = next(match)
   assert m.name == 'provider_enabled'
   m = next(match)
   assert m.name == 'provider_started'
   m = next(match)
   assert m.name == 'provider_selected'
   print("OK")

# vim: sw=3 ts=3
